Package: episomer
Title: Early Detection of Public Health Threats from Social Media Data
Version: 3.0.35
Authors@R: c( 
     person(given = "Laura",
            family = "Espinosa",
            role = c("aut", "fnd", "cre"),
            email = "laura.espinosa@ecdc.europa.eu",
            comment = c(ORCID = "0000-0003-0748-9657", "Project manager, author of the design and concept of the package, and package maintainer")),
     person(given = "Francisco",
            family = "Orchard",
            role = c("aut", "ctr"),
            email = "f.orchard@epiconcept.fr",
            comment = c(ORCID = "0000-0001-5793-3301", "Author of the package and original code")),
     person(given = "Gianfranco",
            family = "Spiteri",
            role = c("ctb", "fnd"),
            email = "gianfranco.spiteri@ecdc.europa.eu"),
     person(given = "Ariana",
            family = "Wijermans",
            role = "ctb",
            email = "ariana.wijermans@ecdc.europa.eu",
            comment = "Contributor to the design and concept of the original version of the package"),
     person(given = "Thomas",
            family = "Mollet",
            role = c("ctb"),
            email = "thomas.mollet@ecdc.europa.eu",
            comment = "Contributor to the design and concept of the original version of the package"),
     person(given = "Adrian",
            family = "Prodan",
            role = "ctb",
            email = "adrian.prodan@ecdc.europa.eu"),
     person(given = "Martin",
            family = "Norling",
            role = "ctb",
            email = "martin.norling@ecdc.europa.eu"),
     person(given = "Enrique",
            family = "Delgado",
            role = "ctb",
            email = "enrique.delgado@ecdc.europa.eu"),
     person(given = "Thomas",
            family = "Czernichow",
            role = "ctb",
            email = "thomas.czernichow@epiconcept.fr"),
     person(given = "Maria",
            family = "Prieto Gonzalez",
            role = "ctb",
            email = "m.gonzalez@epiconcept.fr"),
     person(given = "Esther",
            family = "Kissling",
            role = "ctb",
            email = "e.kissling@epiconcept.fr"), 
     person(given = "Michael",
            family = "Höhle",
            role = "ctb",
            email = "hoehle@math.su.se"),
     person(given = "Yohann",
            family = "Mansiaux",
            role = "ctb",
            email = "y.mansiaux@epiconcept.fr")
     )
Description: It allows you to automatically monitor trends of social media messages by time, place and topic aiming at detecting public health threats early through the detection of signals (i.e., an unusual increase in the number of messages per time, topic and location). It was designed to focus on infectious diseases, and it can be extended to all hazards or other fields of study by modifying the topics and keywords. More information on the original package 'epitweetr' is available in the peer-review publication Espinosa et al. (2022) <doi:10.2807/1560-7917.ES.2022.27.39.2200177>.
License: EUPL
URL: https://github.com/EU-ECDC/episomer
BugReports: https://github.com/EU-ECDC/episomer/issues
Encoding: UTF-8
Imports: dplyr, curl, DT, emayili, future, httr, htmltools, jsonlite,
        keyring, ggplot2, janitor, magrittr, parallel, plotly,
        processx, readxl, rlang, rmarkdown, rnaturalearthdata,
        openxlsx, shiny, stringr, stats, tibble, tools, utils, xtable,
        httr2, lubridate, sf, cli
RoxygenNote: 7.3.3
Suggests: knitr, taskscheduleR, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-08 06:30:49 UTC; fod
Author: Laura Espinosa [aut, fnd, cre] (ORCID:
    <https://orcid.org/0000-0003-0748-9657>, Project manager, author of
    the design and concept of the package, and package maintainer),
  Francisco Orchard [aut, ctr] (ORCID:
    <https://orcid.org/0000-0001-5793-3301>, Author of the package and
    original code),
  Gianfranco Spiteri [ctb, fnd],
  Ariana Wijermans [ctb] (Contributor to the design and concept of the
    original version of the package),
  Thomas Mollet [ctb] (Contributor to the design and concept of the
    original version of the package),
  Adrian Prodan [ctb],
  Martin Norling [ctb],
  Enrique Delgado [ctb],
  Thomas Czernichow [ctb],
  Maria Prieto Gonzalez [ctb],
  Esther Kissling [ctb],
  Michael Höhle [ctb],
  Yohann Mansiaux [ctb]
Maintainer: Laura Espinosa <laura.espinosa@ecdc.europa.eu>
Repository: CRAN
Date/Publication: 2026-05-08 10:50:02 UTC
Built: R 4.5.2; ; 2026-05-08 12:49:36 UTC; unix

Package: episomer
Title: Early Detection of Public Health Threats from Social Media Data
Version: 3.0.35
Authors@R: c( 
     person(given = "Laura",
            family = "Espinosa",
            role = c("aut", "fnd", "cre"),
            email = "laura.espinosa@ecdc.europa.eu",
            comment = c(ORCID = "0000-0003-0748-9657", "Project manager, author of the design and concept of the package, and package maintainer")),
     person(given = "Francisco",
            family = "Orchard",
            role = c("aut", "ctr"),
            email = "f.orchard@epiconcept.fr",
            comment = c(ORCID = "0000-0001-5793-3301", "Author of the package and original code")),
     person(given = "Gianfranco",
            family = "Spiteri",
            role = c("ctb", "fnd"),
            email = "gianfranco.spiteri@ecdc.europa.eu"),
     person(given = "Ariana",
            family = "Wijermans",
            role = "ctb",
            email = "ariana.wijermans@ecdc.europa.eu",
            comment = "Contributor to the design and concept of the original version of the package"),
     person(given = "Thomas",
            family = "Mollet",
            role = c("ctb"),
            email = "thomas.mollet@ecdc.europa.eu",
            comment = "Contributor to the design and concept of the original version of the package"),
     person(given = "Adrian",
            family = "Prodan",
            role = "ctb",
            email = "adrian.prodan@ecdc.europa.eu"),
     person(given = "Martin",
            family = "Norling",
            role = "ctb",
            email = "martin.norling@ecdc.europa.eu"),
     person(given = "Enrique",
            family = "Delgado",
            role = "ctb",
            email = "enrique.delgado@ecdc.europa.eu"),
     person(given = "Thomas",
            family = "Czernichow",
            role = "ctb",
            email = "thomas.czernichow@epiconcept.fr"),
     person(given = "Maria",
            family = "Prieto Gonzalez",
            role = "ctb",
            email = "m.gonzalez@epiconcept.fr"),
     person(given = "Esther",
            family = "Kissling",
            role = "ctb",
            email = "e.kissling@epiconcept.fr"), 
     person(given = "Michael",
            family = "Höhle",
            role = "ctb",
            email = "hoehle@math.su.se"),
     person(given = "Yohann",
            family = "Mansiaux",
            role = "ctb",
            email = "y.mansiaux@epiconcept.fr")
     )
Description: It allows you to automatically monitor trends of social media messages by time, place and topic aiming at detecting public health threats early through the detection of signals (i.e., an unusual increase in the number of messages per time, topic and location). It was designed to focus on infectious diseases, and it can be extended to all hazards or other fields of study by modifying the topics and keywords. More information on the original package 'epitweetr' is available in the peer-review publication Espinosa et al. (2022) <doi:10.2807/1560-7917.ES.2022.27.39.2200177>.
License: EUPL
URL: https://github.com/EU-ECDC/episomer
BugReports: https://github.com/EU-ECDC/episomer/issues
Encoding: UTF-8
Imports: dplyr, curl, DT, emayili, future, httr, htmltools, jsonlite,
        keyring, ggplot2, janitor, magrittr, parallel, plotly,
        processx, readxl, rlang, rmarkdown, rnaturalearthdata,
        openxlsx, shiny, stringr, stats, tibble, tools, utils, xtable,
        httr2, lubridate, sf, cli
RoxygenNote: 7.3.3
Suggests: knitr, taskscheduleR, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-08 06:30:49 UTC; fod
Author: Laura Espinosa [aut, fnd, cre] (ORCID:
    <https://orcid.org/0000-0003-0748-9657>, Project manager, author of
    the design and concept of the package, and package maintainer),
  Francisco Orchard [aut, ctr] (ORCID:
    <https://orcid.org/0000-0001-5793-3301>, Author of the package and
    original code),
  Gianfranco Spiteri [ctb, fnd],
  Ariana Wijermans [ctb] (Contributor to the design and concept of the
    original version of the package),
  Thomas Mollet [ctb] (Contributor to the design and concept of the
    original version of the package),
  Adrian Prodan [ctb],
  Martin Norling [ctb],
  Enrique Delgado [ctb],
  Thomas Czernichow [ctb],
  Maria Prieto Gonzalez [ctb],
  Esther Kissling [ctb],
  Michael Höhle [ctb],
  Yohann Mansiaux [ctb]
Maintainer: Laura Espinosa <laura.espinosa@ecdc.europa.eu>
Repository: CRAN
Date/Publication: 2026-05-08 10:50:02 UTC
Built: R 4.6.0; ; 2026-05-08 11:25:15 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(add_new_social_media)
export(admin_app)
export(bluesky_create_session)
export(calculate_region_alerts)
export(calculate_regions_alerts)
export(check_all)
export(create_map)
export(create_snapshot)
export(create_topchart)
export(create_topwords)
export(dashboard_app)
export(detect_loop)
export(download_dependencies)
export(ears_t_reweighted)
export(episomer_app)
export(fs_loop)
export(generate_alerts)
export(geolocate_text)
export(get_aggregates)
export(get_alerts)
export(get_tasks)
export(health_check)
export(is_detect_running)
export(is_fs_running)
export(is_search_running)
export(missing_search_jobs)
export(register_detect_runner_task)
export(register_fs_monitor)
export(register_fs_runner_task)
export(register_search_runner_task)
export(save_config)
export(search_loop)
export(search_loop_worker)
export(search_posts)
export(setup_config)
export(sm_api_get_token_bluesky)
export(sm_api_search_bluesky)
export(sm_api_set_auth_bluesky)
export(stop_detect_runner_task)
export(stop_fs_runner_task)
export(stop_search_runner_task)
export(trend_line)
export(update_geonames)
export(update_languages)
importFrom(DT,dataTableOutput)
importFrom(DT,dataTableProxy)
importFrom(DT,datatable)
importFrom(DT,formatPercentage)
importFrom(DT,renderDataTable)
importFrom(DT,replaceData)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,dense_rank)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,if_else)
importFrom(dplyr,mutate)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(dplyr,summarize)
importFrom(dplyr,ungroup)
importFrom(future,future)
importFrom(future,multisession)
importFrom(future,plan)
importFrom(ggplot2,aes)
importFrom(ggplot2,coord_fixed)
importFrom(ggplot2,coord_flip)
importFrom(ggplot2,coord_sf)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,element_text)
importFrom(ggplot2,fortify)
importFrom(ggplot2,geom_col)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_polygon)
importFrom(ggplot2,geom_sf)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggsave)
importFrom(ggplot2,labs)
importFrom(ggplot2,margin)
importFrom(ggplot2,scale_fill_manual)
importFrom(ggplot2,scale_size_continuous)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_classic)
importFrom(ggplot2,xlab)
importFrom(grDevices,png)
importFrom(httr2,is_online)
importFrom(httr2,last_response)
importFrom(httr2,req_body_json)
importFrom(httr2,req_error)
importFrom(httr2,req_headers)
importFrom(httr2,req_perform)
importFrom(httr2,req_retry)
importFrom(httr2,req_url_query)
importFrom(httr2,request)
importFrom(httr2,resp_body_json)
importFrom(httr2,resp_check_status)
importFrom(httr2,resp_status)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,rbind_pages)
importFrom(jsonlite,read_json)
importFrom(jsonlite,stream_in)
importFrom(lubridate,as_datetime)
importFrom(magrittr,"%>%")
importFrom(magrittr,`%>%`)
importFrom(plotly,config)
importFrom(plotly,ggplotly)
importFrom(plotly,layout)
importFrom(plotly,plotlyOutput)
importFrom(plotly,renderPlotly)
importFrom(readxl,read_excel)
importFrom(rlang,current_env)
importFrom(rmarkdown,render)
importFrom(sf,st_as_sf)
importFrom(sf,st_crs)
importFrom(sf,st_transform)
importFrom(shiny,actionButton)
importFrom(shiny,checkboxInput)
importFrom(shiny,column)
importFrom(shiny,conditionalPanel)
importFrom(shiny,dateRangeInput)
importFrom(shiny,downloadButton)
importFrom(shiny,downloadHandler)
importFrom(shiny,fileInput)
importFrom(shiny,fluidPage)
importFrom(shiny,fluidRow)
importFrom(shiny,h2)
importFrom(shiny,h3)
importFrom(shiny,h4)
importFrom(shiny,h5)
importFrom(shiny,htmlOutput)
importFrom(shiny,invalidateLater)
importFrom(shiny,navbarPage)
importFrom(shiny,need)
importFrom(shiny,numericInput)
importFrom(shiny,observe)
importFrom(shiny,observeEvent)
importFrom(shiny,passwordInput)
importFrom(shiny,radioButtons)
importFrom(shiny,renderText)
importFrom(shiny,renderUI)
importFrom(shiny,selectInput)
importFrom(shiny,shinyApp)
importFrom(shiny,sliderInput)
importFrom(shiny,span)
importFrom(shiny,tabPanel)
importFrom(shiny,textAreaInput)
importFrom(shiny,textInput)
importFrom(shiny,uiOutput)
importFrom(shiny,updateDateRangeInput)
importFrom(shiny,updateSliderInput)
importFrom(shiny,validate)
importFrom(stats,qt)
importFrom(stats,reorder)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stringr,str_replace_all)
importFrom(tools,md5sum)
importFrom(utils,URLencode)
importFrom(utils,browseURL)
importFrom(utils,download.file)
importFrom(utils,head)
importFrom(utils,tail)
importFrom(utils,unzip)
importFrom(utils,write.csv)

This is the source code for ecdc-twitter-bundle_2.12-1.0.jar. 
To build the jar yourself you have to execute you can do:

On Linux

#make sure the decompressed files are on the current directory
export SPARK_VERSION=3.0.0
sbt package

On Windows
#make sure the decompressed files are on the current directory
set SPARK_VERSION=3.0.0
sbt package

## ----echo = FALSE-------------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(episomer)
package_name <- "episomer"
package_desc <- packageDescription(package_name)
pkg_version <- package_desc$Version
pkg_date <- package_desc$Date
authors_str <- gsub("^ *|(?<= ) |\n| *$", "", package_desc$Authors, perl = TRUE)
author_list <- eval(parse(text = authors_str))
pkg_authors <- paste(
  format(
    author_list,
    include = c("given", "family", "email", "comment"),
    braces = list(
      email = c("<", ">,<br />"),
      comment = c("", "")
    )
  ),
  collapse = "<br /><br />"
)
pkg_maintainer <- package_desc$Maintainer
pkg_license <- package_desc$License
pkg_url <- package_desc$URL

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/dashboard.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/alerts.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/geotag_evaluation.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/data_protection.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/config.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/troubleshoot.png")

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# install.packages(episomer)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# system("java -version")

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# Sys.getenv("JAVA_HOME")

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# library(episomer)
# episomer_app("data_dir")

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# library(episomer)
# episomer_app("data_dir")

## ----echo=FALSE, fig.cap="", out.width = '50%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/run_tasks.png")

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# library(episomer)
# fs_loop("data_dir")

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# library(episomer)
# search_loop("data_dir")

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# library(episomer)
# detect_loop("data_dir")

## ----echo=FALSE, fig.cap="", out.width = '50%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/bluesky_authentication.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/topics.png")

## ----echo=FALSE, fig.cap="", out.width = '80%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/Example_topics_list.png")

## ----echo=FALSE, fig.cap="", out.width = '80%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/Schedule_plan_image.png")

## ----echo=FALSE, fig.cap="", out.width = '80%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/Search_plan_queues.png")

## ----echo=FALSE, fig.cap="", out.width = '80%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/config_languages.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/subscribers.png")

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# episomer_app("data_dir")

## ----eval=FALSE, echo=TRUE----------------------------------------------------
# # Opens only the interactive dashboard
# dashboard_app("data_dir")
# 
# # Opens only the rest of the features with the administrative
# admin_app("data_dir")

## ----echo=FALSE, fig.cap="", out.width = '40%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/topics_countries.png")

## ----echo=FALSE, fig.cap="", out.width = '40%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/period.png")

## ----echo=FALSE, fig.cap="", out.width = '30%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/time_unit.png")

## ----echo=FALSE, fig.cap="", out.width = '30%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/include_quotes.png")

## ----echo=FALSE, fig.cap="", out.width = '30%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/signal_detection_fpr.png")

## ----echo=FALSE, fig.cap="", out.width = '40%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/outlier_dashboard.png")

## ----echo=FALSE, fig.cap="", out.width = '30%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/Bonferroni_correction.png")

## ----echo=FALSE, fig.cap="", out.width = '30%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/days_in_baseline.png")

## ----echo=FALSE, fig.cap="", out.width = '30%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/same_weekday_baseline.png")

## ----echo=FALSE, fig.cap="", out.width = '30%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/hover_trendline.png")

## ----echo=FALSE, fig.cap="", out.width = '60%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/hover_map.png")

## ----echo=FALSE, fig.cap="", out.width = '50%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/top_words.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/alerts.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/geotag_evaluation.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/config.png")

## ----echo=FALSE, fig.cap="", out.width = '50%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/status.png")

## ----echo=FALSE, fig.cap="", out.width = '70%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/detection_pipeline.png")

## ----echo=FALSE, fig.cap="", out.width = '50%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/signal_detection_config.png")

## ----echo=FALSE, fig.cap="", out.width = '40%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/general.png")

## ----echo=FALSE, fig.cap="", out.width = '30%'--------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/region_disclaimer.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/topics.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/languages.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/troubleshoot.png")

## ----echo=FALSE, fig.cap="", out.width = '100%'-------------------------------
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/ebola.png")

#!/usr/bin/env Rscript
# script for running the search loop
args = commandArgs(trailingOnly=TRUE)

# test if the data directory path has been provided
if (length(args)==0) {
  stop("The config file has not been provided", call.=FALSE)
} 

data_dir <- args[1]
episomer::search_loop(data_dir)  

# functions in this script are the generic functions that needs to be implemented for 
# code in episomer call these functions instead of the social media specific passing the plan which contains the name of the social media.
# social media specific functions are identified by naming convention using adding the name of the social media at the end of the function name.

#' parse custom attributes for a plan in this specific social media and returns them as a named list with attributes in the proper type
#' @param network char, the name of the social media 
#' @param ... list, triple dot arguments to be passed as a parameter tu underlying functions 
sm_plan_parse_attributes <- function(network, ...) {
  get(sprintf("sm_plan_parse_attributes_%s", network))(...)
}

#' prepare a plan for serialisation by transforming attributes with non standard serialisation types e.g. dates to string on the format expected by the parse functions
#' @param plan, the plan to format the attributes
sm_plan_format <- function(plan) {
  get(sprintf("sm_plan_format_%s", plan$network))(plan)
}

#' returns a list with the custom social media plan attributes with the default values for a first plan in a topic.
#' @param network char, the name of the social media 
sm_plan_first_attributes <- function(network) {
  get(sprintf("sm_plan_first_attributes_%s", network))()
}

#' returns a list with the custom social media plan attributes with the values for the plan to be run after the plans provided as parameter
#' @param network char, network the name of the social media 
#' @param plans list, list of plans to use for calculate new plan attributes
sm_plan_next_attributes <- function(network, plans) {
  get(sprintf("sm_plan_next_attributes_%s", network))(plans)
}

#' calculates and returns the estimated progress of a plan using custom attributes
#' @param plan list, the plan to obtain the progress from
sm_plan_get_progress <- function(plan) {
  get(sprintf("sm_plan_get_progress_%s", plan$network))(plan)
}

#' Return a message to log what the next query is going to be for the current plan
#' @param plan list, the plan to obtain the message 
sm_plan_next_search_info <- function(plan) {
  get(sprintf("sm_plan_search_info_%s", plan$network))(plan)
}

#' Get social media token
#' @details this function implements the authethication on the respective social media and stores it securely. 
#' It is called by the search loop before starting message collection
#' @param network char, network the name of the social media 
sm_api_get_token <- function(network) {
  get(sprintf("sm_api_get_token_%s", network))()
}

#' Translate a parsed query to the social media query format
#' @details This function receive as a parameter a parsed query from the a topic in the list of topics. 
#' It can return one or many queries that matches the expected results using the underlying social media API
#' @param network char, the name of the social media 
#' @param parsed char, parsed query to translate in one or many native equivalent queries 
sm_api_translate_query <- function(network, parsed) {
  get(sprintf("sm_api_translate_query_%s", network))(parsed)
}

#' Search posts for a given query and produce standardised results
#' @details This function impleents the search of messages for a given query and plan.
#' @param query char, the query to use in the search in nattive format
#' @param token char, the token to authenticate against the underlying social media api 
#' @param plan list, the plan containing the targeted time frame
sm_api_search <- function(query, token, plan) {
  get(sprintf("sm_api_search_%s", plan$network))(query, token, plan)
}

#' Takes a plan after a request has been performed and the standardised results provided as returned by the function sm_api_search_xxx
#' it should return a plan with necessary information updated so next request will continue paginating as expected to obtain
#' targeted messages
#' @param plan list, the plan producing the resusts
#' @param results list, the result of the query in episomer format
sm_api_update_plan_after_request <- function(plan, results) {
  get(sprintf("sm_api_update_plan_after_request_%s", plan$network))(
    plan,
    results
  )
}

#' @title Save credentials for a social media provider and store them securely
#' @description Update configuration object and underlying storage with given bluesky username and password. The password is encrypted.
#' @param network char, the name of the social media
#' @param shiny_input_list list, the shiny input object to extract the information from
sm_api_set_auth <- function(network, shiny_input_list) {
  get(sprintf("sm_api_set_auth_%s", network))(
    shiny_input_list
  )
}

# @function_def_start (do not delete)

#' Get social media token
#' @details this function implements the authethication on the respective social media and stores it securely. 
#' It is called by the search loop before starting message collection
#' @return Token
#' @export
sm_api_get_token_bluesky <- function() {
# @function_def_end (do not delete)
  if (is_secret_set("bluesky_session")) {
    session <- jsonlite::fromJSON(get_secret("bluesky_session"))
    access_jwt <- bluesky_check_token_validity(session$access_jwt)
    if (session$created < strptime(Sys.time() - 3600, "%Y-%m-%d %H:%M:%S")) {
      session <- bluesky_create_session()
      access_jwt <- session$access_jwt
    }
    return(access_jwt)
  }
  return(bluesky_create_session()$access_jwt)
}


# @function_def_start (do not delete)

#' Translate a parsed query to the social media query format
#' @details This function receive as a parameter a parsed query from the a topic in the list of topics. 
#' It can return one or many queries that matches the expected results using the underlying social media API
#' @param parsed list. Named list containing two attributes 'ors' and 'neg'. 
#' 'ors' is a list containing subqueries that should be contatenated with an OR logic 
#'   - each subquery in 'ors' is a list containing terms that should be concatecated with an AND logic
#'       - each term in 'ors subqueries' is a list of synonyms for a particular literal
#' 'neg' is a list of terms that should be removed from any result of the query 
sm_api_translate_query_bluesky <- function(parsed) {
# @function_def_end (do not delete)
  ret = list()
  for(i in 1:length(parsed$ors)) {
     comb = do.call(expand.grid, parsed$ors[[i]])
     if (nrow(comb) > 0) {
       r <- paste(lapply(1:nrow(comb), function(i) do.call(paste, as.list(comb[i, ]))))
       if(length(parsed$neg)>0) 
          r <- paste(r, do.call(paste, as.list(paste0("-", parsed$neg))))
       ret = c(ret, r)
     }
  }
  ret
}


# @function_def_start (do not delete)

#' Search posts for a given query and produce standardised results
#'
#' @param query query to search for
#' @param token Access token
#' @param plan Plan object
#' @param max_retries Maximum number of retries
#' @param verbose Whether to print progress messages
#' @details This function impleents the search of messages for a given query and plan.
#' The query has been already transformed by sm_api_translate_query_xxx so it is expected to be already in the native social media format
#' The token has been produced by the function sm_api_get_token_xxx and shoulb be used to autenticate the request
#' The plan should be used to extract the precise time-span to request messages
#'
#' - Response attributes
#'   - network: string
#'   - posts: [...]
#'   - count: number, 
#'   - pagination:\{...\}
#' - post in posts
#'     - id: string
#'     - uri: string
#'     - created_at: datetime
#'     - user_name: string
#'     - text: string
#'     - lang: string
#'     - quoted_text: string
#'     - quoted_lang: string
#'     - tags: [...]
#'     - urls: [...]
#'     - categories: [...]
#' - tag in post.tags: string
#' - url in post.urls: string
#' - category in post.categories: string
#' - pagination: attributes used for paginating requests
#'
#' This is the implementation of the bluesky search. You can find more information about the bluesky API here: \url{https://docs.bsky.app/docs/api/app-bsky-feed-search-posts}
#' Additionnaly more information about the "post" object is available here: \url{https://atproto.blue/en/latest/atproto/atproto_client.models.app.bsky.feed.defs.html#atproto_client.models.app.bsky.feed.defs.PostView}
#' @return List with posts int the standard post format defined in episomer
#' @export
#' @importFrom httr2 request req_url_query req_headers req_perform resp_body_json is_online resp_status last_response req_retry
#' @importFrom magrittr %>%
#' @importFrom lubridate as_datetime
sm_api_search_bluesky <- function(
  query,
  token,
  plan,
  max_retries = 10,
  verbose = TRUE
) {
# @function_def_end (do not delete)
  search_url = "https://bsky.social/xrpc/app.bsky.feed.searchPosts"
  sort <- "latest"
  number_of_posts_per_request <- 100

  # Make a single request and return results
  req <- httr2::request(search_url) %>%
    httr2::req_url_query(
      q = query,
      limit = number_of_posts_per_request,
      sort = sort
    )

  if (!is.null(plan$plan_min_date)) {
    req <- req %>%
      httr2::req_url_query(since = bluesky_format_date(plan$plan_min_date))
  }
  if (!is.null(plan$current_min_date)) {
    req <- req %>%
      httr2::req_url_query(
        until = bluesky_format_date(plan$current_min_date - 0.000001)
      )
  } else {
    req <- req %>%
      httr2::req_url_query(until = bluesky_format_date(plan$plan_max_date))
  }

  # Must find a way to check for invalid token
  resp <- req %>%
    httr2::req_headers(Authorization = paste("Bearer", token)) %>%
    httr2::req_retry(
      max_tries = max_retries,
      retry_on_failure = TRUE,
      is_transient = bluesky_rate_limited_check,
      after = bluesky_rerun_after_rate_limit
    ) %>%
    httr2::req_perform()
  # Get results
  json <- httr2::resp_body_json(resp)
  bluesky_parse_response(json, plan$current_min_date)
}

# @function_def_start (do not delete)

#' @title Save bluesky credentials (login and password) and store them securely
#' @description Update configuration object and underlying storage with given bluesky username and password. The password is encrypted.
#' @param shiny_input_list List of shiny input values containing the bluesky specific credentials
#' @details Update authentication tokens in configuration object
#' @examples
#' \dontrun{
#'  #Setting the configuration values
#'    sm_api_set_auth_bluesky(
#'      shiny_input_list = list(bluesky_user = "your user here", bluesky_password = "your password"),
#'    )
#' }
#' @seealso
#' \code{\link{save_config}}
#' @return Nothing. Called for side effects (saving credentials)
#' @export
sm_api_set_auth_bluesky <- function(shiny_input_list) {
# @function_def_end (do not delete)
  if(!"bluesky_user" %in% names(shiny_input_list)) {
    stop("bluesky_user is not in the shiny_input_list")
  }
  if(!"bluesky_password" %in% names(shiny_input_list)) {
    stop("bluesky_password is not in the shiny_input_list")
  }
  set_secret("bluesky_user", shiny_input_list$bluesky_user)
  set_secret("bluesky_password", shiny_input_list$bluesky_password)
}

# @function_def_start (do not delete)

#' Takes a plan after a request has been performed and the standardised results provided as returned by the function sm_api_search_xxx
#' it should return a plan with necessary information updated so next request will continue paginating as expected to obtain
#' targeted messages
#' @param plan list, the plan to update
#' @param results list, the obtained results of the plan
sm_api_update_plan_after_request_bluesky <- function(plan, results) {
# @function_def_end (do not delete)
  if (!is.null(results$pagination$min_created_at)) {
    min_obtained = lubridate::as_datetime(results$pagination$min_created_at)
    if(is.null(plan$current_min_date) || min_obtained < plan$current_min_date) {
       plan$current_min_date <- min_obtained
    }
  }
  return(plan)
}

# @function_def_start (do not delete)

#' Receive the results provided by a request and return a logical value indicating whether the request returned results.
#' @details This function needs to be implemented for each social media since it is interpreted as 'No more messages are available for this plan'
#' When FALSE is returned the underlying plan will be ended
#' @param results list, results of the request in the standardised episomer format
sm_api_got_rows_bluesky <- function(results) {
# @function_def_end (do not delete)
  (exists("posts", results) & length(results$posts) > 0)
}


## The rest of the functions in this script are used by sm_api_***_bluesky functiona and are internal 
## They can be used for inspiration for other social media

#' function to check if there has been a rate limit error returned by the bluesky API
#' @param resp, the httr2 response object
bluesky_rate_limited_check <- function(resp) {
  if (httr2::resp_status(resp) == 429) {
    identical(httr2::resp_header(resp, "RateLimit-Remaining"), "0")
  } else if (httr2::resp_status(resp) %in% c(420, 500, 503, 502)) {
    message(sprintf("Error %s encountered.", httr2::resp_status(resp)))
    TRUE
  } else {
    FALSE
  }
}

#' function to extract the number of seconds to wait after a rate limit has been reached
#' @param resp, the httr2 response object
bluesky_rerun_after_rate_limit <- function(resp) {
  if (httr2::resp_status(resp) == 429) {
    message("Rate limit exceeded. Waiting for reset.")
    time <- as.numeric(httr2::resp_header(resp, "RateLimit-Reset"))
    reset_after <- time - unclass(Sys.time())
    message(sprintf("Rate limit reset in %s seconds.", reset_after))
    return(reset_after)
  } else {
    return(NA)
  }
}

#' formats a date in the expected bluesky format
#' @param datetime, the datetime to format
bluesky_format_date <- function(datetime) {
  timezone = "UTC"
  # Full ISO 8601 format with time
  format(datetime, "%Y-%m-%dT%H:%M:%OS6Z", tz = "UTC")
}

#' parse a date as provided by bluesky into a date_time object
#' @param date_input char, the text of a date
bluesky_parse_date <- function(date_input) {
  lubridate::as_datetime(unlist(date_input))
}

#' parses the response of bluesky API into the standardised episomer format. For more details see the vignette.
#' @param response list, the standardised response
#' @param current_min_date datetime, the current min date obtained on the plan to detect if there are messages after this limit and discard them 
bluesky_parse_response <- function(response, current_min_date) {
  first <- function(x) unlist(x[min(1, length(x))])
  res = list(
    network = "bluesky",
    pagination = list(min_created_at = NULL)
  )
  res[["posts"]] = lapply(response$posts, function(post) {
    record <- post$record
    quote <- bluesky_parse_quoted(post)
    features <- bluesky_parse_features(post)
    urls <- unique(c(features$urls, quote$urls))
    ret <- list(
      id = unlist(post$cid),
      uri = sprintf(
        "https://bsky.app/profile/%s/post/%s",
        post$author$did,
        sub(".*/", "", post$uri)
      ),
      created_at = bluesky_format_date(bluesky_parse_date(record$createdAt)),
      user_name = gsub(".bsky.social", "", post$author$handle),
      text = unlist(record$text),
      lang = first(record$langs),
      is_quote = quote$is_quote &&
        !is.null(quote$text) &&
        nchar(quote$text) > 0,
      quoted_text = quote$text,
      quoted_lang = quote$lang,
      tags = features$tags,
      urls = urls,
      categories = features$categories
    )
    if(bluesky_parse_date(record$createdAt) > Sys.time()) {
        warning(sprintf("Ignoring post with creation date in future :%s (%s)", record$createdAt, ret$uri))
        ret <- NULL
    }

    # checking if there are posts after the current min date. If it is the case we remove them since it breaks search loop pagination and can enter into a endless loop.
    if (!is.null(current_min_date) && bluesky_parse_date(record$createdAt) > current_min_date)  {
        warning(sprintf("Ignoring post with creation (%s) after the upper bound: (%s), uri: (%s)", current_min_date, record$createdAt, ret$uri))
        ret <- NULL
    }
    #msg(jsonlite::toJSON(list(text=substr(ret$text, 1, 30), lang = ret$lang, quote=substr(ret$quoted_text, 1, 30), qlang=ret$quoted_lang), auto_unbox = T, null = "null"))
    #if(!is.null(quote$text) && nchar(quote$text) > 0)
    #msg(jsonlite::toJSON(list(text=substr(ret$text, 1, 30), isq = ret$is_quote, quote=substr(ret$quoted_text, 1, 30), qlang=ret$quoted_lang), auto_unbox = T, null = "null"))
    ret
  })
  # removing null posts due to a creation date in future
  
  if (length(res$posts) > 0) {
      res$posts <- res$posts[sapply(res$posts, function(p) !is.null(p))]
  }
  if (length(res$posts) > 0) {
      res$posts <- res$posts[sapply(res$posts, function(p) !is.null(p$lang))]
  }
  if (length(res$posts) > 0) {
    res$pagination$min_created_at <- Reduce(
      function(x, y) if (x < y) x else y,
      lapply(res$posts, `[[`, "created_at")
    )
  } 
  res$count = length(res$posts)
  res
}

#' extracts the quotes information from a post in one of the many possible locations found in bluesky format
#' @param post list, the post to to extract data from in bluesky format
bluesky_parse_quoted <- function(post) {
  # utilitary functions
  first <- function(x) {
    x <- x[!sapply(x, is.null)]
    unlist(x[min(1, length(x))])
  }
  mode <- function(x) first(names(sort(table(unlist(x)), decreasing = TRUE)))

  # extracting embedding data from post
  if ("embed" %in% names(post)) {
    quoted_texts = list()
    embeds = list()
    records = list()
    images = list()
    urls = list()
    is_quote <- FALSE

    if (
      "record" %in% names(post$embed) && "embeds" %in% names(post$embed$record)
    ) {
      embeds = post$embed$record$embeds
    }
    if ("external" %in% names(post$embed)) {
      embeds = c(embeds, list(post$embed))
    }
    found <- FALSE
    # case 1: URLS title and descriptions
    #         post$embed$record$embeds[[1]]$external$(title and description)
    for (embed in embeds) {
      if (
        "external" %in%
          names(embed) &&
          length(intersect(
            names(embed$external),
            c("uri", "title", "description")
          )) >
            0
      ) {
        found <- TRUE
        texts = sapply(
          list("title", "description"),
          function(attr) unlist(embed$external[[attr]])
        )
        quoted_texts = c(
          quoted_texts,
          list(list(
            text = paste(texts[!is.null(texts)], collapse = "\n"),
            lang = mode(post$record$langs)
          ))
        )
        if ("uri" %in% names(embed$external)) {
          urls = c(urls, embed$external$uri)
        }
      }
    }
    # case 2: quoted post
    #         post$embed$record$value$text
    #         post$embed$record$description
    #         post$embed$record$record$value$text
    #         post$embed$record$record$description
    if (
      "record" %in%
        names(post$embed) &&
        (("value" %in%
          names(post$embed$record) &&
          "text" %in% names(post$embed$record$value)) ||
          "description" %in% names(post$embed$record))
    ) {
      records = c(records, list(post$embed$record))
    }
    if (
      "record" %in%
        names(post$embed) &&
        "record" %in% names(post$embed$record) &&
        (("value" %in%
          names(post$embed$record$record) &&
          "text" %in% names(post$embed$record$record$value)) ||
          "description" %in% names(post$embed$record$record))
    ) {
      records = c(records, list(post$embed$record$record))
    }
    for (record in records) {
      found <- TRUE
      is_quote <- TRUE
      quoted_texts = c(
        quoted_texts,
        list(list(
          text = first(c(record$value$text, record$description)),
          lang = first(c(record$value$langs, post$record$langs))
        ))
      )
    }
    # case 3: Images alternative textes
    #         post$embed$images[[1]]$alt
    #         post$embed$media$images[[1]]$alt
    if ("images" %in% names(post$embed)) {
      found <- TRUE
      images = c(images, post$embed$images)
    }
    if (
      "media" %in% names(post$embed) && "images" %in% names(post$embed$media)
    ) {
      found <- TRUE
      images = c(images, post$embed$media$images)
    }
    for (image in images) {
      if ("alt" %in% names(image) && nchar(image$alt) > 0) {
        quoted_texts = c(
          quoted_texts,
          list(list(
            text = image$alt,
            lang = mode(post$record$langs)
          ))
        )
      }
    }
    # case 4: video (not considered)
    if ("playlist" %in% names(post$embed)) {
      found <- TRUE
    }
    # case 5: not Found, blocked or detached (not considered)
    if (
      "record" %in%
        names(post$embed) &&
        length(intersect(
          c("notFound", "blocked", "detached"),
          names(post$embed$record)
        )) >
          0
    ) {
      found <- TRUE
    }
    if (
      "record" %in%
        names(post$embed) &&
        length(intersect(
          c("notFound", "blocked", "detached"),
          names(post$embed$record$record)
        )) >
          0
    ) {
      found <- TRUE
    }

    if (!found) {
      jsonlite::write_json(post, sprintf("%s/badpost.json", conf$data_dir))
      warning(sprintf(
        ":-) 2 Embed information expected but not found in %s",
        jsonlite::toJSON(post, auto_inbox = TRUE)
      ))
    }
    text <- paste(sapply(quoted_texts, `[[`, "text"), collapse = "\n")
    lang = paste(
      substr(mode(sapply(quoted_texts, `[[`, "lang")), 1, 2),
      collapse = "\n"
    )
    list(
      text = if (is.null(text) || nchar(text) == 0) NULL else text,
      lang = if (is.null(lang) || nchar(lang) == 0) NULL else lang,
      urls = unique(urls),
      is_quote = is_quote
    )
  } else {
    list(
      is_quote = FALSE
    )
  }
}

#' extracts features from the post structure returned by the bluesky api 
#' @param post list, the post to extract features from in bluesky format
bluesky_parse_features <- function(post) {
  # extracting features data from post
  ret <- list(tags = list(), urls = list(), categories = list())
  if ("record" %in% names(post) && "facets" %in% names(post$record)) {
    for (facet in post$record$facets) {
      if ("features" %in% names(facet)) {
        for (feature in facet$features) {
          ftype <- feature[["$type"]]
          if (ftype == "app.bsky.richtext.facet#tag" || ftype == "app.bsky.richtext.facet#hashtag") {
            ret$tags <- c(ret$tags, feature[["tag"]])
          } else if (ftype == "app.bsky.richtext.facet#link") {
            ret$urls <- c(ret$urls, feature[["uri"]])
          } else if (
            ftype %in%
              c(
                "app.bsky.richtext.facet#mention",
                "app.bsky.richtext.facet#format",
                "app.twitter.source#id",
                "blue.poll.post.facet#option"
              )
          ) {
            #pass
          } else {
            jsonlite::write_json(
              post,
              sprintf("%s/badpost.json", conf$data_dir)
            )
            warning(sprintf(
              ":-) Unexpected feature type %s in %s",
              ftype,
              jsonlite::toJSON(post, auto_unbox = TRUE)
            ))
          }
        }
      }
    }
  }
  ret
}

#' Get bearer token
#'
#' @param handle bluesky handle
#' @param password bluesky password
#'
#' @return List with access_jwt and did
#' @export
#' @importFrom httr2 request req_body_json req_perform resp_check_status resp_body_json is_online req_headers
#' @details You can find more information about the bluesky API here: \url{https://docs.bsky.app/docs/api/com-atproto-server-create-session}
#' @examples
#' \dontrun{
#' bluesky_create_session()
#' }
#'
bluesky_create_session <- function(handle = NULL, password = NULL) {
  if (is.null(handle) && is_secret_set("bluesky_user")) {
    handle = get_secret("bluesky_user")
  } else if (is.null(handle) && Sys.getenv("bluesky_id") != "")
    handle = Sys.getenv("bluesky_id") else {
    stop("You have to provide a user_name to collect Bluesky posts")
  }

  if (is.null(password) && is_secret_set("bluesky_password")) {
    password = get_secret("bluesky_password")
  } else if (is.null(handle) && Sys.getenv("bluesky_pwd") != "")
    password = Sys.getenv("bluesky_pwd") else {
    stop("You have to provide a password to collect Bluesky posts")
  }
  login_url = "https://bsky.social/xrpc/com.atproto.server.createSession"

  if (!httr2::is_online()) {
    stop("No internet connection")
  }

  resp <- httr2::request(login_url) %>%
    httr2::req_body_json(list(identifier = handle, password = password)) %>%
    httr2::req_retry(
      max_tries = 5,
      retry_on_failure = TRUE,
      is_transient = bluesky_rate_limited_check,
      after = bluesky_rerun_after_rate_limit
    ) %>%
    httr2::req_perform()

  httr2::resp_check_status(resp)

  session <- httr2::resp_body_json(resp)
  session <- list(
    handle = session$handle,
    access_jwt = session$accessJwt,
    did = session$did,
    refreshJwt = session$refreshJwt,
    created = strptime(Sys.time(), "%Y-%m-%d %H:%M:%S")
  )
  json_session = jsonlite::toJSON(session)
  set_secret("bluesky_session", json_session)
  return(jsonlite::fromJSON(json_session))
}


#' check the current bluesky token validity
#' @param access_jwt char, the jwt token provided by the bluesky api
#' @param search_url char, the url to test if the current token is still valid
#' @importFrom httr2 request req_url_query req_headers req_error req_perform resp_status
bluesky_check_token_validity <- function(
  access_jwt,
  search_url = "https://bsky.social/xrpc/app.bsky.feed.searchPosts"
) {
  simple_request <- try(
    httr2::request(search_url) %>%
    httr2::req_url_query(q = "covid19", limit = 1, sort = "latest") %>%
    httr2::req_headers(Authorization = paste("Bearer", access_jwt)) %>%
    httr2::req_retry(
      max_tries = 5,
      retry_on_failure = TRUE,
      is_transient = bluesky_rate_limited_check,
      after = bluesky_rerun_after_rate_limit
    ) %>%
    httr2::req_perform()
  )
  if (httr2::resp_status(last_response()) %in% c(400, 401)) {
    message("Invalid token. Creating a new session.")
    access_jwt <- bluesky_create_session()$access_jwt
  }
  return(access_jwt)
}

# Get the SQL-like expression to extract post geolocation variables and apply prioritisation
# This function is used on aggregate posts for translating variables names into SQL-valid columns of geotagged posts
get_location_var <- function(varname) {
  paste(
    "coalesce(",
    "text_loc.",
    varname,
    ", quoted_text_loc.",
    varname,
    ")",
    sep = ""
  )
}

# It gets columns used for post geolocation. This is used for limiting columns to extract from JSON files
get_location_columns <- function(table) {
  list(
    "text_loc",
    "quoted_text_loc"
  )
}

# Get raw countries data frame from Excel file
get_raw_countries <- function() {
  # obtained from https://gist.github.com/AdrianBinDC/621321d6083b176b3a3215c0b75f8146#file-country-bounding-boxes-md but it had been manually updated
  df <- readxl::read_excel(get_countries_path())
  names(df) <- make.names(names(df), unique = TRUE)
  df$minLat <- as.numeric(df$minLat)
  df$maxLat <- as.numeric(df$maxLat)
  df$minLong <- as.numeric(df$minLong)
  df$maxLong <- as.numeric(df$maxLong)
  df
}


# Getting a list of regions, sub regions and countries for use as a selection
get_country_items <- function(order = "level") {
  `%>%` <- magrittr::`%>%`
  #If countries are already on cache we return them otherwise we calculate them
  if (exists(paste("country_items", order, sep = ""), where = cached)) {
    return(cached[["country_items", order, sep = ""]])
  } else {
    # Getting country data from package embedded excel, ignoring Antarctica
    countries <- get_raw_countries() %>% dplyr::filter(.data$region != "")

    # using intermediate region when available
    countries$sub.region <- mapply(
      function(int, sub) {
        if (is.na(int)) sub else int
      },
      countries$intermediate.region,
      countries$sub.region
    )
    #getting regions and subregions
    regions <- (countries %>%
      dplyr::group_by(.data$region) %>%
      dplyr::summarize(
        minLat = min(.data$minLat),
        minLong = min(.data$minLong),
        maxLat = max(.data$maxLat),
        maxLong = max(.data$maxLong)
      ) %>%
      dplyr::ungroup())
    subregions <- (countries %>%
      dplyr::group_by(.data$region, .data$sub.region) %>%
      dplyr::summarize(
        minLat = min(.data$minLat),
        minLong = min(.data$minLong),
        maxLat = max(.data$maxLat),
        maxLong = max(.data$maxLong)
      ) %>%
      dplyr::ungroup())

    # creating list containing results
    items <- list()
    # adding hardcoded world items
    row <- 1
    items[[row]] <- list(
      name = "World (all)",
      codes = list(),
      level = 0,
      minLat = -90,
      maxLat = 90,
      minLong = -180,
      maxLong = 180
    )
    row <- 2
    items[[row]] <- list(
      name = "World (geolocated)",
      codes = as.list(countries$alpha.2),
      level = 0,
      minLat = -90,
      maxLat = 90,
      minLong = -180,
      maxLong = 180
    )
    for (r in 1:nrow(regions)) {
      row <- row + 1
      region <- regions$region[[r]]
      items[[row]] = list(
        name = region,
        codes = list(),
        level = 1,
        minLat = 90,
        maxLat = -90,
        minLong = 180,
        maxLong = -180
      )
      # filling region item
      for (c in 1:nrow(countries)) {
        if (countries$region[[c]] == region) {
          # adding country codes for region
          items[[row]]$codes[[
            length(items[[row]]$codes) + 1
          ]] <- countries$alpha.2[[c]]
          # calculating region bounding box
          if (countries$minLat[[c]] < items[[row]]$minLat)
            items[[row]]$minLat <- countries$minLat[[c]]
          if (countries$maxLat[[c]] > items[[row]]$maxLat)
            items[[row]]$maxLat <- countries$maxLat[[c]]
          if (countries$minLong[[c]] < items[[row]]$minLong)
            items[[row]]$minLong <- countries$minLong[[c]]
          if (countries$maxLong[[c]] > items[[row]]$maxLong)
            items[[row]]$maxLong <- countries$maxLong[[c]]
        }
      }
      # Adding elements for subregions
      for (s in 1:nrow(subregions)) {
        if (subregions$region[[s]] == region) {
          row <- row + 1
          subregion <- subregions$sub.region[[s]]
          items[[row]] = list(
            name = subregion,
            codes = list(),
            level = 2,
            minLat = 90,
            maxLat = -90,
            minLong = 180,
            maxLong = -180
          )
          # filling sub region item
          for (c in 1:nrow(countries)) {
            if (
              countries$region[[c]] == region &&
                countries$sub.region[[c]] == subregion
            ) {
              # adding country codes for region
              items[[row]]$codes[[
                length(items[[row]]$codes) + 1
              ]] <- countries$alpha.2[[c]]
              # calculating region bounding box
              if (countries$minLat[[c]] < items[[row]]$minLat)
                items[[row]]$minLat <- countries$minLat[[c]]
              if (countries$maxLat[[c]] > items[[row]]$maxLat)
                items[[row]]$maxLat <- countries$maxLat[[c]]
              if (countries$minLong[[c]] < items[[row]]$minLong)
                items[[row]]$minLong <- countries$minLong[[c]]
              if (countries$maxLong[[c]] > items[[row]]$maxLong)
                items[[row]]$maxLong <- countries$maxLong[[c]]
            }
          }
          # adding country items
          for (c in 1:nrow(countries)) {
            if (
              countries$region[[c]] == region &&
                countries$sub.region[[c]] == subregion
            ) {
              row <- row + 1
              country <- countries$name[[c]]
              items[[row]] <- list(
                name = country,
                codes = list(countries$alpha.2[[c]]),
                level = 3,
                minLat = countries$minLat[[c]],
                maxLat = countries$maxLat[[c]],
                minLong = countries$minLong[[c]],
                maxLong = countries$maxLong[[c]]
              )
            }
          }
        }
      }
    }

    # Getting EU classification
    eu_zones <- countries$EU.EEA[!is.na(countries$EU.EEA)] %>% unique
    eu_codes <- lapply(eu_zones, function(r) {
      list(
        name = r,
        codes = countries$alpha.2[
          !is.na(countries$EU.EEA) & countries$EU.EEA == r
        ],
        level = 0.5
      )
    })
    eu_codes <- (if (length(eu_codes) == 0) eu_codes else
      c(
        eu_codes,
        list(list(
          name = paste(eu_zones, collapse = "+"),
          codes = countries$alpha.2[!is.na(countries$EU.EEA)],
          level = 0.5
        ))
      ))

    # Getting WHO classification
    who_zones <- countries$WHO.Region[!is.na(countries$WHO.Region)] %>% unique
    who_codes <- lapply(who_zones, function(r) {
      list(
        name = r,
        codes = countries$alpha.2[
          !is.na(countries$WHO.Region) & countries$WHO.Region == r
        ],
        level = 0.6
      )
    })
    specials <- c(eu_codes, who_codes)
    for (s in 1:length(specials)) {
      row <- row + 1
      items[[row]] = list(
        name = specials[[s]]$name,
        codes = list(),
        level = specials[[s]]$level,
        minLat = 90,
        maxLat = -90,
        minLong = 180,
        maxLong = -180
      )
      for (c in 1:nrow(countries)) {
        if (countries$alpha.2[[c]] %in% specials[[s]]$codes) {
          # adding country codes for region
          items[[row]]$codes[[
            length(items[[row]]$codes) + 1
          ]] <- countries$alpha.2[[c]]
          # calculating region bounding box
          if (countries$minLat[[c]] < items[[row]]$minLat)
            items[[row]]$minLat <- countries$minLat[[c]]
          if (countries$maxLat[[c]] > items[[row]]$maxLat)
            items[[row]]$maxLat <- countries$maxLat[[c]]
          if (countries$minLong[[c]] < items[[row]]$minLong)
            items[[row]]$minLong <- countries$minLong[[c]]
          if (countries$maxLong[[c]] > items[[row]]$maxLong)
            items[[row]]$maxLong <- countries$maxLong[[c]]
        }
      }
    }
    # Items are natively ordered as a hierarchy, but default of this function is ordered by level and name
    if (order == "level")
      items <- items[order(sapply(items, function(i) {
        paste(i$level, i$name)
      }))]

    # Setting the cache
    cached[[paste("country_items", order, sep = "_")]] <- items
    return(items)
  }
}

# Get country codes
# usage
# map <- get_country_code_map()
# map[["FR"]]
get_country_code_map <- function() {
  regions <- get_country_items()
  countries <- regions[sapply(regions, function(i) i$level == 3)]
  return(setNames(
    sapply(countries, function(r) r$name),
    sapply(countries, function(r) r$code)
  ))
}

# Get country codes by name
# usage
# map <- get_country_codes_by_name()
# map[["Asia"]]
get_country_codes_by_name <- function() {
  regions <- get_country_items()
  return(setNames(
    sapply(regions, function(r) r$code),
    sapply(regions, function(r) r$name)
  ))
}

# Get regions data as data frame
get_regions_df <- function() {
  regions <- get_country_items()
  data.frame(
    Name = sapply(regions, function(r) r$name),
    Codes = sapply(regions, function(r) paste(r$codes, collapse = ", ")),
    Level = sapply(regions, function(r) r$level),
    MinLatitude = as.numeric(sapply(regions, function(r) r$minLat)),
    MaxLatitude = as.numeric(sapply(regions, function(r) r$maxLat)),
    MinLongitude = as.numeric(sapply(regions, function(r) r$minLong)),
    MaxLingityde = as.numeric(sapply(regions, function(r) r$maxLong))
  )
}

# Get country indexes
# usage
# map <- get_country_index_map()
# map[["FR"]]
get_country_index_map <- function() {
  regions <- get_country_items()
  all_index <- 1:length(regions)
  indexes <- all_index[sapply(all_index, function(i) regions[[i]]$level == 3)]
  return(setNames(indexes, sapply(indexes, function(i) regions[[i]]$code)))
}

#' @title Updates the local copy of the GeoNames database
#' @description Downloading and indexing a fresh version of the GeoNames database from the provided URL
#' @param tasks Tasks object for reporting progress and error messages, default: get_tasks()
#' @return The list of tasks updated with produced messages
#' @details Run a one shot task to download and index a local copy of the  \href{http://www.geonames.org/}{GeoNames database}.
#' The GeoNames geographical database covers all countries and contains over eleven million place names that are available; Creative Commons Attribution 4.0 License.
#'
#' The URL to download the database from is set on the configuration tab of the Shiny app, in case it changes.
#'
#' The indexing is developed in Spark and Lucene
#'
#' A prerequisite to this function is that the \code{\link{search_loop}} must already have stored collected posts in the search folder and that the task \code{\link{download_dependencies}}
#' has been successfully run.
#'
#' Normally this function is not called directly by the user but from the \code{\link{detect_loop}} function.
#' @examples
#' \dontrun{
#'    library(episomer)
#'    # setting up the data folder
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'
#'    # geolocating last posts
#'    tasks <- update_geonames()
#' }
#' @rdname update_geonames
#' @seealso
#'  \code{\link{download_dependencies}}
#'
#'  \code{\link{detect_loop}}
#'
#'  \code{\link{get_tasks}}
#'
#' @export
#' @importFrom utils download.file unzip
update_geonames <- function(tasks = get_tasks()) {
  stop_if_no_config()
  old <- options()
  on.exit(options(old))
  options(timeout = 900)
  tasks <- tryCatch(
    {
      tasks <- update_geonames_task(
        tasks,
        "running",
        "downloading",
        start = TRUE
      )
      # Create geo folder if it does not exist
      if (!file.exists(file.path(conf$data_dir, "geo"))) {
        dir.create(file.path(conf$data_dir, "geo"))
      }
      # Downloading GeoNames
      temp <- tempfile()
      temp_dir <- tempdir()
      download.file(tasks$geonames$url, temp, mode = "wb")

      # Decompressing file
      tasks <- update_geonames_task(tasks, "running", "decompressing")
      unzip(zipfile = temp, files = "allCountries.txt", exdir = temp_dir)
      file.remove(temp)
      file.rename(
        from = file.path(temp_dir, "allCountries.txt"),
        to = get_geonames_txt_path()
      )

      # Assembling GeoNames datasets
      # using Spark job with entry point updateGeonames part 'ASSEMBLE'
      tasks <- update_geonames_task(tasks, "running", "assembling")
      spark_job(
        paste(
          "updateGeonames",
          conf_geonames_as_arg(),
          "assemble",
          paste("\"TRUE\"", sep = ""),
          "index",
          paste("\"FALSE\"", sep = ""),
          "parallelism",
          conf$spark_cores
        )
      )

      # Indexing GeoNames datasets
      tasks <- update_geonames_task(tasks, "running", "indexing")
      # using spark job with entry point updateGeonames part 'INDEX'
      spark_job(
        paste(
          "updateGeonames",
          conf_geonames_as_arg(),
          "assemble",
          paste("\"FALSE\"", sep = ""),
          "index",
          paste("\"TRUE\"", sep = ""),
          "parallelism",
          conf$spark_cores
        )
      )

      # Setting status to success
      tasks <- update_geonames_task(tasks, "success", "", end = TRUE)
      tasks
    },
    warning = function(error_condition) {
      # Setting status to failed
      tasks <- update_geonames_task(
        tasks,
        "failed",
        paste(
          "failed while",
          tasks$geonames$message,
          " geonames",
          error_condition
        )
      )
      tasks
    },
    error = function(error_condition) {
      # Setting status to failed
      tasks <- update_geonames_task(
        tasks,
        "failed",
        paste(
          "failed while",
          tasks$geonames$message,
          " geonames",
          error_condition
        )
      )
      tasks
    }
  )
  return(tasks)
}


#' @title Updates local copies of languages
#' @description Downloading and indexing a fresh version of language models tagged for update on the Shiny app configuration tab
#' @param tasks Tasks object for reporting progress and error messages, default: get_tasks()
#' @param reuse_downloads logical indicating if the downloads should be reused if they already exist, default: FALSE
#' @return The list of tasks updated with produced messages
#' @details Run a one shot task to download and index a local fasttext \href{https://fasttext.cc/docs/en/crawl-vectors.html}{pretrained models}.
#' A fasttext model is a collection of vectors for a language automatically produced scrolling a big corpus of text that can be used to capture the semantic of a word.
#'
#' The URL to download the vectors from are set on the configuration tab of the Shiny app.
#'
#' This task will also update SVM models to predict whether a word is a location that will be used in the geolocation process.
#'
#' The indexing is developed in SPARK and Lucene.
#'
#' A prerequisite to this function is that the \code{\link{search_loop}} must already have stored collected posts in the search folder and that the tasks \code{\link{download_dependencies}}
#' and \code{\link{update_geonames}} has been run successfully.
#'
#' Normally this function is not called directly by the user but from the \code{\link{detect_loop}} function.
#' @examples
#' \dontrun{
#'    library(episomer)
#'    # setting up the data folder
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'
#'    # updating language tasks
#'    tasks <- update_languages()
#' }
#' @rdname update_languages
#' @seealso
#'  \code{\link{download_dependencies}}
#'
#'  \code{\link{update_geonames}}
#'
#'  \code{\link{detect_loop}}
#'
#'  \code{\link{get_tasks}}
#'
#' @export
#' @importFrom utils download.file
update_languages <- function(tasks = get_tasks(), reuse_downloads = FALSE) {
  stop_if_no_config()
  old <- options()
  on.exit(options(old))
  options(timeout = 900)
  tryCatch2 <- function(exp, handler) {tryCatch(exp, error=handler, warning=handler)}
  tasks <- tryCatch(
    {
      tasks <- update_languages_task(
        tasks,
        "running",
        "downloading",
        start = TRUE
      )

      # Create languages folder if it does not exist
      if (!file.exists(file.path(conf$data_dir, "languages"))) {
        dir.create(file.path(conf$data_dir, "languages"))
      }

      # Executing actions per language
      i <- 1
      while (i <= length(tasks$languages$statuses)) {
        # downloading when necessary
        if (tasks$languages$statuses[[i]] %in% c("to add", "to update")) {
          tasks <- update_languages_task(
            tasks,
            "running",
            paste("downloading", tasks$languages$name[[i]]),
            lang_code = tasks$languages$codes[[i]],
            lang_start = TRUE
          )
          if(!reuse_downloads || !file.exists(tasks$languages$vectors[[i]])) { 
            temp <- sprintf("%s.tmp", tasks$languages$vectors[[i]])
            # downloading the file
	    download.file(tasks$languages$url[[i]], temp, mode = "wb")
	    tryCatch2({
                    file.rename(from = file.path(temp), to = tasks$languages$vectors[[i]])
	        },
	        handler = function(e) {
		    message(sprintf("Failed while trying to rename %s to %s with error : %s", file.path(temp), tasks$languages$vectors[[i]], e))
		    message("going try to copy the file")
                            if(file.copy(from = file.path(temp), to = tasks$languages$vectors[[i]], overwrite=T)) { 
			        message("Copy of file succeeded, going to delete origin file")
			        file.remove(temp)
			    } else {
		                stop("copy failed silently")
			    }
		}
	    )
            tasks <- update_languages_task(
              tasks,
              "running",
              paste("downloaded", tasks$languages$name[[i]]),
              lang_code = tasks$languages$codes[[i]],
              lang_done = TRUE
            )
	  }
        } else if (tasks$languages$statuses[[i]] %in% c("to remove")) {
          # requesting to delete language data
          # deleting language files
          code <- tasks$languages$codes[[i]]
          if (file.exists(get_lang_vectors_path(code)))
            file.remove(get_lang_vectors_path(code))
          if (dir.exists(get_lang_model_path(code)))
            unlink(get_lang_model_path(code))
          if (file.exists(get_lang_stamp_path(code)))
            file.remove(get_lang_stamp_path(code))
          tasks <- update_languages_task(
            tasks,
            "running",
            paste("deleted", tasks$languages$name[[i]]),
            lang_code = code,
            lang_removed = TRUE
          )
          i <- i - 1
        }
        i <- i + 1
      }
      # Indexing languages after changes
      # using the Spark job with entry point updateLanguages
      tasks <- update_languages_task(
        tasks,
        "running",
        "updating training set & indexing"
      )
      update_geotraining_df()
      retrain_languages()
      update_geotraining_df()

      # Setting status to success
      tasks <- update_languages_task(tasks, "success", "", end = TRUE)
      tasks
    },
    warning = function(error_condition) {
      # Setting status to failed
      tasks <- update_languages_task(
        tasks,
        "failed",
        paste(
          "failed while",
          tasks$languages$message,
          " languages",
          error_condition
        )
      )
      tasks
    },
    error = function(error_condition) {
      # Setting status to failed
      tasks <- update_languages_task(
        tasks,
        "failed",
        paste(
          "failed while",
          tasks$languages$message,
          " languages",
          error_condition
        )
      )
      tasks
    }
  )
  return(tasks)
}


#' @title geolocate text in a data frame given a text column and optionally a language column
#' @description extracts geolocaion information on text on a column of the provided data frame and returns a new data frame with geolocation information
#' @param df A data frame containing at least character column with text, a column with the language name can be provided to improve geolocation quality
#' @param text_col character, name of the column on the data frame containing the text to geolocalize, default:text
#' @param lang_col character, name of the column on the data frame containing the language of texts, default: NULL
#' @param min_score, numeric, the minimum score obtained on the Lucene scoring function to accept matches on GeoNames. It has to be empirically set default: NULL
#' @return A new data frame containing the following geolocation columns: geo_code, geo_country_code, geo_country, geo_name, tags
#' @details This function perform a call to the episomer database which includes functionality for geolocating for languages activated and successfully processed on the shiny app.
#'
#' The geolocation process tries to find the best match in GeoNames database \url{https://www.geonames.org/} including all local aliases for words.
#'
#' If no language is associated to the text, all tokens will be sent as a query to the indexed GeoNames database.
#'
#' If a language code is associated to the text and this language is trained on episomer, entity recognition techniques will be used to identify the best candidate in text to contain a location
#' and only these tokens will be sent to the GeoNames query.
#'
#' A custom scoring function is implemented to grant more weight to cities increasing with population to try to perform disambiguation.
#'
#' Rules for forcing the geolocation choices of the algorithms and for tuning performance with manual annotations can be performed on the geotag tab of the Shiny app.
#'
#' A prerequisite to this function is that the tasks \code{\link{download_dependencies}} \code{\link{update_geonames}} and \code{\link{update_languages}} has been run successfully.
#'
#' This function is called from the Shiny app on geolocation evaluation tab but can also be used for manually evaluating the episomer geolocation algorithm.
#' @examples
#' \dontrun{
#'    library(episomer)
#'    # setting up the data folder
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'
#'    # creating a test dataframe
#'    df <- data.frame(text = c("Me gusta Santiago de Chile es una linda ciudad"), lang = c("es"))
#'    geo <- geolocate_text(df = df, text_col = "text", lang_col="lang")
#'
#' }
#' @rdname geolocate_text
#' @seealso
#'  \code{\link{download_dependencies}}
#'
#'  \code{\link{update_geonames}}
#'
#'  \code{\link{detect_loop}}
#'
#' @export
#' @importFrom utils download.file
geolocate_text <- function(
  df,
  text_col = "text",
  lang_col = NULL,
  min_score = NULL
) {
  stop_if_no_config()
  stop_if_no_fs()

  `%>%` <- magrittr::`%>%`
  if (is.null(lang_col)) df[[lang_col]] <- NA
  to_geolocate = df %>%
    dplyr::transmute(
      id = as.character(dplyr::row_number()),
      text = .data[[text_col]],
      lang = .data[[lang_col]]
    ) %>%
    jsonlite::toJSON()
  geo_uri = paste(get_scala_geolocate_text_url(), "?jsonnl=true", sep = "")
  if (!is.null(min_score)) {
    geo_uri <- paste(geo_uri, "&minScore=", min_score, sep = "")
  }
  ret = stream_post(uri = geo_uri, body = to_geolocate) %>%
    dplyr::arrange(as.integer(.data$id))
  for (col in c(
    "geo_code",
    "geo_country_code",
    "geo_country",
    "geo_name",
    "tags"
  ))
    if (!exists(col, where = ret)) ret[[col]] <- NA
  dplyr::select(ret, -which(names(ret) %in% c("text", "id", "lang")))
}

# returns the current geotraining annotations as a data frame
get_geotraining_df <- function() {
  `%>%` <- magrittr::`%>%`
  data_types <- c(
    "text",
    "text",
    "text",
    "text",
    "text",
    "text",
    "text",
    "text",
    "text",
    "text",
    "text",
    "text",
    "text"
  )
  current <- readxl::read_excel(get_geotraining_path(), col_types = data_types)
  current$Text <- stringr::str_trim(current$Text)
  current <- current %>%
    dplyr::distinct(
      .data$`Text`,
      .data$Lang,
      .data$`Location in text`,
      .keep_all = T
    )
  current
}

# returns the current geotraining annotation augmented with posts_to_add posts to annotate
updated_geotraining_df <- function(
  posts_to_add = 100,
  progress = function(a, b) {
  }
) {
  `%>%` <- magrittr::`%>%`
  progress(0.1, "getting current training file")
  current <- get_geotraining_df()
  locations <- current %>%
    dplyr::filter(
      .data$Type == "Location" & .data$`Location yes/no` %in% c("yes", "OK")
    )
  add_locations <- if (nrow(locations) == 0) "&locationSamples=true" else
    "&locationSamples=false"
  non_locations <- current %>%
    dplyr::filter(
      .data$Type == "Location" & .data$`Location yes/no` %in% c("KO", "no")
    )
  non_location_langs <- unique(non_locations$Lang)
  excl_langs <- if (length(non_location_langs) > 0)
    paste(
      "&excludedLangs=",
      paste(non_location_langs, collapse = "&excludedLangs="),
      sep = ""
    ) else ""

  progress(0.3, "obtaining locations from downloaded models and geonames")
  geo_training_uri <- paste(
    get_scala_geotraining_url(),
    "?jsonnl=true",
    add_locations,
    excl_langs,
    sep = ""
  )
  geo_training_url <- url(geo_training_uri)
  geo_training <- jsonlite::stream_in(geo_training_url)
  if (nrow(geo_training) > 0) {
    geo_training <- geo_training %>%
      dplyr::transmute(
        Type = "Location",
        `Text` = .data$word,
        `Location in text` = as.character(NA),
        `Location yes/no` = ifelse(.data$isLocation, "yes", "no"),
        `Associate country code` =  as.character(NA),
        `Associate with` =  as.character(NA),
        `Source` = "Episomer model",
        `Post Id` =  as.character(NA),
        `Lang` = .data$lang,
        `Post part` =  as.character(NA),
        `Episomer match` =  as.character(NA),
        `Episomer country match` =  as.character(NA),
        `Episomer country code match` =  as.character(NA)
      )
    geo_training$Text <- stringr::str_trim(geo_training$Text)
    geo_training <- geo_training %>%
      dplyr::distinct(.data$`Text`, .data$Lang, .keep_all = T)
  }
  # Adding new posts
  untagged <- unique(length(
    (current %>%
      dplyr::filter(
        .data$`Post part` == 'text' & .data$`Location yes/no` == '?'
      ))$`Post Id`
  ))
  to_add <- if (posts_to_add > untagged) posts_to_add else 0

  progress(0.6, "adding new posts")
  to_tag <- search_posts(
    query = paste(
      "lang:",
      lapply(conf$languages, function(l) l$code),
      collapse = " ",
      sep = ""
    ),
    max = to_add
  )
  if (nrow(to_tag) > 0) {
    to_tag$text <- stringr::str_trim(to_tag$text)
    texts <- to_tag %>%
      dplyr::transmute(
        Type = "Text",
        `Text` = .data$text,
        `Location in text` = as.character(NA),
        `Location yes/no` = "?",
        `Associate country code` =  as.character(NA),
        `Associate with` =  as.character(NA),
        `Source` = "Post",
        `Post Id` = .data$id,
        `Lang` = .data$lang,
        `Post part` = "text",
        `Episomer match` =  as.character(NA),
        `Episomer country match` =  as.character(NA),
        `Episomer country code match` =  as.character(NA)
      ) %>%
      dplyr::filter(!.data$`Text` %in% current$`Text`) %>%
      dplyr::distinct(.data$`Text`, .data$Lang, .keep_all = T)

    ret <- jsonlite::rbind_pages(list(
      current,
      geo_training,
      texts
    ))
  } else {
    ret <- jsonlite::rbind_pages(list(
      current,
      geo_training
    ))
  }

  ret <- ret %>%
    dplyr::filter(!is.na(.data$Text) & !.data$Text == "") %>%
    dplyr::distinct(.data$`Text`, .data$Lang, .keep_all = TRUE)

  text_togeo <- ret %>%
    dplyr::transmute(
      Text = ifelse(
        !is.na(.data$`Associate with`),
        .data$`Associate with`,
        .data$Text
      ),
      Lang = ifelse(!is.na(.data$`Associate with`), "all", .data$Lang)
    )
  #getting current geolocation evaluation
  progress(0.7, "geolocating all items")
  tryCatch(
    {
      geoloc = geolocate_text(
        df = text_togeo,
        text_col = "Text",
        lang_col = "Lang"
      )
      ret$`Episomer match` <- geoloc$geo_name
      ret$`Episomer country match` <- geoloc$geo_country
      ret$`Episomer country code match` <- geoloc$geo_country_code
      ret$`Location in text` <- ifelse(
        (is.na(ret$`Location yes/no`) | ret$`Location yes/no` == "?") &
          ret$Type == "Text",
        geoloc$tags,
        ret$`Location in text`
      )
    },
    error = function(e) {
      message(
        "Models are not trained, getting geotraining dataset without evaluation"
      )
      ret$`Episomer match` <- NA
      ret$`Episomer country match` <- NA
      ret$`Episomer country code match` <- NA
    }
  )
  ret %>% arrange(dplyr::desc(.data$Type), .data$`Post part`)
}

# returns the current geotraining annotation augmented with posts_to_add posts to annotate and write the results to the geotraining spreadsheet
update_geotraining_df <- function(
  posts_to_add = 100,
  progress = function(a, b) {
  }
) {
  `%>%` <- magrittr::`%>%`
  training_df <- updated_geotraining_df(
    posts_to_add = 100,
    progress = progress
  )
  update_topic_keywords()
  update_forced_geo()
  update_forced_geo_codes()
  if (is.function(progress)) progress(0.9, "writing result file")
  wbk <- openxlsx::createWorkbook()
  openxlsx::addWorksheet(wbk, "geolocation")
  # writing data to the worksheet
  openxlsx::writeDataTable(
    wbk,
    sheet = "geolocation",
    training_df,
    colNames = TRUE,
    startRow = 1,
    startCol = "A"
  )
  # setting some minimal formatting
  openxlsx::setColWidths(wbk, "geolocation", cols = c(2), widths = c(70))
  openxlsx::setColWidths(
    wbk,
    "geolocation",
    cols = c(3, 11, 12, 13),
    widths = c(25)
  )
  openxlsx::setColWidths(
    wbk,
    "geolocation",
    cols = c(4, 5, 6, 10),
    widths = c(17)
  )
  openxlsx::setRowHeights(wbk, "geolocation", rows = 1, heights = 20)
  openxlsx::addStyle(
    wbk,
    sheet = "geolocation",
    style = openxlsx::createStyle(
      fontSize = 10,
      halign = "center",
      fgFill = "#ff860d",
      border = c("top", "bottom", "left", "right"),
      textDecoration = "bold",
      wrapText = T,
      fontColour = "#222222"
    ),
    rows = 1,
    cols = 1,
    gridExpand = FALSE
  )
  openxlsx::addStyle(
    wbk,
    sheet = "geolocation",
    style = openxlsx::createStyle(
      fontSize = 10,
      halign = "center",
      fgFill = "#ffde59",
      border = c("top", "bottom", "left", "right"),
      textDecoration = "bold",
      wrapText = T,
      fontColour = "#222222"
    ),
    rows = 1,
    cols = 2:6,
    gridExpand = FALSE
  )
  openxlsx::addStyle(
    wbk,
    sheet = "geolocation",
    style = openxlsx::createStyle(
      fontSize = 10,
      halign = "center",
      fgFill = "#f7d1d5",
      border = c("top", "bottom", "left", "right"),
      textDecoration = "bold",
      wrapText = T,
      fontColour = "#222222"
    ),
    rows = 1,
    cols = 7:10,
    gridExpand = FALSE
  )
  openxlsx::addStyle(
    wbk,
    sheet = "geolocation",
    style = openxlsx::createStyle(
      fontSize = 10,
      halign = "center",
      fgFill = "#dedce6",
      border = c("top", "bottom", "left", "right"),
      textDecoration = "bold",
      wrapText = T,
      fontColour = "#222222"
    ),
    rows = 1,
    cols = 11:13,
    gridExpand = FALSE
  )
  openxlsx::addStyle(
    wbk,
    sheet = "geolocation",
    style = openxlsx::createStyle(
      fontSize = 10,
      border = c("top", "bottom", "left", "right")
    ),
    rows = 1:nrow(training_df) + 1,
    cols = 1:13,
    gridExpand = TRUE
  )
  openxlsx::addStyle(
    wbk,
    sheet = "geolocation",
    style = openxlsx::createStyle(
      fontSize = 10,
      border = c("top", "bottom", "left", "right"),
      wrapText = T
    ),
    rows = 2:nrow(training_df) + 1,
    cols = 2,
    gridExpand = TRUE
  )
  openxlsx::freezePane(wbk, "geolocation", firstActiveRow = 2)
  openxlsx::saveWorkbook(wbk, file = get_user_geotraining_path(), overwrite = T)
}

# retrain the language classifiers for entity recognition with the current set of annotations
retrain_languages <- function() {
  `%>%` <- magrittr::`%>%`
  old <- options()
  on.exit(options(old))
  options(timeout = 1800)
  body <- get_geotraining_df() %>% jsonlite::toJSON()
  post_result <- httr::POST(
    url = get_scala_geotraining_url(),
    httr::content_type_json(),
    body = body,
    encode = "raw",
    encoding = "UTF-8"
  )
  if (httr::status_code(post_result) != 200) {
    stop(paste(
      "retrain web service failed with the following output: ",
      substring(httr::content(post_result, "text", encoding = "UTF-8"), 1, 100),
      sep = "\n"
    ))
  } else {
    fileConn <- file(get_geotraining_evaluation_path(), encoding = "UTF-8")
    writeLines(httr::content(post_result, "text", encoding = "UTF-8"), fileConn)
    close(fileConn)
  }
}

# update the topic keywords JSON file. When several location entities are found, the one closest to a topic is chosen
update_topic_keywords <- function() {
  `%>%` <- magrittr::`%>%`
  keywords <- lapply(
    strsplit(
      gsub(
        "\\-\\w*\\b|\"",
        "",
        unlist(lapply(conf$topics, function(t) t$query))
      ),
      "OR|AND|NOT|\\(|\\)"
    ),
    function(t) {
      x <- gsub("^ *| *$", "", t)
      x[x != ""]
    }
  )
  topicKeywords <- setNames(keywords, sapply(conf$topics, function(t) t$topic))
  ret <- list()
  for (i in 1:length(conf$topics))
    if (exists(conf$topics[[i]]$topic, where = ret))
      ret[[conf$topics[[i]]$topic]] <- unique(c(
        ret[[conf$topics[[i]]$topic]],
        topicKeywords[[i]]
      )) else ret[[conf$topics[[i]]$topic]] <- unique(topicKeywords[[i]])
  write_json_atomic(
    ret,
    get_topic_keywords_path(),
    pretty = TRUE,
    force = TRUE,
    auto_unbox = TRUE
  )
}

# update the forced geo JSON file, listing words that will be associated with particular location names, ignoring the geolocation algorithm
update_forced_geo <- function() {
  `%>%` <- magrittr::`%>%`
  df <- get_geotraining_df() %>%
    dplyr::transmute(
      from = ifelse(
        is.na(.data$`Location in text`),
        .data$`Text`,
        .data$`Location in text`
      ),
      to = .data$`Associate with`
    ) %>%
    dplyr::filter(!is.na(.data$to))
  ret <- list()
  if (nrow(df) > 0) {
    for (i in 1:nrow(df)) {
      toReplace <- sapply(strsplit(df$from[[i]], ",")[[1]], function(t) {
        x <- gsub("^ *| *$", "", t)
        x[x != ""]
      })
      for (j in 1:length(toReplace)) ret[[toReplace[[j]]]] <- df$to[[i]]
    }
    write_json_atomic(
      ret,
      get_forced_geo_path(),
      pretty = TRUE,
      force = TRUE,
      auto_unbox = TRUE
    )
  } else {
    if (file.exists(get_forced_geo_path())) file.remove(get_forced_geo_path())
  }
}

# update the forced geo codes JSON file lisiting words that will be associated to particular location codes ignoring the geolocation algorithm
update_forced_geo_codes <- function() {
  `%>%` <- magrittr::`%>%`
  df <- get_geotraining_df() %>%
    dplyr::transmute(
      from = ifelse(
        is.na(.data$`Location in text`),
        .data$`Text`,
        .data$`Location in text`
      ),
      to = .data$`Associate country code`
    ) %>%
    dplyr::filter(!is.na(.data$to))
  ret <- list()
  if (nrow(df) > 0) {
    for (i in 1:nrow(df)) {
      toReplace <- sapply(strsplit(df$from[[i]], ",")[[1]], function(t) {
        x <- gsub("^ *| *$", "", t)
        x[x != ""]
      })
      for (j in 1:length(toReplace)) ret[[toReplace[[j]]]] <- df$to[[i]]
    }
    write_json_atomic(
      ret,
      get_forced_geo_codes_path(),
      pretty = TRUE,
      force = TRUE,
      auto_unbox = TRUE
    )
  } else {
    if (file.exists(get_forced_geo_codes_path()))
      file.remove(get_forced_geo_codes_path())
  }
}

# get the Java path using JAVA_HOME if set
java_path <- function() {
  if (Sys.getenv("JAVA_HOME") == "") "java" else if (
    .Platform$OS.type == "windows"
  )
    paste("\"", Sys.getenv("JAVA_HOME"), "\\bin\\", "java\"", sep = "") else
    paste("\"", Sys.getenv("JAVA_HOME"), "/bin/", "java\"", sep = "")
}

# Setting the blas environment variables depending on chosen configuration and OS
set_blas_env <- function() {
  if (.Platform$OS.type == "windows") {
    if (
      conf$use_mkl &&
        file.exists(
          "C:/Program Files (x86)/IntelSWTools/compilers_and_libraries/windows/redist/intel64_win/mkl"
        )
    ) {
      mkl_path = "C:\\Program Files (x86)\\IntelSWTools\\compilers_and_libraries\\windows\\redist\\intel64_win\\mkl"
      Sys.setenv(PATH = paste(mkl_path, Sys.getenv("PATH"), sep = ";"))
    }
  }
}

# Getting Java options for using MKL BLAS if setup
get_blas_java_opt <- function() {
  if (.Platform$OS.type == "windows") {
    if (
      conf$use_mkl &&
        file.exists(
          "C:/Program Files (x86)/IntelSWTools/compilers_and_libraries/windows/redist/intel64_win/mkl"
        )
    ) {
      "-Dcom.github.fommil.netlib.NativeSystemBLAS.natives=mkl_rt.dll"
    } else ""
  } else ""
}

# getting a string for passing languages as parameter for system calls
conf_languages_as_arg <- function() {
  paste(
    "langCodes \"",
    paste(lapply(conf$languages, function(l) l$code), collapse = ","),
    "\" ",
    "langNames \"",
    paste(lapply(conf$languages, function(l) l$name), collapse = ","),
    "\" ",
    "langPaths \"",
    paste(lapply(conf$languages, function(l) l$vectors), collapse = ","),
    "\" ",
    sep = ""
  )
}

# getting a string for passing geonames as paramater for system calls
conf_geonames_as_arg <- function() {
  return(
    paste(
      "geonamesSource",
      paste(
        "\"",
        file.path(conf$data_dir, "geo", "allCountries.txt"),
        "\"",
        sep = ""
      ),
      "geonamesDestination",
      paste("\"", paste(conf$data_dir, "geo", sep = "/"), "\"", sep = ""),
      "geonamesSimplify",
      paste("\"", conf$geonames_simplify, "\"", sep = "")
    )
  )
}


# making a spark call via system call without returning any value
spark_job <- function(args) {
  # Setting BLAS environmebt variables
  set_blas_env()
  cmd <- paste(
    if (.Platform$OS.type != "windows") {
      #Default non windows environment variables within the command
      "export OPENBLAS_NUM_THREADS=1"
    } else {
      Sys.setenv(
        OPENBLAS_NUM_THREADS = 1,
        HADOOP_HOME = get_winutils_hadoop_home_path(),
        PATH = paste(file.path(get_winutils_hadoop_home_path(), "bin"), Sys.getenv("PATH"), sep = ";")
      ) #Default windows environment variables
      ""
    },
    paste(
      java_path(), #Java executable
      paste(
        "-cp \"",
        get_app_jar_path(), #episomer embedded jar path with
        if (.Platform$OS.type != "windows") ":" else ";",
        get_jars_dest_path(), #Java class path with all dependencies
        "/*\"",
        sep = ""
      ),
      paste("-Xmx", conf$spark_memory, sep = ""), #memory limits
      paste("--add-opens=java.base/java.nio=ALL-UNNAMED"), #allowing proper serialisation in spark of ByteArray
      paste("--add-opens=java.base/java.lang.invoke=ALL-UNNAMED"), #allowing proper serialisation in spark of ByteArray
      paste("--add-opens=java.base/java.util=ALL-UNNAMED"), #allowing proper serialisation in spark of ByteArray
      paste(get_blas_java_opt()), #Java blas specific options
      "org.ecdc.episomer.Episomer", #episomer scala main class
      args #arguments to the call
    ),
    sep = if (.Platform$OS.type == "windows") '\r\n' else '\n'
  )

  #message(cmd) # uncomment this to see the command sent
  res <- system(cmd)
  if (res != 0) stop(paste("Error encountered while executing: ", cmd))
}


# getting a spark data frame by piping a system call
spark_df <- function(args, handler = NULL) {
  # Setting BLAS environmebt variables
  set_blas_env()
  # Setting the memory to the half of allocated since R will use memory for loading the dataframe
  half_mem <- paste(
    ceiling(as.integer(gsub("[^0-9]*", "", conf$spark_memory)) / 2),
    gsub("[0-9]*", "", conf$spark_memory),
    sep = ""
  )
  cmd <- paste(
    if (.Platform$OS.type != "windows")
      #Default non windows environment variables
      "export OPENBLAS_NUM_THREADS=1" else {
      Sys.setenv(
        OPENBLAS_NUM_THREADS = 1,
        HADOOP_HOME = get_winutils_hadoop_home_path()
      ) # Setting windowd default environment variables
      ""
    },
    paste(
      paste(
        if (.Platform$OS.type == "windows") "call " else "",
        java_path(), # Java executable
        sep = ""
      ),
      paste(
        "-cp \"",
        get_app_jar_path(), #episomer embedded jar path with
        if (.Platform$OS.type != "windows") ":" else ";",
        get_jars_dest_path(), #Java class path with all dependencies
        "/*\"",
        sep = ""
      ),
      "-Dfile.encoding=UTF8", #Setting Java encoding to UTF-8
      paste("-Xmx", half_mem, sep = ""), #Setting Java memory
      paste(get_blas_java_opt()),
      "org.ecdc.episomer.Episomer", #episomer scala main class
      args
    ),
    sep = '\n'
  )
  #message(cmd) # uncomment this to see the command sent
  con <- if (.Platform$OS.type == "windows") {
    # on windows we send the results to a temporary files
    t <- tempfile()
    tcmd <- paste(cmd, shQuote(t), sep = " >")
    shell(tcmd)
    file(t, encoding = "UTF-8")
  } else {
    # on other OS we get the results directly from the command line
    pipe(cmd, encoding = "UTF-8")
  }
  # here we have a connection ro read returning json lines
  df <- if (is.null(handler)) {
    #If no custom transformation will be done on data we pass the json lines to jsonlite to interprete it as an R dataframe
    jsonlite::stream_in(con, pagesize = 10000, verbose = FALSE)
  } else {
    # If a custom transformation is going to be done (a handler function has been set)
    # this function will be called on pages of 10k lines using the jsonlite stream_in function
    # the transformed results will be stored on a temporary file that will be read by stream_in
    tmp_file <- tempfile(pattern = "episomer", fileext = ".json")
    #message(tmp_file)
    con_tmp <- file(tmp_file, open = "w", encoding = "UTF-8")
    jsonlite::stream_in(
      con,
      pagesize = 10000,
      verbose = FALSE,
      function(df) handler(df, con_tmp)
    )
    close(con_tmp)
    con_tmp <- file(tmp_file, open = "r", encoding = "UTF-8")
    ret <- jsonlite::stream_in(con_tmp, pagesize = 10000, verbose = FALSE)
    close(con_tmp)
    unlink(tmp_file)
    ret
  }
  if (.Platform$OS.type == "windows") unlink(t)
  df
}

#' @title Updates Java dependencies
#' @description Download Java dependencies of the application mainly related to Apache SPARK, Apache Lucene, and Apache Pekko
#' @param tasks Task object for reporting progress and error messages, default: get_tasks()
#' @return The list of tasks updated with produced messages
#' @details Run a one-shot task consisting of downloading Java and Scala dependencies, this is separated by the following subtasks
#' \itemize{
#'   \item{Download jar dependencies from configuration maven repo to project data folder. This includes, scala, spark, lucene. Packages to be downloaded are defined in package file 'sbt-deps.txt'}
#'   \item{Download winutils from configuration URL to project data folder. For more details on winutils please see
#'     \url{https://issues.apache.org/jira/browse/HADOOP-13223} and \url{https://issues.apache.org/jira/browse/HADOOP-16816}
#'   }
#' }
#'
#' The URLs to download the JAR dependencies (maven package manager) and Winutils are on the configuration tab of the Shiny app.
#'
#' Normally this function is not called directly by the user but from the \code{\link{detect_loop}} function.
#' @examples
#' \dontrun{
#'    library(episomer)
#'    # setting up the data folder
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'
#'    # geolocating last posts
#'    tasks <- download_dependencies()
#' }
#' @rdname download_dependencies
#' @seealso
#'  \code{\link{detect_loop}}
#'
#'  \code{\link{get_tasks}}
#'
#' @export
#' @importFrom utils download.file
download_dependencies <- function(tasks = get_tasks()) {
  tasks <- tryCatch(
    {
      tasks <- update_dep_task(
        tasks,
        "running",
        "getting sbt dependencies",
        start = TRUE
      )
      # downloading SPARK, lucene and other scala dependencies from the provided maven repository
      while (is_fs_running()) {
        tasks <- update_dep_task(
          tasks,
          "running",
          paste(
            "To start the dependencies task, please turn it OFF the episomer database. You can do this by clicking on the 'stop' button next to the database status or manually kill the process."
          )
        )
        Sys.sleep(5)
      }
      tasks <- download_sbt_dependencies(tasks)
      if (.Platform$OS.type == "windows") {
        tasks <- update_dep_task(tasks, "running", "getting winutils")
        # on windows, downloading winutils from the URL set from the shiny app which is a prerequisite of SPARK
        tasks <- download_winutils()
      }
      # ensuring that storage system is running
      while (!is_fs_running()) {
        tasks <- update_dep_task(
          tasks,
          "running",
          paste(
            "To continue you need to ACTIVATE the embedded database. You can activate it by clicking on the 'activate' database service button on the config page ",
            "You can also manually run the fs service by executing the following command on a separate R session. episomer::fs_loop('",
            conf$data_dir,
            "')"
          )
        )
        Sys.sleep(5)
      }
      # Setting status to succes
      tasks <- update_dep_task(tasks, "success", "", end = TRUE)
      tasks
    },
    error = function(error_condition) {
      # Setting status to failed
      message("Failed...")
      message(paste(
        "failed while",
        tasks$dependencies$message,
        " :",
        error_condition
      ))
      tasks <- update_dep_task(
        tasks,
        "failed",
        paste(
          "failed while",
          tasks$dependencies$message,
          " getting dependencies ",
          error_condition
        )
      )
      tasks
    },
    warning = function(error_condition) {
      # Setting status to failed
      message("Failed...")
      message(paste(
        "failed while",
        tasks$dependencies$message,
        " ",
        error_condition
      ))
      tasks <- update_dep_task(
        tasks,
        "failed",
        paste(
          "failed while",
          tasks$dependencies$message,
          " getting dependencies ",
          error_condition
        )
      )
      tasks
    }
  )
  return(tasks)
}

# Download SBT dependencies from maven (scala, spark, lucene, pekko, netlib (BLAS), httpclient) into package folder
download_sbt_dependencies <- function(tasks = get_tasks()) {
  if (!exists("maven_repo", where = tasks$dependencies))
    stop(
      "Before running detect loop you have to manually update 'Java/Scala dependencies' to set the used repository"
    )
  # reading the dependencies to download from embedded list
  con <- file(get_sbt_file_dep_path())
  deps <- readLines(con)
  close(con)

  # transforming the dependencies listed into maven URLs there are two possible formats that will be stored on urls and urls2
  parts <- strsplit(deps, "/")
  versions <- sapply(parts, function(p) {
    l <- nchar(p[[4]])
    substr(p[[5]], l + 2, nchar(p[[5]]))
  })
  versions <-  gsub("\\.([^\\.]*)$", "", versions)
  versions2 <- gsub("-(.*)$", "", versions)
  urls <- paste(
    tasks$dependencies$maven_repo,
    gsub("\\.", "/", sapply(parts, `[[`, 3)),
    sapply(parts, `[[`, 4),
    versions,
    sapply(parts, `[[`, 5),
    sep = "/"
  )
  urls2 <- paste(
    tasks$dependencies$maven_repo,
    gsub("\\.", "/", sapply(parts, `[[`, 3)),
    sapply(parts, `[[`, 4),
    versions2,
    sapply(parts, `[[`, 5),
    sep = "/"
  )

  #Creating destination directory
  if (!dir.exists(get_jars_dest_path())) dir.create(get_jars_dest_path()) else {
    #Deleting existing files
    existing <- list.files(get_jars_dest_path(), full.names = TRUE)
    tryCatch(
      {
        file.remove(existing)
      },
      warning = function(c) {
        warning(paste(
          "Warning during removing JAR files in ",
          get_jars_dest_path(),
          c
        ))
      },
      error = function(c) {
        stop(paste(
          "episomer cannot remove JAR fils files in ",
          get_jars_dest_path(),
          " please remove the files manually.",
          c
        ))
      }
    )
  }

  # destination file
  dest_file <- file.path(
    get_jars_dest_path(),
    paste(
      sapply(parts, `[[`, 3),
      sapply(parts, `[[`, 4),
      sapply(parts, `[[`, 5),
      sep = "_"
    )
  )

  # iterating over all files
  for (i in 1:length(urls)) {
    tryCatch(
      {
        tasks <- update_dep_task(
          tasks,
          "running",
          paste("downloading", urls[[i]])
        )
        # downloading files on first format
        download.file(url = urls[[i]], destfile = dest_file[[i]], mode = "wb")
      },
      warning = function(c) {
        tasks <- update_dep_task(
          tasks,
          "running",
          paste("downloading", urls2[[i]])
        )
        # downloading files on secong format
        download.file(url = urls2[[i]], destfile = dest_file[[i]], mode = "wb")
      },
      error = function(c) {
        tasks <- update_dep_task(
          tasks,
          "running",
          paste("downloading", urls2[[i]])
        )
        download.file(url = urls2[[i]], destfile = dest_file[[i]], mode = "wb")
      }
    )
  }
  return(tasks)
}

# Download winutils (necessary for spark on Windows,
# for more details on winutils please see https://issues.apache.org/jira/browse/HADOOP-13223 and https://issues.apache.org/jira/browse/HADOOP-16816
download_winutils <- function(tasks = get_tasks()) {
  if (!dir.exists(get_winutils_hadoop_home_path()))
    dir.create(get_winutils_hadoop_home_path())
  if (!dir.exists(file.path(get_winutils_hadoop_home_path(), "bin")))
    dir.create(file.path(get_winutils_hadoop_home_path(), "bin"))
  if (!exists("winutils_url", where = tasks$dependencies))
    stop(
      "Before running detect loop you have to manually activate 'Java/Scala dependencies' to set the winutils to use"
    )
  tasks <- update_dep_task(
    tasks,
    "running",
    paste("downloading", tasks$dependencies$winutils_url)
  )
  # downloading winutils
  url = tasks$dependencies$winutils_url
  dest = get_winutils_path()
  download.file(url = url, destfile = dest, mode = "wb" )
  url <- sub("winutils.exe", "hadoop.dll", url)
  dest <- sub("winutils.exe", "hadoop.dll", dest) 
  # downloading hadoop dll
  download.file(url = url, destfile = dest, mode = "wb" )
  return(tasks)
}

# @function_def_start (do not delete)

#' parse custom attributes for a plan in this specific social media and returns them as a named list with attributes in the proper type
#' @param ... list, triple dot arguments to be passed as a parameter tu underlying functions 
sm_plan_parse_attributes_bluesky <- function(...) {
# @function_def_end (do not delete)
  p <- list(...)
  ret <- list(
    "plan_max_date" = if (!is.null(unlist(p$plan_max_date)))
      strptime(unlist(p$plan_max_date), "%Y-%m-%d %H:%M:%OSZ", tz = "UTC") else
      NULL,
    "plan_min_date" = if (!is.null(unlist(p$plan_min_date)))
      strptime(unlist(p$plan_min_date), "%Y-%m-%d %H:%M:%OSZ", tz = "UTC") else
      NULL,
    "current_min_date" = if (!is.null(unlist(p$current_min_date)))
      strptime(
        unlist(p$current_min_date),
        "%Y-%m-%d %H:%M:%OSZ",
        tz = "UTC"
      ) else NULL
  )
  ret
}

# @function_def_start (do not delete)


#' prepare a plan for serialisation by transforming attributes with non standard serialisation types e.g. dates to string on the format expected by the parse functions
#' @param plan list, the plan with the attributes go format
sm_plan_format_bluesky <- function(plan) {
# @function_def_end (do not delete)
  if (!is.null(plan$plan_max_date))
    plan$plan_max_date = strftime(
      plan$plan_max_date,
      "%Y-%m-%d %H:%M:%OS6Z",
      tz = "UTC"
    )
  if (!is.null(plan$plan_min_date))
    plan$plan_min_date = strftime(
      plan$plan_min_date,
      "%Y-%m-%d %H:%M:%OS6Z",
      tz = "UTC"
    )
  if (!is.null(plan$current_min_date))
    plan$current_min_date = strftime(
      plan$current_min_date,
      "%Y-%m-%d %H:%M:%OS6Z",
      tz = "UTC"
    )
  plan
}

# @function_def_start (do not delete)


#' returns a list with the custom social media plan attributes with the default values for a first plan in a topic.
sm_plan_first_attributes_bluesky <- function() {
# @function_def_end (do not delete)
  list(
    plan_max_date = strftime(Sys.time(), "%Y-%m-%d %H:%M:%OS6Z", tz = "UTC"),
    plan_min_date = strftime(Sys.Date() - conf$collect_history_days, "%Y-%m-%d %H:%M:%OS6Z", tz = "UTC"),
    current_min_date = NULL
  )
}

# @function_def_start (do not delete)

#' returns a list with the custom social media plan attributes with the values for the plan to be run after the plans provided as parameter
#' @param plans list, the list with current plans of the given topic to calculate next plan attributes
sm_plan_next_attributes_bluesky <- function(plans) {
# @function_def_end (do not delete)
  list(
    plan_max_date = strftime(Sys.time(), "%Y-%m-%d %H:%M:%OS6Z", tz = "UTC"),
    plan_min_date = strftime(
      plans[[1]]$plan_max_date,
      "%Y-%m-%d %H:%M:%OS6Z",
      tz = "UTC"
    ),
    current_min_date = NULL
  )
}

# @function_def_start (do not delete)

#' calculates and returns the estimated progress of a plan using custom attributes
#' @param plan list, the plan to calculate the progress from
sm_plan_get_progress_bluesky <- function(plan) {
# @function_def_end (do not delete)
  if (
    !is.null(plan$plan_max_date) &&
      !is.null(plan$plan_min_date) &&
      !is.null(plan$current_min_date)
  ) {
    as.numeric(plan$plan_max_date - plan$current_min_date, unit = "secs") /
      as.numeric(plan$plan_max_date - plan$plan_min_date, unit = "secs")
  } else {
    NULL
  }
}

# @function_def_start (do not delete)

#' Return a message to log what the next query is going to be for the current plan
#' @param plan list, the plan to print the information about
sm_plan_search_info_bluesky <- function(plan) {
# @function_def_end (do not delete)
  info = ""
  if (is.null(plan$current_min_date)) {
    info = paste(
      info,
      "before (first)",
      strftime(plan$plan_max_date, "%Y-%m-%d %H:%M:%OS6Z", tz = "UTC")
    )
  } else {
    info = paste(
      info,
      "before",
      strftime(plan$current_min_date, "%Y-%m-%d %H:%M:%OS6Z", tz = "UTC")
    )
  }
  if (!is.null(plan$plan_min_date)) {
    info = paste(
      info,
      "until",
      strftime(plan$plan_min_date, "%Y-%m-%d %H:%M:%OS6Z", tz = "UTC")
    )
  }
  return(info)
}

# check if Java is installed
check_java_present <- function() {
  if (is_java_present()) TRUE else {
    warning(paste(
      "episomer could not detect Java. Please install it and make sure the Java binary is on the path or that JAVA_HOME is set",
      "episomer requires 64-bit Java between ",
      java_min_version(),
      " and ",
      java_max_version()
    ))
  }
}

# check if Java is installed
is_java_present <- function() {
  programs <- "java"
  # If JAVA_HOME is not set, checks if the system can found Java on the path
  if (Sys.getenv("JAVA_HOME") == "" || is.null(Sys.getenv("JAVA_HOME"))) {
    length(grep(programs, Sys.which(programs))) == length(programs)
  } else {
    #checking if the Java binary can be found from JAVA_HOME
    if (.Platform$OS.type == "windows")
      file.exists(file.path(Sys.getenv("JAVA_HOME"), "bin", "java.exe")) else
      file.exists(file.path(Sys.getenv("JAVA_HOME"), "bin", "java"))
  }
}

# get Java version vector
get_java_version_vector <- function() {
  if (.Platform$OS.type == "windows") {
    tmp <- tempfile()
    shell(paste(java_path(), "-version >", tmp, "2>&1"))
    con <- file(tmp)
    jver <- readLines(con)
    close(con)
    file.remove(tmp)
    jver
  } else {
    con <- pipe(paste(java_path(), "-version 2>&1"))
    jver <- readLines(con)
    close(con)
    jver
  }
}

# check if Java is 64 bits
check_java_64 <- function() {
  if (any(grepl("64-Bit", get_java_version_vector()))) TRUE else {
    warning(paste(
      "Your current Java version does not appear to be 64-bit. episomer requires 64-bit Java between ",
      java_min_version(),
      " and ",
      java_max_version()
    ))
    FALSE
  }
}

# get Java version
get_java_version <- function() {
  jver <- get_java_version_vector()
  index <- grep("version", jver)
  if (length(index) == 0) NULL else {
    j_version_parts <- strsplit(
      gsub("[^0-9]", ".", jver[[index[[1]]]]),
      "\\.+"
    )[[1]]
    j_version_parts <- j_version_parts[grepl("[0-9]", j_version_parts)]
    if (length(j_version_parts) > 1 && as.integer(j_version_parts[[1]]) == 1)
      as.integer(j_version_parts[[2]]) else as.integer(j_version_parts[[1]])
  }
}

java_min_version <- function() 17
java_max_version <- function() 21
java_tested_version <- function() 21
# check Java version
check_java_version <- function() {
  java_version <- get_java_version()
  if (is.null(java_version)) {
    warning("episomer cannot identify your Java version")
    FALSE
  } else if (
    java_version >= java_min_version() && java_version <= java_max_version()
  ) {
    TRUE
  } else if (java_version <= java_tested_version()) {
    warning(paste(
      "Your current Java version is",
      java_version,
      ". episomer needs Java versions between ",
      java_min_version(),
      " and ",
      java_max_version(),
      ". Some users have reported that your version also works, but it is recommended to use a previously compatible Java version"
    ))
    FALSE
  } else {
    warning(paste(
      "Your current Java version is",
      java_version,
      ". episomer needs Java versions between ",
      java_min_version(),
      " and ",
      java_max_version()
    ))
    FALSE
    
  }
}


is_64bit <- function() {
  if (.Platform$OS.type == "windows") {
    Sys.getenv("PROCESSOR_ARCHITECTURE") == "AMD64"
  } else {
    grepl("64", system("uname -a", intern = TRUE))
  }
}

# check 64 architecture
check_64 <- function() {
  if (is_64bit()) {
    TRUE
  } else {
    warning("Your R seems to be 32-bit, episomer needs 64-bit R")
    FALSE
  }
}


# check msvcr100 on Windows machines
check_winmsvcr100 <- function() {
  # inspired from sparklyr https://github.com/sparklyr/sparklyr/blob/master/R/install_spark_windows.R
  if (.Platform$OS.type == "windows") {
    systemRoot <- Sys.getenv("SystemRoot")
    systemDir <- "system32"
    msvcr100Path <- normalizePath(
      file.path(systemRoot, systemDir, "msvcr100.dll"),
      winslash = "/",
      mustWork = FALSE
    )
    if (file.exists(msvcr100Path)) TRUE else {
      warning(paste(
        "Running Spark on Windows requires the",
        "Microsoft Visual C++ 2010 SP1 Redistributable Package.",
        "Please download and install from: \nhttps://www.microsoft.com/download/en/details.aspx?id=13523",
        "Then restart R after the installation is completed"
      ))
      FALSE
    }
  } else {
    TRUE
  }
}

# check winutils is installed
check_winutils <- function() {
  if (.Platform$OS.type == "windows") {
    if (file.exists(get_winutils_path())) TRUE else {
      warning(
        "To run SPARK on Windows, you need winutils which is a HADOOP binary. You can download it by running the update dependencies task"
      )
      FALSE
    }
  } else {
    TRUE
  }
}

# check download Java dependencies has been executed
check_java_deps <- function() {
  if (!dir.exists(get_jars_dest_path())) {
    warning(
      "episomer requires the dependencies task to run and download all Java dependencies"
    )
    FALSE
  } else {
    con <- file(get_sbt_file_dep_path())
    deps <- readLines(con)
    close(con)
    downloaded <- list.files(get_jars_dest_path())
    if (length(deps) != length(downloaded)) {
      warning(
        "episomer requires the dependencies task to run and download all Java dependencies"
      )
      FALSE
    } else {
      TRUE
    }
  }
}

# check geonames is downloaded and indexed
check_geonames <- function() {
  if (!file.exists(get_geonames_txt_path())) {
    warning(
      "episomer needs GeoNames to be downloaded and unzipped for geolocating posts. Please run the GeoNames task in the requirements and alert pipeline"
    )
    FALSE
  } else {
    dir <- get_geonames_index_path()
    ret <- (if (!file.exists(dir)) FALSE else {
      dirs <- list.dirs(dir, recursive = FALSE)
      ind_dir <- dirs[grepl(".*/lucIndex[^/]*", dirs)]
      if (length(ind_dir) == 0) FALSE else
        length(list.files(ind_dir, "_.*\\.cfs")) > 0
    })
    if (ret) TRUE else {
      warning("GeoNames has not been indexed. Please execute the geonames task")
      FALSE
    }
  }
}


# check that languages are downloaded and indexed
check_languages <- function() {
  lang_files <- sapply(conf$languages, function(l) l$vectors)
  if (
    length(lang_files) == 0 ||
      !all(sapply(lang_files, function(f) file.exists(f)))
  ) {
    warning(
      "episomer requires languages models to be downloaded for geolocating posts. Please activate at least one language and launch the languages task"
    )
    FALSE
  } else {
    dir <- get_lang_index_path()
    ret <- (if (!file.exists(dir)) FALSE else {
      dirs <- list.dirs(dir, recursive = FALSE)
      ind_dir <- dirs[grepl(".*/lucIndex[^/]*", dirs)]
      if (length(ind_dir) == 0) FALSE else
        length(list.files(ind_dir, "_.*\\.cfs")) > 0
    })
    if (ret) TRUE else {
      warning(
        "The language vector index has not been built. Please activate at least one language and launch the languages task"
      )
      FALSE
    }
  }
}

# check if geolocated files are present (geotag has successfully run)
check_geolocated_present <- function() {
  last_month <- tail(
    list.dirs(
      path = file.path(conf$data_dir, "posts", "geolocated"),
      recursive = FALSE
    ),
    n = 1
  )
  last_file <- tail(list.files(last_month, pattern = "*.json.gz"), n = 1)
  if (length(last_file) > 0) {
    TRUE
  } else {
    warning("No geolocated files found. Please execute the geotag task")
    FALSE
  }
}

# check if aggregated files are present (aggregate has successfully run)
check_series_present <- function(
  series = c("country_counts", "geolocated", "topwords", "tags", "urls")
) {
  rds_counts <- sapply(
    series,
    function(ds)
      length(list.files(
        path = file.path(conf$data_dir, "series"),
        recursive = TRUE,
        pattern = paste(ds, ".*\\.Rds", sep = "")
      ))
  )
  fs_counts <- sapply(
    series,
    function(ds)
      length(list.files(
        path = file.path(conf$data_dir, "fs", ds),
        recursive = TRUE,
        pattern = ".*\\.fdt"
      ))
  )
  if (all(rds_counts > 0) || all(fs_counts > 0)) TRUE else {
    warning(
      "No aggregated series found. Please run the search loop to collect and aggregate data"
    )
    FALSE
  }
}

# check if post files are present & post collection has run
check_posts_present <- function() {
  if (!is.na(last_fs_updates("posts"))) {
    TRUE
  } else {
    warning("No posts found on database. Please execute the Data collection")
    FALSE
  }
}

# check if alert files are present
check_alerts_present <- function() {
  dirs <- list.dirs(
    path = file.path(conf$data_dir, "alerts"),
    recursive = FALSE
  )
  dirs <- dirs[!grepl(".*NA$", dirs)]
  last_year <- tail(dirs, n = 1)
  last_file <- tail(list.files(path = last_year, pattern = "*.json"), n = 1)

  if (length(last_file) > 0) {
    TRUE
  } else {
    warning("No alert files found. Please execute the alerts task")
    FALSE
  }
}

# check if taskScheduleR is installed on Windows
check_scheduler <- function() {
  if (.Platform$OS.type == "windows") {
    installed <- tryCatch(
      {
        func <- taskscheduleR::taskscheduler_create
        TRUE
      },
      warning = function(w) FALSE,
      error = function(e) FALSE
    )
    if (installed) {
      TRUE
    } else {
      warning(
        "To activate tasks, you need to install the taskscheduleR package manually. Please run install.packages('taskscheduleR') on an R session and restart episomer"
      )
      FALSE
    }
  } else {
    TRUE
  }
}

# check if Pandoc is installed
check_pandoc <- function() {
  if (unname(Sys.which("pandoc")) == "") {
    warning(
      "To export the PDF report from the dashboard, Pandoc must be installed."
    )
    FALSE
  } else TRUE
}

# check if Pandoc is installed
check_tex <- function() {
  if (unname(Sys.which("pdflatex")) == "") {
    warning(paste(
      "For exporting the PDF report from the dashboard, a tex distribution has to be installed.",
      "We suggest installing tinytex with the following commands on an R session:",
      "install.packages('tinytex'); tinytex::install_tinytex()"
    ))
    FALSE
  } else TRUE
}

# check search is running
check_search_running <- function() {
  if (is_search_running()) TRUE else {
    warning(paste0(
      "Data collection & processing pipeline is not running. On Windows, you can activate it by clicking on the 'activate' Data collection & processing button on the configuration page ",
      "You can also manually run the Data collection & processing pipeline by executing the following command on a separate R session. episomer::search_loop('",
      conf$data_dir,
      "')"
    ))
    FALSE
  }
}

# check fs is running
check_fs_running <- function() {
  if (is_fs_running()) TRUE else {
    warning(paste0(
      "Embedded episomer database is not running. On Windows, you can activate it by clicking on the 'activate' episomer database button on the configuration page ",
      "You can also manually run the Data collection & processing pipeline by executing the following command on a separate R session. episomer::search_loop('",
      conf$data_dir,
      "')"
    ))
    FALSE
  }
}

# check detect is running
check_detect_running <- function() {
  if (is_detect_running()) TRUE else {
    warning(paste0(
      "Requirements & alerts pipeline is not running. On Windows, you can activate it by clicking on the 'activate' Requirements & alerts button on the configuration page ",
      "You can also manually run the Requirements & alerts pipeline by executing the following command on a separate R session. episomer::detect_loop('",
      conf$data_dir,
      "')"
    ))
    FALSE
  }
}

# check fs is running
check_fs_running <- function() {
  if (is_fs_running()) TRUE else {
    warning(paste0(
      "Embedded database is not running. On Windows, you can activate it by clicking on the 'activate' database service button on the configuration page ",
      "You can also manually run the fs service by executing the following command on a separate R session. episomer::fs_loop('",
      conf$data_dir,
      "')"
    ))
    FALSE
  }
}

# check sm authentication
last_check_sm_auth <- new.env()

check_sm_auth <- function() {
    lastsm = "No active SM"
    ok <- tryCatch(
    {
      ok <- FALSE
      for(sm in active_social_media()) {
	lastsm = sm
        token <- sm_api_get_token(sm)
        ok <- is.character(token) && length(token)  > 0
        if(!ok)
	  break
      }
      last_check_sm_auth$value <- ok
      ok
    },
    warning = function(m) {
      FALSE
    },
    error = function(e) {
      FALSE
    }
  )
  if (ok) TRUE else {
    warning(
      sprintf("Cannot create a token for %s, please choose an authentication method on the configuration page", lastsm)
    )
  }
}
check_last_sm_auth <- function() {
  if (exists("value", last_check_sm_auth) && last_check_sm_auth$value)
    TRUE else {
    warning(
      "Cannot create a social media token, please choose an authentication method on the configuration page"
    )
    FALSE
  }
}

#check manual tasks requested
check_manual_task_request <- function() {
  if (
    is.na(conf$dep_updated_on) ||
      is.na(conf$geonames_updated_on) ||
      is.na(conf$lang_updated_on)
  ) {
    warning(
      "Before running the Requirements & alerts pipeline, you have to click on 'activate' Requirements & alerts pipeline button on configuration page"
    )
    FALSE
  } else TRUE
}

# can move from temp to data directory
check_move_from_temp <- function() {
  ok <- tryCatch(
    {
      temp <- tempfile()
      fileConn <- file(temp)
      writeLines(c("hello temp"), fileConn)
      close(fileConn)
      file.rename(temp, file.path(conf$data_dir, "deleteme.temp"))
      file.remove(file.path(conf$data_dir, "deleteme.temp"))
      TRUE
    },
    warning = function(m) {
      FALSE
    },
    error = function(e) {
      FALSE
    }
  )
  if (ok) TRUE else {
    warning(paste(
      "Cannot move files from temporary folder to data_dir, possible cross link device.",
      "Please set TMPDIR, TMP or TEMP environment variable to the same drive as the data folder",
      conf$data_dir
    ))
    FALSE
  }
}


# can tar gz
check_tar_gz <- function() {
  ok <- tryCatch(
    {
      temp <- tempdir()
      p <- file.path(temp, "test")
      pgz <- file.path(temp, "test.tar.gz")
      fileConn <- file(p)
      writeLines(c("hello temp"), fileConn)
      close(fileConn)

      result = tryCatch(
        {
          processx::run(
            "tar",
            c(
              "-czf",
              normalizePath(pgz, winslash = "/", mustWork = FALSE),
              normalizePath(p, winslash = "/"),
              "--force-local"
            ),
            wd = temp
          )
        },
        error = function(e) {
          processx::run(
            "tar",
            c(
              "-czf",
              normalizePath(pgz, winslash = "/", mustWork = FALSE),
              normalizePath(p, winslash = "/")
            ),
            wd = temp
          )
        }
      )

      TRUE
    },
    warning = function(m) {
      FALSE
    },
    error = function(e) {
      FALSE
    }
  )
  if (ok) TRUE else {
    warning(paste(
      "Cannot build tar.gz. Please ensure that Tar and Gzip are available in the system path and make sure you can execute 'tar -czf test.tar.gz somefile' on a terminal.",
      "The latest versions of Windows 10 may include these tools; otherwise, you would need to install them to create snapshots."
    ))
    FALSE
  }
}


#' @title Run automatic sanity checks
#' @description It runs a set of automated sanity checks for helping the user to troubleshot issues
#' @return Data frame containing the statuses of all realized checks
#' @details This function executes a series of sanity checks, concerning, Java, bitness, task status, dependencies and social media authentication.
#' @examples
#' \dontrun{
#'    #importing episomer
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    #running all tests
#'    check_all()
#' }
#' @rdname check_all
#' @export
check_all <- function() {
  check_sm_auth()
  checks <- list(
    scheduler = check_scheduler,
    sm_auth = check_last_sm_auth,
    search_running = check_search_running,
    database_running = check_fs_running,
    posts = check_posts_present,
    os64 = check_64,
    java = check_java_present,
    java64 = check_java_64,
    java_version = check_java_version,
    winmsvc = check_winmsvcr100,
    detect_activation = check_manual_task_request,
    detection_running = check_detect_running,
    winutils = check_winutils,
    java_deps = check_java_deps,
    move_from_temp = check_move_from_temp,
    tar_gz = check_tar_gz,
    geonames = check_geonames,
    languages = check_languages,
    aggregate = check_series_present,
    alerts = check_alerts_present,
    pandoc = check_pandoc,
    tex = check_tex
  )
  results <- sapply(
    checks,
    function(check)
      tryCatch(
        if (check()) "ok" else "ko",
        warning = function(m) {
          as.character(m$message)
        },
        error = function(e) {
          as.character(e)
        }
      )
  )
  data.frame(
    check = names(checks),
    passed = results == "ok",
    message = ifelse(results == "ok", "", results)
  )
}

# Environment for storing last admin email
checks <- new.env()

#' @title Send email to administrator if a failure of episomer is detected
#' @description It validates if episomer is not collecting posts, aggregating posts or not calculating alerts
#' @param send_mail Boolean. Whether an email should be sent to the defined administrator, default: TRUE
#' @param one_per_day Boolean. Whether a limit of one email per day will be applied, default: TRUE
#' @return A list of health check errors found
#' @details This function sends an email to the defined administrator if episomer is not collecting posts, aggregating posts or not calculating alerts
#' @examples
#' \dontrun{
#'    #importing episomer
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    #sending the email to the administrator if episomer components are not properly working
#'    health_check()
#' }
#' @rdname health_check
#' @export
health_check <- function(send_mail = TRUE, one_per_day = TRUE) {
  #health check is only done if one_per_day limit is disabled or if no email was already sent since yesterday and it is after 8AM.
  start_of_day <- strptime(strftime(Sys.time(), "%Y-%m-%d"), "%Y-%m-%d")
  alerts <- list()
  if (
    !one_per_day ||
      !exists("last_check", checks) ||
      (checks$last_check < start_of_day) &&
        as.numeric(strftime(Sys.time(), "%H")) >= 8
  ) {
    # check 01: if last post collected date is more than 1 hour
    last_collected <- file.mtime(get_search_touch_path())
    if (
      is.na(last_collected) ||
        as.numeric(Sys.time() - last_collected, unit = "hours") >= 1
    ) {
      alerts <- append(
        alerts,
        paste(
          "Posts have not been collected for more than ",
          as.integer(as.numeric(Sys.time() - last_collected, unit = "hours")),
          "hours"
        )
      )
    }

    # Check 02: if geolocated posts has not been done during more than one hour
    collection_times <- last_fs_updates("posts")
    last_geo <- collection_times$posts
    if (
      is.na(last_geo) || as.numeric(Sys.time() - last_geo, unit = "hours") >= 1
    ) {
      alerts <- append(
        alerts,
        paste(
          "Posts have not been geolocated for more than ",
          as.integer(as.numeric(Sys.time() - last_geo, unit = "hours")),
          "hours"
        )
      )
    }
    # Check 03: if any time series has not been processed in more than one hour
    collection_times <- last_fs_updates(c(
      "topwords",
      "country_counts",
      "geolocated"
    ))
    for (i in 1:length(collection_times)) {
      serie <- names(collection_times)[[i]]
      last_collect <- collection_times[[serie]]
      if (
        is.na(last_collect) ||
          as.numeric(Sys.time() - last_collect, unit = "hours") >= 1
      ) {
        alerts <- append(
          alerts,
          paste(
            "Serie,",
            serie,
            "has not been produced for more than ",
            as.integer(as.numeric(Sys.time() - last_collect, unit = "hours")),
            "hours"
          )
        )
      }
    }

    # Check 04: if alert detection has not finished in more than 2 hours + schedule span
    tasks <- get_tasks()
    last_detect_end <- tasks$alert$end_on
    if (
      is.na(last_detect_end) ||
        as.numeric(Sys.time() - last_detect_end, unit = "hours") >=
          (2 + conf$schedule_span / 60)
    ) {
      alerts <- append(
        alerts,
        paste(
          "The alerts tasks has not been finished since",
          as.integer(as.numeric(Sys.time() - last_detect_end, unit = "hours")),
          "hours"
        )
      )
    }

    # Check 05: if any task is on aborted status
    for (i in 1:length(tasks)) {
      if (!is.na(tasks[[i]]$status) && tasks[[i]]$status == "aborted") {
        alerts <- append(
          alerts,
          paste("Task,", tasks[[i]]$task, "is on aborted status")
        )
      }
    }
    if (
      send_mail &&
        length(alerts) > 1 &&
        !is.null(conf$admin_email) &&
        conf$admin_email != "" &&
        conf$smtp_host != ""
    ) {
      # creating the message to send
      msg <- (emayili::envelope() %>%
        emayili::from(conf$smtp_from) %>%
        emayili::to(conf$admin_email) %>%
        emayili::subject("episomer may not be properly running") %>%
        emayili::html(
          paste(
            "<p>The following errors were found during episomer health check, please check your episomer installation</p><p><ul><li>",
            paste(alerts, collapse = "</li><li>"),
            "</li></ul></p>"
          )
        ))
      smtp <- (if (
        is.na(conf$smtp_password) ||
          is.null(conf$smtp_password) ||
          conf$smtp_password == ""
      )
        emayili::server(
          host = conf$smtp_host,
          port = conf$smtp_port,
          insecure = conf$smtp_insecure,
          reuse = FALSE
        ) else
        emayili::server(
          host = conf$smtp_host,
          port = conf$smtp_port,
          username = conf$smtp_login,
          insecure = conf$smtp_insecure,
          password = conf$smtp_password,
          reuse = FALSE
        ))
      message(sprintf("Health check failed, sending email from %s to %s\n%s", conf$smtp_from, conf$admin_email, paste(alerts, collapse="\n")))
      tryCatch(
        smtp(msg),
        warning = function(e)
          message(paste(
            "Warning: Cannot send email. The following error occured",
            e
          )),
        error = function(e)
          message(paste(
            "Warning: Cannot send email. The following error occured",
            e
          ))
      )

      checks$last_check <- Sys.time()
    }
  }
  alerts
}

update_session_info <- function() {
  con <- file(get_session_info_path())
  writeLines(utils::capture.output(utils::sessionInfo()), con)
  close(con)
}


#' @title Snapshot of your episomer installation
#' @description Creates a snapshot file of your episomer installation folder. This can include all or a subset of the data files.
#' @param destination_dir, character(1) vector with the path of the destination folder to produce the snapshot
#' @param types, character vector indicating the types of data to include on a snapshot. Some of: "settings", "dependencies", "machine-learning", "aggregations", "posts", "logs"
#' @param posts_period, date(2) start and end dates to filter posts to include on snapshot (if selected)
#' @param aggregated_period, date(2) start and end dates to filter time series to include on snapshot (if selected)
#' @param compress, logical(1) whether to compress or not the output file
#' @param progress, function to report progress during execution.
#' @return Nothing
#' @details
#' This function can be used to create a a portable file to move your episomer installation in a single file, to backup your data, to archive your old data or to send information to technical team in order to reproduce an observed issue.
#' Different kinds of data can be included on the snapshot depending on the type of parameter. Possible values are:
#' - 'settings': Including all setting files of your installation (excluding passwords)
#' - 'dependencies': All jars and winutils.exe on windows installations
#' - 'machine-learning': All trained models and vectors and training data (this can include post text which is personal data)
#' - 'aggregations': Episomer aggregated time series
#' - 'posts': Posts collected by episomer
#' - 'logs': Log files produced automatically on windows task scheduler tasks.
#' @examples
#' \dontrun{
#'    #importing episomer
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    message('Please choose a destination directory')
#'    dest <- file.choose()
#'    #creating a compressed snapshot for settings and logs
#'    create_snapshot(dest, c("settings","dependencies"), compress = TRUE)
#' }
#' @rdname create_snapshot
#' @export
create_snapshot <- function(
  destination_dir,
  types = c(
    "settings",
    "dependencies",
    "machine-learning",
    "aggregations",
    "posts",
    "logs"
  ),
  posts_period = get_aggregated_period(),
  aggregated_period = get_aggregated_period(),
  compress = TRUE,
  progress = function(v, m) message(paste(round(v * 100, 2), m))
) {
  destination <- normalizePath(
    file.path(
      destination_dir,
      paste0(
        "episomer-snapshot",
        strftime(Sys.time(), "%Y.%m.%d-%H.%M.%S"),
        if (compress) ".tar.gz" else ".tar"
      )
    ),
    winslash = "/",
    mustWork = FALSE
  )
  update_session_info()
  items = list()
  if ("settings" %in% types)
    items <- c(
      items,
      list(
        "properties.json" = get_properties_path(),
        "topics.xlsx" = get_user_topics_path(),
        "tasks.json" = get_tasks_path(),
        "subscribers.xlsx" = get_subscribers_path(),
        "countries.xlsx" = get_countries_path(),
        "users.xlsx" = get_known_users_path(),
        "languages.xlsx" = get_available_languages_path(),
        "topics.json" = get_plans_path(),
        "topic-keywords.json" = get_topic_keywords_path(),
        "session-info.log" = get_session_info_path()
      ),
      episomer_files("collections")
    )
  if ("dependencies" %in% types)
    items <- c(
      items,
      list(
        "hadoop/bin/winutils.exe" = get_winutils_path()
      ),
      as.list(setNames(
        list.files(get_jars_dest_path(), full.names = TRUE),
        file.path("jars", list.files(get_jars_dest_path()))
      ))
    )
  if ("machine-learning" %in% types)
    items <- c(
      items,
      list(
        "alert-training.xlsx" = get_alert_training_path(),
        "geo-training.xlsx" = get_geotraining_path(),
        "geo-training-evaluation.json" = get_geotraining_evaluation_path(),
        "geo/forced-geo.json" = get_forced_geo_path(),
        "geo/forced-geo-codes.json" = get_forced_geo_codes_path(),
        "geo/allCountries.txt" = get_geonames_txt_path()
      ),
      episomer_files(get_cities_parquet_path(relative = TRUE)),
      episomer_files(get_geonames_parquet_path(relative = TRUE)),
      episomer_files(get_geonames_index_path(relative = TRUE)),
      episomer_files(get_lang_index_path(relative = TRUE)),
      episomer_files(get_lang_index_path(relative = TRUE)),
      episomer_files("languages"),
      episomer_files("alert-ml")
    )
  if ("aggregations" %in% types)
    items <- c(
      items,
      episomer_files("alerts", aggregated_period[[1]], aggregated_period[[2]]),
      episomer_files(
        "fs/contexts",
        aggregated_period[[1]],
        aggregated_period[[2]]
      ),
      episomer_files(
        "fs/country_counts",
        aggregated_period[[1]],
        aggregated_period[[2]]
      ),
      episomer_files(
        "fs/entities",
        aggregated_period[[1]],
        aggregated_period[[2]]
      ),
      episomer_files(
        "fs/geolocated",
        aggregated_period[[1]],
        aggregated_period[[2]]
      ),
      episomer_files(
        "fs/hashtags",
        aggregated_period[[1]],
        aggregated_period[[2]]
      ),
      episomer_files(
        "fs/topwords",
        aggregated_period[[1]],
        aggregated_period[[2]]
      ),
      episomer_files(
        "fs/urls",
        aggregated_period[[1]],
        aggregated_period[[2]]
      ),
      episomer_files("series", aggregated_period[[1]], aggregated_period[[2]]),
      episomer_files("stats", aggregated_period[[1]], aggregated_period[[2]])
    )
  if ("posts" %in% types)
    items <- c(
      items,
      list(
        "geo/togeolocate.json" = get_post_togeo_path(),
        "geo/geolocating.json" = get_post_geoing_path(),
        "geo/toaggregate.json" = get_post_toaggr_path(),
        "geo/aggregating.json" = get_post_aggring_path()
      ),
      episomer_files("fs/posts", posts_period[[1]], posts_period[[2]])
    )
  if ("logs" %in% types)
    jobs_dir <- file.path(conf$data_dir, "jobs")
    items <- c(
      items,
      list(
        "detect.log" = file.path(jobs_dir, "detect.log"),
        "fs.log" =  file.path(jobs_dir, "fs.log"),
        "search.log" =  file.path(jobs_dir, "search.log")
      )
    )
  nb <- length(items)
  i <- 0
  progress(0, paste("Building archive for", nb, "files"))
  last_message <- as.integer(Sys.time())
  oldwd <- getwd()
  on.exit(setwd(oldwd))
  setwd(conf$data_dir)

  # adding files from external folders to ensure portability
  to_copy <- file.exists(unlist(items)) &
    (!file.exists(names(items)) |
      names(items) %in% c("detect.log", "fs.log", "search.log"))
  if (length(to_copy) > 0) {
    file.copy(unlist(items[to_copy]), names(items)[to_copy])
  }

  # creating list of files to add
  to_arc <- normalizePath(tempfile("toarc"), winslash = "/", mustWork = FALSE)
  to_arc_con <- file(to_arc, "wb", encoding = "UTF-8")
  fnames <- normalizePath(names(items), winslash = "/", mustWork = FALSE)
  writeLines(fnames[file.exists(fnames)], sep = "\n", to_arc_con)
  close(to_arc_con)

  # launching the compression as a background process
  cb <- function(line, proc) {
    i <<- i + 1
    if (as.integer(Sys.time()) - last_message > 2) {
      progress(
        1.0 * i / nb,
        paste("Building archive for", i, "of", nb, "files")
      )
      last_message <<- as.integer(Sys.time())
    }
  }

  result = tryCatch(
    {
      processx::run(
        "tar",
        c(
          if (compress) "-czvf" else "-cvf",
          destination,
          "-T",
          to_arc,
          "--force-local"
        ),
        stdout_line_callback = cb,
        wd = conf$data_dir
      )
    },
    error = function(e) {
      processx::run(
        "tar",
        c(if (compress) "-czvf" else "-cvf", destination, "-T", to_arc),
        stdout_line_callback = cb,
        wd = conf$data_dir
      )
    }
  )
}

episomer_files <- function(rel_path, dmin = NULL, dmax = NULL) {
  if (!is.null(dmin) && is.na(dmin)) dmin <- NULL
  if (!is.null(dmax) && is.na(dmax)) dmax <- NULL
  ret = list()
  dirs <- list.dirs(file.path(conf$data_dir, rel_path), full.names = FALSE)
  rel_dirs <- ifelse(dirs == "", rel_path, file.path(rel_path, dirs))

  for (d in rel_dirs) {
    f <- file.path(conf$data_dir, d)
    file_names <- setdiff(
      list.files(f),
      list.dirs(f, recursive = FALSE, full.names = FALSE)
    )
    files <- as.list(setNames(
      file.path(conf$data_dir, d, file_names),
      file.path(d, file_names)
    ))
    dates <- as.Date(strptime(
      regmatches(
        names(files),
        gregexpr("[0-9][0-9][0-9][0-9]\\.[0-9][0-9]\\.[0-9][0-9]", names(files))
      ),
      "%Y.%m.%d"
    ))
    fd <- ifelse(
      sapply(dates, function(v) is.null(dmin)),
      TRUE,
      ifelse(is.na(dates), FALSE, dates >= dmin)
    ) &
      ifelse(
        sapply(dates, function(v) is.null(dmax)),
        TRUE,
        ifelse(is.na(dates), FALSE, dates <= dmax)
      )
    weeks <- regmatches(
      names(files),
      gregexpr("[0-9][0-9][0-9][0-9]\\.[0-9][0-9]", names(files))
    )
    weeks <- ifelse(is.na(dates), weeks, NA)
    wmin = if (is.null(dmin)) NULL else strftime(dmin, "%G.%V")
    wmax = if (is.null(dmax)) NULL else strftime(dmax, "%G.%V")
    fw <- ifelse(
      sapply(weeks, function(v) is.null(wmin)),
      TRUE,
      ifelse(is.na(weeks), FALSE, weeks >= wmin)
    ) &
      ifelse(
        sapply(weeks, function(v) is.null(wmax)),
        TRUE,
        ifelse(is.na(weeks), FALSE, weeks <= wmax)
      )
    ret <- c(ret, files[fd | fw])
  }
  ret
}

#' @title Run the episomer Shiny app
#' @description Open the episomer Shiny app, used to setup the Data collection & processing pipeline, the Requirements & alerts pipeline and to visualise the outputs.
#' @param data_dir Path to the 'data directory' containing application settings, models and collected posts.
#' If not provided the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, default: NA
#' @param host The IP if the network interface to run the Shiny app on, default: NULL (will run on 127.0.0.1)
#' @param port The port to run the Shiny app on, default: NULL (will run on a random port)
#' @param profile The profileis to run the Shiny app on, default: "dashboard" (can be "dashboard" or "admin")
#' @return The Shiny server object containing the launched application
#' @details The episomer app is the user entry point to the episomer package. This application will help the user to setup the post collection process, manage all settings,
#' see the interactive dashboard visualisations, export them to Markdown or PDF, and setup the alert emails.
#'
#' All its functionality is described on the episomer vignette.
#' @examples
#' \dontrun{
#'    #Running the episomer app
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    episomer_app()
#' }
#' @seealso
#'  \code{\link{search_loop}}
#'
#'  \code{\link{detect_loop}}
#' @rdname episomer_app
#' @export
#' @importFrom shiny fluidPage fluidRow column selectInput h4 conditionalPanel dateRangeInput radioButtons checkboxInput sliderInput numericInput downloadButton h3 htmlOutput actionButton span textInput textAreaInput h2 passwordInput h5 fileInput uiOutput navbarPage tabPanel observe updateSliderInput updateDateRangeInput downloadHandler invalidateLater renderText observeEvent renderUI validate need shinyApp
#' @importFrom plotly plotlyOutput renderPlotly ggplotly config layout
#' @importFrom DT dataTableOutput renderDataTable datatable formatPercentage replaceData dataTableProxy
#' @importFrom rmarkdown render
#' @importFrom magrittr `%>%`
#' @importFrom stringr str_replace_all
#' @importFrom grDevices png
#' @importFrom ggplot2 ggsave
#' @importFrom dplyr select
#' @importFrom stats setNames
#' @importFrom utils write.csv head
episomer_app <- function(data_dir = NA, profile = c("dashboard", "admin"), host = NULL, port = NULL) {
  old <- options()
  on.exit(options(old))
  options(shiny.host = host, shiny.port = port)
  match.arg(profile, c("dashboard", "admin"))
  # Setting up configuration if not already done
  if (is.na(data_dir)) setup_config_if_not_already() else {
    setup_config(data_dir)
  }
  if (!file.exists(get_properties_path())) {
    save_config(data_dir = conf$data_dir, properties = TRUE, sm_topics = list())
    save_tasks(get_tasks())
  }
  # Loading data for dashboard and configuration
  d <- refresh_dashboard_data()
  if("admin" %in% profile) {
      cd <- refresh_config_data()
  }

  if("dashboard" %in% profile) {
      # Defining dashboard page UI
      ################################################
      ######### DASHBOARD PAGE #######################
      ################################################
      dashboard_page <-
        shiny::fluidPage(
          shiny::fluidRow(
            shiny::column(
              3,
              ################################################
              ######### DASHBOARD FILTERS #####################
              ################################################
              shiny::actionButton("run_dashboard", "Run"),
              shiny::selectInput(
                "sms",
                label = shiny::h4("Social Media"),
                multiple = TRUE,
                choices = d$sms,
                selected = d$sms
              ),
              shiny::selectInput(
                "topics",
                label = shiny::h4("Topics"),
                multiple = FALSE,
                choices = d$topics
              ),
              shiny::selectInput(
                "countries",
                label = shiny::h4("Countries & regions"),
                multiple = TRUE,
                choices = d$countries
              ),
              shiny::selectInput(
                "fixed_period",
                label = shiny::h4("Period"),
                multiple = FALSE,
                choices = list(
                  "Last 7 days" = "last 7 days",
                  "Last 30 days" = "last 30 days",
                  "Last 60 days" = "last 60 days",
                  "Last 180 days" = "last 180 days",
                  "custom"
                ),
                selected = d$fixed_period
              ),
              shiny::conditionalPanel(
                condition = "input.fixed_period == 'custom'",
                shiny::dateRangeInput(
                  "period",
                  label = shiny::h4("Dates"),
                  start = d$date_start,
                  end = d$date_end,
                  min = d$date_min,
                  max = d$date_max,
                  format = "yyyy-mm-dd",
                  startview = "month"
                )
              ),
              shiny::radioButtons(
                "period_type",
                label = shiny::h4("Time unit"),
                choices = list(
                  "Days" = "created_date",
                  "Weeks" = "created_weeknum"
                ),
                selected = "created_date",
                inline = TRUE
              ),
              shiny::h4("Include quotes"),
              shiny::checkboxInput(
                "with_quotes",
                label = NULL,
                value = conf$alert_with_quotes
              ),
              shiny::sliderInput(
                "alpha_filter",
                label = shiny::h4("Signal false positive rate"),
                min = 0,
                max = 0.3,
                value = conf$alert_alpha,
                step = 0.005
              ),
              shiny::sliderInput(
                "alpha_outlier_filter",
                label = shiny::h4("Outlier false positive rate"),
                min = 0,
                max = 0.3,
                value = conf$alert_alpha_outlier,
                step = 0.005
              ),
              shiny::sliderInput(
                "k_decay_filter",
                label = shiny::h4("Outlier downweight strength"),
                min = 0,
                max = 10,
                value = conf$alert_k_decay,
                step = 0.5
              ),
              shiny::h4("Bonferroni correction"),
              shiny::checkboxInput(
                "bonferroni_correction",
                label = NULL,
                value = conf$alert_with_bonferroni_correction
              ),
              shiny::numericInput(
                "history_filter",
                label = shiny::h4("Days in baseline"),
                value = conf$alert_history
              ),
              shiny::h4("Same weekday baseline"),
              shiny::checkboxInput(
                "same_weekday_baseline",
                label = NULL,
                value = conf$alert_same_weekday_baseline
              ),
              shiny::fluidRow(
                shiny::column(4, shiny::downloadButton("export_pdf", "PDF")),
                shiny::column(4, shiny::downloadButton("export_md", "Md"))
              )
            ),
            shiny::column(
              9,
              ################################################
              ######### DASHBOARD PLOTS #######################
              ################################################
              shiny::fluidRow(
                shiny::column(
                  12,
                  shiny::fluidRow(
                    shiny::column(
                      2,
                      shiny::downloadButton("download_line_data", "Data")
                    ),
                    shiny::column(2, shiny::downloadButton("export_line", "image"))
                  ),
                  plotly::plotlyOutput("line_chart")
                )
              ),
              shiny::fluidRow(
                shiny::column(
                  6,
                  shiny::fluidRow(
                    shiny::column(
                      3,
                      shiny::downloadButton("download_map_data", "Data")
                    ),
                    shiny::column(3, shiny::downloadButton("export_map", "image"))
                  ),
                  plotly::plotlyOutput(
                    "map_chart"
                  )
                ),
                shiny::column(
                  6,
                  shiny::fluidRow(
                    shiny::column(
                      3,
                      shiny::downloadButton("download_top1_data", "Data")
                    ),
                    shiny::column(3, shiny::downloadButton("export_top1", "image")),
                    shiny::column(6,"")
                  ),
                  plotly::plotlyOutput("top_chart1")
                )
              ),
              shiny::fluidRow(
                shiny::column(
                  6,
                  shiny::fluidRow(
                    shiny::column(
                      3,
                      shiny::downloadButton("download_top2_data", "Data")
                    ),
                    shiny::column(3, shiny::downloadButton("export_top2", "image")),
                    shiny::column(6, "")
                  ),
                  plotly::plotlyOutput("top_chart2")
                ),
                shiny::column(
                  6,
                  shiny::fluidRow(
                    shiny::column(
                      3,
                      shiny::downloadButton("download_top3_data", "Data")
                    ),
                  ),
                  shiny::fluidRow(
                    shiny::column(12, shiny::htmlOutput("top_table_title"))
                  ),
                  shiny::fluidRow(
                    shiny::column(12, DT::dataTableOutput("top_table"))
                  ),
                  shiny::fluidRow(
                    shiny::column(12, shiny::htmlOutput("top_table_disc"))
                  )
                )
              )
            )
          )
        )
  }

  if("admin" %in% profile) {
      # Defining alerts page UI
      ################################################
      ######### ALERTS PAGE ##########################
      ################################################
      alerts_page <-
        shiny::fluidPage(
          shiny::h3("Find alerts"),
          ################################################
          ######### ALERTS FILTERS #######################
          ################################################
          shiny::fluidRow(
            shiny::column(3, shiny::h4("Detection date")),
            shiny::column(2, shiny::h4("Topics")),
            shiny::column(2, shiny::h4("Countries & regions")),
            shiny::column(2, shiny::h4("Display")),
            shiny::column(1, shiny::h4("Limit")),
          ),
          shiny::fluidRow(
            shiny::column(
              3,
              shiny::dateRangeInput(
                "alerts_period",
                label = "",
                start = d$date_end,
                end = d$date_end,
                min = d$date_min,
                max = d$date_max,
                format = "yyyy-mm-dd",
                startview = "month"
              )
            ),
            shiny::column(
              2,
              shiny::selectInput(
                "alerts_topics",
                label = NULL,
                multiple = TRUE,
                choices = d$topics[d$topics != ""]
              )
            ),
            shiny::column(
              2,
              shiny::selectInput(
                "alerts_countries",
                label = NULL,
                multiple = TRUE,
                choices = d$countries
              )
            ),
            shiny::column(
              2,
              shiny::radioButtons(
                "alerts_display",
                label = NULL,
                choices = list("Posts" = "posts", "Parameters" = "parameters"),
                selected = "parameters",
                inline = TRUE
              ),
            ),
            shiny::column(
              1,
              shiny::selectInput(
                "alerts_limit",
                label = NULL,
                multiple = FALSE,
                choices = list(
                  "None" = "0",
                  "10" = "10",
                  "50" = "50",
                  "100" = "100",
                  "500" = "500"
                )
              )
            ),
          ),
          shiny::fluidRow(
            ################################################
            ######### ALERTS FILTERS #######################
            ################################################
            shiny::column(
              2,
              shiny::actionButton("alerts_search", "Search alerts"),
            ),
            shiny::column(
              2,
              shiny::actionButton("alerts_close", "Hide search"),
              shiny::conditionalPanel(
                condition = "false",
                shiny::textInput(
                  "alerts_show_search",
                  value = "false",
                  label = NULL
                )
              )
            ),
            shiny::column(
              2,
              shiny::actionButton("alertsdb_add", "Add alerts to annotations")
            ),
            shiny::column(6)
          ),
          shiny::fluidRow(
            shiny::column(
              12,
              ################################################
              ######### ALERTS TABLE #########################
              ################################################
              shiny::conditionalPanel(
                condition = "input.alerts_show_search == 'true'",
                DT::dataTableOutput("alerts_table"),
              )
            )
          ),
          shiny::h3("Alerts annotations"),
          shiny::fluidRow(
            shiny::column(
              2,
              shiny::actionButton("alertsdb_search", "Show annotations")
            ),
            shiny::column(
              2,
              shiny::actionButton("alertsdb_close", "Hide annotations")
            ),
            shiny::column(
              2,
              shiny::downloadButton("alertsdb_download", "Download annotations")
            ),
            shiny::column(
              4,
              shiny::fileInput(
                "alertsdb_upload",
                label = NULL,
                buttonLabel = "Upload & evaluate annotations", accept = ".xlsx"
              )
            ),
            shiny::column(
              2,
              shiny::conditionalPanel(
                condition = "false",
                shiny::textInput("alertsdb_show", value = "false", label = NULL)
              )
            )
          ),
          shiny::fluidRow(
            shiny::conditionalPanel(
              condition = "input.alertsdb_show == 'true'",
              shiny::column(
                12,
                ################################################
                ######### ALERT ANNOTATIONS EVALUATION #########
                ################################################
                shiny::h4(
                  "Performance evaluation of alert classification algorithm"
                ),
                DT::dataTableOutput("alertsdb_runs_table"),
                ################################################
                ######### ANNOTATED ALERTS #####################
                ################################################
                shiny::h4(
                  "Database used for training the alert classification algorithm"
                ),
                DT::dataTableOutput("alertsdb_table")
              )
            )
          )
        )
      # Defining configuration
      ################################################
      ######### CONFIGURATION PAGE####################
      ################################################
      config_page <-
        shiny::fluidPage(
          shiny::fluidRow(
            shiny::column(
              4,
              ################################################
              ######### STATUS PANEL #########################
              ################################################
              shiny::h3("Status"),
              shiny::fluidRow(
                shiny::column(3, "episomer database"),
                shiny::column(5, shiny::htmlOutput("fs_running")),
                shiny::column(2, shiny::actionButton("activate_fs", "activate")),
                shiny::column(2, shiny::actionButton("stop_fs", "stop"))
              ),
              shiny::fluidRow(
                shiny::column(3, "Data collection & processing"),
                shiny::column(5, shiny::htmlOutput("search_running")),
                shiny::column(
                  2,
                  shiny::actionButton("activate_search", "activate")
                ),
                shiny::column(2, shiny::actionButton("stop_search", "stop"))
              ),
              shiny::fluidRow(
                shiny::column(3, "Requirements & alerts"),
                shiny::column(5, shiny::htmlOutput("detect_running")),
                shiny::column(
                  2,
                  shiny::actionButton("activate_detect", "activate")
                ),
                shiny::column(2, shiny::actionButton("stop_detect", "stop"))
              ),
              ################################################
              ######### GENERAL PROPERTIES ###################
              ################################################
              shiny::h3("Signal detection"),
              shiny::fluidRow(
                shiny::column(3, "Signal false positive rate"),
                shiny::column(
                  9,
                  shiny::sliderInput(
                    "conf_alpha",
                    label = NULL,
                    min = 0,
                    max = 0.3,
                    value = conf$alert_alpha,
                    step = 0.005
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Outlier false positive rate"),
                shiny::column(
                  9,
                  shiny::sliderInput(
                    "conf_alpha_outlier",
                    label = NULL,
                    min = 0,
                    max = 0.3,
                    value = conf$alert_alpha_outlier,
                    step = 0.005
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Outlier downweight strength"),
                shiny::column(
                  9,
                  shiny::sliderInput(
                    "conf_k_decay",
                    label = NULL,
                    min = 0,
                    max = 10,
                    value = conf$alert_k_decay,
                    step = 0.5
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Days in baseline"),
                shiny::column(
                  9,
                  shiny::numericInput(
                    "conf_history",
                    label = NULL,
                    value = conf$alert_history
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Same weekday baseline"),
                shiny::column(
                  9,
                  shiny::checkboxInput(
                    "conf_same_weekday_baseline",
                    label = NULL,
                    value = conf$alert_same_weekday_baseline
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Include quotes"),
                shiny::column(
                  9,
                  shiny::checkboxInput(
                    "conf_with_quotes",
                    label = NULL,
                    value = conf$alert_with_quotes
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Bonferroni correction"),
                shiny::column(
                  9,
                  shiny::checkboxInput(
                    "conf_with_bonferroni_correction",
                    label = NULL,
                    value = conf$alert_with_bonferroni_correction
                  )
                )
              ),
              shiny::h3("General"),
              shiny::fluidRow(
                shiny::column(3, "Data dir"),
                shiny::column(9, shiny::span(conf$data_dir))
              ),
              shiny::fluidRow(
                shiny::column(3, "Data collection & processing span (min)"),
                shiny::column(
                  9,
                  shiny::numericInput(
                    "conf_collect_span",
                    label = NULL,
                    value = conf$collect_span
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Requirements & alerts span (min)"),
                shiny::column(
                  9,
                  shiny::numericInput(
                    "conf_schedule_span",
                    label = NULL,
                    value = conf$schedule_span
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Launch slots"),
                shiny::column(9, shiny::htmlOutput("conf_schedule_slots"))
              ),
              shiny::fluidRow(
                shiny::column(3, "Password store"),
                shiny::column(
                  9,
                  shiny::selectInput(
                    "conf_keyring",
                    label = NULL,
                    choices = c(
                      "wincred",
                      "macos",
                      "file",
                      "secret_service",
                      "environment"
                    ),
                    selected = conf$keyring
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Spark cores"),
                shiny::column(
                  9,
                  shiny::numericInput(
                    "conf_spark_cores",
                    label = NULL,
                    value = conf$spark_cores
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Spark memory"),
                shiny::column(
                  9,
                  shiny::textInput(
                    "conf_spark_memory",
                    label = NULL,
                    value = conf$spark_memory
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Geolocation threshold"),
                shiny::column(
                  9,
                  shiny::textInput(
                    "geolocation_threshold",
                    label = NULL,
                    value = conf$geolocation_threshold
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "GeoNames URL"),
                shiny::column(
                  9,
                  shiny::textInput(
                    "conf_geonames_url",
                    label = NULL,
                    value = conf$geonames_url
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Simplified GeoNames"),
                shiny::column(
                  9,
                  shiny::checkboxInput(
                    "conf_geonames_simplify",
                    label = NULL,
                    value = conf$geonames_simplify
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Stream database results"),
                shiny::column(
                  9,
                  shiny::checkboxInput(
                    "conf_onthefly_api",
                    label = NULL,
                    value = conf$onthefly_api
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Maven repository"),
                shiny::column(
                  9,
                  shiny::textInput(
                    "conf_maven_repo",
                    label = NULL,
                    value = conf$maven_repo
                  )
                )
              ),
              shiny::conditionalPanel(
                condition = "false",
                shiny::textInput("os_type", value = .Platform$OS.type, label = NULL)
              ),
              shiny::conditionalPanel(
                condition = "input.os_type == 'windows'",
                shiny::fluidRow(
                  shiny::column(3, "Winutils URL"),
                  shiny::column(
                    9,
                    shiny::textInput(
                      "conf_winutils_url",
                      label = NULL,
                      value = conf$winutils_url
                    )
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Region disclaimer"),
                shiny::column(
                  9,
                  shiny::textAreaInput(
                    "conf_regions_disclaimer",
                    label = NULL,
                    value = conf$regions_disclaimer
                  )
                )
              ),
              shiny::h2("Email authentication (SMTP)"),
              shiny::fluidRow(
                shiny::column(3, "Server"),
                shiny::column(
                  9,
                  shiny::textInput(
                    "smtp_host",
                    label = NULL,
                    value = conf$smtp_host
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Port"),
                shiny::column(
                  9,
                  shiny::numericInput(
                    "smtp_port",
                    label = NULL,
                    value = conf$smtp_port
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "From"),
                shiny::column(
                  9,
                  shiny::textInput(
                    "smtp_from",
                    label = NULL,
                    value = conf$smtp_from
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Login"),
                shiny::column(
                  9,
                  shiny::textInput(
                    "smtp_login",
                    label = NULL,
                    value = conf$smtp_login
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Password"),
                shiny::column(
                  9,
                  shiny::passwordInput(
                    "smtp_password",
                    label = NULL,
                    value = conf$smtp_password
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Unsafe certificates"),
                shiny::column(
                  9,
                  shiny::checkboxInput(
                    "smtp_insecure",
                    label = NULL,
                    value = conf$smtp_insecure
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(3, "Admin email"),
                shiny::column(
                  9,
                  shiny::textInput(
                    "admin_email",
                    label = NULL,
                    value = conf$admin_email
                  )
                )
              ),
              shiny::h2("Task registering"),
              shiny::fluidRow(
                shiny::column(3, "Custom date format"),
                shiny::column(
                  9,
                  shiny::textInput(
                    "force_date_format",
                    label = NULL,
                    value = conf$force_date_format
                  )
                )
              )
            ),
            shiny::column(
              8,
              ################################################
              ######### Requirements & alerts PANEL ######################
              ################################################
              shiny::h3("Requirements & alerts pipeline"),
              shiny::h5("Manual tasks"),
              shiny::fluidRow(
                shiny::column(
                  3,
                  shiny::actionButton("update_dependencies", "Run dependencies")
                ),
                shiny::column(
                  3,
                  shiny::actionButton("update_geonames", "Run GeoNames")
                ),
                shiny::column(
                  3,
                  shiny::actionButton("update_languages", "Run languages")
                ),
                shiny::column(
                  3,
                  shiny::actionButton("request_alerts", "Run alerts")                  
                )
              ),
              DT::dataTableOutput("tasks_df"),
              ################################################
              ######### SOCIAL MEDIA PANDEL###################
              ################################################
              shiny::h2("Social media"),
              shiny::fluidRow(
                shiny::column(4, "Social Media"),
                shiny::column(8, "Configuration"),
                style = "font-weight:bold"
              ),
              shiny::fluidRow(
                shiny::column(
                  4,
                  shiny::checkboxInput(
                    "conf_sm_activated_bluesky",
                    label = "Bluesky",
                    value = conf$sm_activated_bluesky
                  )
                ),
                shiny::column(
                  8,
                  shiny::conditionalPanel(
                    condition = "input.conf_sm_activated_bluesky == true",
                    shiny::checkboxInput(
                      "conf_sm_alerts_bluesky",
                      label = "Calculate alerts",
                      value = conf$sm_alerts_bluesky
                    ),
                    shiny::fluidRow(
                      shiny::column(3, "User"),
                      shiny::column(
                        9,
                        shiny::textInput(
                          "bluesky_user",
                          label = NULL,
                          value = if (is_secret_set("bluesky_user"))
                            get_secret("bluesky_user") else NULL
                        )
                      )
                    ),
                    shiny::fluidRow(
                      shiny::column(3, "Password"),
                      shiny::column(
                        9,
                        shiny::passwordInput(
                          "bluesky_password",
                          label = NULL,
                          value = if (is_secret_set("bluesky_password"))
                            get_secret("bluesky_password") else NULL
                        )
                      )
                    )
                  )
                ),
                style = "background-color:#f9f9f9;border-top: 1px solid #b2b2b2;border-bottom: 1px solid #d9d9d9;"
              ),
              ################################################
              ######### TOPICS PANEL ######################
              ################################################
              shiny::h3("Topics"),
              shiny::fluidRow(
                shiny::column(3, shiny::h5("Available topics")),
                shiny::column(
                  2,
                  shiny::downloadButton("conf_topics_download", "Download")
                ),
                shiny::column(
                  2,
                  shiny::downloadButton(
                    "conf_orig_topics_download",
                    "Download default"
                  )
                ),
                shiny::column(
                  3,
                  shiny::fileInput(
                    "conf_topics_upload",
                    label = NULL,
                    buttonLabel = "Upload", accept = ".xlsx"
                  )
                ),
                shiny::column(
                  2,
                  shiny::actionButton(
                    "conf_dismiss_past_posts",
                    "Dismiss past posts"
                  )
                )
              ),
              shiny::fluidRow(
                shiny::column(4, shiny::h5("Limit topic history"))
              ),
              DT::dataTableOutput("config_topics"),
              ################################################
              ######### LANGUAGES PANEL ######################
              ################################################
              shiny::h3("Languages"),
              shiny::fluidRow(
                shiny::column(4, shiny::h5("Available languages")),
                shiny::column(
                  2,
                  shiny::downloadButton("conf_lang_download", "Download")
                ),
                shiny::column(
                  2,
                  shiny::downloadButton(
                    "conf_orig_lang_download",
                    "Download default"
                  )
                ),
                shiny::column(
                  4,
                  shiny::fileInput(
                    "conf_lang_upload",
                    label = NULL,
                    buttonLabel = "Upload", accept = ".xlsx"
                  )
                ),
              ),
              shiny::fluidRow(
                shiny::column(4, shiny::h5("Active languages")),
                shiny::column(6, shiny::uiOutput("lang_items_0")),
                shiny::column(1, shiny::actionButton("conf_lang_add", "+")),
                shiny::column(1, shiny::actionButton("conf_lang_remove", "-")),
              ),
              DT::dataTableOutput("config_langs"),
              ################################################
              ######### IMPORTANT USERS ######################
              ################################################
              shiny::h3("Important users"),
              shiny::fluidRow(
                shiny::column(4, shiny::h5("User file")),
                shiny::column(
                  2,
                  shiny::downloadButton("conf_users_download", "Download")
                ),
                shiny::column(
                  2,
                  shiny::downloadButton(
                    "conf_orig_users_download",
                    "Download default"
                  )
                ),
                shiny::column(
                  4,
                  shiny::fileInput(
                    "conf_users_upload",
                    label = NULL,
                    buttonLabel = "Upload", accept = ".xlsx"
                  )
                ),
              ),
              ################################################
              ######### SUSCRIBERS PANEL #####################
              ################################################
              shiny::h3("Subscribers"),
              shiny::fluidRow(
                shiny::column(4, shiny::h5("Subscribers")),
                shiny::column(
                  2,
                  shiny::downloadButton("conf_subscribers_download", "Download")
                ),
                shiny::column(
                  2,
                  shiny::downloadButton(
                    "conf_orig_subscribers_download",
                    "Download default"
                  )
                ),
                shiny::column(
                  4,
                  shiny::fileInput(
                    "conf_subscribers_upload",
                    label = NULL,
                    buttonLabel = "Upload", accept = ".xlsx"
                  )
                ),
              ),
              DT::dataTableOutput("config_subscribers"),
              ################################################
              ######### COUNTRIES / REGIONS PANEL ############
              ################################################
              shiny::h3("Countries & regions"),
              shiny::fluidRow(
                shiny::column(4, shiny::h5("Countries & regions")),
                shiny::column(
                  2,
                  shiny::downloadButton("conf_countries_download", "Download")
                ),
                shiny::column(
                  2,
                  shiny::downloadButton(
                    "conf_orig_countries_download",
                    "Download defaults"
                  )
                ),
                shiny::column(
                  4,
                  shiny::fileInput(
                    "conf_countries_upload",
                    label = NULL,
                    buttonLabel = "Upload", accept = ".xlsx"
                  )
                ),
              ),
              DT::dataTableOutput("config_regions")
            )
          )
        )
      # Defining geo tuning page UI
      ################################################
      ######### GEOTRAINING PAGE #####################
      ################################################
      geotraining_page <-
        shiny::fluidPage(
          shiny::h3("Geo tagging training"),
          shiny::fluidRow(
            ################################################
            ######### GEO TAG FILTERS ######################
            ################################################
            shiny::column(
              4,
              shiny::numericInput(
                "geotraining_posts2add",
                label = shiny::h4("Posts to add"),
                value = 100
              )
            ),
            shiny::column(
              2,
              shiny::actionButton("geotraining_update", "Geolocate annotations")
            ),
            shiny::column(
              2,
              shiny::downloadButton("geotraining_download", "Download annotations")
            ),
            shiny::column(
              4,
              shiny::fileInput(
                "geotraining_upload",
                label = NULL,
                buttonLabel = "Upload & evaluate annotations", accept = ".xlsx"
              )
            )
          ),
          shiny::fluidRow(
            shiny::column(
              12,
              ################################################
              ######### GEO TAG EVALUATION####################
              ################################################
              shiny::h4(
                "Performance evaluation of geo tagging algorithm (finding location position in text)"
              ),
              DT::dataTableOutput("geotraining_eval_df"),
              ################################################
              ######### GEO TAG TABLE #########################
              ################################################
              shiny::h4("Database used for training the geo tagging algorithm"),
              DT::dataTableOutput("geotraining_table")
            )
          )
        )
      ################################################
      ######### DATA PROTECTION PAGE #################
      ################################################
      dataprotection_page <-
        shiny::fluidPage(
          shiny::h3("Data protection"),
          shiny::fluidRow(
            ################################################
            ######### TWEET SEARCH FILTERS #################
            ################################################
            shiny::column(2, shiny::h4("Topic")),
            shiny::column(2, shiny::h4("Period")),
            shiny::column(2, shiny::h4("Countries & regions")),
            shiny::column(
              2,
              shiny::h4("Users"),
              shiny::radioButtons(
                "data_mode",
                label = NULL,
                choices = list(
                  "Mentioning" = "mentioning",
                  "From User" = "users",
                  "Both" = "all"
                ),
                selected = "all",
                inline = TRUE
              ),
            ),
            shiny::column(1, shiny::h4("Limit")),
            shiny::column(3, shiny::h4("Action"))
          ),
          shiny::fluidRow(
            shiny::column(
              2,
              shiny::selectInput(
                "data_topics",
                label = NULL,
                multiple = TRUE,
                choices = d$topics[d$topics != ""]
              )
            ),
            shiny::column(
              2,
              shiny::dateRangeInput(
                "data_period",
                label = "",
                start = d$date_end,
                end = d$date_end,
                min = d$date_min,
                max = d$date_max,
                format = "yyyy-mm-dd",
                startview = "month"
              )
            ),
            shiny::column(
              2,
              shiny::selectInput(
                "data_countries",
                label = NULL,
                multiple = TRUE,
                choices = d$countries
              )
            ),
            shiny::column(
              2,
              shiny::textInput("data_users", label = NULL, value = NULL)
            ),
            shiny::column(
              1,
              shiny::selectInput(
                "data_limit",
                label = NULL,
                multiple = FALSE,
                choices = list("50" = "50", "100" = "100", "500" = "500"),
                selected = "50"
              )
            ),
            shiny::column(
              3,
              shiny::actionButton("data_search", "Search"),
              shiny::actionButton("data_search_ano", "Anonym Search"),
              shiny::actionButton("data_anonymise", "Anonymize"),
              shiny::actionButton("data_delete", "Delete"),
            ),
          ),
          shiny::fluidRow(
            shiny::column(12, shiny::htmlOutput("data_message"))
          ),
          shiny::fluidRow(
            shiny::column(
              12,
              ################################################
              ######### TWEET SEARCH RESULTS #################
              ################################################
              DT::dataTableOutput("data_search_df")
            )
          )
        )
      ################################################
      ######### TROUBLESHOOT PAGE ####################
      ################################################
      troubleshoot_page <-
        shiny::fluidPage(
          shiny::h3("Create snapshot file"),
          shiny::fluidRow(
            shiny::column(
              6,
              shiny::checkboxGroupInput(
                "snapshot_types",
                "Include:",
                inline = TRUE,
                choices = c(
                  "Settings" = "settings",
                  "Dependencies" = "dependencies",
                  "Machine Learning" = "machine-learning",
                  "Aggregation" = "aggregations",
                  "Posts" = "posts",
                  "Logs" = "logs"
                ),
                selected = c("settings", "logs")
              )
            ),
            shiny::column(
              3,
              shiny::dateRangeInput(
                "snapshot_aggr_period",
                label = shiny::h4("Aggregated data period"),
                start = d$date_min,
                end = d$date_end,
                min = d$date_min,
                max = d$date_max,
                format = "yyyy-mm-dd",
                startview = "month"
              )
            ),
            shiny::column(
              3,
              shiny::dateRangeInput(
                "snapshot_post_period",
                label = shiny::h4("Post data period"),
                start = d$date_min,
                end = d$date_end,
                min = d$date_min,
                max = d$date_max,
                format = "yyyy-mm-dd",
                startview = "month"
              )
            )
          ),
          shiny::fluidRow(
            shiny::column(
              12,
              shiny::textInput(
                "snapshot_folder",
                value = ensure_snapshot_folder(),
                label = "Destination folder"
              ),
              shiny::checkboxInput(
                "snapshot_compress",
                "Compress snapshot",
                value = FALSE
              ),
              shiny::p(
                "It is recommended to stop the search and detect loop before running a snapshopt to ensure data consistency (embedded database must be running)"
              ),
              shiny::actionButton("build_snapshot", "Create snapshot")
            )
          ),
          shiny::h3("Diagnostics"),
          shiny::h5("Automated diagnostic tasks"),
          shiny::fluidRow(
            shiny::column(
              12,
              shiny::actionButton("run_diagnostic", "run diagnostics")
            )
          ),
          shiny::fluidRow(
            shiny::column(
              12,
              ################################################
              ######### DIAGNOSTIC TABLE #####################
              ################################################
              DT::dataTableOutput("diagnostic_table")
            )
          )
        )
  }
  # Defining navigation UI
  ui_app <- function(profile) {
    args = list("episomer", id = "navbar_shinyapp")
    if("dashboard" %in% profile) {
        args = c(args, list(shiny::tabPanel("Dashboard", value = "dashboard_page", dashboard_page)))
    }
    if("admin" %in% profile) {
        args = c(
           args,  
           list(
	        shiny::tabPanel("Alerts", value = "alerts_page", alerts_page),
                shiny::tabPanel("Geotag", value = "geotraining_page", geotraining_page),
                shiny::tabPanel(
                  "Data protection",
                  value = "dataprotection_page",
                  dataprotection_page
                ),
                shiny::tabPanel("Configuration", value = "config_page", config_page),
                shiny::tabPanel(
                  "Troubleshoot",
                  value = "troubleshoot_page",
                  troubleshoot_page
                )
	   )
        ) 
    }
    do.call(shiny::navbarPage, args)
  }
  # Defining server logic
  server_app <- function(input, output, session, profile, ...) {
    `%>%` <- magrittr::`%>%`

    shiny::onStop(function() {
      options("shiny.port" = NULL)
      options("shiny.host" = NULL)
    })
    
    if("dashboard" %in% profile) {

        ################################################
        ######### FILTERS LOGIC ########################
        ################################################

        # updating alpha filter based on selected topic value
        shiny::observe({
          # reading input$topics will automatically trigger the update on change
          val <- {
            if (length(input$topics) == 0 || input$topics == "")
              conf$alert_alpha else
              unname(get_topics_alphas()[stringr::str_replace_all(
                input$topics,
                "%20",
                " "
              )])
          }
          shiny::updateSliderInput(
            session,
            "alpha_filter",
            value = val,
            min = 0,
            max = 0.3,
            step = 0.005
          )
        })

        # upadating outliers alpha filter based on selected topic value
        shiny::observe({
          # reading input$topics will automatically trigger the update on change
          val <- {
            if (length(input$topics) == 0 || input$topics == "")
              conf$alert_alpha_outlier else
              unname(get_topics_alpha_outliers()[stringr::str_replace_all(
                input$topics,
                "%20",
                " "
              )])
          }
          shiny::updateSliderInput(
            session,
            "alpha_outlier_filter",
            value = val,
            min = 0,
            max = 0.3,
            step = 0.005
          )
        })

        # update the date ranges based on type of range selected
        shiny::observe({
          # reading input$fixed_period will automatically trigger the update on change
          # refresh dashboard data is necessary to update active dates
          refresh_dashboard_data(d, input$fixed_period)
          shiny::updateDateRangeInput(
            session,
            "period",
            start = d$date_start,
            end = d$date_end,
            min = d$date_min,
            max = d$date_max
          )
        })

        ################################################
        ######### DASHBOARD LOGIC ######################
        ################################################
        # updating the topics if they change on the configuraton page

	shiny::observe({
	  d$topics_refresh_flag()
	  shiny::isolate(refresh_dashboard_data(d))
	  shiny::updateSelectInput(
             session,
             "topics",
	     choices = d$topics
	  )
	})

	# dashboard actions
	shiny::observeEvent(input$run_dashboard, {
          # rendering line chart
          rep = new.env()
          progress_start("Generating report", rep)
          output$line_chart <- plotly::renderPlotly({
            # Validate if minimal requirements for rendering are met
            progress_set(value = 0.15, message = "Generating line chart", rep)
            shiny::isolate({
              can_render(input, d)
              # getting the chart
              chart <- line_chart_from_filters(
                input$sms,
                input$topics,
                input$countries,
                input$period_type,
                input$period,
                input$with_quotes,
                input$alpha_filter,
                input$alpha_outlier_filter,
                input$k_decay_filter,
                input$history_filter,
                input$bonferroni_correction,
                input$same_weekday_baseline
              )$chart

              # Setting size based on container size
              height <- session$clientData$output_line_chart_height
              width <- session$clientData$output_line_chart_width

              # returning empty chart if no data is found on chart
              chart_not_empty(chart)

              # transforming chart on plotly
              gg <- plotly::ggplotly(
                chart,
                height = height,
                width = width,
                tooltip = c("label")
              ) %>%
                plotly::config(displayModeBar = FALSE)

              # Fixing bad entries on ggplotly chart
              for (i in 1:length(gg$x$data)) {
                if (
                  startsWith(gg$x$data[[i]]$name, "(") &&
                    endsWith(gg$x$data[[i]]$name, ")")
                )
                  gg$x$data[[i]]$name = gsub(
                    "\\(|\\)|,|[0-9]",
                    "",
                    gg$x$data[[i]]$name
                  ) else gg$x$data[[i]]$name = "                        "
                gg$x$data[[i]]$legendgroup = gsub(
                  "\\(|\\)|,|[0-9]",
                  "",
                  gg$x$data[[i]]$legendgroup
                )
              }
              gg
            })
          })

          # rendering map chart
          output$map_chart <- plotly::renderPlotly({
            # Validate if minimal requirements for rendering are met
            progress_set(value = 0.30, message = "Generating map chart", rep)

            shiny::isolate({
              can_render(input, d)
              # Setting size based on container size
              height <- session$clientData$output_map_chart_height
              width <- session$clientData$output_map_chart_width

              # getting the chart
              chart <- map_chart_from_filters(
                input$sms,
                input$topics,
                input$countries,
                input$period,
                input$with_quotes
              )$chart

              # returning empty chart if no data is found on chart
              chart_not_empty(chart)

              # transforming chart on plotly
              gg <- chart %>%
                plotly::ggplotly(
                  height = height,
                  width = width,
                  tooltip = c("label")
                ) %>%
                plotly::layout(
                  title = list(text = paste("<b>", chart$labels$title, "</b>")),
                  margin = list(l = 5, r = 5, b = 50, t = 80),
                  annotations = list(
                    text = chart$labels$caption,
                    font = list(size = 10),
                    showarrow = FALSE,
                    xref = 'paper',
                    x = 0,
                    yref = 'paper',
                    y = -0.15
                  ) #,
                  #legend = list(orientation = 'v')
                ) %>%
                plotly::config(displayModeBar = FALSE)

              # Fixing bad entries on ggplotly chart
              for (i in 1:length(gg$x$data)) {
                if (
                  substring(gg$x$data[[i]]$name, 1, 2) %in%
                    c("a.", "b.", "c.", "d.", "e.", "f.", "g.", "h.")
                )
                  gg$x$data[[i]]$name <- substring(
                    gg$x$data[[i]]$name,
                    4,
                    nchar(gg$x$data[[i]]$name)
                  )
                #gg$x$data[[i]]$legendgroup <- gsub("\\(|\\)|,|[0-9]", "", gg$x$data[[i]]$legendgroup)
              }
              gg
            })
          })

          output$top_chart1 <- plotly::renderPlotly({
            # Validate if minimal requirements for rendering are met
            progress_set(value = 0.45, message = "Generating top chart 1", rep)
            shiny::isolate({
              can_render(input, d)
              # Setting size based on container size
              height <- session$clientData$output_top_chart1_height
              width <- session$clientData$output_top_chart1_width

              # getting the chart
              chart <- top_chart_from_filters(
                input$sms,
                input$topics,
                "tags",
                input$countries,
                input$period,
                input$with_quotes,
                20
              )$chart

              # returning empty chart if no data is found on chart
              chart_not_empty(chart)

              # transforming chart on plotly
              chart %>%
                plotly::ggplotly(height = height, width = width) %>%
                plotly::layout(
                  title = list(
                    text = paste(
                      "<b>",
                      chart$labels$title,
                      "<br>",
                      chart$labels$subtitle
                    ),
                    "</b>"
                  ),
                  margin = list(l = 30, r = 30, b = 100, t = 80),
                  annotations = list(
                    text = chart$labels$caption,
                    font = list(size = 10),
                    showarrow = FALSE,
                    xref = 'paper',
                    x = 0,
                    yref = 'paper',
                    y = -0.4
                  )
                ) %>%
                plotly::config(displayModeBar = FALSE)
            })
          })

          output$top_chart2 <- plotly::renderPlotly({
            # Validate if minimal requirements for rendering are met
            progress_set(value = 0.7, message = "Generating top chart 2", rep)
            shiny::isolate({
              can_render(input, d)
              # Setting size based on container size
              height <- session$clientData$output_top_chart2_height
              width <- session$clientData$output_top_chart2_width

              # getting the chart
              chart <- top_chart_from_filters(
                input$sms,
                input$topics,
                "topwords",
                input$countries,
                input$period,
                input$with_quotes,
                20
              )$chart

              # returning empty chart if no data is found on chart
              chart_not_empty(chart)

              # transforming chart on plotly
              chart %>%
                plotly::ggplotly(height = height, width = width) %>%
                plotly::layout(
                  title = list(
                    text = paste(
                      "<b>",
                      chart$labels$title,
                      "<br>",
                      chart$labels$subtitle
                    ),
                    "</b>"
                  ),
                  margin = list(l = 30, r = 30, b = 100, t = 80),
                  annotations = list(
                    text = chart$labels$caption,
                    font = list(size = 10),
                    showarrow = FALSE,
                    xref = 'paper',
                    x = 0,
                    yref = 'paper',
                    y = -0.4
                  )
                ) %>%
                plotly::config(displayModeBar = FALSE)
            })
          })

          output$top_table_title <- shiny::isolate({
            shiny::renderText({
              topic <- unname(get_topics_labels()[stringr::str_replace_all(
                input$topics,
                "%20",
                " "
              )])
              paste(
                "<h5 style=\"font-weight:bold;font-family:Helvetica;font-size:16px\">Top URLS of messages mentioning",
                topic,
                "from",
                input$period[[1]],
                "to",
                input$period[[2]],
                "</h5>"
              )
            })
          })
          output$top_table <- DT::renderDataTable({
            # Validate if minimal requirements for rendering are met
            progress_set(value = 0.85, message = "Generating top links", rep)

            shiny::isolate({
              can_render(input, d)
              # Setting size based on container size
              height <- session$clientData$output_top_chart2_height
              width <- session$clientData$output_top_chart2_width

              # getting the chart to obtain the table
              chart <- top_chart_from_filters(
                input$sms,
                input$topics,
                "urls",
                input$countries,
                input$period,
                input$with_quotes,
                200
              )$chart

              # returning empty if no data is found on chart
              chart_not_empty(chart)

              data <- chart$data %>%
                dplyr::mutate(
                  top = paste(
                    "<a href=\"",
                    .data$top,
                    "\"  target=\"_blank\">",
                    .data$top,
                    "</a>"
                  )
                ) %>%
                dplyr::rename(Url = .data$top, Frequency = .data$frequency) %>%
                DT::datatable(escape = FALSE)
            })
          })

          output$top_table_disc <- shiny::isolate({
            shiny::renderText({
              progress_close(rep)
            })
          })
        })
        # Saving line chart as png and downloading it from shiny app
        output$export_line <- shiny::downloadHandler(
          filename = function() {
            paste(
              "line_dataset_",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              ".png",
              sep = ""
            )
          },
          content = function(file) {
            chart <-
              line_chart_from_filters(
                input$sms,
                input$topics,
                input$countries,
                input$period_type,
                input$period,
                input$with_quotes,
                input$alpha_filter,
                input$alpha_outlier_filter,
                input$k_decay_filter,
                input$history_filter,
                input$bonferroni_correction,
                input$same_weekday_baseline
              )$chart
            device <- function(..., width, height)
              grDevices::png(
                ...,
                width = width,
                height = height,
                res = 300,
                units = "in"
              )
            ggplot2::ggsave(file, plot = chart, device = device)
          }
        )
        # Saving line chart as CSV and downloading it from shiny app
        output$download_line_data <- shiny::downloadHandler(
          filename = function() {
            paste(
              "line_dataset_",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              ".csv",
              sep = ""
            )
          },
          content = function(file) {
            write.csv(
              line_chart_from_filters(
                input$sms,
                input$topics,
                input$countries,
                input$period_type,
                input$period,
                input$with_quotes,
                input$alpha_filter,
                input$alpha_outlier_filter,
                input$k_decay_filter,
                input$history_filter,
                input$bonferroni_correction,
                input$same_weekday_baseline
              )$data,
              file,
              row.names = FALSE
            )
          }
        )

        # Saving map chart as CSV and downloading it from shiny app
        output$download_map_data <- shiny::downloadHandler(
          filename = function() {
            paste(
              "map_dataset_",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              ".csv",
              sep = ""
            )
          },
          content = function(file) {
            write.csv(
              map_chart_from_filters(
                input$sms,
                input$topics,
                input$countries,
                input$period,
                input$with_quotes
              )$data,
              file,
              row.names = FALSE
            )
          }
        )

        # Saving map chart as png and downloading it from shiny app
        output$export_map <- shiny::downloadHandler(
          filename = function() {
            paste(
              "map_dataset_",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              ".png",
              sep = ""
            )
          },
          content = function(file) {
            chart <- map_chart_from_filters(
              input$sms,
              input$topics,
              input$countries,
              input$period,
              input$with_quotes
            )$chart
            device <- function(..., width, height)
              grDevices::png(
                ...,
                width = width,
                height = height,
                res = 300,
                units = "in"
              )
            ggplot2::ggsave(file, plot = chart, device = device)
          }
        )

        # Saving top chart 1 as CSV and downloading it from shiny app
        output$download_top1_data <- shiny::downloadHandler(
          filename = function() {
            paste(
              "top_dataset_",
              "tags",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              ".csv",
              sep = ""
            )
          },
          content = function(file) {
            write.csv(
              top_chart_from_filters(
                input$sms,
                input$topics,
                "tags",
                input$countries,
                input$period,
                input$with_quotes,
                200
              )$data,
              file,
              row.names = FALSE
            )
          }
        )
        # Saving top chart 2 as CSV and downloading it from shiny app
        output$download_top2_data <- shiny::downloadHandler(
          filename = function() {
            paste(
              "top_dataset_",
              "topwords",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              ".csv",
              sep = ""
            )
          },
          content = function(file) {
            write.csv(
              top_chart_from_filters(
                input$sms,
                input$topics,
                "topwords",
                input$countries,
                input$period,
                input$with_quotes,
                200
              )$data,
              file,
              row.names = FALSE
            )
          }
        )
        # Saving top table as CSV and downloading it from shiny app
        output$download_top3_data <- shiny::downloadHandler(
          filename = function() {
            paste(
              "top_dataset_",
              "urls",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              ".csv",
              sep = ""
            )
          },
          content = function(file) {
            write.csv(
              top_chart_from_filters(
                input$sms,
                input$topics,
                "urls",
                input$countries,
                input$period,
                input$with_quotes,
                200
              )$data,
              file,
              row.names = FALSE
            )
          }
        )

        # Saving top chart 1 as png and downloading it from shiny app
        output$export_top1 <- shiny::downloadHandler(
          filename = function() {
            paste(
              "top_",
              "tags",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              ".png",
              sep = ""
            )
          },
          content = function(file) {
            chart <-
              top_chart_from_filters(
                input$sms,
                input$topics,
                "tags",
                input$countries,
                input$period,
                input$with_quotes,
                50
              )$chart
            device <- function(..., width, height)
              grDevices::png(
                ...,
                width = width,
                height = height,
                res = 300,
                units = "in"
              )
            ggplot2::ggsave(file, plot = chart, device = device)
          }
        )

        # Saving top chart 2 as png and downloading it from shiny app
        output$export_top2 <- shiny::downloadHandler(
          filename = function() {
            paste(
              "top_",
              "topwords",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              ".png",
              sep = ""
            )
          },
          content = function(file) {
            chart <-
              top_chart_from_filters(
                input$sms,
                input$topics,
                "topwords",
                input$countries,
                input$period,
                input$with_quotes,
                50
              )$chart
            device <- function(..., width, height)
              grDevices::png(
                ...,
                width = width,
                height = height,
                res = 300,
                units = "in"
              )
            ggplot2::ggsave(file, plot = chart, device = device)
          }
        )

        # Saving dashboard as PDF downloading it from shiny app
        output$export_pdf <- shiny::downloadHandler(
          filename = function() {
            paste(
              "episomer_dashboard_",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              "_",
              "tags",
              "_",
              "topwords",
              ".pdf",
              sep = ""
            )
          },
          content = function(file) {
            export_dashboard(
              "pdf_document",
              file,
              input$sms,
              input$topics,
              input$countries,
              input$period_type,
              input$period,
              input$with_quotes,
              input$location_type,
              input$alpha_filter,
              input$alpha_outlier_filter,
              input$k_decay_filter,
              input$history_filter,
              input$bonferroni_correction,
              input$same_weekday_baseline,
              "tags",
              "topwords"
            )
          }
        )
        # Saving dashboard as markdown downloading it from shiny app
        output$export_md <- shiny::downloadHandler(
          filename = function() {
            paste(
              "episomer_dashboard_",
              "_",
              paste(input$topics, collapse = "-"),
              "_",
              paste(input$countries, collapse = "-"),
              "_",
              input$period[[1]],
              "_",
              input$period[[2]],
              "_",
              "tags",
              "_",
              "topwords",
              ".md",
              sep = ""
            )
          },
          content = function(file) {
            export_dashboard(
              "md_document",
              file,
              input$sms,
              input$topics,
              input$countries,
              input$period_type,
              input$period,
              input$with_quotes,
              input$location_type,
              input$alpha_filter,
              input$alpha_outlier_filter,
              input$k_decay_filter,
              input$history_filter,
              input$bonferroni_correction,
              input$same_weekday_baseline,
              "tags",
              "topwords"
            )
          }
        )
    }


    if("admin" %in% profile) {
        ################################################
        ######### CONFIGURATION LOGIC ##################
        ################################################
        
        ######### STATUS PANEL LOGIC ##################
        # Timer for updating task statuses
        # each ten seconds config data will be reloaded to capture changes from data collection & processing and requirements & alerts pipelines
        # each ten seconds a process refresh flag will be invalidated to trigger process status recalculation
        shiny::observe({
          # Setting the timer
          shiny::invalidateLater(10000)

          # updating config data
          refresh_config_data(cd, list("tasks", "topics", "langs"))

          # invalidating the process refresh flag
          cd$process_refresh_flag(Sys.time())
        })

        # rendering the Data collection & processing running status
        output$search_running <- shiny::renderText({
          # Adding a dependency to task refresh (each time a task has changed by the Requirements & alerts pipeline)
          cd$tasks_refresh_flag()
          # Adding a dependency to process update (each 10 seconds)
          cd$process_refresh_flag()

          color = if (
            cd$search_running &&
              !is.na(cd$search_diff) &&
              length(cd$missing_search_jobs) == 0
          ) {
            "#348017"
          } else if (!cd$search_running || is.na(cd$search_diff)) {
            "#F75D59"
          } else {
            "#EfAF0D"
          }

          message = ""
          if (!cd$search_running || is.na(cd$search_diff)) {
            message = "Stopped"
          } else {
            message = "Running"
          }
          if (!is.na(cd$search_diff)) {
            message = paste(
              message,
              "( wrote",
              round(cd$search_diff, 2),
              units(cd$search_diff),
              "ago )"
            )
          }
          if (length(cd$missing_search_jobs) > 0) {
            message = paste(
              message,
              "<br/>",
              paste(cd$missing_search_jobs, collapse = ", "),
              " missing"
            )
          }

          # updating the label
          paste(
            "<span",
            " style='color:",
            color,
            "'>",
            message,
            "</span>",
            sep = ""
          )
        })

        # rendering the fs running status
        output$fs_running <- shiny::renderText({
          # Adding a dependency to task refresh (each time a task has changed by the Requirements & alerts pipeline)
          cd$tasks_refresh_flag()
          # Adding a dependency to process update (each 10 seconds)
          cd$process_refresh_flag()

          # updating the label
          paste(
            "<span",
            " style='color:",
            if (cd$fs_running) "#348017'" else "#F75D59'",
            ">",
            if (cd$fs_running) "Running" else "Stopped",
            "</span>",
            sep = ""
          )
        })

        # rendering the Requirements & alerts running status
        output$detect_running <- shiny::renderText({
          # Adding a dependency to task refresh (each time a task has changed by the Requirements & alerts pipeline)
          cd$tasks_refresh_flag()
          # Adding a dependency to process update (each 10 seconds)
          cd$process_refresh_flag()

          # updating the label
          paste(
            "<span",
            " style='color:",
            if (cd$detect_running) "#348017'" else "#F75D59'",
            ">",
            if (cd$detect_running) "Running" else "Stopped",
            "</span>",
            sep = ""
          )
        })

        # rendering the slots each time properties are changed
        output$conf_schedule_slots <- shiny::renderText({
          # Adding a dependency to config refresh
          cd$properties_refresh_flag()

          #rendering the label
          paste(
            head(
              lapply(
                get_task_day_slots(get_tasks()$alerts),
                function(d) strftime(d, format = "%H:%M")
              ),
              -1
            ),
            collapse = ", "
          )
        })

        # registering the Data collection & processing runner after button is clicked
        shiny::observeEvent(input$activate_search, {
          # registering the scheduled task
          issues = loop_run_issues("search")
          if (issues == "") {
            register_search_runner_task()
            # refresh task data to check if tasks are to be updated
            refresh_config_data(cd, list("tasks"))
          } else {
            shiny::showModal(shiny::modalDialog(
              title = "Warning",
              issues,
              footer = shiny::tagList(shiny::modalButton("OK"))
            ))
          }
        })
        shiny::observeEvent(input$stop_search, {
          stop_search_runner_task()
          # refresh task data to check if tasks are to be updated
          refresh_config_data(cd, list("tasks"))
        })

        # registering the fs runner after button is clicked
        shiny::observeEvent(input$activate_fs, {
          issues = loop_run_issues("fs")
          if (issues == "") {
            # registering the scheduled task
            register_fs_runner_task()
            # refresh task data to check if tasks are to be updated
            refresh_config_data(cd, list("tasks"))
          } else {
            shiny::showModal(shiny::modalDialog(
              title = "Warning",
              issues,
              footer = shiny::tagList(shiny::modalButton("OK"))
            ))
          }
        })

        shiny::observeEvent(input$stop_fs, {
          # stop the scheduled task
          stop_fs_runner_task()
          # refresh task data to check if tasks are to be updated
          refresh_config_data(cd, list("tasks"))
        })

        # registering the Requirements & alerts runner after button is clicked
        shiny::observeEvent(input$activate_detect, {
          # Setting values configuration to trigger one shot tasks for the first time
          if (is.na(conf$dep_updated_on))
            conf$dep_updated_on <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
          if (is.na(conf$geonames_updated_on))
            conf$geonames_updated_on <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
          if (is.na(conf$lang_updated_on))
            conf$lang_updated_on <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
          # Forcing refresh of tasks
          cd$tasks_refresh_flag(Sys.time())
          # saving properties to ensure the Requirements & alerts pipeline can see the changes and start running the tasks
          save_config(
            data_dir = conf$data_dir,
            properties = TRUE,
            sm_topics = list()
          )

          # registering the scheduled task
          register_detect_runner_task()

          # refresh task data to check if tasks are to be updated
          refresh_config_data(cd, list("tasks"))
        })

        shiny::observeEvent(input$stop_detect, {
          stop_detect_runner_task()
          # refresh task data to check if tasks are to be updated
          refresh_config_data(cd, list("tasks"))
        })

        # action to activate de dependency download tasks
        shiny::observeEvent(input$update_dependencies, {
          # Setting values on the configuration so the Requirements & alerts pipeline know it has to launch the task
          conf$dep_updated_on <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
          # forcing a task refresh :TODO test if this is still necessary
          cd$tasks_refresh_flag(Sys.time())
          # saving configuration so the Requirements & alerts pipeline will see the changes
          save_config(
            data_dir = conf$data_dir,
            properties = TRUE,
            sm_topics = list()
          )
          # refreshing the tasks data
          refresh_config_data(cd, list("tasks"))
        })

        # action to activate the update geonames task
        shiny::observeEvent(input$update_geonames, {
          # Setting values on the configuration so the Requirements & alerts pipeline know it has to launch the task
          conf$geonames_updated_on <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
          # forcing a task refresh
          cd$tasks_refresh_flag(Sys.time())
          # saving configuration so the Requirements & alerts pipeline will see the changes
          save_config(
            data_dir = conf$data_dir,
            properties = TRUE,
            sm_topics = list()
          )
          # refreshing the tasks data
          refresh_config_data(cd, list("tasks"))
        })

        # action to activate the update languages task
        shiny::observeEvent(input$update_languages, {
          # Setting values on the configuration so the Requirements & alerts pipeline know it has to launch the task
          conf$lang_updated_on <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
          # forcing a task refresh
          cd$tasks_refresh_flag(Sys.time())
          # saving configuration so the Requirements & alerts pipeline will see the changes
          save_config(
            data_dir = conf$data_dir,
            properties = TRUE,
            sm_topics = list()
          )
          # refreshing the tasks data
          refresh_config_data(cd, list("tasks"))
        })

        # action to activate the alerts task
        shiny::observeEvent(input$request_alerts, {
          # Setting values on the configuration so the Requirements & alerts pipeline know it has to launch the task
          conf$alerts_requested_on <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
          # forcing a task refresh
          cd$tasks_refresh_flag(Sys.time())
          # saving configuration so the Requirements & alerts pipeline will see the changes
          save_config(
            data_dir = conf$data_dir,
            properties = TRUE,
            sm_topics = list()
          )
          # refreshing the tasks data
          refresh_config_data(cd, list("tasks"))
        })      

        ######### PROPERTIES LOGIC ##################
        reactive_elements_to_save_in_conf <- shiny::reactive({
          list(
          conf_collect_span = input$conf_collect_span,
          conf_schedule_span = input$conf_schedule_span,
          conf_keyring = input$conf_keyring,
          conf_spark_cores = input$conf_spark_cores,
          conf_spark_memory = input$conf_spark_memory,
          conf_onthefly_api = input$conf_onthefly_api,
          geolocation_threshold = input$geolocation_threshold,
          conf_geonames_url = input$conf_geonames_url,
          conf_maven_repo = input$conf_maven_repo,
          conf_winutils_url = input$conf_winutils_url,
          conf_geonames_simplify = input$conf_geonames_simplify,
          conf_regions_disclaimer = input$conf_regions_disclaimer,
          conf_alpha = input$conf_alpha,
          conf_alpha_outlier = input$conf_alpha_outlier,
          conf_k_decay = input$conf_k_decay,
          conf_history = input$conf_history,
          conf_same_weekday_baseline = input$conf_same_weekday_baseline,
          conf_with_bonferroni_correction = input$conf_with_bonferroni_correction,
          conf_with_quotes = input$conf_with_quotes,
          conf_smtp_host = input$smtp_host,
          conf_smtp_port = input$smtp_port,
          conf_smtp_insecure = input$smtp_insecure,
          conf_admin_email = input$admin_email,
          conf_smtp_from = input$smtp_from,
          conf_smtp_login = input$smtp_login,
          conf_smtp_password = input$smtp_password,
          conf_sm_alerts_bluesky = input$conf_sm_alerts_bluesky,
          conf_sm_activated_bluesky = input$conf_sm_activated_bluesky,
          conf_force_date_format = input$force_date_format
	  )
    })
    debounced_elements_to_save_in_conf <- shiny::debounce(reactive_elements_to_save_in_conf, 3000)

        shiny::observeEvent(debounced_elements_to_save_in_conf(), {          
          update_config_from_input(debounced_elements_to_save_in_conf())

          # Saving Bluesky propertes          
          sm_api_set_auth(
            network = "bluesky",
            shiny_input_list = input
          )                      
          
          # Saving properties.json
          save_config(
            data_dir = conf$data_dir,
          )

          # Forcing update on properties dependant refresh (e.g. time slots)
          cd$properties_refresh_flag(Sys.time())

        }, ignoreInit = TRUE)

        ######### IMPORTANT USERS LOGIC ###########
        # downloading current important users file
        output$conf_users_download <- shiny::downloadHandler(
          filename = function() "users.xlsx",
          content = function(file) {
            file.copy(get_known_users_path(), file)
          }
        )

        # downloading default important users file
        output$conf_orig_users_download <- shiny::downloadHandler(
          filename = function() "users.xlsx",
          content = function(file) {
            file.copy(get_default_known_users_path(), file)
          }
        )

        # uploading new important users
        shiny::observe({
          df <- input$conf_users_upload
          if (!is.null(df) && nrow(df) > 0) {
            uploaded <- df$datapath[[1]]
            file.copy(
              uploaded,
              paste(conf$data_dir, "users.xlsx", sep = "/"),
              overwrite = TRUE
            )
          }
        })

        ######### LANGUAGE LOGIC ##################
        # rendering languages each time a change is registered on languges
        output$config_langs <- DT::renderDataTable({
          # Adding dependency with lang refresh
          cd$langs_refresh_flag()
          DT::datatable(cd$langs)
        })

        # downloading current languages
        output$conf_lang_download <- shiny::downloadHandler(
          filename = function() "languages.xlsx",
          content = function(file) {
            file.copy(get_available_languages_path(), file)
          }
        )

        # downloading original languages file
        output$conf_orig_lang_download <- shiny::downloadHandler(
          filename = function() "languages.xlsx",
          content = function(file) {
            file.copy(get_default_available_languages_path(), file)
          }
        )

        # uploading new language file
        shiny::observe({
          df <- input$conf_lang_upload
          if (!is.null(df) && nrow(df) > 0) {
            uploaded <- df$datapath[[1]]
            file.copy(
              uploaded,
              paste(conf$data_dir, "languages.xlsx", sep = "/"),
              overwrite = TRUE
            )
            refresh_config_data(e = cd, limit = list("langs"))
          }
        })

        # rendering active languages when something has changed on languages
        output$lang_items_0 <- shiny::renderUI({
          # Adding a dependency to lang refresh
          cd$langs_refresh_flag()
          shiny::selectInput(
            "lang_items",
            label = NULL,
            multiple = FALSE,
            choices = cd$lang_items
          )
        })

        # adding the current language
        shiny::observeEvent(input$conf_lang_add, {
          add_config_language(input$lang_items, cd$lang_names[input$lang_items])
          save_config(
            data_dir = conf$data_dir,
            properties = TRUE,
            sm_topics = list()
          )
          refresh_config_data(e = cd, limit = list("langs"))
        })

        #removing a language
        shiny::observeEvent(input$conf_lang_remove, {
          remove_config_language(input$lang_items)
          save_config(
            data_dir = conf$data_dir,
            properties = TRUE,
            sm_topics = list()
          )
          refresh_config_data(e = cd, limit = list("langs"))
        })

        ######### TASKS LOGIC ##################
        # rendering the tasks each time something changes in the tasks
        output$tasks_df <- DT::renderDataTable({
          # Adding dependency with tasks refresh
          cd$tasks_refresh_flag()
          DT::datatable(cd$tasks_df)
        })

        ######### TOPICS LOGIC ##################
        # rendering the topics only on first load update is done on next statement
        output$config_topics <- DT::renderDataTable({
          `%>%` <- magrittr::`%>%`
          cd$topics_df %>%
            DT::datatable(
              colnames = c(
                "Query length" = "QueryLength",
                "Active plans" = "ActivePlans",
                "Progress" = "Progress",
                "Requests" = "Requests",
                "Signal alpha (FPR)" = "Alpha",
                "Outlier alpha (FPR)" = "OutliersAlpha"
              ),
              filter = "top",
              escape = TRUE
            ) %>%
            DT::formatPercentage(columns = c("Progress"))
        })
        # this is for updating the topics
        shiny::observe({
          # Adding a dependency to topics refresh or plans refresh
          cd$topics_refresh_flag()
          cd$plans_refresh_flag()
          DT::replaceData(DT::dataTableProxy('config_topics'), cd$topics_df)
        })

        # download the current topics file
        output$conf_topics_download <- shiny::downloadHandler(
          filename = function() "topics.xlsx",
          content = function(file) {
            file.copy(get_user_topics_path(), file)
          }
        )

        #download the original topics file
        output$conf_orig_topics_download <- shiny::downloadHandler(
          filename = function() "topics.xlsx",
          content = function(file) {
            file.copy(get_default_topics_path(), file)
          }
        )

        # uploading a new topics file
        shiny::observe({
          df <- input$conf_topics_upload
          if (!is.null(df) && nrow(df) > 0) {
            uploaded <- df$datapath[[1]]
            if(file.exists(uploaded)) {
              file.copy(
                uploaded,
                paste(conf$data_dir, "topics.xlsx", sep = "/"),
                overwrite = TRUE
              )
	      file.remove(uploaded)
              refresh_config_data(e = cd, limit = list("topics"))
	      d$topics_refresh_flag(cd$topics_refresh_flag())
	    }
          }
        })
        # action to dismiss past posts
        shiny::observeEvent(input$conf_dismiss_past_posts, {
          # Setting values on the configuration so the search loop knows history needs to be dismissed
          conf$dismiss_past_request <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
          # saving configuration so the Requirements & alerts pipeline will see the changes
          save_config(
            data_dir = conf$data_dir,
            properties = TRUE,
            sm_topics = list()
          )
          # refreshing the tasks data
        })

        ######### SUBSCRIBERS LOGIC ##################
        # rendering subscribers (first run update on next statement)
        output$config_subscribers <- DT::renderDataTable({
          `%>%` <- magrittr::`%>%`
          get_subscribers() %>%
            DT::datatable(
              colnames = c(
                "User" = "User",
                "Email" = "Email",
                "Topics" = "Topics",
                "Excluded topics" = "Excluded Topics",
                "Immediate topics" = "Real time Topics",
                "Regions" = "Regions",
                "Immediate regions" = "Real time Regions",
                "Alert category" = "Alert category",
                "Alert slots" = "Alert Slots"
              ),
              filter = "top",
              escape = TRUE
            )
        })

        # updating the subscriber table on change of subscribers file
        shiny::observe({
          # Adding a dependency to subscribers refresh
          cd$subscribers_refresh_flag()
          DT::replaceData(
            DT::dataTableProxy('config_subscribers'),
            get_subscribers()
          )
        })

        # downloading subscribers default file
        output$conf_orig_subscribers_download <- shiny::downloadHandler(
          filename = function() "subscribers.xlsx",
          content = function(file) {
            file.copy(get_default_subscribers_path(), file)
          }
        )

        # downloading subscriber current file
        output$conf_subscribers_download <- shiny::downloadHandler(
          filename = function() "subscribers.xlsx",
          content = function(file) {
            file.copy(get_subscribers_path(), file)
          }
        )

        # uploading subscribers file
        shiny::observe({
          df <- input$conf_subscribers_upload
          if (!is.null(df) && nrow(df) > 0) {
            uploaded <- df$datapath[[1]]
            file.copy(
              uploaded,
              paste(conf$data_dir, "subscribers.xlsx", sep = "/"),
              overwrite = TRUE
            )
            cd$subscribers_refresh_flag(Sys.time())
          }
        })

        ######### REGIONS LOGIC ##################
        # rendering reguions for the first time (update on next statement)
        output$config_regions <- DT::renderDataTable({
          `%>%` <- magrittr::`%>%`
          get_regions_df() %>%
            DT::datatable(
              colnames = c(
                "Name" = "Name",
                "Codes" = "Codes",
                "Level" = "Level",
                "Min. lat." = "MinLatitude",
                "Max. lat." = "MaxLatitude",
                "Min. long." = "MinLongitude",
                "Max. long." = "MaxLingityde"
              ),
              filter = "top",
              escape = TRUE
            )
        })

        # Updating regions when the new file is uploaded
        shiny::observe({
          # Adding a dependency to country refresh
          cd$countries_refresh_flag()
          DT::replaceData(DT::dataTableProxy('config_regions'), get_regions_df())
        })

        # Download the current country file
        output$conf_countries_download <- shiny::downloadHandler(
          filename = function() "countries.xlsx",
          content = function(file) {
            file.copy(get_countries_path(), file)
          }
        )

        # Dowload the original country file
        output$conf_orig_countries_download <- shiny::downloadHandler(
          filename = function() "countries.xlsx",
          content = function(file) {
            file.copy(get_default_countries_path(), file)
          }
        )

        #Update the new country file
        shiny::observe({
          df <- input$conf_countries_upload
          if (!is.null(df) && nrow(df) > 0) {
            uploaded <- df$datapath[[1]]
            file.copy(
              uploaded,
              paste(conf$data_dir, "countries.xlsx", sep = "/"),
              overwrite = TRUE
            )
            cd$countries_refresh_flag(Sys.time())
          }
        })

        ######### ALERTS LOGIC ##################
        # rendering the alerts
        # updates are launched automatically when any input value changes

        # setting default value for number of alerts to return depending on wether we are displaying posts or not
        shiny::observe({
          # Can also set the label and select items
          shiny::updateSelectInput(
            session,
            "alerts_limit",
            selected = if (input$alerts_display == "posts") "10" else "0"
          )
        })

        shiny::observeEvent(input$alerts_search, {
          shiny::updateTextInput(
            session,
            "alerts_show_search",
            label = NULL,
            value = "true"
          )
          output$alerts_table <- DT::renderDataTable({
            `%>%` <- magrittr::`%>%`
            shiny::isolate({
              topposts <- if (input$alerts_display == "posts") 10 else 0
              progress_start("Getting alerts")
              alerts <- get_alerts(
                topic = input$alerts_topics,
                countries = as.numeric(input$alerts_countries),
                from = input$alerts_period[[1]],
                until = input$alerts_period[[2]],
                topposts = topposts,
                limit = as.integer(input$alerts_limit),
                progress = function(value, message) {
                  progress_set(value = value, message = message)
                }
              )
            })
            shiny::validate(
              shiny::need(
                !is.null(alerts),
                'No alerts generated for the selected period'
              )
            )
            if (topposts == 0) {
            dt <-   alerts %>%
                dplyr::select(
                  "date",
                  "hour",
                  "topic",
                  "country",
                  "episomer_category",
                  "tops",
                  "number_of_posts",
                  "known_ratio",
                  "limit",
                  "no_historic",
                  "bonferroni_correction",
                  "same_weekday_baseline",
                  "rank",
                  "with_quotes",
                  "alpha",
                  "alpha_outlier",
                  "k_decay"
                ) %>%
                DT::datatable(
                  colnames = c(
                    "Date" = "date",
                    "Hour" = "hour",
                    "Topic" = "topic",
                    "Region" = "country",
                    "Category" = "episomer_category",
                    "Tops" = "tops",
                    "Posts" = "number_of_posts",
                    "% from important user" = "known_ratio",
                    "Threshold" = "limit",
                    "Baseline" = "no_historic",
                    "Bonf. corr." = "bonferroni_correction",
                    "Same weekday baseline" = "same_weekday_baseline",
                    "Day rank" = "rank",
                    "With quotes" = "with_quotes",
                    "Alert FPR (alpha)" = "alpha",
                    "Outlier FPR (alpha)" = "alpha_outlier",
                    "Downweight strenght" = "k_decay"
                  ),
                  filter = "top",
                  escape = FALSE
                )
            } else {
              alerts$topposts <- sapply(alerts$topposts, function(postsbylang) {
                if (length(postsbylang) == 0) "" else {
                  paste(
                    "<UL>",
                    lapply(1:length(postsbylang), function(i) {
                      paste(
                        "<LI>",
                        names(postsbylang)[[i]],
                        "<OL>",
                        paste(
                          lapply(postsbylang[[i]], function(t) {
                            paste0(
                              "<LI>",
                              htmltools::htmlEscape(t),
                              "</LI>"
                            )
                          }),
                          collapse = ""
                        ),
                        "</OL>",
                        "</LI>"
                      )
                    }),
                    "</UL>",
                    collapse = ""
                  )
                }
              })

              dt <- alerts %>%
                dplyr::select(
                  "date",
                  "hour",
                  "topic",
                  "country",
                  "episomer_category",
                  "tops",
                  "number_of_posts",
                  "topposts"
                ) %>%
                DT::datatable(
                  colnames = c(
                    "Date" = "date",
                    "Hour" = "hour",
                    "Topic" = "topic",
                    "Region" = "country",
                    "Category" = "episomer_category",
                    "Tops" = "tops",
                    "Posts" = "number_of_posts",
                    "Top posts" = "topposts"
                  ),
                  filter = "top",
                  escape = FALSE
                )
            }
            progress_close()
            dt
          })
        })

        shiny::observeEvent(input$alerts_close, {
          shiny::updateTextInput(
            session,
            "alerts_show_search",
            label = NULL,
            value = "false"
          )
        })

        shiny::observeEvent(input$alertsdb_search, {
          shiny::updateTextInput(
            session,
            "alertsdb_show",
            label = NULL,
            value = "true"
          )
          output$alertsdb_table <- DT::renderDataTable(
            get_alertsdb_html() %>%
              DT::datatable(
                colnames = c(
                  "Date" = "date",
                  "Topic" = "topic",
                  "Region" = "country",
                  "Top words" = "topwords",
                  "Posts" = "number_of_posts",
                  "Top posts" = "topposts",
                  "Given Category" = "given_category",
                  "Episomer Category" = "episomer_category"
                ),
                filter = "top",
                escape = FALSE
              )
          )

          output$alertsdb_runs_table <- DT::renderDataTable(
            get_alertsdb_runs_html() %>%
              DT::datatable(
                colnames = c(
                  "Ranking" = "ranking",
                  "Models" = "models",
                  "Alerts" = "alerts",
                  "Runs" = "runs",
                  "F1Score" = "f1score",
                  "Accuracy" = "accuracy",
                  "Precision By Class" = "precision_by_class",
                  "Sensitivity By Class" = "sensitivity_by_class",
                  "FScore By Class" = "fscore_by_class",
                  "Last run" = "last_run",
                  "Balance classes" = "balance_classes",
                  "Force to use" = "force_to_use",
                  "Active" = "active",
                  "Documentation" = "documentation",
                  "Custom Parameters" = "custom_parameters"
                ),
                filter = "top",
                escape = FALSE
              )
          )
        })
        shiny::observeEvent(input$alertsdb_close, {
          shiny::updateTextInput(
            session,
            "alertsdb_show",
            label = NULL,
            value = "false"
          )
        })
        output$alertsdb_download <- shiny::downloadHandler(
          filename = function() "alert-training.xlsx",
          content = function(file) {
            progress_start("updating geo training dataset for download")
            #updating geotraining file before downloading
            tryCatch(
              {
                if (
                  get_alert_training_path() == get_default_alert_training_path()
                ) {
                  alerts <- get_alerts(
                    topic = input$alerts_topics,
                    countries = as.numeric(input$alerts_countries),
                    from = input$alerts_period[[1]],
                    until = input$alerts_period[[2]],
                    topposts = 10,
                    limit = if (
                      !is.na(input$alerts_limit) &&
                        as.integer(input$alerts_limit) > 0
                    )
                      as.integer(input$alerts_limit) else 20,
                    progress = function(value, message) {
                      progress_set(value = value, message = message)
                    }
                  )
                  if (!is.null(alerts)) {
                    alerts$given_category <- NA
                    write_alert_training_db(alerts)
                    cd$alert_training_refresh_flag(Sys.time())
                    progress_close()
                    file.copy(get_alert_training_path(), file)
                  } else
                    app_error(
                      "Cannot download alerts since no alerts where found with the selected filters",
                      env = cd
                    )
                } else {
                  progress_close()
                  file.copy(get_alert_training_path(), file)
                }
              },
              error = function(w) {
                app_error(w, env = cd)
              }
            )
          }
        )

        shiny::observeEvent(input$alertsdb_add, {
          progress_start("Adding new alerts to the alert database")
          #updating geotraining file before downloading
          tryCatch(
            {
              new_alerts <- get_alerts(
                topic = input$alerts_topics,
                countries = as.numeric(input$alerts_countries),
                from = input$alerts_period[[1]],
                until = input$alerts_period[[2]],
                topposts = 10,
                limit = if (
                  !is.na(input$alerts_limit) && as.integer(input$alerts_limit) > 0
                )
                  as.integer(input$alerts_limit) else 20,
                progress = function(value, message) {
                  progress_set(value = value, message = message)
                }
              )

              existing_alerts <- get_alert_training_df()
              if (!is.null(new_alerts) && nrow(new_alerts) > 0) {
                new_alerts$given_category <- NA
                alerts <- Reduce(
                  x = list(existing_alerts, new_alerts),
                  f = function(df1, df2) {
                    dplyr::bind_rows(df1, df2)
                  }
                )
                write_alert_training_db(alerts)
                cd$alert_training_refresh_flag(Sys.time())
              } else {
                app_error(
                  "Cannot add alerts to classify. No alerts where found with the selected filters",
                  env = cd
                )
              }
            },
            error = function(w) {
              app_error(w, env = cd)
            }
          )
        })

        # uploading alerts training file
        shiny::observe({
          df <- input$alertsdb_upload
          if (!is.null(df) && nrow(df) > 0) {
            progress_start("processing new training set")
            uploaded <- df$datapath[[1]]
            #copying file and making a backup of existing one
            if (file.exists(get_user_alert_training_path()))
              file.copy(
                get_user_alert_training_path(),
                paste(get_user_alert_training_path(), ".bak", sep = ""),
                overwrite = TRUE
              )
            file.copy(uploaded, get_user_alert_training_path(), overwrite = TRUE)
            #updating geotraining df taking in condideration new uploaded file

            tryCatch(
              {
                retrain_alert_classifier(progress_set)
                cd$alert_training_refresh_flag(Sys.time())
              },
              warning = function(w) {
                app_error(w, env = cd)
              },
              error = function(w) {
                app_error(w, env = cd)
              }
            )
          }
        })

        # updating alerts table on change of geotraining file
        shiny::observe({
          # Adding a dependency to alert refresh
          cd$alert_training_refresh_flag()
          DT::replaceData(DT::dataTableProxy('alertsdb_table'), get_alertsdb_html())
          DT::replaceData(
            DT::dataTableProxy('alertsdb_runs_table'),
            get_alertsdb_runs_html()
          )
          progress_close()
        })

        ######### GEOTRAINING EVALUATION DATAFRAME ####
        # rendering a dataframe showing evaluation metrics
        output$geotraining_eval_df <- DT::renderDataTable({
          df <- get_geotraining_eval_df()
          DT::datatable(df, escape = TRUE)
        })

        get_geotraining_eval_df <- function() {
          `%>%` <- dplyr::`%>%`
          ret <- data.frame(
            Category = character(),
            `True positives` = integer(),
            `False positives` = integer(),
            `True negatives` = integer(),
            `False negatives` = integer(),
            `Precision` = double(),
            `Sensitivity` = double(),
            `F1Score` = double(),
            check.names = FALSE
          )
          if (file.exists(get_geotraining_evaluation_path())) {
            df <- jsonlite::fromJSON(
              get_geotraining_evaluation_path(),
              flatten = TRUE
            )
            df <- as.data.frame(df)
            if (nrow(df) > 0) {
              ret <- df %>%
                dplyr::select(
                  .data$test,
                  .data$tp,
                  .data$tn,
                  .data$fp,
                  .data$fn
                ) %>%
                dplyr::group_by(.data$test) %>%
                dplyr::summarise(
                  tp = sum(.data$tp),
                  fp = sum(.data$fp),
                  tn = sum(.data$tn),
                  fn = sum(.data$fn)
                ) %>%
                janitor::adorn_totals("row") %>%
                dplyr::mutate(
                  Precision = round(.data$tp / (.data$tp + .data$fp), 3)
                ) %>%
                dplyr::mutate(
                  Sensitivity = round(.data$tp / (.data$tp + .data$fn), 3)
                ) %>%
                dplyr::mutate(
                  F1Score = round(
                    (2 * .data$tp) / (2 * .data$tp + .data$fn + .data$fp),
                    3
                  )
                ) %>%
                dplyr::rename(Category = .data$test) %>%
                dplyr::rename("True positives" = .data$tp) %>%
                dplyr::rename("False positives" = .data$fp) %>%
                dplyr::rename("True negatives" = .data$tn) %>%
                dplyr::rename("False negatives" = .data$fn)
            }
          }
          ret
        }

        ######### GEOTEST LOGIC ##################
        # rendering geotest data, updates are done automatically whenever an input changes
        output$geotraining_table <- DT::renderDataTable({
          `%>%` <- magrittr::`%>%`
          df <- get_geotraining_df()
          df %>%
            DT::datatable(
              filter = "top",
              escape = TRUE
            )
        })
        # uploading geotagging training file
        shiny::observe({
          df <- input$geotraining_upload
          if (!is.null(df) && nrow(df) > 0) {
            progress_start("processing new training set")
            uploaded <- df$datapath[[1]]
            #copying file and making a backup of existing one
            if (file.exists(get_user_geotraining_path()))
              file.copy(
                get_user_geotraining_path(),
                paste(get_user_geotraining_path(), ".bak", sep = ""),
                overwrite = TRUE
              )
            file.copy(uploaded, get_user_geotraining_path(), overwrite = TRUE)
            #updating geotraining df taking in condideration new uploaded file

            tryCatch(
              {
                #update_geotraining_df(input$geotraining_posts2add, progress = function(value, message) {progress_set(value = value /3, message = message)})
                progress_set(
                  value = 0.5,
                  message = "Retraining models and evaluating"
                )
                retrain_languages()
                update_geotraining_df(
                  input$geotraining_posts2add,
                  progress = function(value, message) {
                    progress_set(value = 0.5 + value / 2, message = message)
                  }
                )
                cd$geotraining_refresh_flag(Sys.time())
                #TO DO : afficher les résultats + observe qui met à jour si de nouveelles données sont uploadées
              },
              error = function(w) {
                app_error(w, env = cd)
              }
            )
          }
        })
        # Updating geolocation data
        shiny::observeEvent(input$geotraining_update, {
          progress_start("updating geo training dataset")
          #updating geotraining df taking in condideration new uploaded file
          tryCatch(
            {
              update_geotraining_df(
                input$geotraining_posts2add,
                progress = function(value, message) {
                  progress_set(value = value, message = message)
                }
              )
              cd$geotraining_refresh_flag(Sys.time())
            },
            error = function(w) {
              app_error(w, env = cd)
            }
          )
        })

        # download geotraining data
        output$geotraining_download <- shiny::downloadHandler(
          filename = function() "geo-training.xlsx",
          content = function(file) {
            progress_start("updating geo training dataset for download")
            #updating geotraining file before downloading
            tryCatch(
              {
                if (get_geotraining_path() == get_default_geotraining_path()) {
                  update_geotraining_df(
                    input$geotraining_posts2add,
                    progress = function(value, message) {
                      progress_set(value = value, message = message)
                    }
                  )
                }
                file.copy(get_geotraining_path(), file)
                cd$geotraining_refresh_flag(Sys.time())
              },
              error = function(w) {
                app_error(w, env = cd)
              }
            )
          }
        )

        # updating geotraining table on change of geotraining file
        shiny::observe({
          # Adding a dependency to subscribers refresh
          cd$geotraining_refresh_flag()
          DT::replaceData(
            DT::dataTableProxy('geotraining_table'),
            get_geotraining_df()
          )
          DT::replaceData(
            DT::dataTableProxy('geotraining_eval_df'),
            get_geotraining_eval_df()
          )
          progress_close()
        })

        ######### DATA PROTECTION LOGIC ##################
        # rendering the data protection search table
        shiny::observeEvent(input$data_anonymise, {
          shiny::showModal(shiny::modalDialog(
            title = "Warning",
            "Please confirm you want to anonymise the posts matching this search criteria, this action cannot be undone",
            footer = shiny::tagList(
              shiny::actionButton("data_perform_anonymise", "Yes, anonymise posts"),
              shiny::modalButton("Cancel")
            )
          ))
        })

        shiny::observeEvent(input$data_delete, {
          shiny::showModal(shiny::modalDialog(
            title = "Warning",
            "Please confirm you want to delete the posts matching this search criteria, this action cannot be undone",
            footer = shiny::tagList(
              shiny::actionButton("data_perform_delete", "Yes, delete posts"),
              shiny::modalButton("Cancel")
            )
          ))
        })

        shiny::observeEvent(input$data_perform_delete, {
          search_posts_exp(hide_users = FALSE, action = "delete")
        })
        shiny::observeEvent(input$data_perform_anonymise, {
          search_posts_exp(hide_users = FALSE, action = "anonymise")
        })

        shiny::observeEvent(input$data_search_ano, {
          search_posts_exp(hide_users = TRUE)
        })

        shiny::observeEvent(input$data_search, {
          search_posts_exp(hide_users = FALSE)
        })
        search_posts_exp <- function(hide_users, action = NULL) {
          output$data_message <- shiny::renderText("")
          output$data_search_df <- DT::renderDataTable({
            `%>%` <- magrittr::`%>%`
            shiny::removeModal()
            query <- (if (!is.null(action) && action == "anonymise")
              "created_at:[0 TO Z] NOT user_name:user" else NULL)
            progress_start("Searching")
            shiny::isolate({
              posts <- search_posts(
                query = query,
                topic = input$data_topics,
                from = input$data_period[[1]],
                to = input$data_period[[2]],
                countries = as.numeric(input$data_countries),
                mentioning = if (
                  input$data_mode %in%
                    c("all", "mentioning") &&
                    input$data_users != ""
                )
                  strsplit(input$data_users, "\\s+")[[1]] else NULL,
                users = if (
                  input$data_mode %in% c("all", "users") && input$data_users != ""
                )
                  strsplit(input$data_users, "\\s+")[[1]] else NULL,
                hide_users = hide_users,
                action = action,
                max = as.integer(input$data_limit)
              )
            })

            if (!is.null(action) && nrow(posts) > 0) {
              processed <- 0
              while (nrow(posts) > 0) {
                processed <- processed + nrow(posts)
                progress_set(
                  value = 0.5,
                  message = paste(
                    "Performing",
                    action,
                    processed,
                    "posts processed"
                  )
                )

                shiny::isolate({
                  posts <- search_posts(
                    query = query,
                    topic = input$data_topics,
                    from = input$data_period[[1]],
                    to = input$data_period[[2]],
                    countries = as.numeric(input$data_countries),
                    mentioning = if (
                      input$data_mode %in%
                        c("all", "mentioning") &&
                        input$data_users != ""
                    )
                      strsplit(input$data_users, "\\s+")[[1]] else NULL,
                    users = if (
                      input$data_mode %in%
                        c("all", "users") &&
                        input$data_users != ""
                    )
                      strsplit(input$data_users, "\\s+")[[1]] else NULL,
                    hide_users = hide_users,
                    action = action,
                    max = as.integer(input$data_limit)
                  )
                })
              }
              progress_close()
            }

            shiny::validate(
              shiny::need(
                !is.null(posts) && nrow(posts) > 0,
                'No posts found for the provided filters'
              )
            )
            output$data_message <- if (posts$totalCount[[1]] > 1000) {
              shiny::renderText(paste(
                "more than ",
                posts$totalCount[[1]],
                " posts found"
              ))
            } else {
              shiny::renderText(paste(posts$totalCount[[1]], " posts found"))
            }

            dt <- {
              posts$country_code <- if ("text_loc" %in% colnames(posts))
                posts$text_loc$geo_country_code else ""
              posts$geo_name <- if ("text_loc" %in% colnames(posts))
                posts$text_loc$geo_name else ""
              cols <- c("id","topic","country_code","geo_name","created_at","text","user_name","quoted_text")
	      for(c in cols) {
	        if(!(c %in% names(posts)))
			posts[c] <- ""
	      }
	      posts %>%
                dplyr::select(cols) %>%
                DT::datatable(
                  colnames = c(
                    "Post Id" = "id",
                    "Topic" = "topic",
                    "Created" = "created_at",
                    "Country in text" = "country_code",
                    "Location in text" = "geo_name",
                    "User" = "user_name",
                    "Text" = "text",
                    "Repost/Quoted text" = "quoted_text"
                  ),
                  filter = "top",
                  escape = TRUE
                )
            }
            progress_close()
            dt
          })
        }

        ######## SNAPSHOT LOGIC ############
        # creating an snapshot of episomer data for backup or debuging purposes
        shiny::observeEvent(input$build_snapshot, {
          `%>%` <- magrittr::`%>%`
          progress_start("Creating snapshot")
          shiny::isolate({
            create_snapshot(
              destination_dir = input$snapshot_folder,
              types = input$snapshot_types,
              posts_period = input$snapshot_post_period,
              aggregated_period = input$snapshot_aggr_period,
              compress = input$snapshot_compress,
              progress = function(value, message) {
                progress_set(value = value, message = message)
              }
            )
          })
          progress_close()
        })

        ######## DIAGNOSTIC LOGIC ############
        # running the diagnostics when the button is clicked
        shiny::observeEvent(input$run_diagnostic, {
          `%>%` <- magrittr::`%>%`

          output$diagnostic_table <- DT::renderDataTable(
            check_all() %>%
              DT::datatable(
                colnames = c(
                  "Check Code" = "check",
                  "Passed" = "passed",
                  "Message" = "message"
                ),
                filter = "top",
                escape = TRUE,
                rownames = FALSE,
                options = list(pageLength = 50)
              ) %>%
              DT::formatStyle(
                columns = c("Passed"),
                valueColumns = c("Passed"),
                color = DT::styleEqual(c(TRUE, FALSE), c("green", "red"))
              )
          )
        })
      }
  }

  # Functions for managing calculation progress
  progress_start <- function(start_message = "processing", env = cd) {
    progress_close(env = env)
    env$progress <- shiny::Progress$new()
    env$progress$set(message = start_message, value = 0)
  }

  progress_close <- function(env = cd) {
    if (exists("progress", where = env) && !is.null(env$progress)) {
      env$progress$close()
      rm("progress", envir = env)
    }
  }

  progress_inc <- function(value, message = "processing", env = cd) {
    if (!is.null(env$progress$set))
      env$progress$set(
        value = env$progress$getValue() + value,
        detail = message
      )
  }

  progress_set <- function(value, message = "processing", env = cd) {
    if (!is.null(env$progress$set))
      env$progress$set(value = value, detail = message)
  }

  app_error <- function(e, env = cd) {
    message("000000000000000000000000000")
    message(class(e))
    message(paste(e))
    message("--------------------------------")

    message(e)
    message("11111111111111111111111111")
    progress_close(env = env)
    message("2222222222222222222222222222")
    shiny::showNotification(paste(e), type = "error")
    message("33333333333333333333333333333")
  }

  # Printing PID
  message(Sys.getpid())
  # Launching the app
  old <- options()
  on.exit(options(old))
  shiny::shinyApp(
    ui = ui_app(profile),
    server = function(input, output, session)
      server_app(input, output, session, profile),
    options = options(
      shiny.fullstacktrace = TRUE
    )
  )
}


#' @rdname episomer_app
#' @param data_dir Path to the 'data directory' containing application settings, models and collected posts.
#' If not provided the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, default: NA
#' @param host The IP if the network interface to run the Shiny app on, default: NULL (will run on 127.0.0.1)
#' @param port The port to run the Shiny app on, default: NULL (will run on a random port)
#' @return The Shiny server object containing the launched application
#' @details The dashboard app is the user entry point to the episomer package for regular users. This application shows the interactive dashboard with visualisations, 
#' allow export them to Markdown or PDF.
#'
#' All its functionality is described on the episomer vignette.
#' @export
dashboard_app <- function(
  data_dir = NA,
  host = NULL,
  port = NULL
) {
  episomer_app(
    data_dir = data_dir,
    profile = "dashboard",
    host = host,
    port = port
  )
}

#' @rdname episomer_app
#' @param data_dir Path to the 'data directory' containing application settings, models and collected posts.
#' If not provided the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, default: NA
#' @param host The IP if the network interface to run the Shiny app on, default: NULL (will run on 127.0.0.1)
#' @param port The port to run the Shiny app on, default: NULL (will run on a random port)
#' @return The Shiny server object containing the launched application
#' @details The admin app is contain all functionality for an administrator. It help the user to setup the post collection process, manage all settings,
#' and setup the alert emails.
#'
#' All its functionality is described on the episomer vignette.
#' @export
admin_app <- function(data_dir = NA, host = NULL, port = NULL) {
  episomer_app(
    data_dir = data_dir,
    profile = "admin",
    host = host,
    port = port
  )
}

# Get or updates data used for dashboard filters based on configuration values and aggregated period
refresh_dashboard_data <- function(e = new.env(), fixed_period = NULL) {
  e$topics <- {
    codes <- unique(sapply(conf$topics, function(t) t$topic))
    names <- get_topics_labels()[codes]
    sort(setNames(c("", codes), c("", unname(names))))
  }
  e$countries <- {
    regions <- get_country_items()
    setNames(1:length(regions), sapply(regions, function(r) r$name))
  }
  e$sms <- {
    sms <- active_social_media()
    setNames(sms, paste0(toupper(substr(sms, 1, 1)), substr(sms, 2, nchar(sms))))
  }
  agg_dates <- get_aggregated_period()
  if (is.na(agg_dates$first) || is.na(agg_dates$last)) {
    agg_dates$first = Sys.Date()
    agg_dates$last = Sys.Date()
  }
  e$date_min <- strftime(agg_dates$first, format = "%Y-%m-%d")
  e$date_max <- strftime(agg_dates$last, format = "%Y-%m-%d")
  collected_days <- agg_dates$last - agg_dates$first
  e$fixed_period <- (if (!is.null(fixed_period)) fixed_period else if (
    is.na(collected_days)
  )
    "custom" else if (collected_days < 7) "custom" else "last 7 days")
  e$date_start <- (if (e$fixed_period == "custom" && exists("date_start", e))
    e$date_start else if (e$fixed_period == "last 7 days" && collected_days > 7)
    strftime(agg_dates$last - 7, format = "%Y-%m-%d") else if (
    e$fixed_period == "last 30 days" && collected_days > 30
  )
    strftime(agg_dates$last - 30, format = "%Y-%m-%d") else if (
    e$fixed_period == "last 60 days" && collected_days > 60
  )
    strftime(agg_dates$last - 60, format = "%Y-%m-%d") else if (
    e$fixed_period == "last 180 days" && collected_days > 180
  )
    strftime(agg_dates$last - 180, format = "%Y-%m-%d") else e$date_min)
  e$date_end <- e$date_max
  
  if (!exists("topics_refresh_flag", envir = e)) {
    e$topics_refresh_flag <- shiny::reactiveVal(0)
  }
  return(e)
}


# Get or update data for config page from configuration, Data collection & processing and Requirements & alerts pipeline files
refresh_config_data <- function(
  e = new.env(),
  limit = list("langs", "topics", "tasks", "geo")
) {
  # Refreshing configuration
  setup_config(data_dir = conf$data_dir, ignore_properties = TRUE)

  #Creating the flag for properties refresh
  if (!exists("properties_refresh_flag", where = e)) {
    e$properties_refresh_flag <- shiny::reactiveVal()
  }
  #Creating the flag for status refredh
  if (!exists("process_refresh_flag", where = e)) {
    e$process_refresh_flag <- shiny::reactiveVal()
  }
  #Creating the flag for subscribers refresh
  if (!exists("subscribers_refresh_flag", where = e)) {
    e$subscribers_refresh_flag <- shiny::reactiveVal()
  }
  #Creating the flag for geotraining refresh
  if (!exists("geotraining_refresh_flag", where = e)) {
    e$geotraining_refresh_flag <- shiny::reactiveVal()
  }
  #Creating the flag for alert training refresh
  if (!exists("alert_training_refresh_flag", where = e)) {
    e$alert_training_refresh_flag <- shiny::reactiveVal()
  }
  #Creating the flag for countries refresh
  if (!exists("countries_refresh_flag", where = e)) {
    e$countries_refresh_flag <- shiny::reactiveVal()
  }

  # Updating language related fields
  if ("langs" %in% limit) {
    langs <- get_available_languages()
    lang_tasks <- get_tasks()$language
    if (!exists("langs_refresh_flag", where = e)) {
      e$langs_refresh_flag <- shiny::reactiveVal(0)
    } else if (
      file.exists(get_tasks_path()) &&
        file.info(get_tasks_path())$mtime > e$langs_refresh_flag()
    ) {
      e$langs_refresh_flag(file.info(get_tasks_path())$mtime)
    } else if (
      file.exists(get_properties_path()) &&
        file.info(get_properties_path())$mtime > e$langs_refresh_flag()
    ) {
      e$langs_refresh_flag(file.info(get_properties_path())$mtime)
    }
    e$lang_items <- setNames(langs$Code, langs$`Full Label`)
    e$lang_names <- setNames(langs$Label, langs$Code)
    e$langs <- data.frame(
      Language = unlist(lang_tasks$names),
      Code = unlist(lang_tasks$codes),
      Status = unlist(lang_tasks$statuses),
      URL = unlist(lang_tasks$urls)
    )
  }
  # Updating topics related fields
  if ("topics" %in% limit) {
    # Updating the reactive value tasks_refresh to force dependencies invalidation
    update <- FALSE
    if (
      !exists("topics_refresh_flag", envir = e) ||
        !exists("plans_refresh_flag", envir = e)
    ) {
      e$topics_refresh_flag <- shiny::reactiveVal(0)
      e$plans_refresh_flag <- shiny::reactiveVal(0)
      update <- TRUE
    }
    if (
      !update &&
        file.exists(get_user_topics_path()) &&
        file.info(get_user_topics_path())$mtime != e$topics_refresh_flag()
    ) {
      update <- TRUE
      e$topics_refresh_flag(file.info(get_user_topics_path())$mtime)
    }
    if (
      !update &&
        file.exists(get_plans_path()) &&
        file.info(get_plans_path())$mtime != e$plans_refresh_flag()
    ) {
      update <- TRUE
      e$plans_refresh_flag(file.info(get_plans_path())$mtime)
    }
    if (update) {
      e$topics_df <- get_topics_df()
    }
  }
  # Updating tasks related fields
  if ("tasks" %in% limit) {
    # Updating the reactive value tasks_refresh to force dependencies invalidation
    update <- FALSE
    e$detect_running <- is_detect_running()
    e$fs_running <- is_fs_running()
    e$search_running <- is_search_running()
    e$missing_search_jobs <- missing_search_jobs()
    e$search_diff <- round(Sys.time() - last_search_time())

    # Detecting if some change has happened since last evaluation
    if (!exists("tasks_refresh_flag", where = e)) {
      e$tasks_refresh_flag <- shiny::reactiveVal()
      update <- TRUE
    } else if (
      file.exists(get_tasks_path()) &&
        file.info(get_tasks_path())$mtime != e$tasks_refresh_flag()
    ) {
      update <- TRUE
    }
    if (update) {
      if (file.exists(get_tasks_path()))
        e$tasks_refresh_flag(file.info(get_tasks_path())$mtime)
      tasks <- get_tasks()
      sorted_tasks <- order(sapply(tasks, function(l) l$order))
      e$tasks <- tasks[sorted_tasks]
      e$tasks_df <- data.frame(
        Task = sapply(e$tasks, function(t) t$task),
        Status = sapply(
          e$tasks,
          function(t) if (in_pending_status(t)) "pending" else t$status
        ),
        Scheduled = sapply(
          e$tasks,
          function(t)
            strftime(
              t$scheduled_for,
              format = "%Y-%m-%d %H:%M:%OS",
              origin = '1970-01-01'
            )
        ),
        `Last Start` = sapply(
          e$tasks,
          function(t)
            strftime(
              t$started_on,
              format = "%Y-%m-%d %H:%M:%OS",
              origin = '1970-01-01'
            )
        ),
        `Last End` = sapply(
          e$tasks,
          function(t)
            strftime(
              t$end_on,
              format = "%Y-%m-%d %H:%M:%OS",
              origin = '1970-01-01'
            )
        ),
        Message = sapply(
          e$tasks,
          function(t)
            if (exists("message", where = t) && !is.null(t$message))
              t$message else ""
        )
      )
      row.names(e$tasks_df) <- sapply(e$tasks, function(t) t$order)
    }
  }
  # Updating geolocation test related fields
  if ("geo" %in% limit) {
  }
  return(e)
}

# validate that dashboard can be rendered
can_render <- function(input, d) {
  shiny::validate(
    shiny::need(
      is_fs_running(),
      'Embedded database service is not running, please make sure you have activated it (running update dependencies is necessary after upgrade from episomer version 0.1+ to episomer 1.0+)'
    ),
    shiny::need(
      file.exists(conf$data_dir),
      'Please go to configuration tab and setup post collection (no data directory found)'
    ),
    shiny::need(
      check_series_present(),
      paste(
        'No aggregated data found on ',
        paste(conf$data_dir, "series", sep = "/"),
        " please make sure the Requirements & alerts pipeline has successfully ran"
      )
    ),
    shiny::need(
      is.null(input) ||
        (!is.na(input$period[[1]]) &&
          !is.na(input$period[[2]]) &&
          (input$period[[1]] <= input$period[[2]])),
      'Please select a start and end period for the report. The start period must be a date earlier than the end period'
    ),
    shiny::need(is.null(input) || (input$topics != ''), 'Please select a topic')
  )
}

# validate that chart is not empty
chart_not_empty <- function(chart) {
  shiny::validate(
    shiny::need(!("waiver" %in% class(chart$data)), chart$labels$title)
  )
}


get_alertsdb_html <- function() {
  alerts <- get_alert_training_df()
  shiny::validate(
    shiny::need(!is.null(alerts), 'No alerts generated for the selected period')
  )
  alerts$topposts <- sapply(alerts$topposts, function(postsbylang) {
    if (length(postsbylang) == 0) "" else {
      paste(
        "<UL>",
        lapply(1:length(postsbylang), function(i) {
          paste(
            "<LI>",
            names(postsbylang)[[i]],
            "<OL>",
            paste(
              lapply(postsbylang[[i]], function(t) {
                paste0(
                  "<LI>",
                  htmltools::htmlEscape(t),
                  "</LI>"
                )
              }),
              collapse = ""
            ),
            "</OL>",
            "</LI>"
          )
        }),
        "</UL>",
        collapse = ""
      )
    }
  })
  alerts
}

get_alertsdb_runs_html <- function() {
  runs <- get_alert_training_runs_df()
  runs$f1score <- format(runs$f1score, digits = 3)
  runs$accuracy <- format(runs$accuracy, digits = 3)
  runs$precision_by_class <- gsub("\n", "<BR>", runs$precision_by_class)
  runs$sensitivity_by_class <- gsub("\n", "<BR>", runs$sensitivity_by_class)
  runs$fscore_by_class <- gsub("\n", "<BR>", runs$fscore_by_class)
  runs$custom_parameters <- sapply(runs$custom_parameters, function(params) {
    if (length(params) == 0) "" else {
      paste(
        "<UL>",
        lapply(1:length(params), function(i) {
          paste(
            "<LI>",
            names(params)[[i]],
            ":",
            params[[i]],
            "</LI>"
          )
        }),
        "</UL>",
        collapse = ""
      )
    }
  })
  runs
}

# getting the default snapshot folder and create it if does not exists
ensure_snapshot_folder <- function() {
  path <- file.path(conf$data_dir, "snapshot")
  if (!dir.exists(path)) dir.create(path, recursive = TRUE)
  path
}

update_config_from_input <- function(input) {
  conf$collect_span <- input$conf_collect_span
  conf$schedule_span <- input$conf_schedule_span
  conf$keyring <- input$conf_keyring
  conf$spark_cores <- input$conf_spark_cores
  conf$spark_memory <- input$conf_spark_memory
  conf$onthefly_api <- input$conf_onthefly_api
  conf$geolocation_threshold <- input$geolocation_threshold
  conf$geonames_url <- input$conf_geonames_url
  conf$maven_repo <- input$conf_maven_repo
  conf$winutils_url <- input$conf_winutils_url
  conf$geonames_simplify <- input$conf_geonames_simplify
  conf$regions_disclaimer <- input$conf_regions_disclaimer
  conf$alert_alpha <- input$conf_alpha
  conf$alert_alpha_outlier <- input$conf_alpha_outlier
  conf$alert_k_decay <- input$conf_k_decay
  conf$alert_history <- input$conf_history
  conf$alert_same_weekday_baseline <- input$conf_same_weekday_baseline
  conf$alert_with_bonferroni_correction <- input$conf_with_bonferroni_correction
  conf$alert_with_quotes <- input$conf_with_quotes
  conf$smtp_host <- input$conf_smtp_host
  conf$smtp_port <- input$conf_smtp_port
  conf$smtp_from <- input$conf_smtp_from
  conf$smtp_login <- input$conf_smtp_login
  conf$smtp_insecure <- input$conf_smtp_insecure
  conf$smtp_password <- input$conf_smtp_password
  conf$admin_email <- input$conf_admin_email
  conf$force_date_format <- input$conf_force_date_format
  conf$sm_alerts_bluesky <- input$conf_sm_alerts_bluesky
  conf$sm_activated_bluesky <- input$conf_sm_activated_bluesky
}

# Environment for storing configuration
conf <- new.env()

# Get package name
get_package_name <- function() environmentName(environment(setup_config))

# Get the keyring for the provided backend or a platform dependent default if backend is null
# this function is used to save the social media and smtp credentials on a secure store when saving properties
# this function uses the keyring R package which provides a common interface for accessing system dependent keyring
# backend: backend name to use it will be guessed based on user OS if not defined.
get_key_ring <- function(backend = NULL) {
  if (!is.null(backend)) {
    Sys.setenv(kr_backend = backend)
  }
  Sys.setenv(kr_name = "ecdc_wtitter_tool_kr_name")
  Sys.setenv(kr_service = "kr_service")
  keyring = Sys.getenv("kr_backend")
  if (keyring == "secret_service") {
    kb <- keyring::backend_secret_service$new()
  } else if (keyring == "wincred") {
    kb <- keyring::backend_wincred$new()
  } else if (keyring == "macos") {
    kb <- keyring::backend_macos$new()
  } else if (keyring == "file") {
    kb <- keyring::backend_file$new()
  } else {
    kb <- keyring::backend_env$new()
  }
  kr_name <- NULL
  # creating the file keyring if requested and not existent
  if (keyring == "file") {
    kr_name <- Sys.getenv("kr_name")
    krs <- kb$keyring_list()
    if (nrow(krs[krs$keyring == kr_name, ]) == 0) {
      kb$keyring_create(kr_name)
    }
    kb$keyring_set_default(kr_name)
  }

  # unlocking keyring if unlocked using environment variable to get the password
  if (
    kb$keyring_is_locked(keyring = kr_name) &&
      Sys.getenv("ecdc_wtitter_tool_kr_password") != ""
  ) {
    kb$keyring_unlock(
      keyring = kr_name,
      password = Sys.getenv("ecdc_wtitter_tool_kr_password")
    )
    # unlocking keyring if unlocked using password prompt
  } else if (kb$keyring_is_locked(keyring = kr_name)) {
    kb$keyring_unlock(keyring = kr_name)
  }
  return(kb)
}

# Set a secret from the chosen secret management backend
set_secret <- function(secret, value) {
  get_key_ring()$set_with_value(
    service = Sys.getenv("kr_service"),
    username = secret,
    password = value
  )
}

# Checks whether a secret is set on the secret management backend
is_secret_set <- function(secret) {
  secrets <- get_key_ring()$list(service = Sys.getenv("kr_service"))
  return(nrow(secrets[secrets$username == secret, ]) > 0)
}

# Get a secret from the chosen secret management backend
get_secret <- function(secret) {
  get_key_ring()$get(service = Sys.getenv("kr_service"), username = secret)
}

# get empty config for initialization, this represents default values for all configuration properties
get_empty_config <- function(data_dir) {
  ret <- list()
  ret$keyring <-
    if (.Platform$OS.type == "windows") "wincred" else if (
      Sys.info()[['sysname']] == "Darwin"
    )
      "macos" else "file"
  ret$data_dir <- data_dir
  ret$collect_span <- 60
  ret$schedule_span <- 90
  ret$schedule_start_hour <- 8
  ret$languages <- list(
    list(
      code = "en",
      name = "English",
      vectors = paste(ret$data_dir, "languages/en.txt.gz", sep = "/"),
      modified_on = strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
    ),
    list(
      code = "fr",
      name = "French",
      vectors = paste(ret$data_dir, "languages/fr.txt.gz", sep = "/"),
      modified_on = strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
    ),
    list(
      code = "es",
      name = "Spanish",
      vectors = paste(ret$data_dir, "languages/es.txt.gz", sep = "/"),
      modified_on = strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
    ),
    list(
      code = "pt",
      name = "Portuguese",
      vectors = paste(ret$data_dir, "languages/pt.txt.gz", sep = "/"),
      modified_on = strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
    )
  )
  ret$lang_updated_on <- NA
  ret$geonames_updated_on <- NA
  ret$dep_updated_on <- NA
  ret$geotag_requested_on <- NA
  ret$aggregate_requested_on <- NA
  ret$alerts_requested_on <- NA
  ret$geonames_url <- "http://download.geonames.org/export/dump/allCountries.zip"
  ret$geolocation_threshold <- 5
  ret$known_users <- list()
  ret$spark_cores <- {
    all_cores <- floor(
      parallel::detectCores(all.tests = FALSE, logical = TRUE) * 0.3
    )
    if (is.na(all_cores)) 1 else if (all_cores < 1) 1 else all_cores
  }
  ret$spark_memory <- "4g"
  ret$onthefly_api <- .Platform$OS.type != "windows"
  ret$topics <- list()
  ret$topics_md5 <- ""
  ret$alert_alpha <- 0.025
  ret$alert_alpha_outlier <- 0.05
  ret$alert_k_decay <- 4
  ret$alert_history <- 7
  ret$alert_same_weekday_baseline <- FALSE
  ret$alert_with_quotes <- FALSE
  ret$alert_with_bonferroni_correction <- TRUE
  ret$use_mkl <- FALSE
  ret$geonames_simplify <- TRUE
  ret$regions_disclaimer <- ""
  ret$smtp_host <- ""
  ret$smtp_port <- 25
  ret$smtp_from <- ""
  ret$smtp_login <- ""
  ret$smtp_password <- ""
  ret$smtp_insecure <- FALSE
  ret$force_date_format <- ""
  ret$maven_repo <- "https://repo1.maven.org/maven2"
  ret$winutils_url <- "https://github.com/steveloughran/winutils/raw/master/hadoop-3.0.0/bin/winutils.exe"
  ret$fs_port <- 8080
  ret$fs_batch_timeout <- 60 * 60
  ret$fs_query_timeout <- 60
  ret$admin_email <- ""
  ret$sm_alerts_bluesky <- TRUE
  ret$sm_activated_bluesky <- TRUE
  ret$dismiss_past_request <- "1971-01-01 00:00:00"
  ret$dismiss_past_done <- "2000-01-01 00:00:00"
  ret$collect_history_days <-30
  return(ret)
}

#' @title Load episomer application settings
#' @description Load episomer application settings from the designated data directory
#' @param data_dir Path to the directory containing the application settings (it must exist).
#' If not provided it takes the value of the latest call to setup_config in the current session, or the value of the EPI_HOME environment variable or episomer subdirectory in the working directory,
#' default: if (exists("data_dir", where = conf)) conf$data_dir else if (Sys.getenv("EPI_HOME") !=
#'    "") Sys.getenv("EPI_HOME") else file.path(tempdir(), "episomer")
#' @param ignore_keyring Whether to skip loading settings from the keyring (social media and SMTP credentials), default: FALSE
#' @param ignore_properties Whether to skip loading settings managed by the Shiny app in properties.json file, Default: FALSE
#' @param ignore_topics Whether to skip loading settings defined in the topics.xlsx file and download plans from topics.json file, default: FALSE
#' @param save_properties_first Whether to save current settings before loading new ones from disk, default: FALSE
#' @param save_topics_first List of topics to save before loading new ones from disk, default: list()
#' @return Nothing
#' @details episomer relies on settings and data stored in a system folder, so before loading the dashboard, collecting posts or detecting alerts the user has to designate this folder.
#' When a user wants to use episomer from the R console they will need to call this function for initialisation.
#' The 'data_folder' can also be given as a parameter for program launch functions \code{\link{episomer_app}}, \code{\link{search_loop}} or \code{\link{detect_loop}}, which will internally call this function.
#'
#' This call will fill (or refresh) a package scoped environment 'conf' that will store the settings. Settings stored in conf are:
#' \itemize{
#'   \item{General properties of the Shiny app (stored in properties.json)}
#'   \item{Download plans from the social media collection process (stored in topics.json merged with data from the topics.xlsx file}
#'   \item{Credentials for social media APIs and SMTP stored in the defined keyring}
#' }
#'
#' When calling this function and the keyring is locked, a password will be prompted to unlock the keyring.
#' This behaviour can be changed by setting the environment variable 'ecdc_social_tool_kr_password' with the password.
#'
#' Changes made to conf can be stored permanently (except for 'data_dir') using:
#' \itemize{
#'   \item{\code{\link{save_config}}, or}
#'    \item{\code{\link{sm_api_set_auth_bluesky}}}
#' }
#' @examples
#' \dontrun{
#'    library(episomer)
#'    #loading system settings
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#' }
#' @seealso
#' \code{\link{save_config}}
#' \code{\link{sm_api_set_auth_bluesky}}
#' \code{\link{episomer_app}}
#' \code{\link{search_loop}}
#' \code{\link{detect_loop}}
#' @rdname setup_config
#' @export
#' @importFrom jsonlite read_json
#' @importFrom tools md5sum
#' @importFrom readxl read_excel
setup_config <- function(
  data_dir = if (exists("data_dir", where = conf)) conf$data_dir else if (
    Sys.getenv("EPI_HOME") != ""
  )
    Sys.getenv("EPI_HOME") else file.path(tempdir(), "episomer"),
  ignore_keyring = FALSE,
  ignore_properties = FALSE,
  ignore_topics = FALSE,
  save_properties_first = FALSE,
  save_topics_first = list()
) {
  #setting the data_dir which is a read only property
  conf$data_dir <- data_dir

  # paths contains two files storing configuration data:
  # props which contains properties set on the Shiny App is stored on data_dir/properties.json
  # and data_dir/topics.json which stores search progress and is updated by the search loop
  props_path = get_properties_path()

  #topics_path is the path to the excel file containing the topics provided by the user or episomer default ones
  user_topics_path <- get_user_topics_path(data_dir)

  # save_first may be used by a function which is responsible for a part of the configuration to save changes on its perimeter before refreshing
  if (save_properties_first || length(save_topics_first) > 0) {
    save_config(
      data_dir = data_dir,
      properties = save_properties_first,
      sm_topics = save_topics_first
    )
  }
  #Loading last created configuration from json file on temp variable if exists or load default empty conf instead, this will ensure new settings are loaded with default values if missing
  temp <- get_empty_config(data_dir)

  if (!ignore_properties) {
    # refreshing properties (if requested)
    if (file.exists(props_path)) {
      #merging default values with those stored in the properties.json file
      temp = merge_configs(list(
        temp,
        jsonlite::read_json(
          props_path,
          simplifyVector = FALSE,
          auto_unbox = TRUE
        )
      ))
    }
    #Setting config  variables filled only from json file
    conf$keyring <- temp$keyring
    conf$collect_span <- temp$collect_span
    conf$schedule_span <- temp$schedule_span
    conf$schedule_start_hour <- temp$schedule_start_hour
    conf$languages <- temp$languages
    for (i in 1:length(conf$languages)) {
      conf$languages[[i]]$vectors <- file.path(
        conf$data_dir,
        "languages",
        paste(conf$languages[[i]]$code, "txt", "gz", sep = ".")
      )
    }
    conf$lang_updated_on <- temp$lang_updated_on
    conf$geonames_updated_on <- temp$geonames_updated_on
    conf$dep_updated_on <- temp$dep_updated_on
    conf$geotag_requested_on <- temp$geotag_requested_on
    conf$aggregate_requested_on <- temp$aggregate_requested_on
    conf$alerts_requested_on <- temp$alerts_requested_on
    conf$geonames_url <- temp$geonames_url
    conf$known_users <- temp$known_users
    conf$spark_cores <- temp$spark_cores
    conf$spark_memory <- temp$spark_memory
    conf$onthefly_api <- temp$onthefly_api
    conf$geolocation_threshold <- temp$geolocation_threshold
    conf$alert_alpha <- temp$alert_alpha
    conf$alert_alpha_outlier <- temp$alert_alpha_outlier
    conf$alert_k_decay <- temp$alert_k_decay
    conf$alert_history <- temp$alert_history
    conf$alert_same_weekday_baseline <- temp$alert_same_weekday_baseline
    conf$alert_with_quotes <- temp$alert_with_quotes
    conf$alert_with_bonferroni_correction <- temp$alert_with_bonferroni_correction
    conf$use_mkl <- temp$use_mkl
    conf$geonames_simplify <- temp$geonames_simplify
    conf$regions_disclaimer <- temp$regions_disclaimer
    conf$smtp_host <- temp$smtp_host
    conf$smtp_port <- temp$smtp_port
    conf$smtp_from <- temp$smtp_from
    conf$smtp_login <- temp$smtp_login
    conf$smtp_insecure <- temp$smtp_insecure
    conf$force_date_format <- temp$force_date_format
    conf$maven_repo <- temp$maven_repo
    conf$winutils_url <- temp$winutils_url
    conf$fs_port <- temp$fs_port
    conf$fs_batch_timeout <- temp$fs_batch_timeout
    conf$fs_query_timeout <- temp$fs_query_timeout
    conf$admin_email <- temp$admin_email
    conf$dismiss_past_request <- temp$dismiss_past_request
    conf$collect_history_days <- temp$collect_history_days 
    conf$sm_alerts_bluesky <- temp$sm_alerts_bluesky
    conf$sm_activated_bluesky <- temp$sm_activated_bluesky
  }
  if (!ignore_topics) {
    plans_path = get_plans_paths()
    # updating plans and topics if requested
    plans_exists <- FALSE
    topics_plans <- lapply(plans_path, function(p) {
      if (file.exists(p)) {
        plans_exists <- TRUE
        tops <- jsonlite::read_json(p, simplifyVector = FALSE, auto_unbox = TRUE)
	for(i in seq(1, length(tops$topics), along.with=tops$topics)) {
	  tops$topics[[i]]$topic = tolower(tops$topics[[i]]$topic)
	}
	tops
      }
      else
        list()
    })
    # merging topics from different SM
    merged_topics <- merge_config_lists(topics_plans, "topics")
    # merging initial + properties.json data with plans
    temp = merge_configs(list(temp, merged_topics))

    #Getting topics from excel topics files if it has changed since last load this is identified by checking the md5 signature
    #If user has not overrriden
    topics_changed <- FALSE
    topics <- {
      t <- list()
      t$md5 <- as.vector(tools::md5sum(user_topics_path))
      if (t$md5 != temp$topics_md5 || !plans_exists) {
        t$df <- readxl::read_excel(user_topics_path)
        topics_changed <- TRUE
      }
      t
    }

    #Merging topics from config json and topic excel topics if this last one has changed
    #Each time a topic is found on file, all its occurrences will be processed at the same time, to ensure consistent multi query topics updates based on position
    if (
      topics_changed &&
        exists("df", where = topics) &&
        length(active_social_media()) > 0
    ) {
      # changing topics to lowercase
      topics$df$Topic <- tolower(topics$df$Topic)
      distinct_topics <- as.list(unique(topics$df$Topic))
      adjusted_topics <- list()
      i_adjusted <- 1
      #For each distinct topic on Excel file
      for (i_topic in 1:length(distinct_topics)) {
        topic <- distinct_topics[[i_topic]]
        if (!grepl("^[A-Za-z_0-9][A-Za-z_0-9 \\-]*$", topic)) {
          stop(paste(
            "topic name",
            topic,
            "is invalid, it must contains only by alphanumeric letters, digits spaces '-' and '_' and not start with spaces, '-' or '_'",
            sep = " "
          ))
        }
        queries <- topics$df[topics$df$Topic == topic, ]
        active_sm <- active_social_media()
        # For each active social media
        for (sm in active_sm) {
          i_tmp <- 1
          #For each distinct query on Excel file on current topic
          for (i_query in 1:nrow(queries)) {
            for (translated_query in translate_query(
              sm,
              queries$Query[[i_query]]
            )) {
              #Looking for the next matching entry in json file
	      while (
                i_tmp <= length(temp$topics) &&
                  ((temp$topics[[i_tmp]]$topic != topic &&
                    temp$topics[[i_tmp]]$network == sm) ||
                    temp$topics[[i_tmp]]$network != sm)
              ) {
                i_tmp <- i_tmp + 1
              }
              if (i_tmp <= length(temp$topics)) {
                #reusing an existing query
                adjusted_topics[[i_adjusted]] <- temp$topics[[i_tmp]]
                adjusted_topics[[i_adjusted]]$network <- sm
                adjusted_topics[[i_adjusted]]$query <- translated_query
                adjusted_topics[[i_adjusted]]$label <- queries$Label[[i_query]]
                adjusted_topics[[i_adjusted]]$alpha <- if (
                  !is.null(queries$Alpha[[i_query]]) &&
                    !is.na(queries$Alpha[[i_query]])
                )
                  queries$Alpha[[i_query]] else conf$alert_alpha
                adjusted_topics[[i_adjusted]]$alpha_outlier <- (if (
                  !is.null(queries$`Outliers Alpha`[[i_query]]) &&
                    !is.na(queries$`Outliers Alpha`[[i_query]])
                )
                  queries$`Outliers Alpha`[[i_query]] else
                  conf$alert_alpha_outlier)
              } else {
                #creating a new query
                adjusted_topics[[i_adjusted]] <- list()
                adjusted_topics[[i_adjusted]]$network <- sm
                adjusted_topics[[i_adjusted]]$query <- translated_query
                adjusted_topics[[i_adjusted]]$topic <- queries$Topic[[i_query]]
                adjusted_topics[[i_adjusted]]$label <- queries$Label[[i_query]]
                adjusted_topics[[i_adjusted]]$alpha <- if (
                  !is.null(queries$Alpha[[i_query]]) &&
                    !is.na(queries$Alpha[[i_query]])
                )
                  queries$Alpha[[i_query]] else conf$alert_alpha
                adjusted_topics[[i_adjusted]]$alpha_outlier <- (if (
                  !is.null(queries$`Outliers Alpha`[[i_query]]) &&
                    !is.na(queries$`Outliers Alpha`[[i_query]])
                )
                  queries$`Outliers Alpha`[[i_query]] else
                  conf$alert_alpha_outlier)
              }
              i_tmp <- i_tmp + 1
              i_adjusted <- i_adjusted + 1
            }
          }
        }
      }
      temp$topics <- adjusted_topics
      temp$topics_md5 <- topics$md5
    }

    #Loading topic related information on config file
    conf$topics_md5 <- temp$topics_md5
    conf$topics <- temp$topics
    conf$dismiss_past_done <- temp$dismiss_past_done
    copy_plans_from(temp)
    # updating topic keywords, necessary for ignoring keywords in top words charts
    if (topics_changed) update_topic_keywords()
  }
  #Getting variables stored on keyring
  #Setting up keyring
  if (!ignore_keyring) {
    kr <- get_key_ring(conf$keyring)
    # Fetching and updating variables from keyring
    if (is_secret_set("smtp_password")) {
      conf$smtp_password <- get_secret("smtp_password")
    }
  }
}

# Copying plans from temporary file (non-typed) to conf, making sure plans have the right type
copy_plans_from <- function(temp) {
  #Copying plans
  if (length(temp$topics) > 0) {
    for (i in 1:length(temp$topics)) {
      if (
        !exists("plan", where = temp$topics[[i]]) ||
          length(temp$topics[[i]]$plan) == 0
      ) {
        conf$topics[[i]]$plan <- list()
      } else {
        conf$topics[[i]]$plan <- (lapply(
          1:length(temp$topics[[i]]$plan),
          function(j)
            do.call(parse_plan_attributes, args = temp$topics[[i]]$plan[[j]])
        ))
      }
    }
  }
}

#' @title Save the configuration changes
#' @description Permanently saves configuration changes to the data folder (excluding social media credentials, but not SMTP credentials)
#' @param data_dir Path to a directory to save configuration settings, Default: conf$data_dir
#' @param properties Whether to save the general properties to the properties.json file, default: TRUE
#' @param sm_topics Whether to save topic download plans to the topics.json file, default: TRUE
#' @return Nothing
#' @details Permanently saves configuration changes to the data folder (excluding social media credentials, but not SMTP credentials)
#' to save social media credentials please use \code{\link{sm_api_set_auth_bluesky}}
#' @examples
#' \dontrun{
#'    library(episomer)
#'    #load configuration
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    #make some changes
#'    #conf$collect_span = 90
#'    #saving changes
#'    save_config()
#' }
#' @rdname save_config
#' @seealso
#' \code{\link{setup_config}}
#' \code{\link{sm_api_set_auth_bluesky}}
#' @export
save_config <- function(
  data_dir = conf$data_dir,
  properties = TRUE,
  sm_topics = NULL
) {
  # creating the data directory if it does not exists
  if (!file.exists(conf$data_dir)) {
    dir.create(conf$data_dir, showWarnings = FALSE)
  }
  if (is.null(sm_topics)) {
    sm_topics <- active_social_media()
  }

  if (properties) {
    # saving properties to the properties.json file
    temp <- list()
    temp$collect_span <- conf$collect_span
    temp$schedule_span <- conf$schedule_span
    temp$schedule_start_hour <- conf$schedule_start_hour
    temp$languages <- conf$languages
    temp$dep_updated_on <- conf$dep_updated_on
    temp$lang_updated_on <- conf$lang_updated_on
    temp$geonames_updated_on <- conf$geonames_updated_on
    temp$geotag_requested_on <- conf$geotag_requested_on
    temp$aggregate_requested_on <- conf$aggregate_requested_on
    temp$alerts_requested_on <- conf$alerts_requested_on
    temp$geonames_url <- conf$geonames_url
    temp$keyring <- conf$keyring
    temp$known_users <- conf$known_users
    temp$spark_cores <- conf$spark_cores
    temp$spark_memory <- conf$spark_memory
    temp$onthefly_api <- conf$onthefly_api
    temp$geolocation_threshold <- conf$geolocation_threshold
    temp$geolocation_threshold <- conf$geolocation_threshold
    temp$alert_alpha <- conf$alert_alpha
    temp$alert_alpha_outlier <- conf$alert_alpha_outlier
    temp$alert_k_decay <- conf$alert_k_decay
    temp$alert_history <- conf$alert_history
    temp$alert_same_weekday_baseline <- conf$alert_same_weekday_baseline
    temp$alert_with_quotes <- conf$alert_with_quotes
    temp$alert_with_bonferroni_correction <- conf$alert_with_bonferroni_correction
    temp$use_mkl <- conf$use_mkl
    temp$geonames_simplify <- conf$geonames_simplify
    temp$regions_disclaimer <- conf$regions_disclaimer
    temp$smtp_host <- conf$smtp_host
    temp$smtp_port <- conf$smtp_port
    temp$smtp_from <- conf$smtp_from
    temp$smtp_login <- conf$smtp_login
    if (!is.null(conf$smtp_password) && conf$smtp_password != "")
      set_secret("smtp_password", conf$smtp_password)
    temp$smtp_insecure <- conf$smtp_insecure
    temp$force_date_format <- conf$force_date_format
    temp$maven_repo <- conf$maven_repo
    temp$winutils_url <- conf$winutils_url
    temp$fs_port <- conf$fs_port
    temp$fs_batch_timeout <- conf$fs_batch_timeout
    temp$fs_query_timeout <- conf$fs_query_timeout
    temp$admin_email <- conf$admin_email
    temp$dismiss_past_request <- conf$dismiss_past_request
    temp$dismiss_past_done <- conf$dismiss_past_done
    temp$collect_history_days <- conf$collect_history_days 
    temp$sm_alerts_bluesky <- conf$sm_alerts_bluesky
    temp$sm_activated_bluesky <- conf$sm_activated_bluesky
    temp$topics_md5 <- conf$topics_md5
    # writing the JSON file
    write_json_atomic(
      temp,
      get_properties_path(),
      pretty = TRUE,
      force = TRUE,
      auto_unbox = TRUE
    )
  }
  if (length(sm_topics) > 0) {
    # saving topics to the topics.json file
    paths = get_plans_paths()
    for (sm in sm_topics) {
      topics_in_sm = conf$topics[sapply(
        1:length(conf$topics),
        function(j) conf$topics[[j]]$network == sm
      )]
      temp <- list()
      temp$topics <- format_topics(topics_in_sm)
      # writing plan files for each social media
      write_json_atomic(
        temp,
        paths[[sm]],
        pretty = TRUE,
        force = TRUE,
        auto_unbox = TRUE
      )
    }
  }
}

format_topics <- function(topics) {
  lapply(topics, function(t) {
    t$plan <- lapply(t$plan, format_plan)
    t
  })
}

# Merging two or more configuration files as a list, the latest takes precedence over earlier ones
merge_configs <- function(configs) {
  if (length(configs) == 0)
    stop("No configurations provided for merging") else if (length(configs) == 1)
    configs[[1]] else {
    first <- configs[[1]]
    rest <- merge_configs(configs[-1])
    keys <- unique(c(names(first), names(rest)))
    as.list(setNames(
      mapply(function(x, y) if (is.null(y)) x else y, first[keys], rest[keys]),
      keys
    ))
  }
}

merge_config_lists <- function(configs, property) {
  nested <- lapply(configs, function(cc) {
    if (exists(property, where = cc)) cc[[property]] else list()
  })
  ret <- list()
  ret[[property]] <- as.list(Reduce(c, nested))
  ret
}

# parse topic query and translate it to particular social media quert language
translate_query <- function(sm, q) {
  r = list(neg=list(), ors=list())
  q = gsub(" or ", "|", q) 
  q = gsub(" OR ", "|", q) 
  q = gsub(" AND ", "&", q) 
  q = gsub(" and ", "&", q) 
  q = gsub("^-| -|&-|\\|-", "+~", q) 
  ors <- trimws(strsplit(q, "\\|")[[1]])
  ors <- ors[ors != ""]
  for(i in 1:length(ors)) {
    r$ors[[i]] <- list()
    ands <- trimws(strsplit(ors[[i]], "&")[[1]])
    ands <- ands[ands != ""]
    for(j in 1:length(ands)) {
       parts <- trimws(strsplit(ands[[j]], "\\+")[[1]])
       parts <- parts[parts != ""]
       for(k in 1:length(parts)) {
          part <- trimws(parts[[k]])
          if(startsWith(part, "~") && nchar(part)>1) {
             neg <- substr(part, 2, 1000)
             r$neg <- c(r$neg,  neg)
	  } else if(nchar(part)>0) {
             pos <- as.list(trimws(strsplit(part, "/")[[1]]))
             r$ors[[i]] <- c(r$ors[[i]], list(pos))
          }
       }
    }
  }
  sm_api_translate_query(network = sm, r)
}


# Get topics data frame as displayed on the Shiny configuration tab
get_topics_df <- function() {
  df <- data.frame(
    Topics = sapply(conf$topics, function(t) t$topic),
    Label = sapply(conf$topics, function(t) t$label),
    Query = sapply(conf$topics, function(t) t$query),
    QueryLength = sapply(conf$topics, function(t) nchar(t$query)),
    ActivePlans = sapply(conf$topics, function(t) length(t$plan)),
    Progress = sapply(conf$topics, function(t) {
      if (length(t$plan) > 0)
        mean(unlist(lapply(t$plan, function(p) as.numeric(p$progress)))) else 0
    }),
    Requests = sapply(conf$topics, function(t) {
      if (length(t$plan) > 0)
        sum(unlist(lapply(t$plan, function(p) as.numeric(p$requests)))) else 0
    }),
    Alpha = sapply(conf$topics, function(t) t$alpha),
    OutliersAlpha = sapply(conf$topics, function(t) t$alpha_outlier),
    stringsAsFactors = FALSE
  )
  if (length(df) == 0)
    stop("The list of topics is empty. Please check your configuration")
  df
}

#Get topic labels as named array that can be used for translation
get_topics_labels <- function() {
  `%>%` <- magrittr::`%>%`
  t <- (get_topics_df() %>%
    dplyr::group_by(.data$Topics) %>%
    dplyr::summarise(label = .data$Label[which(!is.na(.data$Label))[1]]) %>%
    dplyr::ungroup())
  setNames(t$label, t$Topics)
}

#Get topic alphas as named array that can be used for translation
get_topics_alphas <- function() {
  `%>%` <- magrittr::`%>%`
  t <- (get_topics_df() %>%
    dplyr::group_by(.data$Topics) %>%
    dplyr::summarise(alpha = .data$Alpha[which(!is.na(.data$Alpha))[1]]) %>%
    dplyr::ungroup())
  setNames(t$alpha, t$Topics)
}

#Get topic outliers alphas as named array that can be used for translation
get_topics_alpha_outliers <- function() {
  `%>%` <- magrittr::`%>%`
  t <- (get_topics_df() %>%
    dplyr::group_by(.data$Topics) %>%
    dplyr::summarise(
      alpha_outlier = .data$OutliersAlpha[which(!is.na(.data$OutliersAlpha))[1]]
    ) %>%
    dplyr::ungroup())
  setNames(t$alpha_outlier, t$Topics)
}

#Get topic k_decay as named array that can be used for translation
get_topics_k_decays <- function() {
  `%>%` <- magrittr::`%>%`
  t <- (get_topics_df() %>%
    dplyr::group_by(.data$Topics) %>%
    dplyr::summarise(k_decay = conf$alert_k_decay) %>%
    dplyr::ungroup())
  setNames(t$k_decay, t$Topics)
}

# Check config setup before continue
stop_if_no_config <- function(
  error_message = "Cannot continue without setting up a configuration."
) {
  if (!exists("data_dir", where = conf)) {
    stop(paste(
      error_message,
      "Please call setup_config('your data directory here')",
      sep = "\n"
    ))
  }
}

# Call config if necessary
setup_config_if_not_already <- function() {
  if (!exists("data_dir", where = conf)) {
    setup_config()
  }
}

# Get current available languages from the available language Excel file
get_available_languages <- function() {
  readxl::read_excel(get_available_languages_path())
}

# Remove language from the used languages on conf
remove_config_language <- function(code) {
  # Timestamping action
  conf$lang_updated_on <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
  # Removing the language
  conf$languages <- conf$languages[sapply(
    conf$languages,
    function(l) l$code != code
  )]
}

# Add a language to the used languages in conf
add_config_language <- function(code, name) {
  # Timestamping action
  conf$lang_updated_on <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
  index <- (1:length(conf$languages))[sapply(
    conf$languages,
    function(l) l$code == code
  )]
  if (length(index) > 0) {
    # Language code is already in the task list
    conf$languages[[index]]$code <- code
    conf$languages[[index]]$name <- name
    conf$languages[[index]]$vectors = paste(
      conf$data_dir,
      "/languages/",
      code,
      ".txt.gz",
      sep = ""
    )
    conf$languages[[index]]$modified_on <- strftime(
      Sys.time(),
      "%Y-%m-%d %H:%M:%S"
    )
  } else {
    # Language code is not on the list
    conf$languages <- c(
      conf$languages,
      list(list(
        code = code,
        name = name,
        vectors = paste(
          conf$data_dir,
          "/languages/",
          code,
          ".txt.gz",
          sep = ""
        ),
        modified_on = strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
      ))
    )
  }
}

# Get the current known users list from the important users files
get_known_users <- function() {
  gsub("@", "", readxl::read_excel(get_known_users_path())[[1]])
}

active_social_media <- function() {
  props <- names(conf)
  sm_props <- props[grepl("^sm_activated", props)]
  active_sm_props <- sm_props[sapply(
    sm_props,
    function(sm_prop) conf[[sm_prop]]
  )]
  active_sm <- gsub("sm_activated_", "", active_sm_props)
  active_sm
}

alerts_social_media <- function() {
  props <- names(conf)
  sm_props <- props[grepl("^sm_alerts", props)]
  alerts_sm_props <- sm_props[sapply(
    sm_props,
    function(sm_prop) conf[[sm_prop]]
  )]
  alerts_sm <- gsub("sm_alerts_", "", alerts_sm_props)
  alerts_sm
}

# Wrapper for jsonlite write_json ensuring atomic file write it replaces always the existing file. It ignores appends modifiers
write_json_atomic <- function(x, path, ...) {
  file_name <- tail(strsplit(path, "/|\\\\")[[1]], 1)
  dir_name <- substring(path, 1, nchar(path) - nchar(file_name) - 1)
  swap_file <- tempfile(
    pattern = paste("~", file_name, sep = ""),
    tmpdir = dir_name
  )
  jsonlite::write_json(x = x, path = swap_file, ...)
  file.rename(swap_file, path)
}

# registers the data collection task (by writing search.PID file) for the current process or stops if no configuration has been set or if it is already running
register_search_runner <- function(network = NULL) {
  stop_if_no_config(paste(
    "Cannot check running status for search without configuration setup"
  ))
  name <- "search"
  if (!is.null(network)) {
    name <- sprintf("search.%s", network)
  }
  register_runner(name)
}


# registers the fs_monitor (by writing fs.PID file) for the current process or stops if no configuration has been set or if it is already running
register_fs_runner <- function() {
  stop_if_no_config(paste(
    "Cannot check running status for fs without configuration setup"
  ))
  register_runner("fs")
}

# registers the detect_runner (by writing detect.PID file) for the current process or stops if no configuration has been set or if it is already running
register_detect_runner <- function() {
  stop_if_no_config(paste(
    "Cannot check running status for fs without configuration setup"
  ))
  register_runner("detect")
}

#' @title Registers the fs_monitor for the current process or exits
#' @description registers the fs_monitor (by writing fs.monitor.PID file) for the current process or stops if no configuration has been set or if it is already running
#' @return Nothing
#' @details Registers the fs_monitor (by writing fs.monitor.PID file) for the current process or stops if no configuration has been set or if it is already running
#' This function has been exported so it can be properly called from the future package on the database runner, but it is not intended to be directly called by end users.
#' @examples
#' \dontrun{
#'    #getting tasks statuses
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    register_fs_monitor()
#' }
#' @rdname register_fs_monitor
#' @export
register_fs_monitor <- function() {
  stop_if_no_config(paste(
    "Cannot check running status for fs monitor without configuration setup"
  ))
  register_runner("fs_mon")
}

# Registers the a task runner (by writing "name".PID file) for the current process
# or stops if no configuration has been set or if it is already running
register_runner <- function(name) {
  stop_if_no_config(paste(
    "Cannot register ",
    name,
    " without configuration setup"
  ))
  # getting current pid
  pid <- Sys.getpid()
  # getting last running PID
  last_pid <- get_running_task_pid(name)
  if (last_pid >= 0 && last_pid != pid) {
    stop(paste("Runner ", name, "PID:", last_pid, "is already running"))
  }
  # If here means that the current process has to register as the runner service
  pid_path <- paste(conf$data_dir, "/", name, ".PID", sep = "")
  write(pid, file = pid_path, append = FALSE)
}

#' @title Registers the social media message collection task
#' @description registers the social media message collection task or stops if no configuration has been set or if it is already running
#' @return Nothing
#' @details Registers the message collextion task or stops if no configuration has been set or if it is already running.
#' Bluesky authentication needs to be previously set on the shiny app or by calling set_auth().
#' You can test if authentication is working on the shiny app troubleshot page or by calling (with dplyr): episomer::check_all() %>% filter(check == 'bluesky_auth')
#' This function will use the task scheduler on windows and will fall back to launching the runner as a separate process (attached to this session) on Linux.
#' @examples
#' \dontrun{
#'    #getting tasks statuses
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    register_search_runner_task()
#' }
#' @rdname register_search_runner_task
#' @export
register_search_runner_task <- function() {
  register_runner_task("search")
}

#' @title Stops the post collection task
#' @description stops the post collection task
#' @return Nothing
#' @details Stops the post collection task if it is already running
#' This function will try also deactivate the respective scheduled task on Windows.
#' @examples
#' \dontrun{
#'    #getting tasks statuses
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    stop_search_runner_task()
#' }
#' @rdname stop_search_runner_task
#' @export
stop_search_runner_task <- function() {
  stop_runner_task("search")
}


#' @title Registers the alert detection task
#' @description registers the alert detection task or stops if no configuration has been set or if it is already running
#' @return Nothing
#' @details Registers the alert detection task or stops if no configuration has been set or if it is already running.
#' To generate alerts, this task needs the post collection to had successfully run since the last time it ran.
#' This function will use the task scheduler on windows and will fall back to launching the runner as a separate process (attached to this session) on Linux.
#' @examples
#' \dontrun{
#'    #getting tasks statuses
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    register_detect_runner_task()
#' }
#' @rdname register_detect_runner_task
#' @export
register_detect_runner_task <- function() {
  register_runner_task("detect")
}

#' @title Stops the alert detection task
#' @description stops the alert detection task
#' @return Nothing
#' @details Stops the alert detection task if it is already running
#' This function will try also deactivate the respective scheduled task on Windows.
#' @examples
#' \dontrun{
#'    #getting tasks statuses
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    stop_detect_runner_task()
#' }
#' @rdname stop_detect_runner_task
#' @export
stop_detect_runner_task <- function() {
  stop_runner_task("detect")
}

#' @title Registers the episomer database task
#' @description registers the episomer database task or stops if no configuration has been set or if it is already running
#' @return Nothing
#' @details Registers the episomer database task or stops if no configuration has been set or if it is already running.
#' This task need the dependencies, geonames and languages steps to have been successfully ran. This can be done on the shiny app configuration page or by manually running the detect_runner_task.
#' This function will try to use the task scheduler on windows and will fall back to launching the runner as a separate process (attached to this session) on Linux.
#' @examples
#' \dontrun{
#'    #getting tasks statuses
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    register_fs_runner_task()
#' }
#' @rdname register_fs_runner_task
#' @export
register_fs_runner_task <- function() {
  register_runner_task("fs")
}


#' @title Stops the episomer database task
#' @description stops the episomer database task
#' @return Nothing
#' @details Stops the episomer database task if it is already running
#' This function will try also deactivate the respective scheduled task on Windows.
#' @examples
#' \dontrun{
#'    #getting tasks statuses
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    stop_fs_runner_task()
#' }
#' @rdname stop_fs_runner_task
#' @export
stop_fs_runner_task <- function() {
  stop_runner_task("fs")
}

# This task is currently Windows only
register_runner_task <- function(task_name) {
  # Making sure configuration has been set
  stop_if_no_config(paste(
    "Cannot register scheduled task without configuration setup"
  ))

  # Getting the embedded script name to be called in order to make the script run
  script_name <- paste(task_name, "R", sep = ".")

  run_attached <- FALSE
  # Filtering by OS, currently only Windows is supported
  # Getting source script
  script_base <- system.file(
    "extdata",
    script_name,
    package = get_package_name()
  )
  # Getting destination folder which is going to be on user c:/Users/user/episomer or ~/episomer
  script_folder <- file.path(conf$data_dir, "jobs")
  if (!file.exists(script_folder)) {
    dir.create(script_folder, showWarnings = FALSE)
  }
  script <- paste(script_folder, script_name, sep = "/")
  # Copying the script
  file.copy(from = script_base, to = script, overwrite = TRUE)
  taskname = paste("episomer", task_name, "loop", sep = "_")
  rscript_args = paste("\"", conf$data_dir, "\"", sep = "")

  if (.Platform$OS.type == "windows") {
    # testing if taskscheduleR can be loaded
    if (requireNamespace("taskscheduleR", quietly = TRUE)) {
      # Registering the task as a schedule task using taskscheduleR
      # Note that the date format is being guess using %DATE% shell variable
      # This value can be overridden by conf$fore_date_format setting
      taskscheduleR::taskscheduler_create(
        taskname,
        rscript = script,
        rscript_args = rscript_args,
        schedule = "HOUR",
        startdate = if (conf$force_date_format == "")
          tail(strsplit(shell("echo %DATE%", intern = TRUE), " ")[[1]], 1) else
          strftime(Sys.time(), conf$force_date_format),
        schtasks_extra = "/F"
      )
      taskscheduleR::taskscheduler_runnow(taskname)
    } else {
      run_attached <- TRUE
      warning(
        "Please install taskscheduleR package in order to permanently activate the task. This will run the task attached to this process and will end when you close this session."
      )
    }
  } else {
    run_attached <- TRUE
    warning(
      "Task scheduling is not implemented yet on this OS. You can still schedule it manually. Please refer to package vignette. Running the task attached to this process, it will end when you close this session"
    )
  }

  if (run_attached) {
    # running the requested task as a separate thread. This is useful for testing episomer from the shiny app but it does not permanently schedule the task.
    # calculating alerts per topic
    message(paste(
      "Running task",
      task_name,
      "attached. Use this only for testing episomer, you will need to find a permanent way to schedule it on production environments"
    ))
    cm <- paste(script, rscript_args)
    script_log <- paste(task_name, "log", sep = ".")
    log <- file.path(script_folder, script_log)
    system2(
      "Rscript",
      args = c(script, conf$data_dir),
      stdout = log,
      stderr = log,
      wait = FALSE
    )
  }
}

#Stopping a running episomer task
stop_runner_task <- function(task_name) {
  stop_if_no_config(paste(
    "Cannot stop scheduled task without configuration setup"
  ))
  message(paste("Stopping", task_name))
  # Filtering by OS, currently only Windows is supported
  if (.Platform$OS.type == "windows") {
    # Stopping the task if running
    taskname = paste("episomer", task_name, "loop", sep = "_")
    cmd <- sprintf("schtasks /End /TN %s", shQuote(taskname, type = "cmd"))
    system(cmd, intern = FALSE)
    # Disabling the task
    cmd <- sprintf(
      "schtasks /Change /TN %s /Disable",
      shQuote(taskname, type = "cmd")
    )
    system(cmd, intern = FALSE)
  }
  # If processes are still running then forcing stop by killing the processes
  if (task_name == "fs") {
    kill_task("fs", "R")
    kill_task("fs.java", "java")
    kill_task("fs_mon", "R")
  } else if (task_name == "search") {
    kill_task("search", "R")
  } else if (task_name == "detect") {
    kill_task("detect", "R")
    kill_task("detect.java", "java")
  }
}
# Forcing the stop of a PID file by killing the underlying process
kill_task <- function(pidfile, type = "R") {
  path = file.path(conf$data_dir, paste0(pidfile, ".PID"))
  if (file.exists(path)) {
    f = file(description = path, open = "r")
    last_pid = readLines(f)
    close(f)

    if (.Platform$OS.type == "windows") {
      if (type == "R")
        image = "R\\.exe|Rscript\\.exe|rsession\\.exe|Rterm\\.exe" else if (
        type == "java"
      )
        image = "java.exe" else stop("Unsupported type of process")
      if (
        length(grep(
          image,
          system(
            paste('tasklist /nh /fi "pid eq ', last_pid, '"'),
            intern = TRUE
          )
        )) >
          0
      ) {
        message(paste("Forcing process", last_pid, "to stop"))
        system(paste('taskkill /F /pid', last_pid), intern = TRUE)
      } else {
        message(paste("registered pid ", last_pid, "not running anymore"))
      }
    } else {
      if (type == "R") image = "R\\|rsession" else if (type == "java")
        image = "java" else stop("Unsupported type of process")
      if (
        system(
          paste(
            "ps -cax | grep ",
            paste0("'", image, "'"),
            " | grep ",
            last_pid
          ),
          ignore.stdout = TRUE
        ) ==
          0
      ) {
        message(paste("Forcing process", last_pid, "to stop"))
        system(paste('kill ', last_pid), intern = TRUE)
      } else {
        message(paste("registered pid ", last_pid, "not running anymore"))
      }
    }
  }
}

#' @title Missing search jobs if any
#' @description check if search jobs for each activated social media are running and return the list names of social media with missing jobs
#' @return char list of social media with no search job running
#' @details returns a char vector with the name of missing social media
#' @examples
#' \dontrun{
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    missing_search_jobs()
#' }
#' @rdname missing_search_jobs
#' @export
missing_search_jobs <- function() {
  stop_if_no_config(paste(
    "Cannot check running status for search without configuration setup"
  ))
  active <- active_social_media()
  not_running <- active[sapply(
    active,
    function(sm) get_running_task_pid(sprintf("search.%s", sm)) < 0
  )]
  not_running
}

#' @title Check whether the post collection task is running
#' @description gets the post collection execution status
#' @return logical Whether the post collection is running
#' @details returns a logical value being TRUE if the post collection is running
#' @examples
#' \dontrun{
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    is_search_running()
#' }
#' @rdname is_search_running
#' @export
is_search_running <- function() {
  stop_if_no_config(paste(
    "Cannot check running status for search without configuration setup"
  ))
  get_running_task_pid("search") >= 0
}

#' @title Check whether the database is running
#' @description gets the database runner execution status
#' @return logical Whether the database is running
#' @details returns a logical value being TRUE if the database runner is actually running
#' @examples
#' \dontrun{
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    is_fs_running()
#' }
#' @rdname is_fs_running
#' @export
is_fs_running <- function() {
  stop_if_no_config(paste(
    "Cannot check running status for fs without configuration setup"
  ))
  tryCatch(
    httr::GET(url = get_scala_ping_url(), httr::timeout(1.0))$status_code ==
      200,
    error = function(e) FALSE,
    warning = function(w) FALSE
  )
}

# stop with an error uf the embedded database service is not running.
stop_if_no_fs <- function(msg = NULL) {
  if (!is_fs_running()) {
    if (!is.null(msg)) stop(msg) else
      stop(
        "This function needs the episomer database to be running.\nYou can activate it for this session with the command register_fs_runner_task() or on a separate session with fs_loop().\nBe aware that for data to be displayed, the collect task must have ran previously"
      )
  }
}


#' @title Check whether the alert detection task is running
#' @description gets the alert detection runner execution status
#' @return logical Whether the alert detection task is running
#' @details returns a logical value being TRUE if the alert detection task is actually running
#' @examples
#' \dontrun{
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    is_detect_running()
#' }
#' @rdname is_detect_running
#' @export
is_detect_running <- function() {
  stop_if_no_config(paste(
    "Cannot check detect batch status for search without configuration setup"
  ))
  get_running_task_pid("detect") >= 0
}

# Get runner active PID this function will return -1 if no process was detected
get_running_task_pid <- function(name) {
  # Getting PID file name based on task name which stored the PID of last time this task was running
  pid_path <- paste(conf$data_dir, "/", name, ".PID", sep = "")
  if (file.exists(pid_path)) {
    # Getting last runner pid from PID file if exists
    last_pid <- as.integer(readChar(pid_path, file.info(pid_path)$size))
    # Checking if last_pid is still running
    # this check makes a system dependent call to get running processes launch by R* or rsession binaries
    # if the last_pid is found on the list of processes returned then we assume is running
    pid_running <- (if (.Platform$OS.type == "windows") {
      length(grep(
        "R\\.exe|Rscript\\.exe|rsession\\.exe|Rterm\\.exe",
        system(paste('tasklist /nh /fi "pid eq ', last_pid, '"'), intern = TRUE)
      )) >
        0
    } else if (.Platform$OS.type == "mac")
      system(
        paste("ps -cax | grep 'R\\|rsession' | grep ", last_pid),
        ignore.stdout = TRUE
      ) ==
        0 else
      system(
        paste("ps -cax | grep 'R\\|rsession' | grep ", last_pid),
        ignore.stdout = TRUE
      ) ==
        0)
    if (pid_running) last_pid else -1
  } else -1
}


#' @title Get the \code{\link{detect_loop}} task status
#' @description Reads the status of the \code{\link{detect_loop}} tasks and update it with changes requested by the Shiny app
#' @param statuses Character vector for limiting the status of the returned tasks, default: list()
#' @return A named list containing all necessary information to run and monitor the detect loop tasks.
#' @details After reading the tasks.json file and parsing it with jsonlite, this function will update the necessary fields in the
#' tasks for executing and monitoring them.
#' @examples
#' \dontrun{
#'    #getting tasks statuses
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    tasks <- get_tasks()
#' }
#' @seealso
#'  \code{\link{download_dependencies}}
#'
#'  \code{\link{update_geonames}}
#'
#'  \code{\link{update_languages}}
#'
#'  \code{\link{detect_loop}}
#'
#'  \code{\link{generate_alerts}}
#'
#' @rdname get_tasks
#' @export
#' @importFrom jsonlite fromJSON
get_tasks <- function(statuses = list()) {
  stop_if_no_config()
  # get the task path which should be data_folder/tasks.json
  # this file is only written by the detect loop process and read by all (shiny, search)
  tasks_path <- get_tasks_path()

  #Loading tasks from file if exists
  tasks <- if (file.exists(tasks_path)) {
    #if tasks files exists getting tasks from disk
    t <- tryCatch(
      jsonlite::fromJSON(tasks_path, simplifyVector = FALSE, auto_unbox = TRUE),
      error = function(e) {
        message("recovering from error on reading tasks")
        Sys.sleep(5)
        jsonlite::fromJSON(
          tasks_path,
          simplifyVector = FALSE,
          auto_unbox = TRUE
        )
      }
    )
    # casting date time values (they are interpreted as character by jsonlite)
    for (i in 1:length(t)) {
      t[[i]]$scheduled_for <- if (inherits(t[[i]]$scheduled_for, "numeric"))
        as.POSIXlt(t[[i]]$scheduled_for / 1000, origin = "1970-01-01") else NA
      t[[i]]$started_on <- if (inherits(t[[i]]$started_on, "numeric"))
        as.POSIXlt(t[[i]]$started_on / 1000, origin = "1970-01-01") else NA
      t[[i]]$end_on <- if (inherits(t[[i]]$end_on, "numeric"))
        as.POSIXlt(t[[i]]$end_on / 1000, origin = "1970-01-01") else NA
      t[[i]]$status <- if (inherits(t[[i]]$status, "character"))
        t[[i]]$status else NA
    }
    t
  } else {
    # If no tasks file is setup a default is built based on configuration
    list()
  }
  # Setting default dependencies task status if not set
  if (!exists("dependencies", where = tasks)) {
    #get Java & Scala dependencies
    tasks$dependencies <- list(
      task = "dependencies",
      order = 0,
      started_on = NA,
      end_on = NA,
      status = NA,
      scheduled_for = NA
    )
  }
  # Setting default GeoNames task status if not set
  if (!exists("geonames", where = tasks)) {
    #get geonames
    tasks$geonames <- list(
      task = "geonames",
      order = 1,
      started_on = NA,
      end_on = NA,
      status = NA,
      scheduled_for = NA
    )
  }
  # Setting default languages task status if not set
  if (!exists("languages", where = tasks)) {
    #get geonames
    tasks$languages <- list(
      task = "languages",
      order = 2,
      started_on = NA,
      end_on = NA,
      status = NA,
      scheduled_for = NA
    )
  }
  # Setting default languages task status if not set
  if (exists("geotag", where = tasks)) {
    tasks$geotag <- NULL
  }
  # Setting default aggregate task status if not set
  if (exists("aggregate", where = tasks)) {
    tasks$aggregate <- NULL
  }

  # Setting default alerts task status if not set
  if (!exists("alerts", where = tasks)) {
    #alerts
    tasks$alerts <- list(
      task = "alerts",
      order = 3,
      started_on = NA,
      end_on = NA,
      status = NA,
      scheduled_for = NA
    )
  } else if (exists("alerts", where = tasks) && tasks$alerts$order != 3) {
    #dealing with bad order case after migration
    tasks$alerts$order = 3
  }
  # The rest of the function code is to detect changes that has been requested by the shiny app
  # changes are detected by looking at xxx_updated_on or xxx_requested_on settings on the configuration saved from the shiny_app

  # Activating dependencies task if requested by the shiny app
  if (in_pending_status(tasks$dependencies)) {
    tasks$dependencies$maven_repo = conf$maven_repo
    tasks$dependencies$winutils_url = conf$winutils_url
    tasks$dependencies$status <- "pending"
  }

  # Activating geonames task if requested by the shiny app
  if (in_pending_status(tasks$geonames)) {
    tasks$geonames$url = conf$geonames_url
    tasks$geonames$status <- "pending"
  }

  # Activating languages task if requested by the shiny app
  if (in_pending_status(tasks$languages)) tasks$languages$status <- "pending"

  # Getting language information (used to detect if there is an action to perform on languages)
  # The tasks files contain the status of languages processed by the detect loop
  # The difference between the settings from the shiny app and the tasks will allow episomer to detect the actions to perform on languages
  langs <- get_available_languages() # All available languages for episomer
  lcodes <- langs$Code # available language codes
  lurls <- setNames(langs$Url, langs$Code) # available language URLs (to download models)
  lnames <- setNames(langs$Label, langs$Code) # available language names
  conf_codes <- sapply(conf$languages, function(l) l$code) # target languages to use in episomer as requested by setting (shiny app)
  task_codes <- tasks$languages$codes # code of languages already processed by the detect loop
  update_times <- setNames(
    sapply(conf$languages, function(l) l$modified_on),
    sapply(conf$languages, function(l) l$code)
  ) # time where the change on languages was required by the shiny app

  # Calculating languages to add, update and remove by comparing configuration setting  with language information processed by the detect loop
  # remove: languages that are not anymore ion the settings
  to_remove <- lcodes[sapply(
    lcodes,
    function(c) c %in% task_codes && !(c %in% conf_codes)
  )]
  # add: languages added on the settings
  to_add <- lcodes[sapply(
    lcodes,
    function(c) !(c %in% task_codes) && (c %in% conf_codes)
  )]
  # update: languages with an update after the last time the language was update was started by the detect loop
  to_update <- lcodes[sapply(
    lcodes,
    function(c)
      (c %in% task_codes) &&
        (c %in% conf_codes) &&
        (is.na(tasks$languages$started_on) ||
          is.na(tasks$languages$end_on) ||
          (!is.null(update_times[[c]]) &&
            strptime(update_times[[c]], "%Y-%m-%d %H:%M:%S") >
              tasks$languages$started_on) ||
          tasks$languages$started_on > tasks$languages$end_on)
  )]
  # languages with no changes
  done <- lcodes[sapply(
    lcodes,
    function(c)
      (c %in% task_codes) &&
        (c %in% conf_codes) &&
        (!is.na(tasks$languages$started_on) &&
          !is.na(tasks$languages$end_on) &&
          (is.null(update_times[[c]]) ||
            strptime(update_times[c], "%Y-%m-%d %H:%M:%S") <=
              tasks$languages$started_on) &&
          tasks$languages$started_on <= tasks$languages$end_on)
  )]

  # merging four possible language cases to produce an updated version of what the detect loop should do
  tasks$languages$codes = as.list(c(to_remove, to_add, to_update, done))
  tasks$languages$statuses = as.list(c(
    sapply(to_remove, function(l) "to remove"),
    sapply(to_add, function(l) "to add"),
    sapply(to_update, function(l) "to update"),
    sapply(done, function(l) "done")
  ))
  tasks$languages$urls = lapply(tasks$languages$codes, function(c) lurls[c])
  tasks$languages$names = lapply(tasks$languages$codes, function(c) lnames[c])
  tasks$languages$vectors = lapply(tasks$languages$codes, function(c) {
    conf$tasks$languages = sapply(
      conf$languages[sapply(conf$languages, function(l) l$code == c)],
      function(l) l$vectors
    )
  })

  # updating the language status and showing it as pending if we detected work to do
  tasks$languages$status <- (if (
    (!is.na(tasks$languages$status) && tasks$languages$status == "success") &&
      length(to_remove) + length(to_add) + length(to_update) > 0
  )
    "pending" else tasks$languages$status)

  #filtering tasks by status if required and return
  if (length(statuses) == 0) tasks else
    tasks[sapply(tasks, function(t) t$status %in% statuses)]
}

# Getting the scheduler task lists and plan execution by setting the scheduled_for times
# pan If tasks file exits it will read it, or get default tasks otherwise
plan_tasks <- function(statuses = list()) {
  tasks <- get_tasks(statuses)
  # Updating next execution time for each task
  now <- Sys.time()

  #sorting tasks by normal execution order
  sorted_tasks <- order(sapply(tasks, function(l) l$order))
  change <- FALSE
  for (i in sorted_tasks) {
    # Scheduling pending tasks one shot tasks (not recurrent) they will simply be executed in order when pending
    if (tasks[[i]]$task %in% c("dependencies", "geonames", "languages")) {
      if (is.na(tasks[[i]]$status) || tasks[[i]]$status == "pending") {
        tasks[[i]]$status <- "scheduled"
        tasks[[i]]$scheduled_for <- now + (i - 1) / 1000 # The (i-1)/1000 is not needed anymore
        change <- TRUE
        break #Just the first pending task is set to scheduled status
      }
    } else if (tasks[[i]]$task %in% c("alerts")) {
      # dealing with recurrent tasks first if take in consideration the case when a manual request for execution has been performed
      if (in_requested_status(tasks[[i]])) {
        tasks[[i]]$status <- "scheduled"
        tasks[[i]]$scheduled_for <- now
        tasks[[i]]$failures <- 0
        change <- TRUE
        break # Just the first requested task is executed
      } else if (
        is.na(tasks[[i]]$status) ||
          (# dealing with normal recurrent scheduling. This is the case when status is not set (first time)
          tasks[[i]]$status %in%
            c("pending", "success", "scheduled") && #or if status is pending success or scheduled (ready for next execution)
            {
              # this is the task ORDER after the last ended task base
              last_ended <-
                Reduce(
                  x = list(tasks$alerts),
                  f = function(a, b) {
                    if (is.na(a$end_on) && is.na(b$end_on)) {
                      if (a$order > b$order) a else b
                    } else if (is.na(a$end_on)) b else if (is.na(b$end_on))
                      a else if (a$end_on > b$end_on) a else b
                  }
                )
              one_shot_task_count <- 3 #NUMBER OF ONE SHOT TASKS UPDATE IT ON CHANGE
              next_order <- if (
                last_ended$order + 1 <= max(sapply(tasks, `[[`, "order"))
              )
                last_ended$order + 1 else one_shot_task_count
              tasks[[i]]$order == next_order
            })
      ) {
        # In this case this is the next task to be executed on normal scheduling
        tasks[[i]]$status <- "scheduled"
        if (is.na(tasks[[i]]$end_on)) {
          # If taks has never ended then schedule it straight away
          tasks[[i]]$scheduled_for <- now + (i - 1) / 1000
        } else if (tasks[[i]]$end_on >= tasks[[i]]$scheduled_for) {
          #if the task has ended after the last schedule (which should always be the case)
          # the last schedule was already executed a new schedule has to be set
          day_slots <- get_task_day_slots(tasks[[i]]) #Getting all day slots (times when the detect loop can be executed
          next_slots <- day_slots[day_slots > tasks[[i]]$scheduled_for] #getting the future execution slots
          #if next slot is in future set it. If it is in past, set now
          tasks[[i]]$scheduled_for <- if (
            length(next_slots) > 0 && next_slots[[1]] > now
          )
            next_slots[[1]] else now + (i - 1) / 1000
        }
        change <- TRUE
        break # only first task is set to scheduled
      }
    }
  }
  # saving changes done on tasks
  if (change) save_tasks(tasks)

  # filtering the returned tasks by requested status
  if (length(statuses) == 0) tasks else
    tasks[sapply(tasks, function(t) t$status %in% statuses)]
}

# get detect loop day slots based on start hour
# detect loop recurrent tasks will only be executed on this times
get_task_day_slots <- function(task) {
  span <- conf$schedule_span
  start_hour <- conf$schedule_start_hour
  per_day <- as.integer(24 * 60 / span)
  # calculating possible next slot for
  hour_slots <- sort((c(0:(per_day - 1)) * (span / 60) + start_hour) %% 24)
  hour_slots <- c(hour_slots, hour_slots[[1]] + 24) #This adding first slot of next day
  day_slots <- hour_slots * 60 * 60 + (as.POSIXlt(as.Date(task$scheduled_for)))
  day_slots
}

# saves the provided task lists to the data directory.
save_tasks <- function(tasks) {
  stop_if_no_config()
  tasks_path <- get_tasks_path()
  write_json_atomic(
    tasks,
    tasks_path,
    pretty = TRUE,
    force = TRUE,
    auto_unbox = TRUE,
    POSIXt = "epoch"
  )
}

#' @title Runs the detect loop
#' @description Infinite loop ensuring the daily signal detection and email alerts
#' @param data_dir Path to the 'data directory' containing application settings, models and collected posts.
#' If not provided the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, default: NA
#' @return nothing
#' @details The detect loop is composed of three 'one shot tasks' \code{\link{download_dependencies}}, \code{\link{update_geonames}}, \code{\link{update_languages}} ensuring the system has
#' all necessary components and data to run the three recurrent tasks \code{\link{generate_alerts}}
#'
#' The loop report progress on the 'tasks.json' file which is read or created by this function.
#'
#' The recurrent tasks are scheduled to be executed each 'detect span' minutes, which is a parameter set on the Shiny app.
#'
#' If any of these tasks fails it will be retried three times before going to abort status. Aborted tasks can be relaunched from the Shiny app.
#' @examples
#' \dontrun{
#'    #Running the detect loop
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    detect_loop()
#' }
#' @rdname detect_loop
#' @seealso
#'  \code{\link{download_dependencies}}
#'
#'  \code{\link{update_geonames}}
#'
#'  \code{\link{update_languages}}
#'
#'  \code{\link{detect_loop}}
#'
#'  \code{\link{generate_alerts}}
#'
#'  \code{\link{get_tasks}}
#' @export
detect_loop <- function(data_dir = NA) {
  # ensuring that config is set or using the one provided as a parameter
  if (is.na(data_dir)) setup_config_if_not_already() else
    setup_config(data_dir = data_dir)

  # registering the search runner (search.PID) or stopping if it is already running
  register_detect_runner()

  # lass_sleeping_message set to ensure it will be shown if all is done
  last_sleeping_message <- Sys.time() - (conf$schedule_span * 60)
  while (TRUE) {
    # getting tasks to execute with one scheduled
    tasks <- plan_tasks()
    # seeing if there is something to do
    # either there is a scheduled task
    # or a task is in an inconsistency state (running) or failed (failed or aborted)
    # aborted tasks are being taken into consideration since they should stop further taks from being executed
    if (
      length(tasks[sapply(
        tasks,
        function(t)
          !is.na(t$status) &&
            (t$status %in%
              c("failed", "running", "aborted") ||
              (t$status %in% c("scheduled") && t$scheduled_for <= Sys.time()))
      )]) >
        0
    ) {
      # getting the position of the task to execute (first in one of the active status)
      i_next <- order(sapply(
        tasks,
        function(t)
          if (
            !is.na(t$status) &&
              t$status %in% c("aborted", "scheduled", "failed", "running")
          )
            t$order else 999
      ))[[1]]
      # proceeding if task is not already aborted (otherwise keep looping)
      if (tasks[[i_next]]$status != "aborted") {
        # increasing number of failures if failed of running
        if (tasks[[i_next]]$status %in% c("failed", "running")) {
          tasks[[i_next]]$failures = (if (
            !exists("failures", where = tasks[[i_next]])
          )
            1 else tasks[[i_next]]$failures + 1)
        }
        # setting to aborted if the max number of retries has been reached
        if (
          tasks[[i_next]]$status %in%
            c("failed", "running") &&
            tasks[[i_next]]$failures > 1
        ) {
          tasks[[i_next]]$status = "aborted"
          tasks[[i_next]]$message = paste(
            "Max number of retries reached",
            tasks[[i_next]]$message,
            sep = "\n"
          )
          save_tasks(tasks)
          health_check(one_per_day = FALSE)
        } else {
          # if task is not aborted proceeding with execution
          message(paste(Sys.time(), ": Executing task", tasks[[i_next]]$task))
          if (tasks[[i_next]]$task == "dependencies") {
            tasks <- download_dependencies(tasks)
          } else if (tasks[[i_next]]$task == "geonames") {
            tasks <- update_geonames(tasks)
          } else if (tasks[[i_next]]$task == "languages") {
            tasks <- update_languages(tasks)
          } else if (tasks[[i_next]]$task == "alerts") {
            tasks <- generate_alerts(tasks)
          }
          # resetting failures to zero if successfully executed
          if (tasks[[i_next]]$status == "success") tasks[[i_next]]$failures = 0

          # saving tasks
          save_tasks(tasks)
        }
      }
    } else if (
      last_sleeping_message + (conf$schedule_span * 60) < Sys.time() &&
        length(tasks[sapply(
          tasks,
          function(t)
            !is.na(t$status) &&
              (t$status %in% c("scheduled") && t$scheduled_for >= Sys.time())
        )]) >
          0
    ) {
      #nothing to do (and not aborted tasks and 'nothing to do' message has not yet been shown for current detect slot
      i_next <- order(sapply(
        tasks,
        function(t)
          if (!is.na(t$status) && t$status %in% c("scheduled")) t$order else 999
      ))[[1]]
      message(paste(
        Sys.time(),
        ": Nothing else todo. Going to sleep each 5 seconds until ",
        format(tasks[[i_next]]$scheduled_for, tz = Sys.timezone(), usetz = TRUE)
      ))
      last_sleeping_message <- Sys.time()
    }
    # episomer sanity check and sending email in case of issues
    health_check()
    # updating config to capture changes coming from search loop or shiny app
    setup_config(data_dir = conf$data_dir)
    # sleep for 5 seconds
    Sys.sleep(5)
  }
}

# Check if provided task is in pending status which means that a user has requested a manual run and the task is not already running or scheduled
in_pending_status <- function(task) {
  ((is.na(task$status) ||
    task$status %in% c("pending", "success", "failure", "aborted")) &&
    ((task$task == "dependencies" &&
      !is.na(conf$dep_updated_on) &&
      (is.na(task$started_on) ||
        task$started_on <
          strptime(conf$dep_updated_on, "%Y-%m-%d %H:%M:%S"))) ||
      (task$task == "geonames" &&
        !is.na(conf$geonames_updated_on) &&
        (is.na(task$started_on) ||
          task$started_on <
            strptime(conf$geonames_updated_on, "%Y-%m-%d %H:%M:%S"))) ||
      (task$task == "languages" &&
        !is.na(conf$lang_updated_on) &&
        (is.na(task$started_on) ||
          task$started_on <
            strptime(conf$lang_updated_on, "%Y-%m-%d %H:%M:%S"))) ||
      (task$task == "alerts" &&
        !is.na(conf$alerts_requested_on) &&
        (is.na(task$started_on) ||
          task$started_on <
            strptime(conf$alerts_requested_on, "%Y-%m-%d %H:%M:%S")))))
}

# Check if provided task is in requested status which means that a user has requested manual run
in_requested_status <- function(task) {
  (((task$task == "dependencies" &&
    !is.na(conf$dep_updated_on) &&
    (is.na(task$started_on) ||
      task$started_on < strptime(conf$dep_updated_on, "%Y-%m-%d %H:%M:%S"))) ||
    (task$task == "geonames" &&
      !is.na(conf$geonames_updated_on) &&
      (is.na(task$started_on) ||
        task$started_on <
          strptime(conf$geonames_updated_on, "%Y-%m-%d %H:%M:%S"))) ||
    (task$task == "languages" &&
      !is.na(conf$lang_updated_on) &&
      (is.na(task$started_on) ||
        task$started_on <
          strptime(conf$lang_updated_on, "%Y-%m-%d %H:%M:%S"))) ||
    (task$task == "alerts" &&
      !is.na(conf$alerts_requested_on) &&
      (is.na(task$started_on) ||
        task$started_on <
          strptime(conf$alerts_requested_on, "%Y-%m-%d %H:%M:%S")))))
}


# Updating geonames task for reporting progress on geonames refresh
update_geonames_task <- function(
  tasks,
  status,
  message,
  start = FALSE,
  end = FALSE
) {
  if (start) tasks$geonames$started_on = Sys.time()
  if (end) tasks$geonames$end_on = Sys.time()
  tasks$geonames$status = status
  tasks$geonames$message = message
  save_tasks(tasks)
  return(tasks)
}

# Updating task for reporting progress on languages refresh
update_languages_task <- function(
  tasks,
  status,
  message,
  start = FALSE,
  end = FALSE,
  lang_code = character(0),
  lang_start = FALSE,
  lang_done = FALSE,
  lang_removed = FALSE
) {
  if (start) tasks$languages$started_on = Sys.time()
  if (end) tasks$languages$end_on = Sys.time()
  i <- 1
  while (i <= length(tasks$languages$code)) {
    if (tasks$languages$codes[[i]] == lang_code && lang_start)
      tasks$languages$statuses[[i]] = "running"
    if (tasks$languages$codes[[i]] == lang_code && lang_done)
      tasks$languages$statuses[[i]] = "done"
    if (tasks$languages$codes[[i]] == lang_code && lang_removed) {
      tasks$languages$codes <- tasks$languages$codes[-i]
      tasks$languages$statuses <- tasks$languages$statuses[-i]
      tasks$languages$urls <- tasks$languages$urls[-i]
      tasks$languages$names <- tasks$languages$names[-i]
      tasks$languages$vectors <- tasks$languages$vectors[-i]
      i <- i - 1
    }
    i <- i + 1
  }

  tasks$languages$status = status
  tasks$languages$message = message
  save_tasks(tasks)
  return(tasks)
}


# Updating alert task for reporting progress on alert detection
update_alerts_task <- function(
  tasks,
  status,
  message,
  start = FALSE,
  end = FALSE
) {
  if (start) tasks$alerts$started_on = Sys.time()
  if (end) tasks$alerts$end_on = Sys.time()
  tasks$alerts$status = status
  tasks$alerts$message = message
  save_tasks(tasks)
  return(tasks)
}


# Updating dependencies task for reporting progress on downloading Java & Scala dependencies
update_dep_task <- function(
  tasks,
  status,
  message,
  start = FALSE,
  end = FALSE
) {
  if (start) tasks$dependencies$started_on = Sys.time()
  if (end) tasks$dependencies$end_on = Sys.time()
  tasks$dependencies$status = status
  tasks$dependencies$message = message
  save_tasks(tasks)
  return(tasks)
}

loop_run_issues <- function(loop_name) {
  t <- get_tasks()
  if (loop_name == "fs") {
    if (
      !is.na(t$dependencies$status) &&
        t$dependencies$status %in% c("success", "running")
    ) {
      ""
    } else if (!is_detect_running()) {
      "Please activate 'Requirement & alerts' task. To run the embedded database, the dependencies task needs to be running or have been completed successfully"
    } else {
      "Please manually run the 'Run dependencies' step. To run the embedded database, the dependencies task needs to be running or have been completed successfully"
    }
  } else if (loop_name == "search") {
    auth_ok <- check_sm_auth()
    if(!auth_ok) {
      "episomer could not create an authentication token, please make sure you have entered valid credentials in the configuration page and click on 'Save settings'"
    } else if (is.na(t$dependencies$status) || t$dependencies$status != "success") {
      "Please make sure 'Requirement & alerts' task has been activated and successfully completed the tasks for 'Dependencies'"
    } else if (is.na(t$geonames$status) || t$geonames$status != "success") {
      "Please make sure 'Requirement & alerts' task has been activated and successfully completed the tasks for 'Geonames'"
    } else if (is.na(t$languages$status) || t$languages$status != "success") {
      "Please make sure 'Requirement & alerts' task has been activated and successfully completed the tasks for 'Languages'"
    }
      else ""
  } else if (loop_name == "detect") {
    ""
  } else stop(paste("Unlnown loop", loop_name))
}

#' @title Execute the alert task
#' @description Evaluate alerts for the last collected day for all topics and regions and send email alerts to subscribers
#' @param tasks Current tasks for reporting purposes, default: get_tasks()
#' @return The list of tasks updated with produced messages
#' @details This function calculates alerts for the last aggregated day and then sends emails to subscribers.
#'
#' The alert calculation is based on the country_counts time series which stores alerts by country, hour and topics.
#'
#' For each country and region, the process starts by aggregating the last N days. A day is defined as a block of consecutive 24 hours ending before the hour of the last collected post.
#' N is defined by the alert baseline parameter on the configuration tab of the Shiny application (the default is N=7).
#'
#' An alert will be produced when the number of posts observed is above the threshold calculated by the modified version of the EARS algorithm (for more details see the package vignette).
#' The behaviour of the alert detection algorithm is modified by the signal false positive rate (alpha), by downweighting previous alerts and by weekly or daily baseline parameters,
#' as defined in the configuration tab of the Shiny application and the topics file.
#'
#' A prerequisite for this function is that the \code{\link{search_loop}} must already have stored posts in the search folder and that the geotagging and aggregation tasks have been completed.
#' Normally, this function is not called directly by the user, but from the \code{\link{detect_loop}} function.
#' @examples
#' \dontrun{
#'    library(episomer)
#'    # setting up the data folder
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'
#'    # calculating alerts for last day posts and sending emails to subscribers
#'    generate_alerts()
#' }
#' @seealso
#'  \code{\link{detect_loop}}
#'
#' @rdname generate_alerts
#' @importFrom stats sd qt
#' @export
generate_alerts <- function(tasks = get_tasks()) {
  tasks <- tryCatch(
    {
      tasks <- update_alerts_task(tasks, "running", "processing", start = TRUE)
      # detecting alerts for the last aggregated date
      tasks <- do_next_alerts(tasks)

      tasks <- update_alerts_task(tasks, "running", "sending emails")
      # Sending email alerts
      tasks <- send_alert_emails(tasks)
      # Setting status to success
      tasks <- update_alerts_task(tasks, "success", "", end = TRUE)
      tasks
    },
    error = function(error_condition) {
      # Setting status to failed
      message("Failed...")
      message(paste(
        "failed while",
        tasks$alerts$message,
        " :",
        error_condition
      ))
      tasks <- update_alerts_task(
        tasks,
        "failed",
        paste("failed while", tasks$alerts$message, " alerts ", error_condition)
      )
      tasks
    },
    warning = function(error_condition) {
      # Setting status to failed
      message("Failed...")
      message(paste("failed while", tasks$alerts$message, " ", error_condition))
      tasks <- update_alerts_task(
        tasks,
        "failed",
        paste("failed while", tasks$alerts$message, " alerts ", error_condition)
      )
      tasks
    }
  )
  return(tasks)
}


#' @title algorithm for outbreak detection, extends the EARS algorithm
#' @author Michael Hoehle <https://www.math.su.se/~hoehle>
#' @description The simple 7 day running mean version of the Early Aberration Reporting System (EARS)
#' algorithm is extended as follows:
#' \itemize{
#'   \item{proper computation of the prediction interval}
#'   \item{downweighting of previous signals, similar to the approach by Farrington (1996)}
#' }
#' @param ts A numeric vector containing the counts of the univariate
#' time series to monitor. The last time point in ts is
#' investigated
#' @param alpha The alpha is used to compute the upper limit of the prediction interval:
#' (1-alpha) * 100\%, default: 0.025
#' @param alpha_outlier Residuals beyond 1-alpha_outlier quantile of the
#'              the t(n-k-1) distribution are downweighted, default: 0.05
#' @param k_decay Power k in the expression (r_star/r_threshold)^k determining the weight, default: 4
#' @param no_historic Number of previous values i.e -1, -2, ..., no_historic
#' to include when computing baseline parameters, default: 7
#' @param same_weekday_baseline whether to calculate baseline using same weekdays or any day, default: FALSE
#' @return A dataframe containing the monitored time point,
#'         the upper limit and whether a signal is detected or not.
#' @details for algorithm details see package vignette.
#' @examples
#' \dontrun{
#'    library(episomer)
#'    #Running the modifies version of the ears algorithm for a particular data series
#'     ts <- c(150, 130, 122, 160, 155, 128, 144, 125, 300, 319, 289, 277, 500)
#'     show(ears_t_reweighted(ts))
#' }
#' @rdname ears_t_reweighted
#' @export
#' @importFrom magrittr `%>%`
ears_t_reweighted <- function(
  ts,
  alpha = 0.025,
  alpha_outlier = 0.05,
  k_decay = 4,
  no_historic = 7L,
  same_weekday_baseline = FALSE
) {
  `%>%` <- magrittr::`%>%`
  # Times 7 on no_historic if baseline considers only same day of week
  step <- (if (same_weekday_baseline) 7 else 1)

  t(sapply(1:length(ts), function(t) {
    if (t < no_historic * step + 1)
      c(
        time_monitored = t,
        y0 = ts[[t]],
        U0 = NA,
        alarm0 = NA,
        y0bar = NA,
        U0_dw = NA,
        alarm0_dw = NA
      ) else {
      ## Extract current value and historic values
      ## Taking in consideration the step size which will be seven for same day of week comparison
      y_historic <- sapply(no_historic:1, function(i) ts[[t - i * step]])
      ## Calculate empirical mean and standard deviation
      y0 <- ts[[t]]

      # Number of observations
      n <- no_historic
      ## Rank of design matrix - it's 1 for the intercept-only-model
      k <- 1 # qr(matrix(rep(1, no_historic)))$rank

      # Helper function to compute threshold based on weights
      compute_threshold <- function(w) {
        # Ensure weights sum to n
        w <- w / sum(w) * n
        # Compute reweighted estimates
        y0bar <- mean(w * y_historic)
        resid <- (y_historic - y0bar)
        sd0 <- sqrt(sum(w * resid^2) / (n - k))

        ## Upper threshold
        U0 <- y0bar +
          qt(1 - alpha, df = no_historic - 1) * sd0 * sqrt(1 + 1 / no_historic)
        ## Enough evidence to sound an alarm?
        alarm0 <- y0 > U0
        # Return results
        return(list(y0bar = y0bar, sd0 = sd0, U0 = U0, alarm0 = alarm0))
      }

      # Results without reweighting
      res_noreweight <- compute_threshold(w = rep(1, n))

      # Extract fit from inversion  initial w=1 version
      y0bar <- res_noreweight$y0bar

      ## Calculate the raw residuals
      e <- (y_historic - y0bar)
      # Diagonal of hat matrix - analytic
      pii <- rep(1 / no_historic, no_historic)
      # Hat matrix from regression with i'th observation removed is always the same for intercept only model
      Xi <- matrix(rep(1, n - 1))
      #P_noi <- Xi %*% solve( t(Xi) %*% Xi) %*% t(Xi)
      #same for the intercept only model, but probably faster:
      # P_noi <- matrix(1/(n-1), ncol=n-1, nrow=n-1)
      # # Compute \hat{sigma}_{(i)} as in Chatterjee and Hadi (1988), eqn 4.7 on p. 74
      # sigmai <- sqrt(sapply(1:n, function(i) {
      #   Y_noi <- matrix(y_historic[-i])
      #   t(Y_noi) %*% (diag(n-1) - P_noi) %*% Y_noi / (n-k-1)
      # }))

      # Fast version for the intercept-only model
      sigmai <- sapply(1:n, function(i) {
        sd(y_historic[-i])
      })

      # Externally Studentized residuals, see Chaterjee and Hadi (1988), p. 74
      rstar <- e / (sigmai * sqrt(1 - pii))

      # Define threshold for outliers
      reweight_threshold <- qt(1 - alpha_outlier, df = n - k - 1)

      # Weights 1: Drop outliers
      # w_drop <- ifelse( rstar > reweight_threshold, 0, 1)
      # res_drop <- compute_threshold(w_drop)

      # Weights 2: Farrington procedure, down weight
      w_dw <- ifelse(
        rstar > reweight_threshold,
        (reweight_threshold / rstar)^k_decay,
        1
      )
      res_dw <- compute_threshold(w_dw)

      ## Return a vector with the components of the computation
      c(
        time_monitored = t,
        y0 = y0,
        U0 = res_noreweight$U0,
        alarm0 = res_noreweight$alarm0,
        y0bar = res_noreweight$y0bar,
        U0_dw = res_dw$U0,
        alarm0_dw = res_dw$alarm0
      )
    }
  })) %>%
    as.data.frame()
}


# Getting alert daily counts taking in consideration a 24-hour sliding window since last full hour
# the provided dataset is expected to be an hourly count by date with the following columns: posts, known_users, created_date, created_hour
get_reporting_date_counts <- function(
  df,
  topic,
  start,
  end,
  last_day
) {
  `%>%` <- magrittr::`%>%`
  empty = !("created_date" %in% names(df)) || !("created_hour" %in% names(df))
  if(empty  && nrow(df)) {
    message(sprintf("Found incoherent non empty dataset for alerts in %s, %s, %s... ignoring data. This is the content:", topic, start, end))
    print(df) 
  }
  if (!empty && nrow(df) > 0) {
    # Calculating end day hour which is going to be the last fully collected hour when
    # requesting the last collected dates or 23 hours if it is a fully collected day
    last_full_hour <- (if (!is.na(end) && end < last_day) 23 else
      as.integer(strftime(
        max(
          as.POSIXlt(df$created_date) + (as.integer(df$created_hour) - 1) * 3600
        ),
        "%H"
      )))
    #Handling going to previous day if last full hour is negative
    if (last_full_hour < 0) {
      last_full_hour = 23
      last_day <- last_day - 1
    }
    # Setting the reporting date based on the cut-off hour
    df$reporting_date <- ifelse(
      df$created_hour <= last_full_hour,
      df$created_date,
      df$created_date + 1
    )

    # filtering by start date
    df <- (if (is.na(start)) df else
      dplyr::filter(df, .data$reporting_date >= start))
    # filtering by end date
    df <- (if (is.na(end)) df else
      dplyr::filter(df, .data$reporting_date <= end))
    # group by reporting date

    df %>%
      dplyr::group_by(.data$reporting_date) %>%
      dplyr::summarise(
        count = sum(.data$posts),
        known_users = sum(.data$known_users)
      ) %>%
      dplyr::ungroup()
  } else {
    data.frame(
      reporting_date = as.Date(character()),
      count = numeric(),
      known_users = numeric(),
      stringsAsFactors = FALSE
    )
  }
}

#' @title calculate alerts for for a particular region
#' @description calculate alerts for a region and topic using provided parameters to the episomer signal detection function
#' @param sms A given social media for which aggregations have been calculated
#' @param topic Limit the alert detection to the given topic 
#' @param country_codes List with the codes of countries to be aggregated for this alerts
#' @param country_code_cols the name of the column containing the country codes
#' @param start the start date for the period in consideration
#' @param end the end date for the period in consideration
#' @param with_quotes wether to consider quotes in the time series to evaluate
#' @param alpha The alpha is used to compute the upper limit of the prediction interval:
#' (1-alpha) * 100\%, default: 0.025
#' @param alpha_outlier Residuals beyond 1-alpha_outlier quantile of the
#'              the t(n-k-1) distribution are downweighted, default: 0.05
#' @param k_decay Power k in the expression (r_star/r_threshold)^k determining the weight, default: 4
#' @param no_historic Number of previous values i.e -1, -2, ..., no_historic
#' to include when computing baseline parameters, default: 7
#' @param bonferroni_m Number the value to apply the Bonferroni correction; it specifies the global number of timeseries being observed
#' @param same_weekday_baseline whether to calculate baseline using same weekdays or any day, default: FALSE
#' @param logenv an environment variable to store the total number of posts concerned including all country_cols
#' @return A dataframe containing the monitored time point for the given countries and topic,
#'         the upper limit and whether a signal is detected or not.
#' @details for algorithm details see package vignette.
#' @examples
#' \dontrun{
#'    library(episomer)
#'    #Running the alerts for France in topic covid-19
#'    df <-calculate_region_alerts(sms="bluesky",
#'                                 topic="covid-19",
#'                                 country_codes = "FR",
#'                                 start=as.Date("2025-09-20"),
#'                                 end=as.Date("2025-09-30"))
#'    show(df)
#' }
#' @rdname calculate_region_alerts
#' @export
#' @importFrom magrittr `%>%`
calculate_region_alerts <- function(
  sms,
  topic,
  country_codes = list(),
  country_code_cols = "geo_country_code",
  start = NA,
  end = NA,
  with_quotes = FALSE,
  alpha = 0.025,
  alpha_outlier = 0.05,
  k_decay = 4,
  no_historic = 7,
  bonferroni_m = 1,
  same_weekday_baseline = FALSE,
  logenv = NULL
) {
  `%>%` <- magrittr::`%>%`
  read_from_date <- get_alert_count_from(
    date = start,
    baseline_size = no_historic,
    same_weekday_baseline = same_weekday_baseline
  )
  f_topic <- topic
  # Getting data from country counts to perform alert calculation
  df <- get_aggregates(
    dataset = "country_counts",
    filter = list(
      topic = topic,
      period = list(read_from_date, end),
      network = sms
    )
  )
  
  if(nrow(df) > 0) {
    # Adding original if missing
    if(!"original" %in% names(df))
       df["original"] <- sapply(1:nrow(df), function(x) 0)
    if(!"geo_country_code" %in% names(df))
       df["geo_country_code"] <- NA
    if(!"known_quotes" %in% names(df))
       df["known_quotes"] <- sapply(1:nrow(df), function(x) 0)
    if(!"known_original" %in% names(df))
       df["known_original"] <- sapply(1:nrow(df), function(x) 0)


    # filtering by topic
    df <- df %>% dplyr::filter(.data$topic == f_topic)

    # adding quotes if required
    df <- if (with_quotes) {
      df %>%
        dplyr::mutate(
          posts = ifelse(is.na(.data$quotes), 0, .data$quotes) +
            ifelse(is.na(.data$original), 0, .data$original),
          known_users = .data$known_quotes + .data$known_original
        )
    } else {
      df %>%
        dplyr::rename(
          known_users = .data$known_original
        ) %>%
        mutate(posts = .data$original)
    }
    # Setting global variable for storing total number of posts concerned including all country_cols
    if (!is.null(logenv)) {
      total_df <- df %>%
        dplyr::filter(
          (length(country_codes) == 0 |
            (.data$geo_country_code %in% country_codes))
        )
      total_count <- sum(
        (get_reporting_date_counts(total_df, topic, read_from_date, end, end) %>%
          dplyr::filter(.data$reporting_date >= start))$count
      )
      logenv$total_count <- if (exists("total_count", logenv))
        logenv$total_count + total_count else total_count
    }
    # filtering by country codes
    df <- (if (length(country_codes) == 0) df else if (
      length(country_code_cols) == 1
    )
      dplyr::filter(
        df,
        .data$topic == f_topic &
          (!!as.symbol(country_code_cols[[1]])) %in% country_codes
      ) else if (length(country_code_cols) == 2)
      dplyr::filter(
        df,
        .data$topic == f_topic &
          (!!as.symbol(country_code_cols[[1]])) %in% country_codes |
          (!!as.symbol(country_code_cols[[2]])) %in% country_codes
      ) else
      stop("get geo count does not support more than two country code columns"))
  }
  # Getting univariate time series aggregating by day
  counts <- get_reporting_date_counts(df, topic, read_from_date, end, end)
  if (nrow(counts) > 0) {
    # filling missing values with zeros if any
    date_range <- as.numeric(read_from_date):as.numeric(end)
    missing_dates <- date_range[sapply(
      date_range,
      function(d) !(d %in% counts$reporting_date)
    )]
    if (length(missing_dates > 0)) {
      missing_dates <- data.frame(reporting_date = missing_dates, count = 0)
      counts <- dplyr::bind_rows(counts, missing_dates) %>%
        dplyr::arrange(.data$reporting_date)
    }
    # Calculating alerts
    alerts <- ears_t_reweighted(
      counts$count,
      alpha = alpha / bonferroni_m,
      alpha_outlier = alpha_outlier,
      k_decay = k_decay,
      no_historic = no_historic,
      same_weekday_baseline
    )
    counts$alert <- alerts$alarm0_dw
    counts$limit <- alerts$U0_dw
    counts$baseline <- alerts$y0bar
    counts <- if (is.na(start)) counts else
      counts %>% dplyr::filter(.data$reporting_date >= start)
    counts
  } else {
    counts$alert <- logical()
    counts$limit <- numeric()
    counts$baseline <- numeric()
    counts
  }
}

#' Calculating alerts for a set of regions and a specific period
#' @param regions List with the codes of regions to be aggregated for this alerts
#' @param date_type the type of date to be used for the period in consideration
#' @param date_min the start date for the period in consideration
#' @param date_max the end date for the period in consideration
#' @param bonferroni_correction whether to apply the bonferroni correction to the alerts, default: FALSE
#' @return A dataframe containing the monitored time point for the given regions and topic,
#'         the upper limit and whether a signal is detected or not.
#' @examples
#' \dontrun{
#'    library(episomer)
#'    #Running the alerts for World in topic covid-19
#'    df <-calculate_regions_alerts(
#'                                 sms="bluesky",
#'                                 topic="covid-19",                                 
#'                                 date_min=as.Date("2025-09-20"),
#'                                 date_max=as.Date("2025-09-30")
#' )
#'    show(df)
#' }
#' @rdname calculate_region_alerts
#' @export
#' @importFrom magrittr `%>%`
calculate_regions_alerts <- function(
  sms,
  topic,
  regions = c(1),
  date_type = c("created_date"),
  date_min = as.Date("1900-01-01"),
  date_max = as.Date("2100-01-01"),
  with_quotes = FALSE,
  alpha = 0.025,
  alpha_outlier = 0.05,
  k_decay = 4,
  no_historic = 7,
  bonferroni_correction = FALSE,
  same_weekday_baseline = FALSE,
  logenv = NULL
) {
  #Importing pipe operator
  `%>%` <- magrittr::`%>%`

  # Getting region details
  all_regions <- get_country_items()
  # Setting world as region if no region is selected
  if (length(regions) == 0) regions = c(1)

  # Calculating alerts for each provided region
  series <- lapply(1:length(regions), function(i) {
    #message(paste("calculating for ", topic , all_regions[[regions[[i]]]]$name, i))
    bonferroni_m <-
      if (bonferroni_correction) {
        length(all_regions[unlist(lapply(
          all_regions,
          function(r) r$level == all_regions[[regions[[i]]]]$level
        ))])
      } else 1
    alerts <-
      calculate_region_alerts(
        sms = sms,
	topic = topic,
        country_codes = all_regions[[regions[[i]]]]$codes,
        country_code_cols = "geo_country_code",
        start = as.Date(date_min),
        end = as.Date(date_max),
        with_quotes = with_quotes,
        no_historic = no_historic,
        alpha = alpha,
        alpha_outlier = alpha_outlier,
        k_decay = k_decay,
        bonferroni_m = bonferroni_m,
        same_weekday_baseline = same_weekday_baseline,
        logenv = logenv
      )
    alerts <- dplyr::rename(alerts, date = .data$reporting_date)
    alerts <- dplyr::rename(alerts, number_of_posts = .data$count)
    alerts$date <- as.Date(alerts$date, origin = '1970-01-01')
    if (nrow(alerts) > 0) alerts$country <- all_regions[[regions[[i]]]]$name
    alerts$known_ratio = alerts$known_users / alerts$number_of_posts
    alerts
  })

  # Joining alerts generated for each region
  df <- Reduce(x = series, f = function(df1, df2) {
    dplyr::bind_rows(df1, df2)
  })
  if (date_type == "created_weeknum") {
    df <- df %>%
      dplyr::group_by(week = strftime(.data$date, "%G%V"), .data$country) %>%
      dplyr::summarise(
        c = sum(.data$number_of_posts),
        a = max(.data$alert),
        d = min(.data$date),
        ku = sum(.data$known_users),
        kr = sum(.data$known_users) / sum(.data$number_of_posts),
        l = 0,
        b = 0
      ) %>%
      dplyr::ungroup() %>%
      dplyr::select(
        .data$d,
        .data$c,
        .data$ku,
        .data$a,
        .data$l,
        .data$country,
        .data$kr,
        .data$l,
        .data$b
      ) %>%
      dplyr::rename(
        date = .data$d,
        number_of_posts = .data$c,
        alert = .data$a,
        known_users = .data$ku,
        known_ratio = .data$kr,
        limit = .data$l,
        baseline = .data$b
      )
  }
  return(df)
}


# get alert count start depending on the baseline type: any day or same weekday
get_alert_count_from <- function(date, baseline_size, same_weekday_baseline) {
  if (!same_weekday_baseline) date - (baseline_size + 2) else
    date - (7 * baseline_size + 2)
}

# getting parameters for current alert generation
# This will be based on last successfully aggregated date and it will only be generated if once each scheduling span
do_next_alerts <- function(tasks = get_tasks()) {
  `%>%` <- magrittr::`%>%`
  # Getting period for last alerts
  alert_to <- NA
  alert_to_hour <- NA
  wait_for <- 0
  while (is.na(alert_to) || is.na(alert_to_hour)) {
    last_agg <- get_aggregated_period()
    alert_to <- last_agg$last
    alert_to_hour <- last_agg$last_hour
    if (is.na(alert_to) || is.na(alert_to_hour)) {
      message(paste(
        "Cannot determine the last aggregated period for alert detection, trying again in",
        wait_for,
        " seconds"
      ))
      Sys.sleep(wait_for)
      wait_for <- wait_for + 10
    }
    if (wait_for > 30)
      stop(
        "Cannot determine the last aggregated period for alert detection. Please check that series have been aggregated"
      )
  }

  # Determining whether we should produce alerts for current hour (if aggregation has produced new records since last alert generation
  do_alerts <- FALSE
  alert_file <- get_alert_file(alert_to)
  if (!file.exists(alert_file)) {
    do_alerts <- TRUE
  } else {
    f <- file(alert_file, "rb")
    existing_alerts <- jsonlite::stream_in(f, verbose = FALSE)
    close(f)
    if (
      nrow(existing_alerts) == 0 ||
        nrow(
          existing_alerts %>%
            dplyr::filter(.data$date == alert_to & .data$hour == alert_to_hour)
        ) ==
          0
    )
      do_alerts <- TRUE
  }
  if (do_alerts) {
    # Getting region details
    regions <- get_country_items()
    cores <- as.numeric(conf$spark_cores)
    cl <- parallel::makePSOCKcluster(cores, outfile = "")
    on.exit(parallel::stopCluster(cl))
    data_dir <- conf$data_dir
    topics <- unique(lapply(conf$topics, function(t) t$topic))
    alphas = get_topics_alphas()
    alpha_outliers = get_topics_alpha_outliers()
    k_decays = get_topics_k_decays()
    parallel::clusterExport(
      cl,
      list(
        "data_dir",
        "topics",
	"alphas",
	"alpha_outliers",
	"k_decays",
        "regions",
        "alert_to",
        "get_country_items",
        "update_alerts_task",
        "cores",
        "tasks",
	"conf"
      ),
      envir = rlang::current_env()
    )

    # Get active social media platforms for alerts
    sms <- alerts_social_media()
    # active_sms <- "bluesky"
    if(length(sms)==0)
      stop("Cannot calculate alerts if no social media has been tagged to be included for alerts")
    # calculating alerts per social media platform and topic
    
    # Parallel processing for topics within each social media platform
    alerts <- parallel::parLapply(cl, 1:length(topics), function(i) {
      topic <- topics[[i]]
      episomer::setup_config(data_dir)
      m <- paste("Getting alerts for", paste(sms, collapse = ", "), topic, i, alert_to, (Sys.time()))
      message(m)
      if (i %% cores == 0) {
        tasks <- update_alerts_task(tasks, "running", m)
      }
      episomer::calculate_regions_alerts(
        sms = sms,
        topic = tolower(topic),
        regions = 1:length(regions),
        # regions = 1,
        date_type = "created_date",
        date_min = alert_to,
        date_max = alert_to,
        # date_min = "2025-08-23",
        # date_max = "2025-09-22",
        with_quotes = conf$alert_with_quotes,
        alpha = as.numeric(alphas[[topic]]),
	alpha_outlier = as.numeric(alpha_outliers[[topic]]),
	k_decay = as.numeric(k_decays[[topic]]),
	no_historic = as.numeric(conf$alert_history),
        bonferroni_correction = conf$alert_with_bonferroni_correction,
        same_weekday_baseline = conf$alert_same_weekday_baseline
      ) %>%
        dplyr::mutate(topic = topic) %>%
        dplyr::filter(!is.na(.data$alert) & .data$alert == 1)
    })

    # Joining all day alerts on a single data frame
    alerts <- Reduce(x = alerts, f = function(df1, df2) {
      dplyr::bind_rows(df1, df2)
    })

    # If alert prediction model has been trained, alerts can be predicted
    if (nrow(alerts) > 0) {
      # Adding top items
      codeMap <- get_country_codes_by_name()
      ts <- unique(alerts$topic)
      for (serie in list(
        list(name = "topwords", col = "token"),
        list(name = "tags", col = "tag"),
        list(name = "urls", col = "url") #,
        # list(name = "contexts", col = "context"),
        # list(name = "entities", col = "entity")
      )) {
        m <- paste("Adding", serie$name)
        message(m)
        alert_to_minus = as.Date(alert_to)-1
        data <- get_aggregates(
          serie$name,
          filter = list(topic = ts, period = c(alert_to_minus, alert_to), network = sms  )
        )
        data$frequency <- data$original
        data$item <- data[[serie$col]]
        new_col <- mapply(
          function(t, d, r) {
            codes <- codeMap[[r]]
            if (is.null(codes)) {
              ""
            } else {
              tps <- (
                data %>%
                  dplyr::filter(
                    .data$topic == t &
                      .data$created_date %in% c(d-1, d) &
                      (if (length(codes) == 0) TRUE else
                        .data$geo_country_code %in% codes)
                  ) %>%
                  dplyr::filter(!is.na(.data$frequency)) %>%
                  dplyr::group_by(.data$item) %>%
                  dplyr::summarize(frequency = sum(.data$frequency)) %>%
                  dplyr::ungroup() %>%
                  dplyr::arrange(-.data$frequency) %>%
                  head(10) #%>%
                #dplyr::mutate(item = reorder(.data$item, .data$frequency))
              )
              if (nrow(tps) == 0) "" else
                paste(paste(
                  tps$item,
                  " (",
                  tps$frequency,
                  ")",
                  collapse = ", ",
                  sep = ""
                ))
            }
          },
          alerts$topic,
          alerts$date,
          alerts$country
        )
        alerts[[serie$name]] <- new_col
      }
      # Adding used parameters
      alerts <- alerts %>%
        dplyr::mutate(
          hour = alert_to_hour,
          with_quotes = conf$alert_with_quotes,
          alpha = as.numeric(get_topics_alphas()[.data$topic]),
          alpha_outlier = as.numeric(get_topics_alpha_outliers()[.data$topic]),
          k_decay = as.numeric(get_topics_k_decays()[.data$topic]),
          no_historic = as.numeric(conf$alert_history),
          no_historic = as.numeric(conf$alert_history),
          bonferroni_correction = conf$alert_with_bonferroni_correction,
          same_weekday_baseline = conf$same_weekday_baseline,
        )
      alert_training <- get_alert_training_df()
      if (length(unique(alert_training$episomer_category)) > 1) {
        # Adding posts to alerts
        m <- paste("Adding topposts")
        message(m)
        tasks <- update_alerts_task(tasks, "running", m)
        alerts <- add_topposts(alerts, 10)
        m <- paste("Classifying alerts topposts")
        message(m)
        tasks <- update_alerts_task(tasks, "running", m)
        alerts <- classify_alerts(
          alerts %>%
            dplyr::mutate(test = FALSE, augmented = FALSE, deleted = FALSE),
          retrain = FALSE
        )
        # Removing top posts
        alerts$topposts <- NULL
        alerts$id <- NULL
      }
      f <- file(alert_file, open = "ab")
      jsonlite::stream_out(alerts, f, auto_unbox = TRUE)
      close(f)
    }
  } else {
    message("no alerts to recalculate")
  }
  tasks
}


#' @title Getting signals produced by the task \code{\link{generate_alerts}} of \code{\link{detect_loop}}
#' @description Returns a data frame of signals produced by the \code{\link{detect_loop}}, which are stored on the signal folder.
#' @param topic Character vector. When it is not empty it will limit the returned signals to the provided topics, default: character()
#' @param countries Character vector containing the names of countries or regions or a numeric vector containing the indexes of countries
#' as displayed at the Shiny App to filter the signals to return., default: numeric()
#' @param from Date defining the beginning of the period of signals to return, default: '1900-01-01'
#' @param until Date defining the end of the period of signals to return, default: '2100-01-01'
#' @param topposts Integer number of top posts to be added to the alert. These are obtained from the post index based on topwords and Lucene score, default: 0
#' @param limit Maximum number of alerts returned, default: 0
#' @param duplicates Character, action to decide what to do with alerts generated on the same day. Options are "all" (keep all alerts), "first" get only first alert and "last" for getting only the last alert
#' @param progress Function, function to report progress it should receive two parameter a progress between 0 and 1 and a message, default: empty function
#' @return a data frame containing the calculated alerts for the period. If no alerts are found then NULL is returned
#' @details For more details see the package vignette.
#' @examples
#' \dontrun{
#'    library(episomer)
#'    # setting up the data folder
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'
#'    # Getting signals produced for last 30 days for a particular country
#'    get_alerts(
#'      countries = c("Chile", "Australia", "France"),
#'      from = as.Date(Sys.time())-30,
#'      until = as.Date(Sys.time())
#'    )
#' }
#' @seealso
#' \code{\link{generate_alerts}}
#'
#' \code{\link{detect_loop}}
#' @rdname get_alerts
#' @export
#' @importFrom magrittr `%>%`
#' @importFrom jsonlite stream_in
#' @importFrom dplyr filter arrange group_by mutate ungroup bind_rows
#' @importFrom stats reorder sd qt
#' @importFrom utils head
get_alerts <- function(
  topic = character(),
  countries = numeric(),
  from = "1900-01-01",
  until = "2100-01-01",
  topposts = 0,
  limit = 0,
  duplicates = "all",
  progress = function(a, b) {
  }
) {
  `%>%` <- magrittr::`%>%`
  # preparing files and renaming variables with names as column on data frame to avoid conflicts
  regions <- get_country_items()
  t <- topic
  cnames <- (if (class(countries) %in% c("numeric", "integer"))
    lapply(countries, function(i) regions[[i]]$name) else countries)
  alert_files <- list.files(
    file.path(conf$data_dir, "alerts"),
    recursive = TRUE,
    full.names = TRUE
  )
  if (length(alert_files) == 0) {
    NULL
  } else {
    file_dates <- lapply(alert_files, function(f) {
      list(
        path = f,
        date = as.Date(
          substr(tail(strsplit(f, "/")[[1]], n = 1), 1, 10),
          format = "%Y.%m.%d"
        )
      )
    })
    files_in_period <- lapply(
      file_dates[sapply(
        file_dates,
        function(fd) fd$date >= as.Date(from) && fd$date <= as.Date(until)
      )],
      function(fd) fd$path
    )
    alerts <- lapply(files_in_period, function(alert_file) {
      f <- file(alert_file, "rb")
      df <- jsonlite::stream_in(f, verbose = FALSE)
      close(f)
      # Adding default value for same_weekday_baseline if does not exists
      if (!("same_weekday_baseline" %in% colnames(df)))
        df$same_weekday_baseline <- sapply(df$topic, function(t) FALSE)
      # Adding default value for alpha_outlier if does not exists
      if (!("alpha_outlier" %in% colnames(df)))
        df$alpha_outlier <- as.numeric(sapply(df$topic, function(t) NA))
      # Adding default value for k_decay if does not exists
      if (!("k_decay" %in% colnames(df)))
        df$k_decay <- as.numeric(sapply(df$topic, function(t) NA))
      # Adding default hashtags
      if (!("tags" %in% colnames(df)))
        df$tags <- sapply(df$topic, function(t) NA)
      # Adding default urls
      if (!("urls" %in% colnames(df)))
        df$urls <- sapply(df$topic, function(t) NA)
      # Adding default entities
      if (!("entities" %in% colnames(df)))
        df$entities <- sapply(df$topic, function(t) NA)
      # Adding default contexts
      if (!("contexts" %in% colnames(df)))
        df$contexts <- sapply(df$topic, function(t) NA)

      df <- df %>%
        dplyr::filter(
          (if (length(cnames) == 0) TRUE else .data$country %in% cnames) &
            (if (length(t) == 0) TRUE else .data$topic %in% t)
        ) %>%
        dplyr::arrange(
          .data$topic,
          .data$country,
          .data$hour,
          .data$number_of_posts
        ) %>%
        dplyr::group_by(.data$topic, .data$country) %>%
        dplyr::mutate(rank = rank(.data$hour, ties.method = "first")) %>%
        dplyr::ungroup()

      if (duplicates == "last") {
        df <- df %>%
          dplyr::arrange(
            .data$topic,
            .data$country,
            dplyr::desc(.data$hour)
          ) %>%
          dplyr::distinct(.data$topic, .data$country, .keep_all = TRUE)
      } else if (duplicates == "first") {
        df <- df %>% dplyr::filter(.data$rank == 1)
      }
      df
    })

    df <- Reduce(x = alerts, f = function(df1, df2) {
      dplyr::bind_rows(df1, df2)
    })
    if (!is.null(df)) {
      if (limit > 0 && limit < nrow(df)) {
        df <- df[sample(nrow(df), limit), ]
      }
      # Adding top posts if required
      if (topposts > 0) {
        df <- add_topposts(df, topposts, progress = progress)
      }
      # Calculating top columns
      df$tops <- paste(
        "<UL>",
        ifelse(
          !is.na(df$topwords),
          paste("<li><b>Top Words</b>: ", df$topwords, "</li>\n"),
          ""
        ),
        ifelse(
          !is.na(df$tags),
          paste("<li><b>Top tags</b>: ", df$tags, "</li>\n"),
          ""
        ),
        ifelse(
          !is.na(df$entities),
          paste("<li><b>Top Entities</b>: ", df$entities, "</li>\n"),
          ""
        ),
        ifelse(
          !is.na(df$contexts),
          paste("<li><b>Top Contexts</b>: ", df$contexts, "</li>\n"),
          ""
        ),
        ifelse(
          !is.na(df$urls) & nchar(df$urls) > 0,
          paste(
            "<li><b>Top Urls</b>: ",
            sapply(strsplit(df$urls, ", "), function(s) {
              if (!is.null(s) && !all(is.na(s)) && length(s) > 0 && all(nchar(s) > 0)) {
                urls <- sapply(strsplit(s, " \\("), function(f) f[[1]])
                counts <- sapply(
                  strsplit(s, " \\("),
                  function(f)
                    paste(" (", if (length(f) > 1) f[[2]] else "-", sep = "")
                )
                paste(
                  sapply(
                    1:length(urls),
                    function(i)
                      paste(
                        "<a target = \"_blank\" href = \"",
                        urls[[i]],
                        "\">",
                        urls[[i]],
                        counts[[i]],
                        "</a>",
                        sep = ""
                      )
                  ),
                  collapse = ", "
                )
              } else ""
            }),
            "</li>\n"
          ),
          ""
        ),
        "</UL>",
        sep = ""
      )
      if (nrow(df) > 0 && !"episomer_category" %in% colnames(df))
        df$episomer_category <- NA

      # overriding classification if provided by users$
      alert_training <- get_alert_training_df()
      given <- setNames(
        alert_training$given_category,
        tolower(paste(
          alert_training[["date"]],
          alert_training[["topic"]],
          alert_training[["country"]],
          sep = "@@"
        ))
      )
      given <- given[!is.na(given)]
      ngiven <- names(given)

      if (nrow(df) > 0) {
        df$episomer_category <- sapply(1:nrow(df), function(i) {
          key = tolower(paste(
            df[["date"]][[i]],
            df[["topic"]][[i]],
            df[["country"]][[i]],
            sep = "@@"
          ))
          if (key %in% ngiven) given[[key]] else df$episomer_category[[i]]
        })
      }
      progress(1, "Alerts obtained")
      tibble::as_tibble(df)
    } else NULL
  }
}

# Add top posts from the post index to the alert data frame
# A query is done to the index for each topic alert to consider top words detected
add_topposts <- function(
  df,
  topposts,
  ignore_place = FALSE,
  progress = function(a, b) {
  }
) {
  codes <- get_country_codes_by_name()
  topwords <- sapply(
    strsplit(df$topwords, "\\W+|[0-9]\\W"),
    function(v) v[nchar(v) > 0]
  )

  df$topposts <- if (nrow(df) == 0) character() else {
    sapply(1:nrow(df), function(i) {
      created_from = (if (df$hour[[i]] == 23) {
        df$date[[i]]
      } else if (df$hour[[i]] < 9) {
        paste0(as.character(as.Date(df$date[[i]]) - 1), "T0", df$hour[[i]] + 1)
      } else
        paste0(as.character(as.Date(df$date[[i]]) - 1), "T", df$hour[[i]] + 1))
      created_to = (if (df$hour[[i]] == 23) {
        paste0(df$date[[i]], "TZ")
      } else if (df$hour[[i]] < 10) {
        paste0(df$date[[i]], "T0", df$hour[[i]], "Z")
      } else paste0(df$date[[i]], "T", df$hour[[i]], "Z"))
      match_codes <- codes[[df$country[[i]]]]
      if (is.null(match_codes)) "" else {
        progress(i / nrow(df), "Getting alerts posts")
        posts <- lapply(conf$languages, function(lang) {
          res <- search_posts(
            query = paste0(
              "created_at:[",
              created_from,
              " TO ",
              created_to,
              "] AND ",
              paste0("lang:", lang$code, " AND "),
              (if (length(match_codes) == 0 || ignore_place) "" else
                paste0(
                  "text_loc.geo_country_code:",
                  paste0(match_codes, collapse = ";"),
                  " AND "
                )),
              (if (length(topwords[[i]]) == 0) "" else
                paste0(
                  "(",
                  paste0(paste0("\"", topwords[[i]], "\""), collapse = " OR "),
                  ") AND "
                )),
              "is_quote:false"
            ),
            topic = df$topic[[i]],
            from = as.character(as.Date(df$date[[i]]) - 1),
            to = df$date[[i]],
            max = topposts,
            by_relevance = TRUE
          )
          if (length(res) == 0) character(0) else res$text
        })
        jsonlite::toJSON(setNames(
          posts,
          lapply(conf$languages, function(l) l$code)
        ))
      }
    })
  }
  df$topposts <- lapply(df$topposts, function(json) jsonlite::fromJSON(json))
  df
}


# get the default or user defined subscribed user list
get_subscribers <- function() {
  df <- readxl::read_excel(get_subscribers_path())
  df$Topics <- as.character(df$Topics)
  df$User <- as.character(df$User)
  df$Email <- as.character(df$Email)
  df$Regions <- as.character(df$Regions)
  df$`Excluded Topics` <- as.character(df$`Excluded Topics`)
  df$`Real time Topics` <- as.character(df$`Real time Topics`)
  df$`Alert category` <- if (!"Alert category" %in% colnames(df)) NA else
    as.character(df$`Alert category`)
  df$`Alert Slots` <- as.character(df$`Alert Slots`)
  df$`One post alerts` <- if (!"One post alerts" %in% colnames(df)) TRUE else
    ifelse(
      as.character(df$`One post alerts`) %in% c("0", "no", "No", "FALSE"),
      FALSE,
      TRUE
    )
  df$`Topics ignoring 1 post alerts` <- if (
    !"Topics ignoring 1 post alerts" %in% colnames(df)
  )
    NA else as.character(df$`Topics ignoring 1 post alerts`)
  df$`Regions ignoring 1 post alerts` <- if (
    !"Regions ignoring 1 post alerts" %in% colnames(df)
  )
    NA else as.character(df$`Regions ignoring 1 post alerts`)
  df
}

# Send email alerts to subscribers based on newly generated alerts and subscribers' configuration
send_alert_emails <- function(tasks = get_tasks()) {
  `%>%` <- magrittr::`%>%`
  task <- tasks$alerts
  # Creating sending email statistics if any
  if (!exists("sent", tasks$alerts)) tasks$alerts$sent <- list()
  # Getting subscriber users
  subscribers <- get_subscribers()
  # iterating for each subscriber
  if (nrow(subscribers) > 0 && conf$smtp_host != "") {
    for (i in 1:nrow(subscribers)) {
      # getting values subscriber settings
      user <- subscribers$User[[i]]
      topics <- if (is.na(subscribers$Topics[[i]])) NA else
        strsplit(subscribers$Topics[[i]], ";")[[1]]
      dest <- if (is.na(subscribers$Email[[i]])) NA else
        strsplit(subscribers$Email[[i]], ";")[[1]]
      regions <- if (is.na(subscribers$Regions[[i]])) NA else
        strsplit(subscribers$Regions[[i]], ";")[[1]]
      excluded <- if (is.na(subscribers$`Excluded Topics`[[i]])) NA else
        strsplit(subscribers$`Excluded Topics`[[i]], ";")[[1]]
      realtime_topics <- if (is.na(subscribers$`Real time Topics`[[i]])) NA else
        strsplit(subscribers$`Real time Topics`[[i]], ";")[[1]]
      realtime_regions <- if (is.na(subscribers$`Real time Regions`[[i]]))
        NA else strsplit(subscribers$`Real time Regions`[[i]], ";")[[1]]
      alert_categories <- if (
        !"Alert category" %in% colnames(subscribers) ||
          is.na(subscribers$`Alert category`[[i]])
      )
        NA else strsplit(subscribers$`Alert category`[[i]], ";")[[1]]
      slots <- as.integer(
        if (is.na(subscribers$`Alert Slots`[[i]])) NA else {
          times <- strsplit(subscribers$`Alert Slots`[[i]], ";")[[1]]
          times <- sapply(
            times,
            function(t) {
              parts <- as.integer(strsplit(t, "[^0-9]")[[1]])
              if (length(parts) <= 1) parts else parts[[1]] + parts[[2]] / 60
            }
          )
        }
      )
      one_post_alerts <- subscribers$`One post alerts`[[i]]
      ignore_one_post_topics <- if (
        is.na(subscribers$`Topics ignoring 1 post alerts`[[i]])
      )
        NA else
        strsplit(subscribers$`Topics ignoring 1 post alerts`[[i]], ";")[[1]]
      ignore_one_post_regions <- if (
        is.na(subscribers$`Regions ignoring 1 post alerts`[[i]])
      )
        NA else
        strsplit(subscribers$`Regions ignoring 1 post alerts`[[i]], ";")[[1]]

      # Adding users' statistics if these do not exist already
      if (!exists(user, where = tasks$alerts$sent))
        tasks$alerts$sent[[user]] <- list()
      # Getting last day alerts date period
      agg_period <- get_aggregated_period()
      alert_date <- agg_period$last
      alert_hour <- agg_period$last_hour

      # Getting last day alerts for subscribed user and regions
      user_alerts <- get_alerts(
        topic = if (all(is.na(topics))) NULL else if (
          !all(is.na(realtime_topics))
        )
          unique(c(topics, realtime_topics)) else topics,
        countries = if (all(is.na(regions))) NULL else if (
          !all(is.na(realtime_regions))
        )
          unique(c(regions, realtime_regions)) else regions,
        from = alert_date,
        until = alert_date
      )

      # Limiting to alerts on selected alert_categories
      if (!is.null(user_alerts) && !all(is.na(alert_categories))) {
        user_alerts <- user_alerts %>%
          dplyr::filter(.data$episomer_category %in% alert_categories)
      }

      if (!is.null(user_alerts)) {
        # Excluding alerts for topics ignored for this user
        user_alerts <- (if (!all(is.na(excluded)))
          user_alerts %>% dplyr::filter(!(.data$topic %in% excluded)) else
          user_alerts)

        # Getting last metrics per alert by region and topic
        last_metrics <- list()
        if (nrow(user_alerts) > 0) {
          user_alerts$key = paste(
            user_alerts$topic,
            user_alerts$country,
            sep = "-"
          )
          for (i in 1:nrow(user_alerts)) {
            key = user_alerts$key[[i]]
            # Getting the maximum values to report for that alert
            if (
              !exists(
                paste(key, "number_of_posts", sep = "-"),
                last_metrics
              ) ||
                last_metrics[[paste(key, "number_of_posts", sep = "-")]] <
                  user_alerts$number_of_posts[[i]]
            ) {
              last_metrics[[paste(
                key,
                "number_of_posts",
                sep = "-"
              )]] <- user_alerts$number_of_posts[[i]]
              last_metrics[[paste(
                key,
                "baseline",
                sep = "-"
              )]] <- user_alerts$baseline[[i]]
              last_metrics[[paste(
                key,
                "known_users",
                sep = "-"
              )]] <- user_alerts$known_users[[i]]
            }
            # Getting the last 1 post alert per region and topic so the next can be reported as first in day
            if (
              !exists(
                paste(key, "last_one_post_rank", sep = "-"),
                last_metrics
              )
            )
              last_metrics[[paste(key, "last_one_post_rank", sep = "-")]] <- -1
            if (
              user_alerts$number_of_posts[[i]] == 1 &&
                last_metrics[[paste(key, "last_one_post_rank", sep = "-")]] <
                  user_alerts$rank[[i]]
            )
              last_metrics[[paste(
                key,
                "last_one_post_rank",
                sep = "-"
              )]] <- user_alerts$rank[[i]]
          }
          # Excluding alerts that are not the first in the day for the specific topic and region
          # This is done because only the first alert on the day is reported
          # One exception is done for alerts that have been increased from 1 post
          user_alerts <- user_alerts %>%
            dplyr::filter(
              .data$rank == 1 |
                .data$rank ==
                  (unlist(last_metrics[paste(
                    .data$key,
                    "last_one_post_rank",
                    sep = "-"
                  )]) +
                    1)
            )

          # Removing potential duplicates added by the increased from 1 post exception
          user_alerts <- user_alerts %>%
            dplyr::arrange(
              .data$topic,
              .data$country,
              dplyr::desc(.data$hour)
            ) %>%
            dplyr::distinct(.data$topic, .data$country, .keep_all = TRUE)

          # Now metrics are updated to last observed alert
          user_alerts <- user_alerts %>%
            dplyr::mutate(
              number_of_posts = unlist(last_metrics[paste(
                .data$key,
                "number_of_posts",
                sep = "-"
              )]),
              baseline = unlist(last_metrics[paste(
                .data$key,
                "baseline",
                sep = "-"
              )]),
              known_users = unlist(last_metrics[paste(
                .data$key,
                "known_users",
                sep = "-"
              )])
            )
          #Removing one post alerts depending on user options
          user_alerts <- user_alerts %>%
            dplyr::filter(
              .data$number_of_posts > 1 |
                (one_post_alerts &
                  (all(is.na(ignore_one_post_topics)) |
                    !(.data$topic %in% ignore_one_post_topics)) &
                  (all(is.na(ignore_one_post_regions)) |
                    !(.data$country %in% ignore_one_post_regions)))
            )
        }

        # Defining if this slot corresponds to a user slot
        current_minutes <- as.numeric(strftime(Sys.time(), format = "%M"))
        current_hour <- as.numeric(strftime(Sys.time(), format = "%H")) +
          current_minutes / 60

        # Getting the current slot
        current_slot <- (if (all(is.na(slots)))
          as.integer(current_hour) else if (
          length(slots[slots <= current_hour]) == 0
        )
          tail(slots, 1) else max(slots[slots <= current_hour]))

        # Evaluating if slot alerts should be sent
        send_slot_alerts <- (all(is.na(slots)) ||
          (length(slots[slots <= current_hour]) > 0 && # After first slot in day AND
            (
              length(slots) == 1 || #Either there is only one slot for the user
                !exists("last_slot", where = tasks$alerts$sent[[user]]) || #or there is no register of previous slot executed
                tasks$alerts$sent[[user]]$last_slot != current_slot # or there are multiple slots and the current slot is different than the previous one
            )))

        # Filtering out alerts that are not instant if these do not respect the defined slots
        instant_alerts <- user_alerts %>%
          dplyr::filter(
            (!all(is.na(realtime_topics)) | !all(is.na(realtime_regions))) &
              ((all(is.na(realtime_topics)) |
                .data$topic %in% realtime_topics) |
                (all(is.na(realtime_regions)) |
                  .data$country %in% realtime_regions))
          )

        # Excluding instant alerts produced before the last alert sent to user
        instant_alerts <- (if (
          exists("date", where = tasks$alerts$sent[[user]]) &&
            exists("hour_instant", where = tasks$alerts$sent[[user]]) &&
            as.Date(tasks$alerts$sent[[user]]$date, format = "%Y-%m-%d") ==
              alert_date
        )
          instant_alerts %>%
            dplyr::filter(
              .data$hour > tasks$alerts$sent[[user]]$hour_instant
            ) else instant_alerts)

        # getting slot alerts
        slot_alerts <- user_alerts %>%
          dplyr::filter(
            ((all(is.na(realtime_topics)) & all(is.na(realtime_regions))) |
              ((!all(is.na(realtime_topics)) &
                !(.data$topic %in% realtime_topics)) &
                (!all(is.na(realtime_regions)) &
                  !(.data$country %in% realtime_regions)))) &
              send_slot_alerts
          )

        # Excluding slot alerts produced before the last alert sent to user
        slot_alerts <- (if (
          exists("date", where = tasks$alerts$sent[[user]]) &&
            exists("hour_slot", where = tasks$alerts$sent[[user]]) &&
            as.Date(tasks$alerts$sent[[user]]$date, format = "%Y-%m-%d") ==
              alert_date
        )
          slot_alerts %>%
            dplyr::filter(.data$hour > tasks$alerts$sent[[user]]$hour_slot) else
          slot_alerts)
        # Joining back instant and slot alert for email send

        user_alerts <- dplyr::union_all(slot_alerts, instant_alerts) %>%
          dplyr::mutate(topic = get_topics_labels()[.data$topic])
        # Sending alert email & registering last sent dates and hour to user

        if (nrow(user_alerts) > 0) {
          # Calculating top alerts for title
          top_alerts <- head(
            user_alerts %>%
              dplyr::arrange(
                .data$topic,
                dplyr::desc(.data$number_of_posts)
              ) %>%
              dplyr::group_by(.data$topic) %>%
              dplyr::mutate(rank = rank(.data$topic, ties.method = "first")) %>%
              dplyr::filter(.data$rank < 3) %>%
              dplyr::summarize(
                top = paste(.data$country, collapse = ", "),
                max_posts = max(.data$number_of_posts)
              ) %>%
              dplyr::arrange(dplyr::desc(.data$max_posts)),
            3
          )

          tasks <- update_alerts_task(
            tasks,
            "running",
            paste("sending alert to", dest)
          )
          message(paste("Sending alert to ", dest))

          # Getting the title
          title <- paste(
            "[episomer] found",
            nrow(user_alerts),
            "signals. Top alerts: ",
            paste(
              top_alerts$topic,
              " (",
              top_alerts$top,
              ")",
              collapse = ", ",
              sep = ""
            )
          )
          # Creating email body

          # Getting the html table for each region
          non_one_alert_tables <- get_alert_tables(
            user_alerts %>% dplyr::filter(.data$number_of_posts > 1),
            group_topics = TRUE
          )
          one_alert_table <- get_alert_tables(
            user_alerts %>% dplyr::filter(.data$number_of_posts == 1),
            group_topics = FALSE,
            ungrouped_title = "1 Post alerts"
          )

          # Applying email template
          t_con <- file(
            get_email_alert_template_path(),
            open = "rb",
            encoding = "UTF-8"
          )
          html <- paste(readLines(t_con, encoding = "UTF-8"), collapse = "\n")
          html <- gsub("@title", title, html)
          html <- gsub(
            "@alerts",
            paste(c(non_one_alert_tables, one_alert_table), collapse = "\n"),
            html
          )
          close(t_con)

          # creating the message to send
          msg <- (emayili::envelope() %>%
            emayili::from(conf$smtp_from) %>%
            emayili::to(dest) %>%
            emayili::subject(title) %>%
            emayili::html(html))

          smtp <- (if (
            is.na(conf$smtp_password) ||
              is.null(conf$smtp_password) ||
              conf$smtp_password == ""
          )
            emayili::server(
              host = conf$smtp_host,
              port = conf$smtp_port,
              insecure = conf$smtp_insecure,
              reuse = FALSE
            ) else
            emayili::server(
              host = conf$smtp_host,
              port = conf$smtp_port,
              username = conf$smtp_login,
              insecure = conf$smtp_insecure,
              password = conf$smtp_password,
              reuse = FALSE
            ))
          tryCatch(
            {
              smtp(msg)
            },
            error = function(e) {
              message(paste("Error when sending email", e))
              stop(e)
            }
          )

          # Storing last day sending alerts for current user
          # If new dates resetting hours to 0
          if (
            exists("date", where = tasks$alerts$sent[[user]]) &&
              tasks$alerts$sent[[user]]$date != alert_date
          ) {
            tasks$alerts$sent[[user]]$hour_slot <- 0
            tasks$alerts$sent[[user]]$hour_instant <- 0
          }

          # Updating date to current date
          tasks$alerts$sent[[user]]$date <- alert_date

          # Updating slot hour if slots alerts are sent
          if (nrow(slot_alerts) > 0) {
            tasks$alerts$sent[[user]]$hour_slot <- alert_hour
            tasks$alerts$sent[[user]]$last_slot <- current_slot
          }

          # Updating instant hour if instant alerts are sent
          if (nrow(instant_alerts) > 0)
            tasks$alerts$sent[[user]]$hour_instant <- alert_hour
        }
      }
    }
  }
  tasks
}

# Produce html tables for alerts provided. This is used on for email alerts.
# Tables can be grouped by topic or left ungrouped.
get_alert_tables <- function(
  alerts,
  group_topics = TRUE,
  ungrouped_title = "Alerts"
) {
  # Getting an array of countries with alerts sorted on descending order by the total number of posts
  topics <- (alerts %>%
    dplyr::group_by(.data$topic) %>%
    dplyr::summarize(posts = sum(.data$number_of_posts)) %>%
    dplyr::ungroup() %>%
    dplyr::arrange(dplyr::desc(.data$posts)))$topic
  if (!group_topics) topics = NA

  alert_tables <- (if (nrow(alerts) > 0) {
    tables <- sapply(topics, function(t) {
      paste(
        "<h5>",
        (if (!group_topics) ungrouped_title else t),
        "</h5>",
        alerts %>%
          dplyr::filter(is.na(t) | .data$topic == t) %>%
          dplyr::arrange(dplyr::desc(.data$number_of_posts)) %>%
          dplyr::select(
            `Date` = .data$`date`,
            `Hour` = .data$`hour`,
            `Topic` = .data$`topic`,
            `Region` = .data$`country`,
            `Tops` = .data$`tops`,
            `Category` = .data$`episomer_category`,
            `Posts` = .data$`number_of_posts`,
            `% from important user` = .data$`known_ratio`,
            `Threshold` = .data$`limit`,
            `Baseline` = .data$`no_historic`,
            `Alert confidence` = .data$`alpha`,
            `Outliers confidence` = .data$`alpha_outlier`,
            `Downweight strength` = .data$`k_decay`,
            `Bonf. corr.` = .data$`bonferroni_correction`,
            `Same weekday baseline` = .data$`same_weekday_baseline`,
            `Day_rank` = .data$`rank`,
            `With reposts` = .data$`with_quotes`
          ) %>%
          xtable::xtable() %>%
          print(
            type = "html",
            file = tempfile(),
            sanitize.text.function = function(f) f
          )
      )
    })
    # Applying table inline format
    tables <- gsub(
      "<table ",
      "<table style=\"border: 1px solid black;width: 100%;border-collapse: collapse;\" ",
      tables
    )
    tables <- gsub(
      "<th ",
      "<th style=\"border: 1px solid black;background-color: #3d6806;color: #ffffff;text-align: center;\" ",
      tables
    )
    tables <- gsub(
      "<th> ",
      "<th style=\"border: 1px solid black;background-color: #3d6806;color: #ffffff;text-align: center;\">",
      tables
    )
    tables <- gsub("<td ", "<td style=\"border: 1px solid black;\" ", tables)
    tables <- gsub("<td> ", "<td style=\"border: 1px solid black;\">", tables)
    tables
  } else c(""))
  alert_tables
}

# Read the alert training dataset generated by the system and updated by the user
get_alert_training_df <- function() {
  `%>%` <- magrittr::`%>%`
  data_types <- c(
    "text",
    "text",
    "text",
    "text",
    "numeric",
    "text",
    "text",
    "text"
  )
  current <- readxl::read_excel(
    get_alert_training_path(),
    col_types = data_types,
    sheet = "Alerts"
  )
  current <- current %>%
    dplyr::transmute(
      date = .data$`Date`,
      topic = .data$`Topic`,
      country = .data$`Region`,
      topwords = .data$`Top words`,
      number_of_posts = .data$`Posts`,
      topposts = .data$`Top Posts`,
      given_category = .data$`Given Category`,
      episomer_category = .data$`Episomer Category`
    )
  current$`topposts` <- lapply(
    current$`topposts`,
    function(json) jsonlite::fromJSON(json)
  )
  current %>%
    dplyr::arrange(.data$date, .data$topic, .data$country, .data$topwords)
}

get_alert_balanced_df <- function(
  alert_training_df = get_alert_training_df(),
  progress = function(value, message) {
  }
) {
  allcat_df <- alert_training_df %>%
    dplyr::filter(!is.na(.data$given_category)) %>%
    dplyr::group_by(.data$given_category) %>%
    dplyr::count()
  # we choose 25 percent of rows to be used as test set. These rows will not be augmented
  split <- jsonlite::rbind_pages(lapply(1:nrow(allcat_df), function(i) {
    x = 1:allcat_df$n[[i]]
    alert_training_df %>%
      dplyr::filter(.data$given_category == allcat_df$given_category[[i]]) %>%
      dplyr::mutate(
        test = sample(x, length(x), replace = FALSE) > length(x) * 0.75
      )
  }))

  test_df <- split %>%
    dplyr::filter(.data$test) %>%
    dplyr::mutate(deleted = FALSE, augmented = FALSE)
  training_df <- split %>% dplyr::filter(!.data$test)
  cat_df <- training_df %>%
    dplyr::filter(!is.na(.data$given_category)) %>%
    dplyr::group_by(.data$given_category) %>%
    dplyr::count()
  cat_df$missing <- max(cat_df$n) - cat_df$n
  cat_to_augment <- cat_df %>% dplyr::filter(.data$missing > 0)
  na_cat <- alert_training_df %>%
    dplyr::filter(is.na(.data$given_category)) %>%
    dplyr::mutate(test = FALSE, deleted = FALSE, augmented = FALSE)

  # getting the alerts to augment which are categories with less annotations, we do not consider location for this purpose
  alerts_to_augment <- training_df %>%
    dplyr::filter(.data$given_category %in% cat_to_augment$given_category) %>%
    dplyr::distinct(
      .data$topic,
      .data$date,
      .data$given_category,
      .keep_all = TRUE
    )
  alerts_to_augment$hour <- sapply(alerts_to_augment$date, function(v) 23)
  # adding extra posts ignoring location to build augmented alerts
  alerts_to_augment <- add_topposts(
    alerts_to_augment,
    1000,
    ignore_place = TRUE,
    progress = function(value, message) progress(value * 0.9, message)
  )

  # Iterating over categories to augment
  # Creating alerts to augment underrepresented categories, using same settings than existing alerts but looking for more posts not limiting by country
  alerts_for_augment <- list()
  smallest_cat <- Inf
  if (nrow(cat_to_augment) > 0) {
    progress(0.9, "Augmenting categories")
    for (icat in 1:nrow(cat_to_augment)) {
      cat <- cat_to_augment$given_category[[icat]]
      to_add <- cat_to_augment$missing[[icat]]
      size_after_augment <- cat_to_augment$n[[icat]]
      posts_exhausted <- FALSE
      found_posts <- FALSE
      ialert <- 1
      # Looping over all alerts with extra posts for this category until no more new posts are found or the expected number of alerts to add is reached
      while (to_add > 0 && !posts_exhausted) {
        # looping over alerts to augment which have
        if (alerts_to_augment$given_category[[ialert]] == cat) {
          topposts <- alerts_to_augment$topposts[[ialert]]
          # getting the index of posts to extract from 5 to five per language, it will start on 10 since these have already been used by original alerts
          if (!exists("ipost", where = topposts)) topposts$ipost <- 7
          # getting posts to add
          langs <- names(topposts)
          langs <- langs[langs != "ipost"]
          i0 <- topposts$ipost + 1
          i1 <- i0 + 5
          # getting new posts
          new_posts <- lapply(langs, function(l) {
            if (length(topposts[[l]]) >= i1) topposts[[l]][i0:i1] else if (
              length(topposts[[l]]) > i0
            )
              topposts[[l]][i0:length(topposts[[l]])] else list()
          })
          topposts$ipost <- i1
          alerts_to_augment$topposts[[ialert]] <- topposts
          # counting new posts
          tcount <- sum(sapply(new_posts, function(t) length(t)))
          # message(paste("cat", cat, "to_add",to_add, "ialert", ialert, "i0", i0, "i1", i1, "tcount", tcount))
          if (tcount > 0) {
            found_posts <- TRUE
            new_posts <- setNames(new_posts, langs)
            alerts_for_augment[[length(alerts_for_augment) + 1]] <- (list(
              date = alerts_to_augment$date[[ialert]],
              topic = alerts_to_augment$topic[[ialert]],
              country = alerts_to_augment$country[[ialert]],
              topwords = alerts_to_augment$topwords[[ialert]],
              number_of_posts = alerts_to_augment$number_of_posts[[ialert]],
              topposts = new_posts,
              given_category = alerts_to_augment$given_category[[ialert]],
              episomer_category = NA,
              augmented = TRUE,
              test = FALSE
            ))
            size_after_augment <- size_after_augment + 1
            to_add <- to_add - 1
          }
        }
        if (ialert == nrow(alerts_to_augment)) {
          ialert <- 1
          if (!found_posts) posts_exhausted <- TRUE
          found_posts <- FALSE
        } else {
          ialert <- ialert + 1
        }
      }
      if (size_after_augment < smallest_cat) smallest_cat <- size_after_augment
    }
  }

  # Adding the new alerts to the original set
  training_df$augmented <- FALSE
  if (length(alerts_for_augment) > 0) {
    augmented_alerts <- jsonlite::rbind_pages(list(
      training_df,
      tibble::tibble(
        date = sapply(alerts_for_augment, function(r) r$date),
        topic = sapply(alerts_for_augment, function(r) r$topic),
        country = as.character(sapply(
          alerts_for_augment,
          function(r) r$country
        )),
        topwords = sapply(alerts_for_augment, function(r) r$topwords),
        number_of_posts = sapply(
          alerts_for_augment,
          function(r) r$number_of_posts
        ),
        topposts = lapply(alerts_for_augment, function(r) r$topposts),
        given_category = sapply(
          alerts_for_augment,
          function(r) r$given_category
        ),
        episomer_category = as.character(sapply(
          alerts_for_augment,
          function(r) r$episomer_category
        )),
        augmented = as.logical(sapply(
          alerts_for_augment,
          function(r) r$augmented
        )),
        test = as.logical(sapply(alerts_for_augment, function(r) r$test))
      )
    ))
  } else augmented_alerts = training_df
  balanced_training <- jsonlite::rbind_pages(lapply(
    1:nrow(cat_df),
    function(i) {
      augmented_in_cat <- augmented_alerts %>%
        dplyr::filter(.data$given_category == cat_df$given_category[[i]])
      frac <- smallest_cat / nrow(augmented_in_cat)
      x = 1:nrow(augmented_in_cat)
      augmented_in_cat %>%
        dplyr::mutate(
          deleted = sample(x, length(x), replace = FALSE) > length(x) * frac
        )
    }
  ))

  test_df$augmented <- FALSE
  test_df$deleted <- FALSE
  balanced_alerts <- jsonlite::rbind_pages(list(
    balanced_training,
    test_df,
    na_cat
  ))
  balanced_alerts
}

# Read the alert training runs that will determine the algorithms to test for alert classification
get_alert_training_runs_df <- function() {
  `%>%` <- magrittr::`%>%`
  # Reading runs without type to detect if the balance and force columns are present since added on an intermediate version
  current <- readxl::read_excel(get_alert_training_path(), sheet = "Runs")
  if ("Force to use" %in% colnames(current))
    data_types <- c(
      "numeric",
      "text",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "text",
      "text",
      "text",
      "text",
      "logical",
      "logical",
      "logical",
      "text",
      "text"
    ) else
    data_types <- c(
      "numeric",
      "text",
      "numeric",
      "numeric",
      "numeric",
      "numeric",
      "text",
      "text",
      "text",
      "text",
      "logical",
      "text",
      "text"
    )

  current <- readxl::read_excel(
    get_alert_training_path(),
    col_types = data_types,
    sheet = "Runs"
  )
  # Adding default values to new columns if not present
  if (!("Force to use" %in% colnames(current))) {
    current$`Force to use` <- FALSE
    current$`Balance classes` <- TRUE
  }
  # giving snake_case format to column names
  current <- current %>%
    dplyr::transmute(
      ranking = .data$`Ranking`,
      models = .data$`Models`,
      alerts = .data$`Alerts`,
      runs = .data$`Runs`,
      f1score = .data$`F1Score`,
      accuracy = .data$`Accuracy`,
      precision_by_class = .data$`Precision by Class`,
      sensitivity_by_class = .data$`Sensitivity by Class`,
      fscore_by_class = .data$`FScore by Class`,
      last_run = .data$`Last run`,
      balance_classes = .data$`Balance classes`,
      force_to_use = .data$`Force to use`,
      active = .data$`Active`,
      documentation = .data$`Documentation`,
      custom_parameters = .data$`Custom Parameters`
    )
  current$`custom_parameters` <- lapply(
    current$`custom_parameters`,
    function(json) jsonlite::fromJSON(json)
  )
  current
}

# Classify the provided alerts retraining the models if requested
classify_alerts <- function(
  alerts,
  retrain = FALSE,
  progress = function(value, message) {
  }
) {
  `%>%` <- magrittr::`%>%`
  if (is.null(alerts) || nrow(alerts) == 0) alerts else if (
    retrain &&
      min(
        (alerts %>%
          dplyr::group_by(.data$given_category) %>%
          dplyr::summarise(n = dplyr::n()) %>%
          dplyr::filter(!is.na(.data$given_category)))$n
      ) <
        10
  ) {
    warning(
      "In order to train alerts' classifier all categories must have at least 10 observations"
    )
    alerts
  } else {
    runs <- get_alert_training_runs_df()
    augment_classes <- nrow(
      runs %>% dplyr::filter(.data$active & .data$balance_classes)
    ) >
      0
    if (augment_classes && retrain) {
      progress(
        value = 0.2,
        message = "Balancing classes as requested per parameters"
      )
      alerts <- get_alert_balanced_df(
        alerts,
        progress = function(value, message) {
          progress(0.2 + value * 0.5, message)
        }
      )
    } else {
      alerts$augmented <- FALSE
      alerts$deleted <- FALSE
      alerts$test <- FALSE
    }
    alerts <- alerts %>% dplyr::mutate(`id` = as.character(1:nrow(alerts)))
    body = paste(
      "{",
      paste(
        "\"alerts\":",
        jsonlite::toJSON(alerts, pretty = T, auto_unbox = FALSE)
      ),
      if (retrain)
        paste(
          ",\"runs\":",
          jsonlite::toJSON(runs, pretty = T, auto_unbox = TRUE)
        ) else "",
      "}",
      sep = "\n"
    )
    progress(value = 0.7, message = "Training models")
    post_result <- httr::POST(
      url = get_scala_alert_training_url(),
      httr::content_type_json(),
      body = body,
      encode = "raw",
      encoding = "UTF-8"
    )
    if (httr::status_code(post_result) != 200) {
      stop(paste(
        "retrain web service failed with the following output: ",
        substring(
          httr::content(post_result, "text", encoding = "UTF-8"),
          1,
          100
        ),
        sep = "\n"
      ))
    } else {
      res <- jsonlite::fromJSON(httr::content(
        post_result,
        "text",
        encoding = "UTF-8"
      ))
      alerts$episomer_category <- (alerts %>%
        dplyr::left_join(res$alerts, by = "id"))[[
        if ("episomer_category" %in% colnames(alerts))
          "episomer_category.y" else "episomer_category"
      ]]

      if (retrain) {
        nruns = tibble::as_tibble(res$runs)
        nruns$custom_parameters <- jsonlite::fromJSON(
          jsonlite::toJSON(nruns$custom_parameters),
          simplifyVector = F
        )
        list(alerts = alerts, runs = nruns)
      } else alerts
    }
  }
}

# Retrain the alert classifiers with current alerts
retrain_alert_classifier <- function(
  progress = function(value, message) {
  }
) {
  `%>%` <- magrittr::`%>%`
  progress(value = 0.1, message = "Getting alerts to retrain")
  alerts = get_alert_training_df()
  ret <- classify_alerts(alerts, retrain = TRUE, progress = progress)
  if (exists("alerts", where = ret)) {
    progress(value = 0.9, message = "Writing results")
    write_alert_training_db(
      alerts = ret$alerts %>% dplyr::filter(!.data$augmented),
      runs = ret$runs
    )
  }
}


# Write the provided alerts and runs to the alert annotation Excel
write_alert_training_db <- function(
  alerts,
  runs = get_alert_training_runs_df()
) {
  `%>%` <- magrittr::`%>%`
  wb <- openxlsx::createWorkbook()
  openxlsx::addWorksheet(wb, "Alerts")
  openxlsx::addWorksheet(wb, "Runs")
  # writing data to the worksheet
  alerts <- alerts %>%
    dplyr::transmute(
      `Date` = .data$date,
      `Topic` = .data$topic,
      `Region` = .data$country,
      `Top words` = .data$topwords,
      `Posts` = .data$number_of_posts,
      `Top Posts` = lapply(
        .data$topposts,
        function(l)
          gsub("\", \"", "\",\n        \"", jsonlite::toJSON(l, pretty = TRUE))
      ),
      `Given Category` = .data$given_category,
      `Episomer Category` = .data$episomer_category
    )
  openxlsx::writeDataTable(
    wb,
    sheet = "Alerts",
    alerts,
    colNames = TRUE,
    startRow = 1,
    startCol = "A"
  )
  runs <- runs %>%
    dplyr::transmute(
      `Ranking` = .data$ranking,
      `Models` = .data$models,
      `Alerts` = .data$alerts,
      `Runs` = .data$runs,
      `F1Score` = .data$f1score,
      `Accuracy` = .data$accuracy,
      `Precision by Class` = .data$precision_by_class,
      `Sensitivity by Class` = .data$sensitivity_by_class,
      `FScore by Class` = .data$fscore_by_class,
      `Last run` = .data$last_run,
      `Balance classes` = .data$balance_classes,
      `Force to use` = .data$force_to_use,
      `Active` = .data$active,
      `Documentation` = .data$documentation,
      `Custom Parameters` = lapply(
        .data$custom_parameters,
        function(l) jsonlite::toJSON(l, auto_unbox = TRUE)
      )
    )
  openxlsx::writeDataTable(
    wb,
    sheet = "Runs",
    runs,
    colNames = TRUE,
    startRow = 1,
    startCol = "A"
  )

  # setting some minimal formatting
  openxlsx::setColWidths(wb, "Alerts", cols = c(4, 6), widths = c(50))
  openxlsx::setRowHeights(wb, "Alerts", rows = 1, heights = 20)
  openxlsx::setRowHeights(wb, "Runs", rows = 1, heights = 20)
  openxlsx::addStyle(
    wb,
    sheet = "Alerts",
    style = openxlsx::createStyle(
      fontSize = 10,
      halign = "center",
      fgFill = "#ff860d",
      border = c("top", "bottom", "left", "right"),
      textDecoration = "bold",
      wrapText = T,
      fontColour = "#222222"
    ),
    rows = 1,
    cols = 1:ncol(alerts),
    gridExpand = FALSE
  )
  openxlsx::addStyle(
    wb,
    sheet = "Runs",
    style = openxlsx::createStyle(
      fontSize = 10,
      halign = "center",
      fgFill = "#ffde59",
      border = c("top", "bottom", "left", "right"),
      textDecoration = "bold",
      wrapText = T,
      fontColour = "#222222"
    ),
    rows = 1,
    cols = 1:ncol(runs),
    gridExpand = FALSE
  )
  openxlsx::addStyle(
    wb,
    sheet = "Alerts",
    style = openxlsx::createStyle(
      fontSize = 10,
      border = c("top", "bottom", "left", "right")
    ),
    rows = 1:nrow(alerts) + 1,
    cols = 1:ncol(alerts),
    gridExpand = TRUE
  )
  openxlsx::addStyle(
    wb,
    sheet = "Runs",
    style = openxlsx::createStyle(
      fontSize = 10,
      border = c("top", "bottom", "left", "right")
    ),
    rows = 1:nrow(runs) + 1,
    cols = 1:ncol(runs),
    gridExpand = TRUE
  )
  openxlsx::freezePane(wb, "Alerts", firstActiveRow = 2)
  openxlsx::freezePane(wb, "Runs", firstActiveRow = 2)
  openxlsx::saveWorkbook(wb, get_user_alert_training_path(), overwrite = TRUE)
}

# @title get_plan S3 class constructor
# @description Create a new 'get plan' for importing posts using a social media API
# @param network Character, name of the social media targeted by this plan
# @param expected_end Character(\%Y-\%m-\%d \%H:\%M:\%S) establishing the target end datetime of this plan
# @param scheduled_for Character(\%Y-\%m-\%d \%H:\%M:\%S) establishing the expected datetime for next execution, default: Sys.time()
# @param start_on Character(\%Y-\%m-\%d \%H:\%M:\%S) establishing the datetime when this plan was first executed, default: NULL
# @param end_on Character(\%Y-\%m-\%d \%H:\%M:\%S) establishing the datetime when this plan has finished, default: NULL
# @param requests Integer, number of requests successfully executed, default: 0
# @param got_rows Boolean, if the plan has collected rows, default: FALSE
# @param progress Numeric, percentage of progress of current plan defined when since_target_id is known or when a request returns no more results, default: 0
# @return The get_plan object defined by input parameters
# @details A plan is an S3 class representing a commitment to download posts from different social media API
# It targets a specific time frame defined from the last post collected by the previous plan, if any, and the last post collected on its first request
# This commitment will be valid during a period of time defined from the time of its first execution until the end_on parameter
# a plan will perform several requests to the underlysing API and each time a request is performed the number of requests will be increased.
# The field scheduled_for indicates the time when the next request is expected to be executed.
# @examples
# \dontrun{
#  #creating the default plan
#  parse_plan_attributes()
# }
# @seealso
#  \code{\link[bit64]{as.integer64.character}}
# @rdname get_plan
# @importFrom bit64 as.integer64
parse_plan_attributes <- function(
  network,
  expected_end,
  scheduled_for = Sys.time(),
  start_on = NULL,
  end_on = NULL,
  requests = 0,
  got_rows = FALSE,
  progress = 0.0,
  ...
) {
  common_attributes <- list(
    "network" = network,
    "expected_end" = if (!is.null(unlist(expected_end)))
      strptime(unlist(expected_end), "%Y-%m-%d %H:%M:%S") else NULL,
    "scheduled_for" = if (!is.null(unlist(scheduled_for)))
      strptime(unlist(scheduled_for), "%Y-%m-%d %H:%M:%S") else NULL,
    "start_on" = if (!is.null(unlist(start_on)))
      strptime(unlist(start_on), "%Y-%m-%d %H:%M:%S") else NULL,
    "end_on" = if (!is.null(unlist(end_on)))
      strptime(unlist(end_on), "%Y-%m-%d %H:%M:%S") else NULL,
    "requests" = unlist(requests),
    "got_rows" = unlist(got_rows),
    "progress" = unlist(progress)
  )
  specific_attributes <- sm_plan_parse_attributes(network, ...)
  me <- c(common_attributes, specific_attributes)
  return(me)
}

format_plan <- function(p) {
  sm_plan_format(p)
}

# Update a plan after a search request is done
# If first request, started and max will be set
# If results are non-empty the current social media cursor is updated
# If no results are obtained the search is supposed to be finished
update_plan_after_request <- function(plan, results) {
  # increasing the number of requests
  plan$requests <- plan$requests + 1
  # setting the start date if not set
  if (is.null(plan$start_on)) {
    plan$start_on = Sys.time()
  }

  # check if the current request got rows
  req_got_rows <- results$count > 0
  if (req_got_rows) {
    # set got rows for this plan
    plan$got_rows <- TRUE

    # updating plan by network parameters
    plan <- sm_api_update_plan_after_request(plan, results)

    # setting plan progress
    plan$progress <- sm_plan_get_progress(plan)
  } else {
    # if no rows have been obtained then we consider the plan has ended
    plan$end_on <- Sys.time()
    plan$progress <- 1.0
  }
  return(plan)
}

# receive two plans and return a new including proprerties from both. If a property exists in both plans values are taken from the first
merge_plans <- function(p1, p2) {
  keys <- unique(c(names(p1), names(p2)))
  as.list(setNames(
    mapply(function(x, y) if (is.null(y)) x else y, p1[keys], p2[keys]),
    keys
  ))
}

# @title Update plans schedule
# @description Updating plans for a particular topic
# @param plans The existing plans for the topic, default: list()
# @param schedule_target target minutes for finishing a plan
# @return updated list of 'get_plan'
# @details
# This function will update the plan list of a topic taking in consideration the search span
# This function is called at the beginning of each search loop iteration applying the following rules
# If no plans are set, a new plan for getting all possible posts will be set
# If current plan has started and the expected end has passed, a new plan will be added for collecting new posts (previous plan will be stored for future execution if possible)
# Any finished plans after the first will be discharged. Note that a maximum of 100 plans are kept)
# @returns the updated list of plans
# @examples
# \dontrun{
#  #Getting default plan
#  update_plans(plans = list(), schedule_span = 120)
#  #Updating topics for first topic
#  update_plans(plans = conf$topics[[1]]$plan, schedule_span = conf$collect_span)
# }
# @rdname update_plans
update_plans_schedule <- function(plans = list(), network, schedule_span) {
  # Testing if there are plans present
  if (length(plans) == 0) {
    # Getting default plan for when no existing plans are present setting the expected end
    elems <- list(
      network = network,
      expected_end = strftime(
        Sys.time() + 60 * schedule_span,
        "%Y-%m-%d %H:%M:%S"
      )
    )
    elems <- c(
      elems,
      sm_plan_first_attributes(network)
    )
    new_plan <- do.call(parse_plan_attributes, elems)
    return(list(new_plan))
  } else if (plans[[1]]$requests > 0 && plans[[1]]$expected_end < Sys.time()) {
    # creating a new plan if expected end has passed
    elems <- list(
      network = plans[[1]]$network,
      # expected end should be 'schedule_span' minutes after previous plan expected en unless this is in the past in which case the span would cound from now
      expected_end = if (
        Sys.time() > plans[[1]]$expected_end + 60 * schedule_span
      ) {
        strftime(Sys.time() + 60 * schedule_span, "%Y-%m-%d %H:%M:%S")
      } else {
        strftime(
          plans[[1]]$expected_end + 60 * schedule_span,
          "%Y-%m-%d %H:%M:%S"
        )
      }
    )
    elems <- c(
      elems,
      sm_plan_next_attributes(network, plans)
    )
    first <- do.call(parse_plan_attributes, elems)

    # removing ended plans
    non_ended <- plans[sapply(plans, function(x) is.null(x$end_on))]
    # removing plans if more of 100 plans are activeff
    return(append(
      list(first),
      if (length(non_ended) < 100) non_ended else non_ended[1:100]
    ))
  } else {
    first <- plans[[1]]
    rest <- plans[-1]
    # removing ended plans
    non_ended <- rest[unlist(sapply(rest, function(x) is.null(x$end_on)))]
    # removing ended plans
    return(append(
      list(first),
      if (length(non_ended) < 100) non_ended else non_ended[1:100]
    ))
  }
}


# finish the provided plans
finish_plans <- function(plans = list()) {
  # Testing if there are plans present
  if (length(plans) == 0) {
    list()
  } else {
    # creating a new plan if expected end has passed
    lapply(plans, function(p) {
      attr <- list(
        network = p$network,
        expected_end = strftime(
          if (is.null(p$end_on)) {
            Sys.time() - conf$schedule_span * 60
          } else {
            p$end_on
          },
          "%Y-%m-%d %H:%M:%S"
        ),
        scheduled_for = strftime(p$scheduled_for, "%Y-%m-%d %H:%M:%S"),
        start_on = strftime(
          if (is.null(p$start_on)) {
            Sys.time() - conf$schedule_span * 60
          } else {
            p$start_on
          },
          "%Y-%m-%d %H:%M:%S"
        ),
        end_on = strftime(
          if (is.null(p$end_on)) {
            Sys.time() - conf$schedule_span * 60
          } else {
            p$end_on
          },
          "%Y-%m-%d %H:%M:%S"
        ),
        requests = p$requests,
        progress = 1.0
      )

      updated <- do.call(parse_plan, attrs)
      merge_plans(p, updated)
    })
  }
}

#Environment for storing cached data
cached <- new.env()


#' @title Getting already aggregated time series produced by \code{\link{detect_loop}}
#' @description Reads and returns the required aggregated dataset for the selected period and topics defined by the filter.
#' @param dataset A character(1) vector with the name of the series to request, it must be one of 'country_counts', 'geolocated', 'topwords', 'hashtags', 'entities', 'urls', 'contexts', default: 'country_counts'
#' @param cache Whether to use the cache for lookup and storing the returned dataframe, default: TRUE
#' @param filter A named list defining the filter to apply on the requested series, it should be in the form of a named list; e.g., list(post_geo_country_code=list('FR', 'DE')) default: list()
#' @param top_field Name of the top field used with top_frequency to enable optimisation for getting only most frequent elements. It will only keep top 500 items after first 50k lines on reverse index order
#' @param top_freq character, Name of the frequency fields used with top_field to enable optimisation for retrieving only most frequent elements.
#' It will only keep top 500 items after first 50k rows on reverse index order
#' @return A data frame containing the requested series for the requested period
#' @details This function returns data aggregated by episomer. The data is found on the 'series' folder, which contains Rds files per weekday and type of series.
#' starting on v 1.0.x it will also look on Lucene indexes situated on fs folder. Names of files and folders are parsed to limit the files to be read.
#' When using Lucene indexes, filters are directly applied on read. This is an improvement compared 'series' folder where filters are applied
#' after read. All returned rows are joined into a single dataframe.
#' If no filter is provided, the entire data series is returned, which can result in millions of rows depending on the time series.
#' To limit by period, the filter list must include an element 'period' containing a date vector or list with two dates representing the start and end of the request.
#'
#' To limit by topic, the filter list must include an element 'topic' containing a non-empty character vector or list with the names of the topics to return.
#'
#' The available time series are:
#' \itemize{
#'   \item{"country_counts" counting posts and reposts by posted date, hour and country}
#'
#'   \item{"geolocated" counting posts and reposts by posted date and the smallest possible geolocated unit (city, administrative level or country)}
#'
#'   \item{"topwords" counting posts and reposts by posted date, country and the most popular words, (this excludes words used in the topic search)}
#' }
#' The returned dataset can be cached for further calls if requested. Only one dataset per series is cached.
#' @examples
#' \dontrun{
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    # Getting all country posts between 2020-jan-10 and 2020-jan-31 for all topics
#'    df <- get_aggregates(
#'      dataset = "country_counts",
#'      filter = list(period = c("2020-01-10", "2020-01-31"))
#'    )
#'
#'    # Getting all country posts for the topic dengue
#'    df <- get_aggregates(dataset = "country_counts", filter = list(topic = "dengue"))
#'
#'    # Getting all country posts between 2020-jan-10 and 2020-jan-31 for the topic dengue
#'     df <- get_aggregates(
#'         dataset = "country_counts",
#'          filter = list(topic = "dengue", period = c("2020-01-10", "2020-01-31"))
#'     )
#' }
#' @seealso
#'  \code{\link{detect_loop}}
#'  \code{\link{fs_loop}}
#' @rdname get_aggregates
#' @export
#' @importFrom magrittr `%>%`
#' @importFrom dplyr filter
#' @importFrom jsonlite rbind_pages
#' @importFrom utils tail
#' @importFrom utils URLencode
get_aggregates <- function(
  dataset = "country_counts",
  cache = TRUE,
  filter = list(),
  top_field = NULL,
  top_freq = NULL
) {
  `%>%` <- magrittr::`%>%`
  # getting the name for cache lookup dataset dependency
  last_filter_name <- paste("last_filter", dataset, sep = "_")

  # Setting the last aggregate filter
  last_agg <- get_aggregated_period()
  last_agg <- paste(last_agg$last, last_agg$last_hour)
  filter$last_aggregate <- last_agg

  # checking whether we can reuse the cache
  reuse_filter <- (cache &&
    exists(dataset, where = cached) &&
    exists(last_filter_name, where = cached) && #cache entry exists for that dataset
    (exists("last_aggregate", where = cached[[last_filter_name]]) &&
      cached[[last_filter_name]]$last_aggregate == filter$last_aggregate) && # No new aggregation has been finished
    (length(setdiff(names(cached[[last_filter_name]]), names(filter))) == 0) && # All filters in the cache are also present in the query
    (!exists("period", cached[[last_filter_name]]) || # all periods are cached or
      (exists("period", cached[[last_filter_name]]) && # there are some periods cached
        exists("period", filter) && # there is a period on the filter
        cached[[last_filter_name]]$period[[1]] <= filter$period[[1]] && # and filtered period is contained on cached period
        cached[[last_filter_name]]$period[[2]] >= filter$period[[2]])))
  for (field in names(filter)) {
    if (!(field %in% c("last_aggregate", "period"))) {
      reuse_filter <- reuse_filter &&
        (!exists(field, cached[[last_filter_name]]) || # all values are cached or,
          (
            exists(field, cached[[last_filter_name]]) && # there are some values cached
              all(filter[[field]] %in% cached[[last_filter_name]][[field]]) # all filtered values are cached
          ))
    }
  }
  if (reuse_filter && !"created_date" %in% names(cached[[dataset]])) {
      reuse_filter <- FALSE
  }
  if (reuse_filter) {
    # On cache hit returning from cache
    ret <- (cached[[dataset]] %>%
      dplyr::filter(
        (if (exists("period", where = filter))
          .data$created_date >= filter$period[[1]] &
            .data$created_date <= filter$period[[2]] else TRUE)
      ))
    for (field in names(filter)) {
      if (
        exists(field, where = filter) &&
          !(field %in% c("last_aggregate", "period")) &&
	  nrow(ret) > 0
      ) {
        filt <- filter[[field]]
        ret <- ret %>%
          dplyr::filter(!!as.symbol(field) %in% filt)
      }
    }
    ret
  } else {
    # Overriding current filter when there is no cache hit
    cached[[last_filter_name]] <- filter
    # getting the aggregated data frame from the storage system
    q_url <- paste0(
      get_scala_aggregate_url(),
      "?jsonnl=true&serie=",
      URLencode(dataset, reserved = T)
    )
    if (!is.null(top_field) && !is.null(top_freq))
      q_url <- paste0(
        q_url,
        "&topField=",
        top_field,
        "&topFrequency=",
        top_freq
      )
    for (field in names(filter)) {
      if (field == "topic")
        q_url <- paste0(
          q_url,
          "&topic=",
          URLencode(paste0(filter$topic, collapse = ";"), reserved = T)
        ) else if (field == "period")
        q_url <- paste0(
          q_url,
          "&from=",
          URLencode(
            strftime(filter$period[[1]], format = "%Y-%m-%d"),
            reserved = T
          ),
          "&to=",
          URLencode(
            strftime(filter$period[[2]], format = "%Y-%m-%d"),
            reserved = T
          )
        ) else if (field != "last_aggregate")
        q_url <- paste0(
          q_url,
          "&filter=",
          URLencode(field, reserved = T),
          ":",
          URLencode(paste0(filter[[field]], collapse = ";"), reserved = T)
        )
    }

    measure_time <- function(f) {
      start.time <- Sys.time()
      ret <- f()
      end.time <- Sys.time()
      time.taken <- end.time - start.time
      message(time.taken)
      ret
    }
    #message(q_url)
    agg_df = (if (conf$onthefly_api)
      jsonlite::stream_in(url(q_url), verbose = FALSE) else {
      if (!file.exists(file.path(conf$data_dir, "tmp")))
        dir.create(file.path(conf$data_dir, "tmp"))
      dest = file.path(
        conf$data_dir,
        "tmp",
        paste0("aggregate_", as.integer(stats::runif(1, 1, 99999)), ".json")
      )
      download.file(url = q_url, destfile = dest, quiet = TRUE)
      on.exit(if (file.exists(dest)) file.remove(dest))
      df <- jsonlite::stream_in(file(dest), verbose = FALSE)
      df
    })
    # Calculating the created week
    if (nrow(agg_df) > 0) {
      agg_df$created_week <- strftime(
        as.Date(agg_df$created_date, format = "%Y-%m-%d"),
        format = "%G.%V"
      )
      agg_df$created_weeknum <- as.integer(strftime(
        as.Date(agg_df$created_date, format = "%Y-%m-%d"),
        format = "%G%V"
      ))
      agg_df$created_date <- as.Date(agg_df$created_date, format = "%Y-%m-%d")
    } else {
      agg_df$created_week <- NULL
      agg_df$created_weeknum <- NULL
      agg_df$created_date <- NULL
    }
    #add any possible missing column that may have been removed by JSON null management
    ret <- add_missing(agg_df, dataset)

    cached[[dataset]] <- ret
    ret
  }
}

# This function registers the aggregated series that are computed by episomer.
# Each series is defined by a name, a date column, primary keys columns, variables columns, group by columns and sources expressions
# Each registered series uses the set_aggregated_posts function
# This function is called periodically by the search loop.
register_series <- function() {
  `%>%` <- magrittr::`%>%`
  #geolocated"
  set_aggregated_posts(
    name = "geolocated",
    dateCol = "created_date",
    pks = list(
      "created_date",
      "topic",
      "geo_country_code",
      "geo_code",
      "geo_name",
      "geo_name",
      "network"
    ),
    aggr = list(
      longitude = "avg",
      latitude = "avg",
      quotes = "sum",
      posts = "sum"
    ),
    sources_exp = c(
      "network",
      "topic",
      list("date_format(created_at, 'yyyy-MM-dd') as created_date", "is_quote"),
      get_location_columns()
    ),
    vars = list(
      paste("avg(", get_location_var("geo_longitude"), ") as longitude"),
      paste("avg(", get_location_var("geo_latitude"), ") as latitude"),
      "cast(sum(case when is_quote then 1 else 0 end) as Integer) as quotes",
      "cast(sum(case when is_quote then 0 else 1 end) as Integer) as posts"
    ),
    group_by = list(
      "network",
      "topic",
      "created_date",
      paste(get_location_var("geo_country_code"), "as geo_country_code"),
      paste(get_location_var("geo_code"), "as geo_code"),
      paste(get_location_var("geo_name"), "as geo_name")
    )
  )
  #topwords
  # Getting topic words to exclude
  topic_word_to_exclude <- unlist(sapply(1:length(conf$topics), function(i) {
    terms <- strsplit(conf$topics[[i]]$query, " |OR|\"|AND|,|\\.| |'|-|_")[[1]]
    terms <- terms[terms != ""]
    paste(conf$topics[[i]]$topic, "_", terms, sep = "")
  }))
  lang_stop_words <- paste(
    "'",
    unlist(lapply(
      conf$languages,
      function(l)
        lapply(get_stop_words(l$code), function(t) paste(l$code, t, sep = "_"))
    )),
    "'",
    sep = "",
    collapse = ","
  )
  # Getting top words aggregation
  set_aggregated_posts(
    name = "topwords",
    dateCol = "created_date",
    pks = list(
      "created_date",
      "topic",
      "post_geo_country_code",
      "token",
      "network"
    ),
    aggr = list(frequency = "sum", original = "sum", quotes = "sum"),
    sources_exp = c(
      list(
        "network",
        "topic",
        "date_format(created_at, 'yyyy-MM-dd') as created_date",
        "is_quote",
        "lang",
        "explode(split(text,'[^a-zA-Z\\'\\\\p{L}]')) as token"
      ),
      get_location_columns()
    ),
    #, sort_by = list(
    #  "topic"
    #  , "post_geo_country_code"
    #  , "created_at"
    #)
    filter_by = list(
      "length(token) > 1",
      "lower(token) not in ('via', 'rt', 'http', 'www', 'https', 'co', 't')",
      paste(
        "lower(concat(topic, '_', token)) not in (",
        paste(
          "'",
          sapply(topic_word_to_exclude, function(t) tolower(t)),
          "'",
          collapse = ",",
          sep = ''
        ),
        ")",
        sep = ""
      ),
      paste("concat(lang, '_', token) not in (", lang_stop_words, ")")
    ),
    vars = list(
      "network",
      "topic",
      "token",
      "created_date",
      paste(get_location_var("geo_country_code"), "as geo_country_code"),
      "count(1) as frequency",
      "sum(case when is_quote then 0 else 1 end) as original",
      "sum(case when is_quote then 1 else 0 end) as quotes"
    ),
    group_by = list(
      "network",
      "topic",
      "token",
      "created_date",
      paste(get_location_var("geo_country_code"), "as geo_country_code")
    )
  )
  #country_counts
  # Getting the expression for known users and writing it as a file so it can be read and applied by spark on query
  known_user <-
    paste(
      "user_name in ('",
      paste(get_known_users(), collapse = "','"),
      "') ",
      sep = ""
    )
  params <- list(
    known_quotes = paste(
      "cast(sum(case when is_quote and ",
      known_user,
      "then 1 else 0 end) as Integer) as known_quotes"
    ),
    known_original = paste(
      "cast(sum(case when not is_quote and ",
      known_user,
      "then 1 else 0 end) as Integer) as known_original"
    )
  )

  # Aggregation at the country level
  set_aggregated_posts(
    name = "country_counts",
    dateCol = "created_date",
    pks = list(
      "created_date",
      "topic",
      "created_hour",
      "geo_country_code",
      "network"
    ),
    aggr = list(
      quotes = "sum",
      posts = "sum",
      know_posts = "sum",
      know_original = "sum"
    ),
    sources_exp = c(
      list("topic", "created_at", "is_quote", "user_name", "network"),
      get_location_columns()
    ),
    group_by = list(
      "network",
      "topic",
      "date_format(created_at, 'yyyy-MM-dd') as created_date",
      "date_format(created_at, 'HH') as created_hour",
      paste(get_location_var("geo_country_code"), "as geo_country_code")
    ),
    vars = list(
      "cast(sum(case when is_quote then 1 else 0 end) as Integer) as quotes",
      "cast(sum(case when is_quote then 0 else 1 end) as Integer) as original",
      "@known_quotes",
      "@known_original"
    ),
    params = params
  )
  # Getting tags aggregation
  set_aggregated_posts(
    name = "tags",
    dateCol = "created_date",
    pks = list(
      "created_date",
      "topic",
      "geo_country_code",
      "entity",
      "network"
    ),
    aggr = list(frequency = "sum", original = "sum", reposts = "sum"),
    sources_exp = c(
      list(
        "network",
        "topic",
        "date_format(created_at, 'yyyy-MM-dd') as created_date",
        "is_quote",
        "explode(tags) as tag"
      ),
      get_location_columns()
    ),
    vars = list(
      "network",
      "topic",
      "tag",
      "created_date",
      paste(get_location_var("geo_country_code"), "as geo_country_code"),
      "count(1) as frequency",
      "sum(case when is_quote then 0 else 1 end) as original",
      "sum(case when is_quote then 1 else 0 end) as quotes"
    ),
    group_by = list(
      "network",
      "topic",
      "tag",
      "created_date",
      paste(get_location_var("geo_country_code"), "as geo_country_code")
    )
  )
  # Getting tags aggregation
  set_aggregated_posts(
    name = "urls",
    dateCol = "created_date",
    pks = list(
      "created_date",
      "topic",
      "geo_country_code",
      "entity",
      "network"
    ),
    aggr = list(frequency = "sum", original = "sum", quotes = "sum"),
    sources_exp = c(
      list(
        "network",
        "topic",
        "date_format(created_at, 'yyyy-MM-dd') as created_date",
        "is_quote",
        "explode(urls) as url"
      ),
      get_location_columns()
    ),
    vars = list(
      "network",
      "topic",
      "url",
      "created_date",
      paste(get_location_var("geo_country_code"), "as geo_country_code"),
      "count(1) as frequency",
      "sum(case when is_quote then 0 else 1 end) as original",
      "sum(case when is_quote then 1 else 0 end) as quotes"
    ),
    group_by = list(
      "network",
      "topic",
      "url",
      "created_date",
      paste(get_location_var("geo_country_code"), "as geo_country_code")
    )
  )
  # Getting tags aggregation
  set_aggregated_posts(
    name = "categories",
    dateCol = "created_date",
    pks = list(
      "created_date",
      "topic",
      "geo_country_code",
      "entity",
      "network"
    ),
    aggr = list(frequency = "sum", original = "sum", quotes = "sum"),
    sources_exp = c(
      list(
        "network",
        "topic",
        "date_format(created_at, 'yyyy-MM-dd') as created_date",
        "is_quote",
        "explode(categories) as category"
      ),
      get_location_columns()
    ),
    vars = list(
      "network",
      "topic",
      "category",
      "created_date",
      paste(get_location_var("geo_country_code"), "as geo_country_code"),
      "count(1) as frequency",
      "sum(case when is_quote then 0 else 1 end) as original",
      "sum(case when is_quote then 1 else 0 end) as quotes"
    ),
    group_by = list(
      "network",
      "topic",
      "category",
      "created_date",
      paste(get_location_var("geo_country_code"), "as geo_country_code")
    )
  )
}

# getting last aggregation date or NA if first time
# date is obtained by sorting and reading first and last file on the series folder and in the fs folder containing Lucene indexes

get_aggregated_period <- function() {
  if (
    !exists("last_agg_request", cached) ||
      as.numeric(Sys.time() - cached$last_agg_request, units = "secs") > 60
  ) {
    cached$last_agg_request_value <- {
      fs_period <- tryCatch(
        {
          ret <- jsonlite::fromJSON(
            url(paste0(get_scala_period_url(), "?serie=country_counts")),
            simplifyVector = T
          )
          ret$first <- if (exists("first", where = ret)) {
            as.Date(strptime(ret$first, format = "%Y-%m-%d"))
          } else {
            NA
          }
          ret$last <- if (exists("last", where = ret)) {
            as.Date(strptime(ret$last, format = "%Y-%m-%d"))
          } else {
            NA
          }
          ret$last_hour <- if (exists("last_hour", where = ret)) {
            as.integer(ret$last_hour)
          } else {
            NA
          }
          ret
        },
        warning = function(w) {
          warning(w)
          list(first = NA, last = NA)
        },
        error = function(e) {
          warning(e)
          list(first = NA, last = NA)
        }
      )

      if (is.na(fs_period$first))
        list(first = NA, last = NA, last_hour = NA) else
        list(
          first = as.Date(fs_period$first),
          last = as.Date(fs_period$last),
          last_hour = fs_period$last_hour
        )
    }
    cached$last_agg_request <- Sys.time()
    if (is.na(cached$last_agg_request_value$last))
      rm("last_agg_request", envir = cached)
  }
  cached$last_agg_request_value
}

# Utility function to ask episomer to recalculate hashes of stored posts in Lucene indexes.
# This function is experimental for testing parallel scan of indexes
# This function is deprecated since no significant performance gain was observed
recalculate_hash <- function() {
  message("recalculating hashes")
  post_result <- httr::POST(
    url = get_scala_recalc_hash_url(),
    httr::content_type_json(),
    body = "",
    encode = "raw",
    encoding = "UTF-8"
  )
  if (httr::status_code(post_result) != 200) {
    stop(paste(
      "recalc hash web service failed with the following output: ",
      substring(httr::content(post_result, "text", encoding = "UTF-8"), 1, 100),
      sep = "\n"
    ))
  } else {
    message("hashes recalculated")
  }
}

# Adds any possible missing column to a dataset produced by an aggregated series
# This is necessary to ensure that all expected columns are present for data produced
# in old versions of episomer
add_missing <- function(df, dataset) {
  return(df)
}

#' @title Plot the trendline report of episomer dashboard
#' @description Generates a trendline chart of number of posts by region for a single topic, including alerts detected using the reweighted version of the EARS algorithm
#' @param sms Social media
#' @param topic Character(1) containing the topic to use for the report
#' @param countries Character vector containing the name of the countries and regions to plot or their respective indexes on the Shiny app select, default: c(1)
#' @param date_type Character vector specifying the time granularity of the report either 'created_weeknum' or 'created_date', default: 'created_date'
#' @param date_min Date indicating start of the reporting period, default: "1900-01-01"
#' @param date_max Date indicating end of the reporting period, default: "2100-01-01"
#' @param with_quotes Logical value indicating whether to include reposts in the time series, default: FALSE
#' @param alpha Numeric(1) value indicating the alert detection confidence, default: 0.025
#' @param alpha_outlier Numeric(1) value indicating the outliers detection confidence for downweighting, default: 0.05
#' @param k_decay Strength of outliers downweighting, default: 4
#' @param no_historic Number of observations to build the baseline for signal detection, default: 7
#' @param bonferroni_correction Logical value indicating whether to apply the Bonferroni correction for signal detection, default: FALSE
#' @param same_weekday_baseline Logical value indicating whether to use same day of weeks for building the baseline or consecutive days, default: FALSE
#' @return A named list containing two elements: 'chart' with the ggplot2 figure and 'data' containing the data frame that was used to build the chart.
#' @details Produces a multi-region line chart for a particular topic of number of posts collected based on the provided parameters.
#' Alerts will be calculated using a modified version of the EARS algorithm that applies a Farrington inspired downweighting of previous outliers.
#'
#' Days in this function are considered as contiguous blocks of 24 hours, starting from the previous hour preceding the last collected post.
#'
#' This function requires that \code{\link{search_loop}} and \code{\link{detect_loop}} have already run successfully in order to show results.
#' @examples
#' \dontrun{
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    #Getting trendline for dengue for South America for the last 30 days
#'    trend_line(
#'      topic = "dengue",
#'      countries = "South America",
#'      date_min = as.Date(Sys.time())-30,
#'      date_max=as.Date(Sys.time())
#'    )
#' }
#' @seealso
#'  \code{\link{create_map}}
#'  \code{\link{create_topwords}}
#'  \code{\link{generate_alerts}}
#'  \code{\link{detect_loop}}
#'  \code{\link{search_loop}}
#' @rdname trend_line
#' @export
#' @importFrom stringr str_replace_all
trend_line <- function(
  sms,
  topic,
  countries = c(1),
  date_type = "created_date",
  date_min = "1900-01-01",
  date_max = "2100-01-01",
  with_quotes = FALSE,
  alpha = 0.025,
  alpha_outlier = 0.05,
  k_decay = 4,
  no_historic = 7,
  bonferroni_correction = FALSE,
  same_weekday_baseline = FALSE
) {
  `%>%` <- magrittr::`%>%`
  stop_if_no_config()
  stop_if_no_fs()
  # If countries are names they have to be changes to region indexes
  if (is.character(countries) && length(countries) > 0) {
    reg <- get_country_items()
    countries = (1:length(reg))[sapply(
      1:length(reg),
      function(i) reg[[i]]$name %in% countries
    )]
  }
  # defining the environment variable for returning complementary data
  logenv <- new.env()

  # getting the data with counts and alerts from country counts
  df <-
    calculate_regions_alerts(
      sms = sms,
      topic = tolower(topic),
      regions = countries,
      date_type = date_type,
      date_min = date_min,
      date_max = date_max,
      with_quotes = with_quotes,
      alpha = alpha,
      alpha_outlier = alpha_outlier,
      k_decay = k_decay,
      no_historic = no_historic,
      bonferroni_correction = bonferroni_correction,
      same_weekday_baseline = same_weekday_baseline,
      logenv = logenv
    )
  # checking if some data points have been returned or return empty char
  if (nrow(df %>% dplyr::filter(.data$number_of_posts > 0)) > 0) {
    topic <- unname(get_topics_labels()[stringr::str_replace_all(
      topic,
      "%20",
      " "
    )])
    df$topic <- topic
    plot_trendline(
      df = df,
      sms = sms,
      countries = countries,
      topic = topic,
      date_min = date_min,
      date_max = date_max,
      date_type = date_type,
      alpha = alpha,
      alpha_outlier = alpha_outlier,
      k_decay = k_decay,
      total_count = logenv$total_count
    )
  } else {
    get_empty_chart("No data found for the selected topic, region and period")
  }
}


# Plot the trend_line chart for shiny app
plot_trendline <- function(
  df,
  sms,
  countries,
  topic,
  date_min,
  date_max,
  date_type,
  alpha,
  alpha_outlier,
  k_decay,
  total_count = NA
) {
  #Importing pipe operator
  `%>%` <- magrittr::`%>%`
  # getting regions and countries
  regions <- get_country_items()

  #turning off the scientific pen
  old <- options()
  on.exit(options(old))
  options(scipen = 999)

  # Calculate alert ranking to avoid alert symbol overlapping
  # this will be used for stacking points
  df <- df %>%
    dplyr::arrange(.data$alert, .data$date, .data$country) %>%
    dplyr::group_by(.data$alert, .data$date) %>%
    dplyr::mutate(rank = rank(.data$country, ties.method = "first")) %>%
    dplyr::ungroup()

  # Getting an alert only dataset for alert markers
  time_alarm <- data.frame(
    date = df$date[which(df$alert == 1)],
    country = df$country[which(df$alert == 1)],
    y = vapply(
      which(df$alert == 1),
      function(i) df$rank[[i]] * (max(df$limit, na.rm=TRUE)) / 30,
      double(1)
    ),
    # adding hover text for alerts
    Details = vapply(
      which(df$alert == 1),
      function(i) {
        paste(
          "\nAlert detected",
          "\nRegion:",
          df$country[[i]],
          "\nNumber of messages: ",
          df$number_of_posts[[i]],
          "\nBaseline: ",
          round(df$baseline[[i]]),
          "\nThreshold: ",
          round(df$limit[[i]]),
          "\nDate:",
          df$date[[i]],
          "\nKnown users posts: ",
          df$known_users[[i]],
          "\nKnown users ratio: ",
          round(df$known_ratio[[i]] * 100, 2),
          "%",
          sep = ""
        )
      },
      character(1)
    )
  )

  # adding hover text for alerts
  df$Details <-
    paste(
      "\nRegion:",
      df$country,
      "\nAlert: ",
      ifelse(df$alert == 1, "yes", "no"),
      "\nNumber of message: ",
      df$number_of_posts,
      "\nBaseline: ",
      round(df$baseline),
      "\nThreshold: ",
      round(df$limit),
      "\nDate:",
      df$date,
      "\nKnown users posts: ",
      df$known_users,
      "\nKnown users ratio: ",
      round(df$known_ratio * 100, 2),
      "%",
      "\nAlpha: ",
      alpha,
      "\nAlpha outliers: ",
      alpha_outlier,
      "\nK decay: ",
      k_decay,
      sep = ""
    )

  # Calculating minimum limit boundary #TODO: changing this behaviour for avoiding shadow reversing
  df$lim_start <- 2 * df$baseline - df$limit
  df$lim_start <- ifelse(df$lim_start < 0, 0, df$lim_start)

  # Calculating breaks of y axis
  y_breaks <- unique(floor(pretty(seq(
    0,
    (max(df$limit, df$number_of_posts, na.rm = TRUE, 0) + 1) * 1.1
  ))))

  # Calculating post location scope count message
  scope_count <- format(
    sum(df$number_of_posts),
    big.mark = " ",
    scientific = FALSE
  )
  total_count <- if (is.na(total_count)) NA else
    format(total_count, big.mark = " ", scientific = FALSE)
  location_message <- (paste("(n=", scope_count, ")", sep = ""))

  # plotting
  fig_line <- ggplot2::ggplot(
    df,
    ggplot2::aes(
      x = .data$date,
      y = .data$number_of_posts,
      label = .data$Details
    )
  ) +
    # Line
    ggplot2::geom_line(ggplot2::aes(colour = .data$country)) +
    {
      # Alert Points
      if (nrow(time_alarm) > 0)
        ggplot2::geom_point(
          data = time_alarm,
          mapping = ggplot2::aes(
            x = .data$date,
            y = .data$y,
            colour = .data$country
          ),
          shape = 2,
          size = 2
        )
    } +
    {
      # Line shadow
      if (length(df$lim_start[!is.na(df$lim_start)]) > 0)
        ggplot2::geom_ribbon(
          ggplot2::aes(
            ymin = .data$lim_start,
            ymax = .data$limit,
            colour = .data$country,
            fill = .data$country
          ),
          linetype = 2,
          alpha = 0.1
        )
    } +
    # Title
    ggplot2::labs(
      title = ifelse(
        length(countries) == 1,
        paste0(
          "Number of messages mentioning ",
          topic,
          " from ",
          date_min,
          " to ",
          date_max,
          "\n in ",
          if (as.integer(countries) == 1) "the world" else
            regions[[as.integer(countries)]]$name,
          " ",
          location_message
        ),
        paste0(
          "Number of messages mentioning ",
          topic,
          " from ",
          date_min,
          " to ",
          date_max,
          "\n in multiples regions ",
          location_message
        )
      ),
      fill = "Countries / Regions",
      color = "Countries / Regions"
    ) +
    ggplot2::xlab(paste(
      if (date_type == "created_weeknum") "Posted week" else "Posted date"
    )) + #, "(days are 24 hour blocks ening on last aggregated post in period)")) +
    ggplot2::ylab('Number of messages') +
    ggplot2::scale_y_continuous(
      breaks = y_breaks,
      limits = c(0, max(y_breaks)),
      expand = c(0, 0)
    ) +
    ggplot2::scale_x_date(
      date_labels = {
        x = df$date
        # custom logic for x axis
        days <- as.numeric(max(x) - min(x))
        weeks <- days / 7
        years <- days / 365
        # Date format if less or equal then 15 days
        if (days < 15 && date_type != "created_weeknum") {
          "%Y-%m-%d"
          # Week format if period is between 16 days and 20 weeks
          #  - Case for period ending on same day of week than period start
        } else if (weeks <= 20) {
          "%Y-%m-%d"
          #"%G-w%V"
          # Month format day of month in period start if period is less or equal to 2 years but more than 20 weeks
        } else if (years <= 2) {
          "%Y-%b"
          # Year format if more than 2 years
        } else {
          "%Y"
        }
      },
      expand = c(0, 0),
      breaks = function(x) {
        # custom logic for x axis
        days <- as.numeric(max(x) - min(x))
        weeks <- days / 7
        years <- days / 365
        # One label per day if period is less or equal then 15 days
        if (days < 15) {
          seq.Date(from = min(x), to = max(x), by = "1 days")
          # One label per day of week of first day in period and one for last day in period if period is between 16 days and 20 weeks
          #  - Case for period ending on same day of week than period start
        } else if (days %% 7 == 0 && weeks <= 10) {
          seq.Date(from = min(x), to = max(x), by = "7 days")
          #  - Case for period ending on same different of week than period start (last day should be added
        } else if (weeks <= 20) {
          c(seq.Date(from = min(x), to = max(x), by = "7 days"), max(x))
          # One label per month for day of month in period start if period is less or equal to 2 years but more than 20 weeks
          #  - Case for period ending on different day of month than period start
        } else if (
          strftime(min(x), format = "%d") != strftime(max(x), format = "%d") &&
            years <= 2
        ) {
          c(seq.Date(from = min(x), to = max(x), by = "1 month"), max(x))
        } else if (years <= 2) {
          #  - Case for period ending on same day of month than period start (last day should be added
          seq.Date(from = min(x), to = max(x), by = "1 month")
          # One label per year for day of year in period start if period is more than 2 years
          #  - Case for period ending on different day of year than period start
        } else if (
          strftime(min(x), format = "%m-%d") !=
            strftime(max(x), format = "%m-%d")
        ) {
          c(seq.Date(from = min(x), to = max(x), by = "1 year"), max(x))
        } else {
          #  - Case for period ending on same day of month than period start (last day should be added
          seq.Date(from = min(x), to = max(x), by = "1 year")
        }
      }
    ) +
    ggplot2::theme_classic(base_family = get_font_family()) +
    {
      if (length(unique(df$country)) == 1)
        ggplot2::scale_color_manual(values = c("#65B32E"))
    } +
    {
      if (length(unique(df$country)) == 1)
        ggplot2::scale_fill_manual(values = c("#65B32E"))
    } +
    ggplot2::theme(
      plot.title = ggplot2::element_text(
        hjust = 0.5,
        size = 12,
        face = "bold",
        lineheight = 0.9
      ),
      axis.text = ggplot2::element_text(colour = "black", size = 8),
      axis.text.x = ggplot2::element_text(
        angle = 45,
        hjust = 1,
        vjust = 0.5,
        margin = ggplot2::margin(-15, 0, 0, 0)
      ),
      axis.title.x = ggplot2::element_text(
        margin = ggplot2::margin(30, 0, 0, 0),
        size = 10
      ),
      axis.title.y = ggplot2::element_text(
        margin = ggplot2::margin(-25, 0, 0, 0),
        size = 10
      ),
      legend.position = ifelse(length(countries) < 2, "none", "right")
    )

  df <- dplyr::rename(df, "Country" = .data$country)

  df <- dplyr::rename(
    df,
    "Number of messages" = .data$number_of_posts,
    "Message date" = .data$date,
    "Topic" = .data$topic
  )
  # returning data and chart
  list("chart" = fig_line, "data" = df)
}

#' @title Plot the map report on the episomer dashboard
#' @description Generates a bubble map plot of number of posts by country for a single topic
#' @param sms Social media
#' @param topic Character(1) containing the topic to use for the report
#' @param countries Character vector containing the name of the countries and regions to plot or their respective indexes on the Shiny app, default: c(1)
#' @param date_min Date indicating start of the reporting period, default: "1900-01-01"
#' @param date_max Date indicating end of the reporting period, default: "2100-01-01"
#' @param with_quotes Logical value indicating whether to include reposts in the time series, default: FALSE
#' @param caption Character(1) vector indicating a caption to print at the bottom of the chart, default: ""
#' @param proj Parameter indicating the CRS (Coordinate Reference System) to use on PROJ4 format, default NULL
#' If null and all countries are selected +proj=robin is used (Robinson projection) otherwise the Lambert azimuthal equal-area projection will be chosen, default: NULL
#' @param forplotly Logical(1) parameter indicating whether some hacks are activated to improve plotly rendering, default: FALSE
#' @return A named list containing two elements: 'chart' with the ggplot2 figure and 'data' containing the dataframe that was used to build the map.
#' @details Produces a bubble chart map for a particular topic on number of posts collected based on the provided parameters.
#' The map will display information at country level if more than one country is selected, otherwise it will display bubbles at the smallest possible location identified for each post within the period
#' which could be any administrative level or city level.
#'
#' Posts associated with a country but with no finer granularity are omitted when displaying a single country.
#'
#' When an aggregated zone is requested, all countries in that zone are included.
#'
#' This functions requires that \code{\link{search_loop}} and \code{\link{detect_loop}} have already been run successfully to show results.
#' @examples
#' \dontrun{
#'    #Getting bubble chart for dengue for South America for last 30 days
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    create_map(
#'      topic = "dengue",
#'      countries = "South America",
#'      date_min = as.Date(Sys.time())-30,
#'      date_max=as.Date(Sys.time())
#'    )
#' }
#' @seealso
#'  \code{\link{trend_line}}
#'  \code{\link{create_topwords}}
#'  \code{\link{detect_loop}}
#'  \code{\link{search_loop}}
#'  \code{\link[sp]{spTransform}},\code{\link[sp]{coordinates}},\code{\link[sp]{is.projected}},\code{\link[sp]{CRS-class}}
#'  \code{\link[ggplot2]{fortify}},\code{\link[ggplot2]{geom_polygon}},\code{\link[ggplot2]{geom_point}}
#' @rdname create_map
#' @export
#' @importFrom magrittr `%>%`
#' @importFrom dplyr filter rename select bind_rows group_by summarize ungroup mutate dense_rank if_else
#' @importFrom ggplot2 fortify theme element_text element_blank element_rect ggplot geom_polygon aes geom_point scale_size_continuous scale_fill_manual coord_fixed labs theme_classic geom_sf coord_sf
#' @importFrom sf st_crs st_transform  st_as_sf
#' @importFrom stats setNames
create_map <- function(
  sms,
  topic = c(),
  countries = c(1),
  date_min = "1900-01-01",
  date_max = "2100-01-01",
  with_quotes = FALSE,
  caption = "",
  proj = NULL,
  forplotly = FALSE
) {
  stop_if_no_config()
  stop_if_no_fs()
  # Importing pipe operator
  `%>%` <- magrittr::`%>%`
  # Setting the scientific pen off
  old <- options()
  on.exit(options(old))
  options(scipen = 999)

  # getting all regions
  regions <- get_country_items()

  # If countries are names, they have to be changed to region indexes
  if (is.character(countries) && length(countries) > 0) {
    countries = (1:length(regions))[sapply(
      1:length(regions),
      function(i) regions[[i]]$name %in% countries
    )]
  }
  country_codes <- Reduce(
    function(l1, l2) {
      unique(c(l1, l2))
    },
    lapply(as.integer(countries), function(i) unlist(regions[[i]]$codes))
  )

  # Setting a variable for choosing if going to subnational level when only one country is required
  detailed <- length(country_codes) == 1

  # Getting the aggregated data based on the national or subnational level
  df <- (if (!detailed)
    get_aggregates(
      dataset = "country_counts",
      filter = list(topic = tolower(topic), period = list(date_min, date_max), network = sms )
    ) else
    get_aggregates(
      dataset = "geolocated",
      filter = list(topic = tolower(topic), period = list(date_min, date_max), network = sms)
    ))

  # retunrning empty chart if no data is found
  if (nrow(df) == 0) {
    return(get_empty_chart(
      "No data found for the selected topic, region and period"
    ))
  }

  #filtering data by topic and date and country_codes
  f_topic <- tolower(topic)

  df <- (df %>%
    dplyr::filter(
      .data$topic == f_topic &
        (!is.na(.data$geo_country_code)) &
        .data$created_date >= date_min &
        .data$created_date <= date_max &
        ((if (length(country_codes) == 0) TRUE else
          .data$geo_country_code %in% country_codes))
    ))

  # Adding reposts if requested
  if (!detailed) {
    if (with_quotes) {
      df$posts <- ifelse(is.na(df$quotes), 0, df$quotes) +
        ifelse(is.na(df$original), 0, df$original)
    } else {
      df$posts <- df$original
    }
  } else {
    if (with_quotes) {
      df$posts <- ifelse(is.na(df$quotes), 0, df$quotes) +
        ifelse(is.na(df$posts), 0, df$posts)
    } else {
      df$posts <- df$posts
    }
  }

  #
  total_count <- (if (!detailed) {
    sum(df$posts)
  } else {
    sum(
      (df %>%
        dplyr::filter(
          (!(.data$geo_country_code %in% country_codes)) &
            !(.data$geo_country_code %in% country_codes)
        ))$posts
    )
  })

  # Setting country codes as requested location types as requested
  # this is to deal with location type and aggregation level (national vs subnational)
  df <- df %>% dplyr::rename(country_code = .data$geo_country_code)

  #Applying country filter after country type
  df <- (df %>%
    dplyr::filter(
      !is.na(.data$country_code) &
        (if (length(country_codes) == 0) TRUE else
          .data$country_code %in% country_codes) &
        .data$posts > 0 &
        (if (detailed) !(.data$geo_code %in% country_codes) else TRUE)
    ))
  # aggregating by country or geo code depending on national or subnational level
  df <- (if (detailed)
    df %>%
      dplyr::group_by(.data$country_code, .data$geo_code) %>%
      dplyr::summarize(
        count = sum(.data$posts),
        Long = mean(.data$longitude),
        Lat = mean(.data$latitude),
        geo_name = max(.data$geo_name, na.rm = TRUE)
      ) %>%
      dplyr::ungroup() else
    df %>%
      dplyr::group_by(.data$country_code) %>%
      dplyr::summarize(count = sum(.data$posts)) %>%
      dplyr::ungroup())
  # returning an empty chart if no rows are found
  if (nrow(df) == 0) {
    return(get_empty_chart(
      "No data found for the selected topic, region and period"
    ))
  }

  topic <- unname(get_topics_labels()[stringr::str_replace_all(
    topic,
    "%20",
    " "
  )])
  df$topic <- topic

  # Adding country properties (bounding boxes and country names)
  regions <- get_country_items()
  map <- get_country_index_map()
  df$Country <- sapply(
    unname(map[df$country_code]),
    function(i) if (!is.na(i)) regions[[i]]$name else NA
  )
  if (!detailed) {
    df$Long <- sapply(
      unname(map[df$country_code]),
      function(i)
        if (!is.na(i)) mean(c(regions[[i]]$minLong, regions[[i]]$maxLong)) else
          NA
    )
    df$Lat <- sapply(
      unname(map[df$country_code]),
      function(i)
        if (!is.na(i)) mean(c(regions[[i]]$minLat, regions[[i]]$maxLat)) else NA
    )
  }
  df$MinLat <- sapply(
    unname(map[df$country_code]),
    function(i) if (!is.na(i)) regions[[i]]$minLat else NA
  )
  df$MaxLat <- sapply(
    unname(map[df$country_code]),
    function(i) if (!is.na(i)) regions[[i]]$maxLat else NA
  )
  df$MinLong <- sapply(
    unname(map[df$country_code]),
    function(i) if (!is.na(i)) regions[[i]]$minLong else NA
  )
  df$MaxLong <- sapply(
    unname(map[df$country_code]),
    function(i) if (!is.na(i)) regions[[i]]$maxLong else NA
  )

  #Calculating the center of the map
  min_long <- min(df$MinLong, na.rm = TRUE)
  max_long <- max(df$MaxLong, na.rm = TRUE)
  min_lat <- min(df$MinLat, na.rm = TRUE)
  max_lat <- max(df$MaxLat, na.rm = TRUE)
  lat_center <- mean(c(min_lat, max_lat))
  long_center <- mean(c(min_long, max_long))

  # Getting the projection to use which will center the global bounding box
  full_world <- (1 %in% countries || 2 %in% countries)
  proj <- (if (!is.null(proj)) proj else if (full_world)
    # Using Robinson projection for world map
    "+proj=robin" else
    # Using projection Lambert Azimuthal Equal Area for partial maps
    paste(
      "+proj=laea",
      " +lon_0=",
      long_center,
      " +lat_0=",
      lat_center,
      sep = ""
    ))
  # Projecting the country counts data frame on the target coordinate system this projected data frame contains the bubble X,Y coordinates

  # We create a sf object from the data containing messages locations
  # They are in WGS84
  proj_df_wgs84 <- sf::st_as_sf(
    df %>% dplyr::filter(!is.na(.data$Long) & !is.na(.data$Lat)),
    coords = c("Long", "Lat"),
    crs = 4326 # WGS84
  )
  proj_df <- proj_df_wgs84 %>%
    sf::st_transform(proj)

  countries_geo_wgs84 <- rnaturalearthdata::countries50

  countries_geo_wgs84_filtered <- countries_geo_wgs84 %>%
    dplyr::filter(
      .data$label_x >= min_long - 20 &
        .data$label_x <= max_long + 20 &
        .data$label_y >= min_lat - 20 &
        .data$label_y <= max_lat + 20
    ) %>%
    # We have weird codes for some countries (e.g FRANCE has -99)
    mutate(iso_a2_clean = dplyr::if_else(iso_a2 == "-99", iso_a2_eh, iso_a2)) %>%
    # LAKES ARE MISSING
    dplyr::mutate(
      selected = ifelse(
        .data$iso_a2_clean %in% country_codes,
        "a. Selected",
        "b. Excluded"
      )
    )

  countries_geo_proj_filtered <- countries_geo_wgs84_filtered %>%
    sf::st_transform(proj)

  map_limits <- countries_geo_proj_filtered %>%
    # Reminder: these coordinates are in WGS84
    dplyr::filter(
      .data$label_x >= (min_long - 15) &
        .data$label_x <= (max_long + 15) &
        .data$label_y >= (min_lat - 15) &
        .data$label_y <= (max_lat + 15)
    )

  minX <- min(map_limits$label_x)
  maxX <- max(map_limits$label_x)
  minY <- min(map_limits$label_y)
  maxY <- max(map_limits$label_y)

  df_min_max_wgs84 <- data.frame(
    x = c(minX, maxX),
    y = c(minY, maxY)
  ) %>%
    sf::st_as_sf(coords = c("x", "y"), crs = sf::st_crs(4326))

  # We reproject the min and max coordinates to the target projection
  # and reextract min max coordinates
  df_min_max_proj <- df_min_max_wgs84 %>%
    sf::st_transform(proj)

  new_min_max <- sf::st_coordinates(df_min_max_proj)
  minX <- min(new_min_max[, 1])
  maxX <- max(new_min_max[, 1])
  minY <- min(new_min_max[, 2])
  maxY <- max(new_min_max[, 2])

  # Calculating counts groups for Legend
  maxCount <- max(df$count)
  cutsCandidates <- unique(sapply(
    c(maxCount / 50, maxCount / 20, maxCount / 5, maxCount),
    function(v)
      max(1, ceiling((v / (10^floor(log10(v))))) * (10^floor(log10(v))))
  ))
  proj_df$countGroup <- sapply(
    proj_df$count,
    function(c) min(cutsCandidates[c <= cutsCandidates])
  )
  proj_df$plotlycuts <- paste(
    letters[dplyr::dense_rank(proj_df$countGroup) + 3],
    ". ",
    proj_df$countGroup,
    sep = ""
  )
  cuts <- sort(unique(proj_df$countGroup))
  plotlycuts <- sort(unique(proj_df$plotlycuts))
  #Creating tooltip
  countries_geo_proj_filtered$Details <-
    paste(
      "\nRegion:",
      countries_geo_proj_filtered$name,
      sep = ""
    )

  proj_df$Details <-
    paste(
      "\nRegion:",
      proj_df$Country,
      "\nNumber of Posts:",
      proj_df$count,
      if (detailed) "\nLocation: " else "",
      if (detailed) proj_df$geo_name else "",
      sep = ""
    )

  # Defining the chart theme
  theme_opts <- list(ggplot2::theme(
    plot.title = ggplot2::element_text(hjust = 0.5, size = 12, face = "bold"),
    plot.subtitle = ggplot2::element_text(hjust = 0.5, size = 12),
    axis.text = ggplot2::element_text(colour = "black", size = 8),
    panel.grid.minor = ggplot2::element_blank(),
    panel.grid.major = ggplot2::element_blank(),
    panel.background = ggplot2::element_blank(),
    plot.background = ggplot2::element_rect(fill = "white"),
    panel.border = ggplot2::element_blank(),
    axis.line = ggplot2::element_blank(),
    axis.text.x = ggplot2::element_blank(),
    axis.text.y = ggplot2::element_blank(),
    axis.ticks = ggplot2::element_blank(),
    axis.title.x = ggplot2::element_blank(),
    axis.title.y = ggplot2::element_blank(),
    legend.direction = "vertical",
    legend.position = "right"
  ))

  fig <- ggplot2::ggplot(df) +
    ggplot2::geom_sf(data = countries_geo_proj_filtered, aes(label = Details)) +
    ggplot2::geom_sf(
      data = countries_geo_proj_filtered,
      aes(fill = selected, label = Details)
    ) +
    (if (forplotly)
      # customistion for plotly legend
      ggplot2::geom_sf(
        data = proj_df,
        aes(size = count, fill = plotlycuts, label = Details),
        color = "#65B32E",
        alpha = I(8 / 10)
      ) else
      ggplot2::geom_sf(
        data = proj_df,
        aes(size = .data$count),
        fill = "#65B32E",
        color = "#65B32E",
        alpha = I(8 / 10)
      )) +
    ggplot2::coord_sf(
      crs = sf::st_crs(countries_geo_proj_filtered),
      datum = sf::st_crs(countries_geo_proj_filtered),
      xlim = c(minX, maxX),
      ylim = c(minY, maxY)
    ) +
    ggplot2::scale_size_continuous(
      name = "Number of messages",
      breaks = {
        x = cuts
        x[length(x)] = maxCount
        x
      },
      labels = cuts
    ) +
    ggplot2::scale_fill_manual(
      values = c(
        "#C7C7C7",
        "#E5E5E5",
        "white",
        sapply(cuts, function(c) "#65B32E")
      ),
      breaks = c("a. Selected", "b. Excluded", "c. Lakes", plotlycuts),
      guide = "none"
    ) +

    ggplot2::labs(
      title = paste0(
        "Geographical distribution of messages mentioning \n",
        topic,
        " from ",
        date_min,
        " to ",
        date_max
      ),
      caption = paste(caption, ". Projection: ", proj, sep = "")
    ) +
    ggplot2::theme_classic(base_family = get_font_family()) +
    theme_opts

  # returning the chart and the data
  list("chart" = fig, "data" = df)
}

#' @title Plot the top words report on the episomer dashboard
#' @description Generates a bar plot of most popular words in posts, for one topic
#' @param sms Social media
#' @param topic Character(1) containing the topic to use for the report
#' @param country_codes Character vector containing the ISO 3166-1 alpha-2 countries to plot, default: c()
#' @param date_min Date indicating start of the reporting period, default: "1900-01-01"
#' @param date_max Date indicating end of the reporting period, default: "2100-01-01"
#' @param with_quotes Logical value indicating whether to include reposts in the time series, default: FALSE
#' @param top numeric(1) Parameter indicating the number of words to show, default: 25
#' @return A named list containing two elements: 'chart' with the ggplot2 figure and 'data' containing the data frame that was used to build the map.
#' @details Produces a bar chart showing the occurrences of the most popular words in the collected posts based on the provided parameters.
#'
#' This report may be empty for combinations of countries and topics with very few posts since for performance reasons, the calculation of top words is an approximation using chunks of 10.000 posts.
#'
#' This function requires that \code{\link{search_loop}} and \code{\link{detect_loop}} have already been run successfully to show results.
#' @examples
#' \dontrun{
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    #Getting topword chart for dengue for France, Chile, Australia for last 30 days
#'    create_topwords(
#'      topic = "dengue",
#'      country_codes = c("FR", "CL", "AU"),
#'      date_min = as.Date(Sys.time())-30,
#'      date_max=as.Date(Sys.time())
#'    )
#'  }
#' @seealso
#'  \code{\link{trend_line}}
#'  \code{\link{create_map}}
#'  \code{\link{detect_loop}}
#'  \code{\link{search_loop}}
#' @export
#' @importFrom magrittr `%>%`
#' @importFrom dplyr filter group_by summarize ungroup arrange mutate
#' @importFrom ggplot2 ggplot aes geom_col xlab coord_flip labs scale_y_continuous theme_classic theme element_text margin element_blank
#' @importFrom stats reorder
#' @importFrom utils head
#'
create_topwords <- function(
  sms,
  topic,
  country_codes = c(),
  date_min = "1900-01-01",
  date_max = "2100-01-01",
  with_quotes = FALSE,
  top = 25
) {
  create_topchart(
    sms = sms,
    topic = topic,
    serie = "topwords",
    country_codes = country_codes,
    date_min = date_min,
    date_max = date_max,
    top = top
  )
}

#' @title Plot the top elements for a specific series on the episomer dashboard
#' @description Generates a bar plot of most popular elements in posts, for one topic. Top elements among ("topwords", "hashtags", "entities", "contexts", "urls")
#' @param sms Social media
#' @param topic Character(1) containing the topic to use for the report
#' @param serie Character(1) name of the series to be used for the report. It should be one of ("topwords", "hashtags", "entities", "contexts", "urls")
#' @param country_codes Character vector containing the ISO 3166-1 alpha-2 countries to plot, default: c()
#' @param date_min Date indicating start of the reporting period, default: "1900-01-01"
#' @param date_max Date indicating end of the reporting period, default: "2100-01-01"
#' @param with_quotes Logical value indicating whether to include reposts in the time series, default: FALSE
#' @param top numeric(1) Parameter indicating the number of words to show, default: 25
#' @return A named list containing two elements: 'chart' with the ggplot2 figure and 'data' containing the data frame that was used to build the map.
#' @details Produces a bar chart showing the occurrences of the most popular words in the collected posts based on the provided parameters.
#'
#' This functions requires that \code{\link{search_loop}} and \code{\link{detect_loop}} have already been run successfully to show results.
#' @examples
#' \dontrun{
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    #Getting topword chart for dengue for France, Chile, Australia for last 30 days
#'    create_topchart(
#'      topic = "dengue",
#'      serie = "topwords",
#'      country_codes = c("FR", "CL", "AU"),
#'      date_min = as.Date(Sys.time())-30,
#'      date_max=as.Date(Sys.time())
#'    )
#'  }
#' @seealso
#'  \code{\link{trend_line}}
#'  \code{\link{create_map}}
#'  \code{\link{detect_loop}}
#'  \code{\link{search_loop}}
#' @export
#' @importFrom magrittr `%>%`
#' @importFrom dplyr filter group_by summarize ungroup arrange mutate
#' @importFrom ggplot2 ggplot aes geom_col xlab coord_flip labs scale_y_continuous theme_classic theme element_text margin element_blank
#' @importFrom stats reorder
#' @importFrom utils head
#'
create_topchart <- function(
  sms,
  topic,
  serie,
  country_codes = c(),
  date_min = "1900-01-01",
  date_max = "2100-01-01",
  with_quotes = FALSE,
  top = 25
) {
  stop_if_no_config()
  stop_if_no_fs()
  #Importing pipe operator
  `%>%` <- magrittr::`%>%`
  f_topic <- tolower(topic)
  dataset <- serie

  # Getting the right top field depending on the series

  match.arg(serie, c("topwords", "tags", "urls"))
  top_field <- (if (serie == "topwords") {
    "token"
  } else if (serie == "tags") {
    "tag"
  } else if (serie == "urls") {
    "url"
  })  

  # getting the data from topwords series
  filter <- (if (length(country_codes) > 0)
    list(
      topic = f_topic,
      period = list(date_min, date_max),
      geo_country_code = country_codes, network = sms
    ) else list(topic = f_topic, period = list(date_min, date_max), network = sms))

  df <- try(get_aggregates(
    dataset = dataset,
    filter = filter,
    top_field = top_field,
    top_freq = "frequency"
  ))

  if (inherits(df, "try-error") || nrow(df) == 0 || is.null(df)) {
    return(get_empty_chart(
      "No data found for the selected topic, region and period"
    ))
  }

  # renaming the series depending column to "top"
  df <- df %>% dplyr::rename(top = !!as.symbol(top_field))

  # getting the series dependant title part
  serie_title <- (if (serie == "topwords") "words" else serie)

  #filtering data by country
  df <- (df %>%
    dplyr::filter(
      .data$topic == f_topic &
        .data$created_date >= date_min &
        .data$created_date <= date_max &
        (if (length(country_codes) == 0) TRUE else
          .data$geo_country_code %in% country_codes)
    ))

  # dealing with reposts if requested

  if (!with_quotes) {
    df$frequency <- df$original
  } else {
    df$frequency <- ifelse(is.na(df$quotes), 0, df$quotes) +
      ifelse(is.na(df$original), 0, df$original)
  }

  # grouping by top and limiting as requested
  df <- (df %>%
    dplyr::filter(!is.na(.data$frequency)) %>%
    dplyr::group_by(.data$top) %>%
    dplyr::summarize(frequency = sum(.data$frequency)) %>%
    dplyr::ungroup() %>%
    dplyr::arrange(-.data$frequency) %>%
    head(top) %>%
    dplyr::mutate(top = reorder(.data$top, .data$frequency)))
  if (nrow(df) == 0) {
    return(get_empty_chart(
      "No data found for the selected topic, region and period"
    ))
  }

  topic <- unname(get_topics_labels()[stringr::str_replace_all(
    topic,
    "%20",
    " "
  )])
  df$topic <- topic

  # Calculating breaks for y axis
  y_breaks <- unique(floor(pretty(seq(0, (max(df$frequency) + 1) * 1.1))))

  # removing scientific pen
  old <- options()
  on.exit(options(old))
  options(scipen = 999)

  # plotting
  fig <- (df %>%
    ggplot2::ggplot(ggplot2::aes(x = .data$top, y = .data$frequency)) +
    ggplot2::geom_col(fill = "#65B32E") +
    ggplot2::xlab(NULL) +
    ggplot2::coord_flip(expand = FALSE) +
    ggplot2::labs(
      y = "Count",
      title = paste("Top", serie_title, "of messages mentioning", topic),
      subtitle = paste("from", date_min, "to", date_max)
    ) +
    ggplot2::scale_y_continuous(
      labels = function(x) format(x, scientific = FALSE),
      breaks = y_breaks,
      limits = c(0, max(y_breaks)),
      expand = c(0, 0)
    ) +
    ggplot2::theme_classic(base_family = get_font_family()) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(hjust = 0.5, size = 12, face = "bold"),
      plot.subtitle = ggplot2::element_text(hjust = 0.5, size = 12),
      axis.text = ggplot2::element_text(colour = "black", size = 8),
      axis.text.x = ggplot2::element_text(
        angle = 45,
        hjust = 1,
        vjust = 0.5,
        margin = ggplot2::margin(0, 0, 0, 0)
      ),
      axis.title.x = ggplot2::element_text(
        margin = ggplot2::margin(30, 0, 0, 0),
        size = 10
      ),
      axis.line.y = ggplot2::element_blank(),
      axis.ticks.y = ggplot2::element_blank()
    ))
  # returning chart and data
  list("chart" = fig, "data" = df)
}


# Helper function to get the font family
get_font_family <- function() {
  if (
    .Platform$OS.type == "windows" &&
      is.null(grDevices::windowsFonts("Helvetica")[[1]])
  )
    grDevices::windowsFonts(Helvetica = grDevices::windowsFont("Helvetica"))
  "Helvetica"
}

# Returns an empty chart with the provided message on title
get_empty_chart <- function(title) {
  chart <- ggplot2::ggplot() +
    ggplot2::theme_minimal(base_family = get_font_family()) +
    ggplot2::labs(title = title)
  df <- chart$data
  list("chart" = chart, "data" = df)
}

utils::globalVariables(c(
  ".data",
  "limit",
  "URLencode",
  "conf",
  "get_scala_posts_url",
  "is_secret_set",
  "last_fs_updates",
  "resp_header",
  "search_endpoint",
  "tail",
  "twitter_get",
  "iso_a2",
  "iso_a2_eh",
  "Details",
  "selected",
  "count",
  "app_error",
  "attrs",
  "browseURL",
  "cd",
  "get_token",
  "parse_plan",
  "progress_close",
  "progress_start"
))

#' internal function for exporting code of necessary functions for creating templates for a new social media
#' @param file_content char, lines of the script to extract content from
keep_roxygen_and_function_declarations <- function(file_content) {
  text_to_keep <- character(0)
  function_body <- FALSE

  for (l in 1:length(file_content)) {
    l_clean <- trimws(file_content[l])
    is_function_start <- grepl("@function_def_start", l_clean)
    is_function_end <- grepl("@function_def_end", l_clean)

    if (function_body & !is_function_start &!is_function_end) {
      text_to_keep <- c(text_to_keep, l_clean)
    }

    if (is_function_start) {
      function_body <- TRUE
      text_to_keep <- c(text_to_keep, l_clean)
    }
    if (is_function_end) {
      function_body <- FALSE
      text_to_keep <- c(text_to_keep, l_clean)
      text_to_keep <- c(text_to_keep, "}\n")
    }    
  }
  return(text_to_keep)
}

#' internal function for producing the code template for a new social media
#' @param new_media char, name of the social media to add
#' @param ref_media char, name of the social media to use as a reference to produce template
#' @importFrom utils browseURL
create_api_and_plan_files_for_new_social_media <- function(
  new_media,
  ref_media = "bluesky"
) {
  is_r_folder_present()

  api_file <- paste0("R/sm-", ref_media, "-api.R")
  api_file_content <- readLines(api_file)
  api_file_cleaned <- keep_roxygen_and_function_declarations(
    api_file_content
  )
  api_file_cleaned <- gsub(ref_media, new_media, api_file_cleaned)
  writeLines(api_file_cleaned, paste0("R/sm-", new_media, "-api.R"))
  cli::cli_alert_success(paste0(
    "R file for api functions for ",
    new_media,
    " created - please look at the file ",
    paste0("R/sm-", new_media, "-api.R")
  ))

  plan_file <- paste0("R/sm-", ref_media, "-plan.R")
  plan_file_content <- readLines(plan_file)
  plan_file_cleaned <- keep_roxygen_and_function_declarations(
    plan_file_content
  )
  plan_file_cleaned <- gsub(ref_media, new_media, plan_file_cleaned)
  writeLines(plan_file_cleaned, paste0("R/sm-", new_media, "-plan.R"))
  cli::cli_alert_success(paste0(
    "R file for plan functions for ",
    new_media,
    " created - please look at the file ",
    paste0("R/sm-", new_media, "-plan.R")
  ))
  cli::cli_alert_info("To keep the documentation up to date, please run devtools::document()")

  return(invisible(TRUE))
}

#' Create new social media template files, so a developer can extend episomer and support a new social media.
#' this function has to be called within episomer subfolder containing thr R subfolder after executing devtools::load_all()
#' @param social_media Name of the new social media
#' @param social_media_ref Name of the reference social media 
#' @return TRUE, called for side effects (creation of new social media files).
#' @export
add_new_social_media <- function(
  social_media,
  social_media_ref = "bluesky"
) {
  create_api_and_plan_files_for_new_social_media(social_media, social_media_ref)
  cli::cli_alert_success("Added new social media files in the R/ folder successfully")
  cli::cli_alert_info("For next steps, please look at the vignette vignettes/add_new_social_media.Rmds")
  return(invisible(TRUE))
}

#' check whether the R folder is present in the current directory
is_r_folder_present <- function() {
  R_files <- list.files("R")
  if (length(R_files) == 0) {
    stop(
      "No R files found in the R directory - you should probably run devtools::load_all() first and set the current directory to the episomer folder (with the R subfolder)"
    )
  }
  return(TRUE)
}

# This script includes all dedicated logic for interacting with the Spark/Scala API via HTTP REST requests

get_scala_api_endpoint <- function() {
  paste("http://localhost:", conf$fs_port, "/", sep = "")
}

get_scala_geotraining_url <- function() {
  paste(get_scala_api_endpoint(), "geotraining-set", sep = "")
}

get_scala_alert_training_url <- function() {
  paste(get_scala_api_endpoint(), "evaluate-alerts", sep = "")
}

get_scala_posts_url <- function() {
  paste(get_scala_api_endpoint(), "posts", sep = "")
}

get_scala_geolocated_posts_url <- function() {
  paste(get_scala_api_endpoint(), "geolocated-posts", sep = "")
}

get_scala_commit_url <- function() {
  paste(get_scala_api_endpoint(), "commit", sep = "")
}


get_scala_geolocate_text_url <- function() {
  paste(get_scala_api_endpoint(), "geolocate-text", sep = "")
}

get_scala_aggregate_url <- function() {
  paste(get_scala_api_endpoint(), "aggregate", sep = "")
}

get_scala_period_url <- function() {
  paste(get_scala_api_endpoint(), "period", sep = "")
}

get_scala_status_url <- function() {
  paste(get_scala_api_endpoint(), "period?serie=country_counts", sep = "")
}

get_scala_ping_url <- function() {
  paste(get_scala_api_endpoint(), "ping", sep = "")
}

get_scala_recalc_hash_url <- function() {
  paste(get_scala_api_endpoint(), "recalculate-hash", sep = "")
} 


#' @title Runs the episomer embedded database loop
#' @description Infinite loop ensuring that the episomer embedded database is running (Lucene + akka-http)
#' @param data_dir Path to the 'data directory' containing application settings, models and collected posts.
#' If not provided, the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, default: NA
#' @return nothing
#' @details Launches the episomer embedded database which is accessed via a REST API located on localhost at port 8080. You can test that the database is running by accesssing the local URL and adding "/ping". 
#' the REST API provide episomer a way to send and retrieve data related with posts and time series and to trigger geolocation or aggregation
#' The database is implemented using Apache Lucene indexes allowing episomer to access its data as a search engine but also as a tabular database.
#' \code{\link{health_check}} called each 60 seconds on a background process to send alerts to the administrator if some episomer components fail.
#' @examples
#' \dontrun{
#'    #Running the detect loop
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    fs_loop()
#' }
#' @rdname fs_loop
#' @seealso
#'  \code{\link{detect_loop}}
#'
#'  \code{\link{search_loop}}
#'
#'  \code{\link{health_check}}
#' @importFrom future multisession plan future
#' @importFrom rlang current_env
#' @export
fs_loop <- function(data_dir = NA) {
  if (is.na(data_dir)) setup_config_if_not_already() else
    setup_config(data_dir = data_dir)
  data_dir <- conf$data_dir

  # calculating alerts per topic
  future::plan(future::multisession)
  monitor <- future::future(
    {
      message("Monitoring episomer")
      setup_config(data_dir = data_dir)
      x = 60
      step = 2
      while (TRUE) {
        register_fs_monitor()
        if (x == 0) {
          health_check()
          x = 60
        }
        x = x - step
        Sys.sleep(step)
      }
      0
    },
    globals = c(
      "data_dir",
      "setup_config",
      "health_check",
      "register_fs_monitor"
    ),
    envir = rlang::current_env()
  )

  message("Launching fs")
  # Registering the fs runner using current PID and ensuring no other instance of the search is actually running.
  register_fs_runner()

  # Infinite loop calling the fs runner
  while (TRUE) {
    tryCatch(
      {
        spark_job(
          paste(
            "fsService",
            "epiHome",
            conf$data_dir
          )
        )
      },
      error = function(e) {
        message(paste(
          "The episomer Scala API was stopped with the following error",
          e,
          "launching it again within 5 seconds",
          sep = "\n"
        ))
        Sys.sleep(5)
      }
    )
  }
}

#' @title perform full text search on posts collected with episomer
#' @description  perform full text search on posts collected with episomer (posts migrated from episomer v<1.0.x are also included)
#' @param query character. The query to be used if a text it will match the post text. To see how to match particular fields please see details, default:NULL
#' @param topic character, Vector of topics to include on the search default:NULL
#' @param from an object which can be converted to ‘"POSIXlt"’ only posts posted after or on this date will be included, default:NULL
#' @param to an object which can be converted to ‘"POSIXlt"’ only posts posted before or on this date will be included, default:NULL
#' @param countries character or numeric, the position or name of episomer regions to be included on the query, default:NULL
#' @param mentioning character, limit the search to the posts mentioning the given users, default:NULL
#' @param users character, limit the search to the posts created by the provided users, default:NULL
#' @param hide_users logical, whether to hide user names on output replacing them by the USER keyword, default:FALSE
#' @param action character, an action to be performed on the search results respecting the max parameter. Possible values are 'delete' or 'anonymise' , default:NULL
#' @param max integer, maximum number of posts to be included on the search, default:100
#' @param by_relevance logical, whether to sort the results by relevance of the matching query or by indexing order, default:FALSE
#' If not provided the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, default: NA
#' @return a data frame containing the posts matching the selected filters, the data frame contains the following columns: user_name, created_date, is_geo_located, is_quote, text, 
#' text_loc, user_id, hash, quote_lang, totalCount, created_at, topic_post_id, topic, lang, user_name, quoted_text, id, quopted_text_loc, tags 
#' @details
#' episomer translate the query provided by all parameters into a single query that will be executed on post indexes which are weekly indexes.
#' The q parameter should respect the syntax of the Lucene classic parser \url{https://lucene.apache.org/core/8_5_0/queryparser/org/apache/lucene/queryparser/classic/QueryParser.html}
#' So other than the provided parameters, multi field queries are supported by using the syntax field_name:value1;value2
#' AND, OR and -(for excluding terms) are supported on q parameter.
#' Order by week is always applied before relevance so even if you provide by_relevance = TRUE all of the matching posts of the first week will be returned first
#' @examples
#' \dontrun{
#'    #Running the detect loop
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    setup_config(file.choose())
#'    df <- search_posts(
#'         q = "vaccination",
#'         topic="COVID-19",
#'         countries=c("Chile", "Australia", "France"),
#'         from = Sys.Date(),
#'         to = Sys.Date()
#'    )
#'    df$text
#' }
#' @rdname search_posts
#' @seealso
#'  \code{\link{search_loop}}
#'
#'  \code{\link{detect_loop}}
#' @importFrom utils URLencode
#' @importFrom jsonlite stream_in
#' @export
search_posts <- function(
  query = NULL,
  topic = NULL,
  from = NULL,
  to = NULL,
  countries = NULL,
  mentioning = NULL,
  users = NULL,
  hide_users = FALSE,
  action = NULL,
  max = 100,
  by_relevance = FALSE
) {
  u <- paste(
    get_scala_posts_url(),
    "?jsonnl=true&estimatecount=true&by_relevance=",
    tolower(by_relevance),
    sep = ""
  )
  if (hide_users) {
    u <- paste(u, "&hide_users=true", sep = "")
  }
  if (!is.null(action)) {
    u <- paste(u, "&action=", action, sep = "")
  }
  if (!is.null(query)) {
    u <- paste(u, "&q=", utils::URLencode(query, reserved = T), sep = "")
  }
  if (!is.null(topic)) {
    u <- paste(
      u,
      "&topic=",
      utils::URLencode(paste(topic, sep = "", collapse = ";"), reserved = T),
      sep = "",
      collapse = ""
    )
  }
  if (!is.null(from)) {
    u <- paste(u, "&from=", strftime(from, format = "%Y-%m-%d"), sep = "")
  }
  if (!is.null(to)) {
    u <- paste(u, "&to=", strftime(to, format = "%Y-%m-%d"), sep = "")
  }
  if (
    !is.null(countries) &&
      length(countries) > 0 &&
      (is.character(countries) || any(countries != 1))
  ) {
    # getting all regions
    regions <- get_country_items()
    # If countries are names they have to be changes to region indexes
    if (is.character(countries)) {
      countries = (1:length(regions))[sapply(
        1:length(regions),
        function(i) regions[[i]]$name %in% countries
      )]
    }
    country_codes <- Reduce(
      function(l1, l2) {
        unique(c(l1, l2))
      },
      lapply(as.integer(countries), function(i) unlist(regions[[i]]$codes))
    )
    u <- paste(
      u,
      "&country_code=",
      paste(country_codes, collapse = ";", sep = ""),
      sep = "",
      collapse = ""
    )
  }
  if (!is.null(mentioning)) {
    u <- paste(
      u,
      "&mentioning=",
      paste(mentioning, sep = "", collapse = ";"),
      sep = ""
    )
  }
  if (!is.null(users)) {
    u <- paste(u, "&user=", paste(users, sep = "", collapse = ";"), sep = "")
  }

  if (!is.null(max)) {
    u <- paste(u, "&max=", max, sep = "")
  }
  #message(u)
  u <- url(u)
  posts <- jsonlite::stream_in(u, verbose = FALSE)
  posts
}


# Registers a query to define custom aggregations on posts
# it is generic enough to choose variables, aggregation and filters
# deals with post deduplication at topic level
# regexp: regexp to limit the geolocation and search files to read
# vars: variable to read (evaluated after aggregation if required)
# sort_by: order to apply to the returned data frame
# filter_by: expressions to use for filtering posts
# sources_exp: variables to limit the source files to read (setting this will improve reading performance)
# handler: function that to perform a custom R based transformation on data returned by SPARK
# params: definition of custom param files to enable big queries
set_aggregated_posts <- function(
  name,
  dateCol,
  pks,
  aggr,
  vars = list("*"),
  group_by = list(),
  sort_by = list(),
  filter_by = list(),
  sources_exp = list(),
  params = list()
) {
  stop_if_no_config(paste("Cannot get posts without configuration setup"))
  post_result <- httr::POST(
    url = get_scala_aggregate_url(),
    httr::content_type_json(),
    body = jsonlite::toJSON(
      list(
        name = name,
        dateCol = dateCol,
        pks = pks,
        aggr = aggr,
        aggregation = list(
          columns = vars,
          groupBy = group_by,
          sortBy = sort_by,
          filterBy = filter_by,
          sourceExpressions = sources_exp,
          params = if (length(params) == 0) list(fake_param_epi = "") else
            params
        )
      ),
      simplify_vector = T,
      auto_unbox = T
    ),
    encode = "raw",
    encoding = "UTF-8"
  )
  if (httr::status_code(post_result) != 200) {
    print(substring(
      httr::content(post_result, "text", encoding = "UTF-8"),
      1,
      100
    ))
    stop()
  }
}


# Utility function to interact with a POST endpoint and process data on a streaming way provided by handler
stream_post <- function(uri, body, handler = NULL) {
  cum <- new.env()
  cum$tail <- c()
  cum$dfs <- list()

  if (!is.null(handler)) {
    # If a custom transformation is going to be done (a handler function has been set)
    # this function will be called on pages of 10k lines using the jsonlite stream_in function
    # the transformed results will be stored on a temporary file that will be read by stream_in
    tmp_file <- tempfile(pattern = "episomer", fileext = ".json")
    #message(tmp_file)
    con_tmp <- file(tmp_file, open = "w", encoding = "UTF-8")
  }

  pipeconnection <- function(x, handler = handler) {
    cuts <- which(x == 0x0a)
    if (length(cuts) == 0) cum$tail <- c(cum$tail, x) else {
      if (tail(cuts, n = 1) != length(x)) {
        bytes <- c(cum$tail, x[1:tail(cuts, n = 1)])
        cum$tail <- x[(tail(cuts, n = 1) + 1):length(x)]
      } else {
        bytes <- c(cum$tail, x)
        cum$tail <- c()
      }
      con <- rawConnection(bytes, "rb")
      on.exit(close(con))
      if (is.null(handler)) {
        cum$dfs[[length(cum$dfs) + 1]] <- jsonlite::stream_in(
          con,
          verbose = FALSE
        )
      } else
        jsonlite::stream_in(
          con,
          function(df) handler(df, con_tmp),
          verbose = FALSE
        )
    }
  }
  # Doing the post request
  parts = gregexpr("//[^/]+/", uri)
  split <- parts[[1]][[1]] + attr(parts[[1]], "match.length")
  #h <- httr::handle(substr(uri, 1, split-1))
  #path <- substr(uri, split, nchar(uri))
  #on.exit(httr::handle_reset(h))

  post_result <- tryCatch(
    {
      httr::POST(
        url = uri,
        httr::write_stream(function(x) {
          pipeconnection(x, handler)
        }),
        body = body,
        httr::content_type_json(),
        encode = "raw",
        encoding = "UTF-8"
      )
    },
    error = function(e) {
      closeAllConnections()
      stop(e)
    }
  )
  if (httr::status_code(post_result) != 200) {
    #message(uri)
    message(httr::status_code(post_result))
    message(httr::content(post_result, "text", encoding = "UTF-8"))
    stop()
  }

  #Applying last line if necessary
  if (length(cum$tail) > 0) {
    if (cum$tail[[length(cum$tail)]] != 0x0a)
      cum$tail[[length(cum$tail) + 1]] <- as.raw(0x0a)
    con <- rawConnection(cum$tail, "rb")
    if (is.null(handler)) {
      cum$dfs[[length(cum$dfs) + 1]] <- jsonlite::stream_in(
        con,
        verbose = FALSE
      )
    } else
      cum$dfs[[length(cum$dfs) + 1]] <- jsonlite::stream_in(
        con,
        function(df) handler(df, con_tmp),
        verbose = F
      )
    close(con)
  }
  #Transforming single lines responses in data frames
  #lapply(cum$dfs, function(l) if(typeof(l) == "list") data.frame(lapply(l, function(x) t(data.frame(x)))) else l)
  #joining response
  if (is.null(handler)) {
    jsonlite::rbind_pages(cum$dfs)
  } else {
    close(con_tmp)
    con_tmp <- file(tmp_file, open = "r", encoding = "UTF-8")
    ret <- jsonlite::stream_in(con_tmp, pagesize = 10000, verbose = FALSE)
    close(con_tmp)
    unlink(tmp_file)
    ret
  }
}

# Get time difference since last request
# This function is used from the Shiny app to report when was the last time that a request saved posts
# This is done by taking the last modified date of current year post files
last_fs_updates <- function(
  collections = c("posts", "topwords", "country_counts", "geolocated")
) {
  times <- lapply(collections, function(collection) {
    folders <- sort(list.files(
      path = paste(conf$data_dir, "fs", collection, sep = "/"),
      full.names = T
    ))
    if (length(folders) > 0) {
      files <- list.files(c(folders), full.names = TRUE, recursive = TRUE)
      files <- files[!grepl("write.lock$", files)]
      max(file.mtime(files), na.rm = T)
    } else {
      NA
    }
  })
  setNames(times, collections)
}

commit_posts <- function() {
  post_result <- httr::POST(
    url = get_scala_commit_url(),
    httr::content_type_json(),
    encode = "raw",
    encoding = "UTF-8"
  )
  if (httr::status_code(post_result) != 200) {
    stop(print(substring(
      httr::content(post_result, "text", encoding = "UTF-8"),
      1,
      100
    )))
  }
}

#' @title Runs the search loop
#' @description Infinite loop ensuring the permanent collection of social media messages
#' @param data_dir optional path to the 'data directory' containing application settings, models and collected posts. If not provided it will reuse the last set on the current session.
#' If not provided the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, Default: NA
#' @param sandboxed logical indicating if the search loop should be run in sandboxed mode. If TRUE, the search loop will not send or commit posts to the database. Default: FALSE
#' @param log_to_file logical indicating if the search loop should log to a file. Default: TRUE
#' @param max_requests integer indicating the maximum number of requests to perform. If 0, the search loop will run indefinitely. Default: 0
#' @return Nothing
#' @details The detect loop is a pure R function designed for downloading posts from the social media search API. It can handle several topics ensuring that all of them will be downloaded fairly using a
#' round-robin philosophy and respecting social media API rate-limits.
#'
#' The progress of this task is reported on the 'topics.*.json' files which are created by this function. This function will try to collect posts respecting a 'collect_span' window
#' in minutes, which is defined on the Shiny app and defaults to 60 minutes.
#'
#' To see more details about the collection algorithm please see episomer vignette.
#'
#' In order to work, this task needs Bluesky credentials, which can be set on the Shiny app or using \code{\link{sm_api_set_auth_bluesky}}
#'
#' @examples
#' \dontrun{
#'    #Running the search loop
#'    library(episomer)
#'    message('Please choose the episomer data directory')
#'    search_loop(file.choose())
#' }
#' @rdname search_loop
#' @seealso
#' \code{\link{sm_api_set_auth_bluesky}}
#' @export
search_loop <- function(
  data_dir = NA,
  sandboxed = FALSE,
  log_to_file = TRUE,
  max_requests = 0
) {
  # Setting or reusing the data directory
  if (is.na(data_dir)) {
    setup_config_if_not_already()
  } else {
    setup_config(data_dir = data_dir)
  }
  if (!sandboxed) {
    # Registering the search runner using current PID and ensuring no other instance of the search is actually running.
    register_search_runner()
  }

  if(log_to_file) {
      jobs_dir <- file.path(conf$data_dir, "jobs")
      if (!file.exists(jobs_dir)) {
        dir.create(jobs_dir, showWarnings = FALSE)
      }
  }
  sms <- active_social_media()
  cores <- as.numeric(length(sms))
  cl <- parallel::makePSOCKcluster(cores, outfile="")
  on.exit(parallel::stopCluster(cl))
  data_dir <- conf$data_dir
  parallel::clusterExport(cl, list("log_to_file", "data_dir", "sandboxed"), envir=rlang::current_env())
  
  # running search loops
  parallel::parLapply(cl, sms, function(sm) {
      # redirecting output to the sm file
      if(log_to_file) {
          jobs_dir <- file.path(data_dir, "jobs")
          outcon <- file(file.path(jobs_dir, sprintf("search.%s.log", sm)), open = "a")
          sink(outcon)
          sink(outcon, type = "message")
      }
      episomer::setup_config(data_dir)
      m <- paste("Running search agent for ",sm, Sys.time())
      message(m)
      episomer::search_loop_worker(sm, data_dir, sandboxed, max_requests)
  })
  stop("All search loop proccesses ended, which is unexpected")
  
}

#' @export
#' @rdname search_loop
#' @param network The social media network to search on.
#' @param data_dir The data directory to use.
#' @param sandboxed Whether to run in sandboxed mode.
#' @param max_requests The maximum number of requests to perform.
#' @return Nothing
#' @details The search loop worker is a function that runs the search loop for a given social media network.
#' It is used to run the search loop in a separate process.
search_loop_worker <- function(
  network,
  data_dir = NA,
  sandboxed = FALSE,
  max_requests = 0
) {
  message(sprintf("Starting runner for %s", network))
  # Setting or reusing the data directory
  if (is.na(data_dir)) {
    setup_config_if_not_already()
  } else {
    setup_config(data_dir = data_dir)
  }

  if (!sandboxed) {
    # Registering the search runner using current PID and ensuring no other instance of the search is actually running.
    register_search_runner(network)
  }
  # Infinite loop for getting posts if it is successfully registered as the search runner
  req2Commit <- 0
  last_check <- Sys.time()
  last_save <- Sys.time()
  first_call <- Sys.time()
  requests_done <- 0

  while (max_requests == 0 || requests_done <= max_requests) {
    loop_start <- Sys.time()

    # Waiting until database system will be running
    if (!sandboxed) {
      while (!is_fs_running()) {
        msg("episomer database is not yet running waiting for 5 seconds")
        Sys.sleep(5)
      }
      # Dismissing history if required from shiny app
      if (conf$dismiss_past_request > conf$dismiss_past_done) {
        if (
          length(conf$topics[sapply(conf$topics, function(t) {
            length(t$plan) == 0 || t$plan[[1]]$requests == 0
          })]) >
            0
        ) {
          msg(
            "Dismissing history has to wait until all plans have at least one plan with one request"
          )
        } else {
          msg("Dismissing history requested")

          for (i in 1:length(conf$topics)) {
            conf$topics[[i]]$plan <- finish_plans(plans = conf$topics[[i]]$plan)
          }
          conf$dismiss_past_done <- strftime(Sys.time(), "%Y-%m-%d %H:%M:%S")
        }
      }
    }

    # On each iteration this loop will perform one request for each active plans having the minimum number of requests
    # Creating plans for each topic if collect span is expired and calculating next execution time for each olan.
    for (i in 1:length(conf$topics)) {
      conf$topics[[i]]$plan <- update_plans_schedule(
        network = network,
        plans = conf$topics[[i]]$plan,
        schedule_span = conf$collect_span
      )
    }

    # Calculating how the time episomer should wait before executing each active plan. If bigger than zero then episomer will wait.
    # If waiting happens here, it means that episomer is able to collect all posts under current social media rate limits, so it could collect more topics or sooner.
    if (!sandboxed) {
      wait_for <- min(unlist(lapply(1:length(conf$topics), function(i) {
        can_wait_for(plans = conf$topics[[i]]$plan)
      })))
      if (wait_for > 0) {
        msg(paste(
          Sys.time(),
          ": All done! going to sleep for until",
          Sys.time() + wait_for,
          "during",
          wait_for,
          "seconds. Consider reducing the schedule_span for getting posts sooner"
        ))
        commit_posts()
        save_config(
          data_dir = conf$data_dir,
          sm_topics = list(network),
          properties = FALSE
        )
        Sys.sleep(wait_for)
      }
    }

    min_requests <- .Machine$integer.max
    #getting the min request on current non ended plans
    for (i in 1:length(conf$topics)) {
      for (j in 1:length(conf$topics[[i]]$plan)) {
        p = conf$topics[[i]]$plan[[j]]
        if (is.null(p$end_on) && p$requests < min_requests)
          min_requests <- p$requests
      }
    }

    if (!sandboxed) {
      #updating series to aggregate
      register_series()
    }

    #msg(paste("iterating in topics", length(conf$topics), min_requests))
    #performing search only for plans with a minimum number of requests (round robin)

    for (i in 1:length(conf$topics)) {
      for (j in 1:length(conf$topics[[i]]$plan)) {
        if (max_requests == 0 || requests_done <= max_requests) {
          plan <- conf$topics[[i]]$plan[[j]]
          if (
            plan$requests <= min_requests &&
              is.null(plan$end_on) &&
              plan$network == network
          ) {
            requests_done <- requests_done + 1
            #if search is performed on the first plan and we get an too old error, we will retry without time limit
            tryCatch(
              {
                conf$topics[[i]]$plan[[j]] = search_topic(
                  plan = plan,
                  query = conf$topics[[i]]$query,
                  topic = conf$topics[[i]]$topic,
                  sandboxed = sandboxed
                )
              },
              error = function(e) {
                if (j == 1 && e$message == "too-old") {
                  msg("Recovering from too-old request")
                  plan$since_id <- NULL
                  plan$since_target <- NULL
                  conf$topics[[i]]$plan[[j]] = search_topic(
                    plan = plan,
                    query = conf$topics[[i]]$query,
                    topic = conf$topics[[i]]$topic
                  )
                } else if (e$message == "too-old") {
                  msg("Canceling too-old request")
                  conf$topics[[i]]$plan[[j]]$end_on <- Sys.time()
                } else {
                  stop(paste(e$message, e))
                }
              }
            )
            req2Commit <- req2Commit + 1
            if (!sandboxed) {
              if (req2Commit > 100) {
		# setting a timestamp for indicating proper run of search loop
	        search_touch() 
		# commiting posts
                commit_posts()
                req2Commit <- 0
              }
            }
          }
        }
      }
    }
    if (!sandboxed) {
      #msg("iteration end")
      #checking at most once per 10 minutes
      if (difftime(Sys.time(), last_check, units = "mins") > 10) {
        # episomer sanity check and sendig email in case of issued
        #msg("health checked")
        last_check <- Sys.time()
        health_check()
      }
    }
    #Updating config to take in consideration possible changes on topics or other settings (plans are saved before reloading config) at most once every 10 seconds
    if (difftime(Sys.time(), last_save, units = "secs") > 10) {
      last_save <- Sys.time()
      setup_config(data_dir = conf$data_dir, save_topics_first = list(network))
      #msg("config saved and refreshed")
    }

    if (requests_done == 0) {
      message(
        "No requests performed on loop.... something may be wrong, sleeping 1 second"
      )
      Sys.sleep(1)
    }
  }
}

#' updates the modificatin date of a flag file indicating that messages have been sent to the embedded database.
search_touch <- function() {
  p <- get_search_touch_path()
  if(!file.exists(p)) {
      write("", file = p, append = FALSE)
  }
  Sys.setFileTime(p, Sys.time())
}

#' performs a search following the given plan, query and topic.
#' @param plan list, the plan that defines the timespan to search for
#' @param query char, the query to use when searching for the topic
#' @param topic char, the topic associated to this search
#' @param sandboxed logical, whether to un in sandboxed mode which prevent sending data to the embedded database.
search_topic <- function(
  plan,
  query,
  topic,
  sandboxed = FALSE
) {
  token <- sm_api_get_token(plan$network)
  info <- sm_plan_next_search_info(plan)

  msg(paste(
    "searching for topic",
    topic,
    info
  ))

  # doing the post search and storing the response object to obtain details on resp
  results <- sm_api_search(
    query = query,
    token = token,
    plan = plan
  )
  if (!sandboxed) {
    # Interpreting the content as JSON and storing the results on json (nested list with dataframes)
    # interpreting is necessary to know the number of obtained posts and the id of the oldest post found and to keep post collecting stats
    # Saving uninterpreted content as a gzip archive
    tries <- 3
    done <- FALSE
    while (!done) {
      tries <- tries - 1
      tryCatch(
        {
          post_result <- httr::POST(
            url = paste0(
              get_scala_posts_url(),
              "?topic=",
              curl::curl_escape(topic),
              "&network=",
              plan$network,
              "&geolocate=true"
            ),
            httr::content_type_json(),
            body = jsonlite::toJSON(
              results$posts,
              auto_unbox = T,
              null = "null"
            ),
            encode = "raw",
            encoding = "UTF-8",
            httr::timeout((4 - tries) * 5)
          )
          if (httr::status_code(post_result) != 200) {
            print(substring(
              httr::content(
                post_result,
                "text",
                encoding = "UTF-8"
              ),
              1,
              500
            ))
            stop()
          }
          done = TRUE
        },
        error = function(e) {
          msg(paste("Error found while sending posts", e))
          if (tries < 0) {
            stop("too many retries")
          }
        }
      )
    }
  }
  update_plan_after_request(plan, results)
}


# Helper function to parse social media date as provided by the social media API
parse_date <- function(str_date) {
  curLocale <- Sys.getlocale("LC_TIME")
  on.exit(Sys.setlocale("LC_TIME", curLocale))
  Sys.setlocale("LC_TIME", "C")
  strptime(str_date, format = "%a %b %d %H:%M:%S +0000 %Y", tz = "UTC")
}


# Calculating how long in seconds should episomer wait before executing one of the plans in the list which would be the case only if all plans are finished before the end of the search span
can_wait_for <- function(plans) {
  plans <- if (is.list(plans)) plans else list(plans)
  non_ended <- plans[sapply(plans, function(x) is.null(x$end_on))]
  if (length(non_ended) == 0) {
    expected_end <- Reduce(min, lapply(plans, function(x) x$expected_end))
    return(ceiling(as.numeric(difftime(
      expected_end,
      Sys.time(),
      units = "secs"
    ))))
  } else {
    return(0)
  }
}

# Last search time on stored in the embedded database
last_search_time <- function() {
  last_fs_updates(c("posts"))$posts
}

# prints message with date
msg <- function(m) {
  message(paste0(Sys.time(), " [INFO]: -------> ", m))
}

# Defining line chart from shiny app filters
line_chart_from_filters <- function(
  sms,
  topics,
  countries,
  period_type,
  period,
  with_quotes,
  alpha,
  alpha_outlier,
  k_decay,
  no_history,
  bonferroni_correction,
  same_weekday_baseline
) {
  trend_line(
    sms = sms,
    topic = topics,
    countries = if (length(countries) == 0) c(1) else as.integer(countries),
    date_type = period_type,
    date_min = period[[1]],
    date_max = period[[2]],
    with_quotes = with_quotes,
    alpha = alpha,
    alpha_outlier = alpha_outlier,
    k_decay = k_decay,
    no_historic = no_history,
    bonferroni_correction = bonferroni_correction,
    same_weekday_baseline = same_weekday_baseline
  )
}
# Defining line chart from shiny app filters
map_chart_from_filters <- function(sms, topics, countries, period, with_quotes) {
  create_map(
    sms = sms,
    topic = topics,
    countries = if (length(countries) == 0) c(1) else as.integer(countries),
    date_min = period[[1]],
    date_max = period[[2]],
    with_quotes = with_quotes,
    caption = conf$regions_disclaimer,
    forplotly = TRUE
  )
}
# Defining top words chart from shiny app filters
top_chart_from_filters <- function(
  sms,
  topics,
  serie,
  fcountries,
  period,
  with_quotes,
  top
) {
  fcountries = if (length(fcountries) == 0 || 1 %in% fcountries) c(1) else
    as.integer(fcountries)
  regions <- get_country_items()
  countries <- Reduce(
    function(l1, l2) {
      unique(c(l1, l2))
    },
    lapply(fcountries, function(i) unlist(regions[[i]]$codes))
  )
  create_topchart(
    sms = sms,
    topic = topics,
    serie = serie,
    country_codes = countries,
    date_min = period[[1]],
    date_max = period[[2]],
    with_quotes = with_quotes,
    top
  )
}

  # Rmarkdown dasboard export bi writing the dashboard on the provided file$
  # it uses the markdown template inst/rmarkdown/dashboard.Rmd
  export_dashboard <- function(
    format,
    file,
    sms,
    topics,
    countries,
    period_type,
    period,
    with_quotes,
    location_type,
    alpha,
    alpha_outlier,
    k_decay,
    no_historic,
    bonferroni_correction,
    same_weekday_baseline,
    top_type1,
    top_type2
  ) {
    tryCatch(
      {
        progress_start("Exporting dashboard")
        r <- rmarkdown::render(
          system.file(
            "rmarkdown",
            "dashboard.Rmd",
            package = get_package_name()
          ),
          output_format = format,
          output_file = file,
          params = list(
            "sms" = sms,
            "topics" = topics,
            "countries" = countries,
            "period_type" = period_type,
            "period" = period,
            "with_quotes" = with_quotes,
            "location_type" = location_type,
            "alert_alpha" = alpha,
            "alert_alpha_outlier" = alpha_outlier,
            "alert_k_decay" = k_decay,
            "alert_historic" = no_historic,
            "bonferroni_correction" = bonferroni_correction,
            "same_weekday_baseline" = same_weekday_baseline,
            "top_type1" = "tags",
            "top_type2" = "topwords"

          ),
          quiet = TRUE
        )
        progress_close()
        r
      },
      error = function(w) {
        app_error(w, env = cd)
      }
    )
  }

# Load stop words from language files which will be a fixed set of twitter stop words plus first 500 more popular words on language
get_stop_words <- function(language_code) {
  #If language stopwords are already cached we rerun them
  stop_id <- paste("stop_words", language_code, sep = "_")
  if (exists(stop_id, where = cached)) {
    return(cached[[stop_id]])
  } else {
    for (i in 1:length(conf$languages)) {
      if (
        conf$languages[[i]]$code == language_code &&
          file.exists(conf$languages[[i]]$vectors)
      ) {
        con <- gzfile(conf$languages[[i]]$vectors)
        fixed <- c("rt", "RT", "http", "https", "t.co", "amp", "via")
        stop_words <- tail(
          gsub(" .*", "", readLines(con, n = 500, encoding = "UTF-8")),
          -1
        )
        close(con)
        if (length(stop_words) > 0) {
          stop_words <- c(fixed, stop_words)
          stop_words <- stop_words[grep("'|\\\\", stop_words, invert = TRUE)]
          cached[[stop_id]] <- stop_words
          return(stop_words)
        } else {
          return(vector(mode = "character", length = 0))
        }
      }
    }
  }
}

# Get path of search files (written by search loop)
get_search_path <- function() file.path(conf$data_dir, "posts", "search")

# Get path of search files (written by search loop)
get_search_touch_path <- function() file.path(conf$data_dir, "search.touched")

# Get path of geolocated files (written by detect loop)
get_geo_path <- function() file.path(conf$data_dir, "posts", "geolocated")

# Get path of search files (written by search loop)
get_search_archive_path <- function()
  file.path(conf$data_dir, "posts", "search_archive")

# Get path of properties files (written by Shiny app)
get_properties_path <- function() file.path(conf$data_dir, "properties.json")

# Get path of plan base file (written by search loop)
get_plans_path <- function() file.path(conf$data_dir, "topics.json")

# Get paths of plans by sm (written by search loop)
get_plans_paths <- function() {
  sms <- active_social_media()
  if (length(sms) > 0) {
    ret <- sms
    names(ret) = ret
    sapply(ret, function(sm) {
      gsub(".json", paste0(".", sm, ".json"), get_plans_path())
    })
  } else {
    character(0)
  }
}

# Get task path (written by detect pipeline)
get_tasks_path <- function() file.path(conf$data_dir, "tasks.json")


# Get available alert training file path
get_alert_training_path <- function() {
  path <- get_user_alert_training_path()
  if (!file.exists(path)) path <- get_default_alert_training_path()
  path
}

# Get default alert training file path
get_default_alert_training_path <- function()
  system.file("extdata", "alert-training.xlsx", package = get_package_name())

# Get user alert training file
get_user_alert_training_path <- function() {
  paste(conf$data_dir, "alert-training.xlsx", sep = "/")
}

# Get available geotraining file path
get_geotraining_path <- function() {
  path <- get_user_geotraining_path()
  if (!file.exists(path)) path <- get_default_geotraining_path()
  path
}

# Get default languages file path
get_default_geotraining_path <- function()
  system.file("extdata", "geo-training.xlsx", package = get_package_name())

# Get user geo training file
get_user_geotraining_path <- function() {
  paste(conf$data_dir, "geo-training.xlsx", sep = "/")
}

# Get geo training evaluation file
get_geotraining_evaluation_path <- function() {
  paste(conf$data_dir, "geo-training-evaluation.json", sep = "/")
}
# Get default languages file path
get_default_available_languages_path <- function()
  system.file("extdata", "languages.xlsx", package = get_package_name())

# Get available languages file path (written by Shiny app)
get_available_languages_path <- function() {
  path <- paste(conf$data_dir, "languages.xlsx", sep = "/")
  if (!file.exists(path)) path <- get_default_available_languages_path()
  path
}

# Get default topics file path
get_default_known_users_path <- function()
  system.file("extdata", "users.xlsx", package = get_package_name())

# Get topics file path either from user or package location (written by Shiny app)
get_known_users_path <- function(data_dir = conf$data_dir) {
  users_path <- paste(data_dir, "users.xlsx", sep = "/")
  if (!file.exists(users_path)) users_path <- get_default_known_users_path()
  return(users_path)
}

# Get default topics files path
get_default_topics_path <- function()
  system.file("extdata", "topics.xlsx", package = get_package_name())

# Get topics file path either from user or package location (written by Shiny app)
get_user_topics_path <- function(data_dir = conf$data_dir) {
  topics_path <- paste(data_dir, "topics.xlsx", sep = "/")
  if (!file.exists(topics_path)) topics_path <- get_default_topics_path()
  return(topics_path)
}

# Get default country files path
get_default_countries_path <- function()
  system.file("extdata", "countries.xlsx", package = get_package_name())

# Get countries file path either from user or package location (written by Shiny app)
get_countries_path <- function(data_dir = conf$data_dir) {
  countries_path <- paste(data_dir, "countries.xlsx", sep = "/")
  if (!file.exists(countries_path))
    countries_path <- get_default_countries_path()
  return(countries_path)
}

# Get default subscribers file path
get_default_subscribers_path <- function()
  system.file("extdata", "subscribers.xlsx", package = get_package_name())

# Get the path for default or user defined subscribed user file (written by Shiny app)
get_subscribers_path <- function() {
  path <- paste(conf$data_dir, "subscribers.xlsx", sep = "/")
  if (!file.exists(path)) path <- get_default_subscribers_path()
  path
}

# Get email template path
get_email_alert_template_path <- function()
  system.file("extdata", "mail.html", package = get_package_name())

# Get Scala Building Tools (SBT) dependencies file
get_sbt_file_dep_path <- function()
  system.file("extdata", "sbt-deps.txt", package = get_package_name())

# Get JAR directory
get_jars_dest_path <- function() file.path(conf$data_dir, "jars")

# Get application JAR (embedded on package)
get_app_jar_path <- function()
  system.file(
    "java",
    "ecdc-episomer-bundle_2.13-1.0.jar",
    package = get_package_name()
  )

# Get hadoop home path for winutils
get_winutils_hadoop_home_path <- function() file.path(conf$data_dir, "hadoop")

# Get winutils path
get_winutils_path <- function()
  file.path(get_winutils_hadoop_home_path(), "bin", "winutils.exe")

# Get JSON file name for alert on given date
get_alert_file <- function(date) {
  alert_folder <- file.path(conf$data_dir, "alerts")
  if (!file.exists(alert_folder)) dir.create(alert_folder)
  alert_folder <- file.path(alert_folder, strftime(date, format = "%Y"))
  if (!file.exists(alert_folder)) dir.create(alert_folder)
  alert_file <- file.path(
    alert_folder,
    paste(strftime(date, format = "%Y.%m.%d"), "-alerts.json", sep = "")
  )
}

# Get geonames txt file
get_geonames_txt_path <- function() {
  file.path(conf$data_dir, "geo", "allCountries.txt")
}

# Get geonames parquet path
get_geonames_parquet_path <- function(relative = FALSE) {
  if (!relative) file.path(conf$data_dir, "geo", "all-geos.parquet") else
    file.path("geo", "all-geos.parquet")
}

# Get geonames cities parquet path
get_cities_parquet_path <- function(relative = FALSE) {
  if (!relative) file.path(conf$data_dir, "geo", "all-cities.parquet") else
    file.path("geo", "all-cities.parquet")
}

# Get geonames index path
get_geonames_index_path <- function(relative = FALSE) {
  if (!relative) file.path(conf$data_dir, "geo", "all-geos.parquet.index") else
    file.path("geo", "all-geos.parquet.index")
}

# Get path of the topics-keyword.json file used for ignoring topic in topwords
get_topic_keywords_path <- function() {
  geo_folder <- file.path(conf$data_dir, "geo")
  if (!file.exists(geo_folder)) dir.create(geo_folder)
  file.path(geo_folder, "topic-keywords.json")
}

# 
get_forced_geo_path <- function() {
  file.path(conf$data_dir, "geo", "forced-geo.json")
}

get_forced_geo_codes_path <- function() {
  file.path(conf$data_dir, "geo", "forced-geo-codes.json")
}

# Get languages index path
get_lang_index_path <- function(relative = FALSE) {
  if (!relative) file.path(conf$data_dir, "geo", "lang_vectors.index") else
    file.path("geo", "lang_vectors.index")
}

# Get language vectors path
get_lang_vectors_path <- function(code) {
  file.path(conf$data_dir, "languages", paste(code, ".txt.gz", sep = ""))
}

# Get language model path
get_lang_model_path <- function(code) {
  file.path(conf$data_dir, "languages", paste(code, ".txt.gz.model", sep = ""))
}

# Get language model stamp
get_lang_stamp_path <- function(code) {
  file.path(conf$data_dir, "languages", paste(code, ".txt.gz.stamp", sep = ""))
}

# Get post collection path
get_post_togeo_path <- function() {
  file.path(conf$data_dir, "geo", "togeolocate.json")
}

# Get path of the geolocating json file
get_post_geoing_path <- function() {
  file.path(conf$data_dir, "geo", "geolocating.json")
}

# Get path of posts to aggregate
get_post_toaggr_path <- function() {
  file.path(conf$data_dir, "geo", "toaggregate.json")
}
# Get path of posts being aggregated
get_post_aggring_path <- function() {
  file.path(conf$data_dir, "geo", "aggregating.json")
}
# Get path of the session info log file
get_session_info_path <- function() {
  file.path(conf$data_dir, "session-info.log")
}

---
title: "Add a new social media"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{add_new_social_media}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(episomer)
```
## Introduction

### Common post format
In order to support multiple social media, epiSoMeR transforms each response containing posts into a standardised common format. This format simplifies the handling of distinc types of social media. In order to implement this a set of functions needs to be implemented for each social media following a specific namingo convention.

The common post format is defined as follow

- Response attributes
  - network: string
  - posts: [...]
  - count: number, 
  - pagination":\{...\}
- post in posts
    - id: string
    - uri: string
    - created\_at: datetime
    - user\_name: string
    - text: string
    - lang: string
    - quoted\_text: string
    - quoted\_lang: string
    - tags: [...]
    - urls: [...]
    - categories: [...]
- tag in post.tags: string
- url in post.urls: string
- category in post.categories: string
- pagination: attributes used for paginating requests

## Pre-requisites

Before trying to add a new social media, you need to ensure that you are in "development mode", i.e:

- the package sources are cloned in your computer
- your working directory is the root of the R package
- you have made a call to `devtools::load_all()` in your R console

## Add a new social media

### Creates API and plan files

In order to add a new social media, you need to follow the following steps:

- call the `add_new_social_media()` function:

```{r, eval = FALSE}
add_new_social_media(social_media = "new_media")
```

2 files are expected to be created:

- `R/sm-new_media-api.R`: API functions for the new social media
- `R/sm-new_media-plan.R`: Plan functions for the new social media

Functions are provided as "skeleton" to be filled, it's advised to look at the existing functions for the reference social media (i.e "bluesky") and adapt the functions to the new social media.

### Edit `config.R`

In `config.R`, please look at any reference to the reference social media (i.e "bluesky") and adapt the functions to the new social media.

The following functions are expected to be adapted:

- `save_config()`

example:

```{r, eval = FALSE}
temp$sm_alerts_bluesky <- conf$sm_alerts_bluesky
temp$sm_activated_bluesky <- conf$sm_activated_bluesky
```

needs to be completed with:

```{r, eval = FALSE}
temp$sm_alerts_new_media <- conf$sm_alerts_new_media
temp$sm_activated_new_media <- conf$sm_activated_new_media
```

- `get_empty_config()`

example:

```{r, eval = FALSE}
ret$sm_alerts_bluesky <- TRUE
ret$sm_activated_bluesky <- TRUE
```

needs to be completed with:

```{r, eval = FALSE}
ret$sm_alerts_new_media <- TRUE
ret$sm_activated_new_media <- TRUE
```

- `setup_config()`

example:

```{r, eval = FALSE}
conf$sm_alerts_bluesky <- temp$sm_alerts_bluesky
conf$sm_activated_bluesky <- temp$sm_activated_bluesky
```

needs to be completed with:

```{r, eval = FALSE}
conf$sm_alerts_new_media <- temp$sm_alerts_new_media
conf$sm_activated_new_media <- temp$sm_activated_new_media
```

The following should be completed, depending on the credentials you are using (user/password, token ...):

```{r, eval = FALSE}
for (v in c("bluesky_user", "bluesky_password")) {
  if (is_secret_set(v)) {
    conf$auth[[v]] <- get_secret(v)
  }
}
```

### Adapt the shiny app

In the `R/shiny-app.R` file, please look at any reference to the reference social media (i.e "bluesky") and adapt the functions to the new social media.

In particular, be vigilant to the way of configuring the credentials for the new social media.

Below, an example of the current behavior for the bluesky social media, on the UI side of the app:

```{r, eval = FALSE}             
shiny::textInput(
  "bluesky_user",
  label = NULL,
  value = if (is_secret_set("bluesky_user")) {
    get_secret("bluesky_user")
  } else {
    NULL
  }
)
```

```{r, eval = FALSE}
shiny::passwordInput(
  "bluesky_password",
  label = NULL,
  value = if (is_secret_set("bluesky_password")) {
    get_secret("bluesky_password")
  } else {
    NULL
  }
)
```

Note that the elements names used to set up the credentials, here `bluesky_user` and `bluesky_password`, should match the ones used in the `config.R` file.

On the server side of the app, you should adapt the call to `sm_api_set_auth()` to use the new social media.

```{r, eval = FALSE}
sm_api_set_auth(
  network = "new_media",
  shiny_input_list = input
)
```

`sm_api_set_auth()` is a generic function that will call the appropriate function to set the credentials for the new social media. It expects the function `set_new_media_auth()` to be defined in the `R/sm-new_media-api.R` file.

---
title: "episomer: user documentation"
author: "European Centre for Disease Prevention and Control (ECDC)"
output: 
  - rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{episomer: user documentation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: "episomer_refs.bib"
---

```{r, echo = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

library(episomer)
package_name <- "episomer"
package_desc <- packageDescription(package_name)
pkg_version <- package_desc$Version
pkg_date <- package_desc$Date
authors_str <- gsub("^ *|(?<= ) |\n| *$", "", package_desc$Authors, perl = TRUE)
author_list <- eval(parse(text = authors_str))
pkg_authors <- paste(
  format(
    author_list,
    include = c("given", "family", "email", "comment"),
    braces = list(
      email = c("<", ">,<br />"),
      comment = c("", "")
    )
  ),
  collapse = "<br /><br />"
)
pkg_maintainer <- package_desc$Maintainer
pkg_license <- package_desc$License
pkg_url <- package_desc$URL
```


## Description

`Episomer` has been created by building on `epitweetr` objectives and functionalities. The `episomer` package allows you to automatically monitor trends of posts by time, place and topic in social media platforms. This automated monitoring aims at the early detection of public health threats through the detection of signals (e.g. an unusual increase in the number of posts for a specific time, place and topic). This release of `episomer` collects posts from Bluesky using its API (https://docs.bsky.app/), and can be extended to support other social media platforms. The `episomer` package was designed to focus on infectious diseases, and it can be extended to cover all hazards or other fields of study by modifying the topics and keywords.

The general principle behind `episomer` is that it collects posts and related metadata according to specified topics. Posts are stored on your computer in a database that can be used to calculate statistics or as a search engine. `episomer` geolocalises the posts and collects information on keywords, URLs and hashtags within a post. Posts are aggregated according to topic and geographical location. Next, a signal detection algorithm identifies when the number of posts (by topic and geographical location) exceeds what is expected for a given day. If the number of posts exceeds what is expected, `episomer` sends out email alerts to notify those who need to further investigate these signals following the epidemic intelligence processes (filtering, validation, analysis and preliminary assessment).

The package includes an interactive web application (Shiny app) with six pages: (i) dashboard, where a user can visualise and explore posts (Fig 1), (ii) alerts page, where you can view the current alerts and train machine learning models for alert classification on user-defined categories (Fig 2), (iii) geotag page, where you can evaluate the geolocation algorithm and provide annotations for improving its performance (Fig 3), (iv) data protection page, where you can search, anonymise and delete posts from the episomer database to support data deletion requests (Fig 4), (v) configuration page, where you can change settings and check the status of the underlying processes (Fig 5), and (vi) troubleshoot page, which provides automatic checks and hints for using `episomer` and all its functionalities (Fig 6). 

On the dashboard, users can view the aggregated number of posts over time, the location of these posts on a map and different most frequent elements found in or extracted from these posts (words, hashtags and URLs). These visualisations can be filtered by the topic, location and time period you are interested in. Other filters are available and include the possibility to adjust the time unit of the timeline, whether quotes should be included, the sensitivity of the prediction interval for the signal detection, and the number of days used to calculate the threshold for signals. This information is also downloadable directly from this interface in the form of data, pictures, and/or reports.

More information on the methodology used is available in the [episomer peer-review publication](https://www.eurosurveillance.org/content/10.2807/1560-7917.ES.2022.27.39.2200177). In addition, you can also visit the [general post](https://github.com/EU-ECDC/episomer/discussions/7) in the discussion forum of the GitHub episomer repository for additional materials and training.

**Shiny app dashboard:**

<center>Fig 1: Shiny app dashboard figure</center>

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/dashboard.png")
```


**Shiny app alerts page:**

<center>Fig 2: Shiny app alerts page</center>

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/alerts.png")
```


**Shiny app geotag evaluation page:**

<center>Fig 3: Shiny app geotag evaluation page</center>

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/geotag_evaluation.png")
```

**Shiny app data protection page:**

<center>Fig 4: Shiny app data protection page</center>

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/data_protection.png")
```

**Shiny app configuration page:**

<center>Fig 5: Shiny app configuration page</center>

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/config.png")
```


**Shiny app troubleshoot page:**

<center>Fig 6: Shiny app troubleshoot page</center>

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/troubleshoot.png")
```

## Background
### Epidemic Intelligence at ECDC
    
Article 3 of the European Centre for Disease Prevention and Control (ECDC) funding regulation and the Regulation (EU) 2022/2371 on serious cross-border threats to health have established the detection of public health threats as a core activity of ECDC.

ECDC performs Epidemic Intelligence (El) activities aiming at rapidly detecting and assessing public health threats, focusing on infectious diseases, to ensure EU's health security. Social media is one of the sources that ECDC uses for the early detection of signals of public health threats. Until 2020 when epitweetr was published, the monitoring of social media was mainly performed through the screening and analysis of posts from pre-selected experts or organisations, mainly in Twitter and Facebook.

More information and an online tutorial are available:

[EI sources](https://www.ecdc.europa.eu/en/training-and-tools/surveillance-and-outbreak-tools/information-sources-epidemic-intelligence-and)

[EI tutorial](https://learning.ecdc.europa.eu/enrol/index.php?id=519%3Fmtm_campaign%3Devaredirection) 


### Monitoring social media trends

Some signals are not detected at all or are not detected early enough by the methods explained above. The automated monitoring of social media (e.g. analysis of social media trends) allows the detection of signals that may not be detected by the monitoring of pre-selected accounts in social media, and it improves the timeliness of signal detection.

The analysis of trends from social media by topic, time and place generates relevant signals for early detection.

In 2019, ECDC developed a prototype of a manual R-based tool for the early detection of public health threats from X (formerly known as Twitter) data. In October 2020, ECDC published `epitweetr`, an R package for collecting data from X, achieving a wider geolocation of posts, increased performance and greater automation. `epitweetr` was discontinued because of changes in the X API terms and conditions and archived on CRAN. `episomer` is an extension of `epitweetr`, designed to support multiple social media platforms and consequently achieving higher reliability.

### Objectives of episomer

The primary objective of `episomer` is to use Social Media APIs in order to detect early signals of potential threats by topic and by geographical unit.

Its secondary objective is to enable users, through an interactive web interface, to explore the trend of posts by time, geographical location and topic, including information on top words and numbers of posts from trusted users, using charts and tables.

### Repository of episomer material and training
More information on episomer is available in the [episomer GitHub discussions](https://github.com/EU-ECDC/episomer/discussions). This [post](https://github.com/EU-ECDC/episomer/discussions/7) contains a summary of relevant links and materials to new users.

## Hardware requirements

The minimum and suggested hardware requirements for the computer are in the table below:

  Hardware requirements                 Minimum        Suggested
  ------------------------------------- -------------- ------------------
  RAM Needed                            8GB            16GB recommended
  CPU Needed                            4 cores        12 cores
  Space needed for 3 years of storage   3TB            5TB

The CPU and RAM usage can be configured in the Shiny app configuration page (see section *The interactive user application (Shiny app)>The configuration page*). The RAM, CPU and space required may depend on the number and popularity of the topics included in the collection process.

## Installation

`episomer` is designed to be platform-independent, working on Windows, Linux and Mac. We recommend that you use `episomer` on a computer that can be run continuously. You can switch the computer off, but you may miss some posts if the downtime is significant. This significant downtime will impact the alert detection. 

**If you need to upgrade or reinstall episomer after activating its tasks, you must stop the tasks from the Shiny app or restart the machine running episomer first.**

### Installation steps

You can find below a summary of the steps required to install `episomer`. Further detailed information is available in the corresponding sections.

1. Ensure all pre-requisites are installed
2. Install episomer (CRAN version or different version using tar.gz file)
3. Select the folder (or create a new folder) for `episomer`
4. Launch the `episomer` Shiny app (ensure to indicate the full path to your data directory)
5. Check the troubleshoot page
6. Modify the parameters in the configuration page as needed. You must provide the following information to enable all functionalities: Bluesky credentials, SMTP for the email sending alert emails and status emails and list of subscribers. The remaining parameters have default values that you can modify if needed. Settings are auto-saved after changing them.
7. Activate 'Requirements & alerts' pipeline in the configuration page
5. When requested in the dependencies task, activate 'episomer database'
6. After the task languages is completed, activate 'Data collection & processing'
7. Alerts task may show an error if posts have not been aggregated yet. Wait few minutes and click on 'Run alerts'

### Prerequisites 

#### Mandatory for running episomer

Before using `episomer`, the following items need to be installed:

-   R version 3.6.3 or higher

-   Java 17-21 64-bit, e.g. [*https://learn.microsoft.com/en-us/java/openjdk/download*](https://learn.microsoft.com/en-us/java/openjdk/download) refers to the Microsoft releases of java for multiple platforms including Windows and Mac.

-   If you are running it on Windows, you will also need Microsoft Visual C++ which in most cases is likely to be pre-installed:

    -   Microsoft Visual C++ 2010 Redistributable Package (x64) [*https://www.microsoft.com/en-us/download/details.aspx?id=26999*]

#### Mandatory for some of the functionalities in episomer

-   Pandoc, for exporting PDFs and Markdown

    -   [*https://pandoc.org/installing.html*]
-   Tex installation (TinyTeX or MiKTeX) (or other TeX installation) for exporting PDFs

    -   Easiest: [*https://yihui.org/tinytex/*] install from R, logoff/logon required after installation

    -   [*https://miktex.org/download*] full installation required, logoff/logon required after installation

-   Machine learning optimisation (only for advanced users)

    -   Open Blas (BLAS optimizer), which will speed up some of the geolocation processes: [*https://www.openblas.net/*] Installation instructions: [*https://github.com/fommil/netlib-Java*]

    -   **or** Intel MKL ([*https://www.intel.com/content/www/us/en/developer/tools/oneapi/onemkl.html*])

-   A scheduler

    -   If using Windows, you need to install the R package: taskscheduleR

    -   If using Linux, you need to plan the tasks manually
    
    -   If using a Mac, you need to plan the tasls manually

#### Only for R developers

If you would like to develop `episomer` further, then the following development tools are needed:

-   Git (source code control) [*https://git-scm.com/downloads/*]

-   Sbt (compiling scala code) [*https://www.scala-sbt.org/download/*]

-   If you are using Windows, then you will additionally need Rtools: [*https://cran.r-project.org/bin/windows/Rtools/*]

### External dependencies

`episomer` will need to download some dependencies in order to work. The tool will perform this automatically the first time the alert detection process is launched. The Shiny app configuration page will allow you to change the target URLs of these dependencies, which are the following:

- *CRAN JARs*: Transitive dependencies for running Spark, Lucene, Pekko and embedded scala code. *[https://repo1.maven.org/maven2]*

- *Winutils.exe* (Windows only) This is a Hadoop binary necessary for running SPARK locally on Windows *[https://github.com/steveloughran/winutils/raw/master/hadoop-3.0.0/bin/winutils.exe]*. 

Please note that during the dependencies download, you will be prompted: first, to stop the embedded database and then, to enable it again. If you are on a Windows machine and you have activated the tasks using the 'activate' buttons on the configuration page, you can perform this tasks by disabling and enabling the tasks in the 'Windows Task Scheduler'. For more information, see the section 'Setting up post collection and the alert detection loop'.

### Installing episomer from CRAN

After installing all required dependencies listed in the section “Prerequisites for running episomer”, you can install `episomer`:

```{r, eval=FALSE, echo=TRUE}
install.packages(episomer)
```

#### Environment variables

Additionally, if you want to use a Java version different than your system's default, the Java installation home should be accessible to the R environment. To check this, type in the R console:

```{r, eval=FALSE, echo=TRUE}
system("java -version")
```

This command should return the version of the java version available to R. In case this command returns something different than a version between 17 and 21, you have yo provide the location of the right version of java by setting the JAVA\_HOME environment variable. Please follow your specific operating system (OS) instructions for doing so. You should be able to detect the current value of JAVA\_HOME by running this instruction:

```{r, eval=FALSE, echo=TRUE}
Sys.getenv("JAVA_HOME")
```
The Java binary (executable) should be located in the subfolder "bin" os JAVA\_HOME.

The first time you run `episomer`, if it cannot identify a secure password store provided by the operating system, you will see a pop-up window requesting a keyring password (Linux and Mac). This is a password necessary for storing encrypted Bluesky credentials. You will be asked for this password each time you run `episomer`. You can avoid this by setting a system environment variable named ecdc\_twitter\_tool\_kr\_password containing the chosen password.

#### Launching the episomer Shiny app

You can launch the `episomer` Shiny app from the R session by typing in the R console. Replace “data\_dir” with the designated data directory (full path) which is a local folder you choose to store posts, time series configuration files and logs in:

```{r, eval=FALSE, echo=TRUE}
library(episomer)
episomer_app("data_dir")
```

Please note that the data directory entered in R should have '/' instead of '\\' (an example of a correct path would be 'C:/user/name/episomer'). This applies especially in Windows if you copy the path from the File Explorer.

Alternatively, you can use a launcher: In an executable .bat or shell file type the following (replacing “data\_dir” with the designated data directory):

R --vanilla -e episomer::episomer\_app("data\_dir")

You can check that all requirements are properly installed in the troubleshoot page. More information is available in section *The interactive user application (Shiny app)>Dashboard:The interactive user interface for visualisation>The troubleshoot page*

### Setting up data collection and alert detection loop

In order to use `episomer`, you will need to collect and process posts from social media, run the 'episomer database' and 'Requirements & alerts' pipelines. Further details are also available in subsequent sections of this user documentation. The main steps needed are as follows:

-   Launch the Shiny app (from the R console) replacing 'data\_dir' with the full path of your data folder e.g. 'C:/Users/name/episomer'
```{r, eval=FALSE, echo=TRUE}
library(episomer)
episomer_app("data_dir")
```

-   In the configuration page of the Shiny app, in the manual tasks of the "Requirements & alerts pipeline", click on "Run dependencies", "Run GeoNames" and "Run languages" (their status will change to “pending”). This allows the 'Requirements & alerts' pipeline to download the elements needed. As long as no languages are added and no updates are available in geonames.org, these tasks have to be run only the first time you install `episomer`.

```{r echo=FALSE, fig.cap="", out.width = '50%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/run_tasks.png")
```

-   Set up the episomer authentication using your Bluesky username and password.

-   Activate the embedded database 

    -   Windows: Click on the “Episomer database” activate button

    -   Other operating systems: In a new R session, run the following command
    
```{r, eval=FALSE, echo=TRUE}
library(episomer)
fs_loop("data_dir")
```
   -   You can confirm that the embeded database is running if the 'episomer database' status is "Running" in the Shiny app configuration page and "true" in the Shiny app troubleshoot page.

-   Activate the post collection and data processing

    -   Windows: Click on the “Data collection & processing” activate button

    -   Other operating systems: In a new R session, run the following command
    
```{r, eval=FALSE, echo=TRUE}
library(episomer)
search_loop("data_dir")
```

   -   You can confirm that the post collection is running, if the 'Data collection & processing' status is "Running" in the Shiny app configuration page (green text in screenshot above) and "true" in the Shiny app troubleshoot page.
   
-   Activate the 'Requirements & alerts' pipeline:

    -   Windows: Click on the “Requirements & alerts pipeline” activation button
    -   Other operating systems: In a new R session run the following command
```{r, eval=FALSE, echo=TRUE}
library(episomer)
detect_loop("data_dir")
```

   -   You can confirm that the 'Requirements & alerts' pipeline is running, if the 'Requirements & alerts' pipeline status is "Running" in the Shiny app configuration page and "true" in the Shiny app troubleshoot page.

-   You will be able to visualise posts after 'Data collection & processing' and 'episomer database' are activated and the languages task has finished successfully.

-   You can start working with the generated signals. **Happy signal detection!**

For more details, you can go through the section *How does it work? General architecture behind `episomer`*, which describes the underlying processes behind the post collection and the signal detection. Also, the section *"The interactive Shiny application (Shiny app)>The configuration page"* describes the different settings available in the configuration page.


## How does it work? General architecture behind episomer

The following sections describe in detail the above general principles. The settings of many of these elements can be configured in the Shiny app configuration page, which is explained in the section *The interactive Shiny application (Shiny app)>The configuration page*.

###  Collection of posts

#### Use of the Bluesky search API version

`episomer` uses the Bluesky search API 'searchPost' endpoint [*https://docs.bsky.app/docs/api/app-bsky-feed-search-posts*]. This service enable authenticated users to access published messages free of charge. There are no mentions on exhaustivieness of this endpoint in the documentaton but it appears to return same results as Bluesky search feature in the app. There is no limit on the number period of messaged that can be obtained but by default, episomer will limit to last 30 days (this parameter can be changed in the configuration page). There is no limitation on the period for the search, but `episomer` limits by default the data collection on the last 30 days (this parameter can be change in the configuration page).

Other attributes of the `searchPost` endpoint includes:

-   A maximum of 3000 requests every 5 minutes are supported by the Bluesky API

-   Each request returns a maximum of 100 posts and/or reposts

#### Bluesky authentication

In order to collect data from Bluesky you need to provide user credentials to `episomer`. This is done by entering your loggin and password as used on the login page of Bluesky. These settings are provided in the congiguration page.

```{r echo=FALSE, fig.cap="", out.width = '50%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/bluesky_authentication.png")
```

#### Topics and post collection queries

After authentication, you need to specify a list of topics in `episomer` to indicate which posts to collect. For each topic, you have one or more queries that `episomer` uses to collect the relevant posts (e.g. several queries for a topic using different terminology and/or languages). 

A query consists of keywords and operators that are used to match post attributes. Since episomer is designed to support multiple social media, it has its own structure for writinh queries which is translated to match each provider syntax. The expected format of queries is defined by the follwoing rules.
-   No parenthesis are supported
-   Sub queries be sepparated by the OR keyword e.g. *measles OR sarampion*
-   If a subquery contain multiple terms they have to be concatenated with the AND keyword e.g. *measles AND feaver OR sarampion AND fiebre*
-   Excluded terms can be included at the end of a subquery or as a separate subquery using a minus sign. They are always applied to all subqueriese.g.  *measles AND feaver OR sarampion AND fiebre -vaccination - vacunas* 
-   You can provide synonyms on terms of a subquery using the "/" operator e.g. the previous example can be rewritten as  *measles/sarampion AND feaver/fiebre -vaccination - vacunas*. Use synonyms with caution. When the OR syntax is not supported natively the number of underlying queries grows as an geometric progression.
-   Other special characters are passed to the query provider without modification e.g. exact match using double quotes.
-  To avoid issues in social media providers a best practice is to limit your query to 10 keywords and operators and limit complexity of the query. If a query surpasses this limit, it is recommended to split the topic in several queries. 

`episomer` comes with a default list of topics as used by the ECDC Epidemic Intelligence team. You can view details of the list of topics in the Shiny app configuration page (see screenshot below). In addition, the colour coding in the downloadable file allows users to see if the query for a topic is too long (red colour) and the topic should be split in several queries.

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/topics.png")
```

In the configuration page, you can also download the list of topics, modify and upload it to `episomer`. The new list of topics will then be used for data collection and visible in the Shiny app. The list of topics is an Excel file (\*.xlsx) as it handles user-specific regional settings (e.g. delimiters) and special characters well. You can create your own list of topics and upload it too, noting that the structure should include at least:

-   The name of the topic, with the header “Topic” in the Excel spreadsheet. The name of the topic is not visible in reports and it is the identifier of the topic. Modifying this field its equivalent to remove and create a new topic. This name should include alphanumeric characters, spaces, dashes and underscores only. Note that it should start with a letter and it is case insensitive.

-   The label of the topic, with the header “Label” in the Excel spreadsheet. The label of the topic is shown on reports and can be changed wihout impact in data collection. This label should include alphanumeric characters, spaces, dashes and underscores only. Note that it should start with a letter.

-   The query, with the header “Query” in the Excel spreadsheet. This is the query `episomer` uses in its requests to obtain posts from the social media API's. See above for syntax and constraints of queries.

The topics.xlsx file additionally includes the following fields:

-   An ID, with the header “\#” in the Excel spreadsheet, this number is just a reference for you and has no impact in the application.

-   An alpha parameter, with the header “Signal alpha (FPR)” in the Excel spreadsheet. FPR stands for "false positive rate". Increasing the alpha will decrease the threshold for signal detection, resulting in an increased sensitivity and possibly obtaining more signals. Setting this alpha can be done empirically and according to the importance and nature of the topic.

-   "Length\_charact" is an automatically generated field that calculates the length of all characters used in the query. This field is helpful as a request should not exceed 500 characters.

-   "Length\_word" indicates the number of words used in a request, including operators. Best practice is to limit your number of keywords to 10.

-   An alpha parameter, with the header “Outlier alpha (FPR)” in the Excel spreadsheet. FPR stands for “false positive rate”. This alpha sets the false positive rate for determining what an outlier is when downweighting previous outliers/signals. The lower the value, the fewer previous outliers will potentially be included. A higher value will potentially include more previous outliers.

-   "Rank" is the number of queries per topic

```{r echo=FALSE, fig.cap="", out.width = '80%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/Example_topics_list.png")
```

When uploading your own file, please modify the topic and query fields, but do not modify the column titles.


#### Scheduled plans to collect posts

As a reminder, `episomer` is scheduled to respect the rate limits of underlying data providers and performs multiple requests to obtain published messages. Each request return 100 posts. The requests return posts and quotes. These are returned in JSON format, which is a light-weighted data format.

In order to collect the maximum number of posts, given the API limitations, and in order for popular topics not to prevent other topics from being adequately collected, `episomer` uses “search plans” for each query.

The first “search plan” for a query will collect posts from the current date-time backwards until 30 days (this parameter can be changed in the configuration but depends also on API limitations) before the current “search plan” was implemented. The first “search plan” is the biggest, as no posts have been collected so far.

All subsequent “search plans” are done in scheduled intervals that are set up in the configuration page of the `episomer` Shiny app (see section *The interactive Shiny app > the configuration page > General*). For illustration purposes, let us consider the search plans are scheduled at four-hour intervals. The plans collect posts for a specific query from the current date-time back until four hours before the date-time when the current “search plan” is implemented (see image below). `episomer` will make as many requests (each returning up to 100 posts) during the four-hour interval as needed to obtain all posts created within that four-hour interval.

For example, if the “search plan” begins at 4 am on the 10^th^ of November 2021, `episomer` will launch requests for posts corresponding to its queries for the four-hour period from 4 am to midnight on the 10^th^ of November 2021. `episomer` starts by collecting the most recent posts (the ones from 4 am) and continues backwards. If during the four-hour time period between 4 am and midnight the API does not return any more results, the “search plan” for this query is considered completed.

```{r echo=FALSE, fig.cap="", out.width = '80%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/Schedule_plan_image.png")
```

However, if topics are very popular (e.g. COVID-19 in 2020 and 2021), then the “search plan” for a query in a given four-hour window may not be completed. If this happens, `episomer` will move on to the “search plans” for the subsequent four-hour window, and put any previous incomplete “search plan” in a queue to execute when “search plans” for this new four-hour window are completed.

Each “search plan” stores the following information:

  --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
  Field                 Type        Description
  --------------------- ----------- ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
  Network               Text        The name of the social media associated to this plan

  expected\_end         Timestamp   End DateTime of the current search window

  scheduled\_for        Timestamp   The scheduled DateTime for the next request. On plan creation this will be the current DateTime and after each request this value will be set to a future DateTime. To establish the future DateTime, the application will estimate the number of requests necessary to finish. If it estimates that N requests are necessary, the next schedule will be in 1/N of the remaining time.

  start\_on             Timestamp   The DateTime when the first request of the plan was finished

  end\_on               Timestamp   The DateTime when the last request of the plan was finished if that request reached a 100% plan progress.

  plan\_max\_date       Timestamp   The latest date for which this plan will collect messages, it will be defined after the first request.

  plan\_min\_date       Timestamp   The oldest date for which this plan will collect messages it will defined after the first request. The next plan will start collecting posts before this value.

  current\_min\_date    Timestamp   The oldest date for which this plan has actually collected messages. This value is updated after each requests and allow next request to keep going backwards until the plan\_min\_date

  requests              Int         Number of requests performed as part of the plan

  progress              Double      Progress of the current plan as a percentage. It is calculated as (current$plan\_max\_date - current\$current\_min\_date)/(current\$plan\_max\_date - current\$plan\_min\_date). If the underlying API returns no posts the progress is set to 100%. This only applies for non error responses containing an empty list of posts.
  --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

`episomer` will execute plans according to these rules:

-   `episomer` will detect the newest unfinished plan for each search query with the scheduled\_for variable located in the past.

-   `episomer` will execute the plans with the minimum number of requests already performed. This ensures that all scheduled plans perform the same number of requests.

-   As a result of the two previous rules, requests for topics with less messages will end first and will produce higher progress than topics with higher volume of tweets. Particularly when a topic has more messaged than those allowed to be downloaded by the API rate limits, these plans will be paused as newer plans are scheduled for future completion. The rationale behind this is that topics with such a large number of posts that the 4-hour search window is not sufficient to collect them, are likely to already be a known topic of interest. Therefore, priority should be given to smaller topics and possibly less well-known topics.

An example was the COVID-19 pandemic in 2020. In early 2020, there was limited information available regarding COVID-19, which allowed detecting signals with meaningful information or updates (e.g. new countries reporting cases or confirming that it was caused by a coronavirus). However, throughout the pandemic, this topic became more popular and the broad topic of COVID-19 was not effective for signal detection and was taking up a lot of time and requests for `epitweetr`. In such a case it is more relevant to prioritise the collection of smaller topics such as sub-topics related to COVID-19 (e.g. vaccine AND COVID-19), or to make sure you do not miss other events with less social media attention.

If search plans cannot be finished, several search plans per query may be in a queue:

```{r echo=FALSE, fig.cap="", out.width = '80%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/Search_plan_queues.png")
```

This design can have the draw back of slowing down big topics collection since `episomer` is trying to rebuilt last 30 days of history. If you are not interested in rebuilding history on a particular point of time, you can click on the "Dismiss past posts" button which will discard all previous/historical plans and will start collecting new data.

### Geolocating locations referenced in message content

In a parallel process to the collection of posts, `episomer` attempts to geolocate all collected posts using a supervised machine learning process. This process runs automatically after posts are collected. `episomer` stores location if mentioned in the text of a post (or a reposted or quoted post). 

The post location is extracted and stored by `episomer` based on the geolocation information found within a post text. In case a quoted post, it will extract the geolocation information from the original content and the text accompagning the quote. If neither are available, no post location is stored based on post text.

`episomer` identifies if a post text contains reference to a particular location by breaking down the post text into sets of words and evaluating those which are more likely to be a location by using a machine learning model. If several parts of the text are likely to be a location, `episomer` will chose the one closest to a topici based on keywoeds provided. After the location candidate has been identified `episomer` matches these words against a reference database, which is geonames.org. This is a geographical database available and accessible through various web services, under a Creative Commons attribution licence. The GeoNames.org database contains over 25,000,000 geographical names. `episomer` uses by default those with a known population (so just over 500,000 names). You can change this default parameter in the Shiny app configuration page, by unchecking "Simplified geonames". The database also contains longitude and latitude attributes of localities and variant spellings (cross-references), which are useful for finding purposes, as well as non-Roman script spellings of many of these names.

The matches can be performed at any level of administrative geography. The matching is powered by Apache Lucene, which is an open-source high-performance full-featured text search engine library.

To validate the candidate against geonames, a score is associated with the probability that a match is correct. A score is:

-   Higher if unusual parts of the name are matched

-   Higher if several administrative levels are matched

-   Higher if location population is bigger

-   Higher for countries and cities vs administrative levels

-   Higher for capital letter acronyms like NY

-   Lower for words that are more likely to be other kinds of words (non-geographical). For example, “Fair Play” town in Colorado. This is achieved by using language models provided by fasttext.cc.

You can select which languages you would like to check for other kinds of words by selecting the active language desired within the configuration page of the Shiny app and clicking on the "+" icon:

```{r echo=FALSE, fig.cap="", out.width = '80%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/config_languages.png")
```

In addition, you can unselect languages by selecting the language within the configuration page of the Shiny app and clicking on the "-"  icon. 

**At least one language must be downloaded before adding new languages or deleting any of the default languages.**


A minimum score (i.e., "geolocation threshold") can be globally set in the general settings on the configuration page to reduce the number of false positives (see image). All geolocations with a smaller score than the geolocation threshold will be discarded by the algorithm as post location. If there is more than one match over the minimum score, then the match with the highest score will be chosen.

The threshold is empirically chosen and can be evaluated against a human read of posts and post locations, in the geotag evaluation page.

#### Improving and evaluating geolocation performance

In the Geotag page, `episomer` will allow you to download the data that was used to build the classifiers for location identificaion. The predefined annotations are based on location vs non-location words. Locations are extracted from geonames.org database. Non-location words are obtained from common words in the downloaded models that are not present in GeoNames. You will be able to add posts to the annotation database and to manually annotate them until you reach the expected level of performance.

To help you on the evaluation process, `episomer` will calculate standard machine learning metrics for evaluating its capacity to identify the right location of geolocation words..


#### Stored geolocated post information

The geolocation of the match is stored as a country code (using the ISO 3166 standard) and as a longitude and latitude associated with the exact geolocation in the aggregated data.

### Most frequent elements found in and extracted from posts

`Episomer` counts trhree types of elements in posts

  - Words: Words present in posts. Words are then aggregated in the topwords time series

  - Hashtags: hashtags present in posts. Hashtags are then aggregated in the tags time series

  - URLs: URLs present in posts. URLs are then aggregated in the URLs time series

#### Aggregation of data

The aggregation process produces data on five subfolders in "fs" folder: geolocated, country\_counts, topwords, urls, tags, entities and contexts. These folders are splitted on week subfolders and each contains a Lucene index with the aggregated information. This data can be extracted as a dataframe through the public package function 'get\_aggregates'

In the geolocated time series, the number of posts or reposts are stored by topic, date, post text, post geolocation, post longitude and post latitude. Each of these entries also has the country associated with the post text geolocation. Note that posts without geolocation information are also included.

The country\_counts serie is used to create the trend line in the Shiny app. This is a smaller time serie, without the longitude and latitude information, and includes the number of posts by hour within a day, by country (according to post location or user location), topic (see screenshot), and whether a post was a quote or not. The known\_reposts and known\_original fields give the number of posts or reposts from a list of “important users”. In this file, posts without geolocation are also included. Including posts without geolocation information enables you to view all posts when selecting “world” as a region, regardless of whether geolocation was successful or not.

The aggregation by top element is stored in the topwords, URLs and tags subfolders in the fs folder which contain the number of posts and/or reposts by topic, top element, date, country of post location and whether a post was a repost or not (see screenshot).

### Signal detection

The main objective of `episomer` is to detect signals in the observed data streams, i.e. counts in the aggregated time series that exceed what is expected. For detecting signals, `episomer` uses an extended version of the EARS (Early Aberration Reporting System) algorithm [@fricker_comparing_2008], which in what follows is denoted by `eears` (extended EARS). This algorithm is part of the R package `surveillance` [@salmon_monitoring_2016].

As a default it uses a moving window of the past seven days to calculate a threshold. If the count for the current day exceeds this threshold, then a signal is generated.

#### Details of the algorithm underlying signal detection

The `eears` algorithm is applied on the counts from the past seven 24-hour blocks prior to the current 24 hour block of the signal detection. The running mean and the running standard deviation are calculated:

$$
\overline{y}_{0} = \frac{1}{7}\sum_{t=-7}^{-1} y_{t} \quad\text{and}\quad  s_{0}^{2} = \frac{1}{7 - 1}\sum_{t=-7}^{-1}{(y_{t} - \overline{y}_{0})}^{2},
$$

where $y_{t}, t=\ldots, -2, -1, 0$ denotes the observed count data time series with time index $0$ denoting the current block. Furthermore, the time index $-7,\ldots, -1$ denote the seven blocks prior to the current block.

Under the null hypothesis of no spikes, it is assumed that the $y_t$ are identically and independently $N(\mu, \sigma^2)$ distributed with unknown mean $\mu$ and unknown variance $\sigma^2$. Hence, the upper limit of a simple one-sided $(1-\alpha)\times$ 100% plug-in prediction interval for $y_0$ based on $y_{-7},\ldots,y_{-1}$ is given as 
$$
U_{0} = {\overline{y}}_{0} + z_{1 - a} \times s_{0},
$$
where $z_{1 - a}$ is the (1 − α)- quantile of the standard normal distribution. An alert is raised if $y_{0} > U_{0}$ . If one uses α=0.025, then this corresponds to investigating, if $y_{0}$ exceeds the estimate for the mean plus 1.96 times the standard deviation. However, as pointed out by @allevius_prospective_2017, the correct approach would be to compare the observation to the upper limit of a two-sided 95% prediction interval for $y_{0}$, because this respects both the sampling variation of a new observation *and* the uncertainty originating from the parameter estimation of the mean and variance. Hence, the statistical appropriate form is to compute the upper limit by
$$
U_{0} = \overline{\ y_{0}} + t_{1 - a}(7 - 1)\times s_{0} \times   \sqrt{1 + \frac{1}{7}}.
$$

where $t_{1 - a}(k - 1)$ denotes the 1 − α quantile of the t-distribution with k − 1 degrees of freedom.

#### Downweighting previous signals

If previous signals are included without modification in the historic values when calculating the running mean and standard deviation for the signal detection, then the estimated mean and standard deviation might become too large. This may mean that important current signals will not be detected. To address this issue, `episomer` downweights previous signals, such that the mean and standard deviation estimation is adjusted for such outliers using an approach similar to that used in the @farrington_statistical_1996. Historic values that are not identified as previous signals are given a weight of "1". Similarly, historic values identified as signals are given a weight lower than one and a new fit is performed using these weights (scaled s.t. they again sum to 7 observations). Details on the downweighting procedure can be found in [Annex I](#annexI) of this user documentation.


#### Timing of signal detection

Signal detection is carried out based on “days”, which are moving windows of 24 hours, moving according to the detect span (see also section *The interactive user application (Shiny app) > The configuration page > General*). The baseline is calculated on these “days” from -1 to -8 (if the current "day" is zero).

Signals are generated according to the detect span (see section *The interactive user application (Shiny app) > The configuration page > General*), with

-   general email alerts sent following this detect span (e.g. if the detect span was four hours, the email alerts will be sent every four hours)

-   email alerts sent in real-time. 

The different types of email alerts for each user can be specified in the configuration page (see section *The interactive user application (Shiny app) > The configuration page > General*).

#### The alpha parameter: the false positive rate of the signal detection

A key attribute of signal detection is the ability of an algorithm to detect true positives (true threats or events) without overloading the episomer analysts with too many false positives. In this way, the alpha parameter determines the threshold of the detection interval. If the alpha is high, then more potential signals are generated and if the alpha is low, fewer potential signals are generated (but potential threats or events could be missed). The setting of the alpha is often done empirically, and depends also on the resources of those investigating the signals and the importance of missing a potential threat or event.

There is a global alpha, that can be set/changed in the `episomer` configuration page under "Signal false positive rate" (see section *The interactive user application (Shiny app) > The configuration page > General*). Additionally, the default alpha can be overridden in the topics list. Here, if you like, you can associate each topic with a specific alpha, depending on the estimated public health importance of the topic or potential associated event or threat.

#### Bonferroni correction

To account for multiple testing, for country-specific signal detection, as a default, the alpha is divided by the number of countries. For continent-specific signal detection, the alpha is divided by the number continents. This is a Bonferroni correction for multiple testing.

To override this, you can uncheck "Bonferroni correction" in the "Signal detection" part of the configuration page in the Shiny app.


#### Using same weekdays as baseline

It is possible that there is a "day of the week effect", where more posts may be posted on a given day of the week (e.g. Monday) than on other days. To avoid this, you can also choose to calculate the baseline not on consecutive days, but on the past N days that correspond to the same 24 hour window N days back. This way if N = 7, the baseline is calculated using the “days” from -7, -14, -21, -28, -35, -42, -49 and -56 (if the current "day" is zero).

This option is on the configuration page of the Shiny app "Default same weekday baseline".

#### Sending email alerts

Emails containing a list of signals detected are sent automatically by `episomer` according to the detect span and the subscribers list. Due to the time necessary to collect, geolocate and aggregate the posts, email alerts will miss the most recent posts that have not yet gone to these processes. The lag between posts and alerts is expected to be less than (2 * ( collect\_span ) + detect\_span) which should be 3h30 using default values.

The email alerts will include the following information on the signals for each topic:

-   The date and hour the signal was detected

-   The geographical location(s) where the signal was detected

-   The most frequent elements (top words, URLs, hashtags, contexts and entities) in the posts

-   The number of posts and the threshold

-   The percentage of posts from important users

-   Information on the settings, such as: was the Bonferroni correction used, was the same weekday baseline used, were reposts included, etc.

-   The alert category estimated by `episomer` based on user's annotations

This information is also available in the alerts page of the Shiny app. 

The subscribers can receive real-time alerts (i.e. as soon as the detection loop is finalised) or scheduled alerts (e.g. once or twice a day). The subscribers list can be changed in the configuration page by downloading the Excel spreadsheet. This file has the following variables:

  - “User”: name of the subscriber (e.g. Jane Doe).

  - “Email”: email of the subscriber (e.g. jane.doe@email.com).

  - “Topics”: list of topics for which the subscriber will receive scheduled alerts. The names used must match the column “Topic” in the list of topics.

  - “Excluded”: topic for which the subscribers will not receive scheduled alerts.

  - “Real time Topics”: list of topics for which the subscriber will receive real-time alerts.

  - “Regions”: list of regions for which the subscriber will receive scheduled alerts.

  - “Real time Regions”: list of regions for which the subscriber will receive real-time alerts.

  - “Alert Slots”: these are the detection loop slots after which the subscriber will receive the scheduled alert. Available slots can be taken from “Launch slots” in the “General” section of the configuration page. If no value is included, the subscriber will receive real-time alerts for all topics and regions, even if there are real-time topics or regions specified in the Excel spreadsheet.

  - “One post alerts” (yes/no): Whether you want to receive alerts containing just one post.
  
  - “Topics ignoring one post alerts”: Topics that will be ignored for one-post alerts.
  
  - “Regions ignoring one post alerts”: Regions that will eb ignored for one-post alerts.
  
When including more than one topic and/or region in the subscribers list, these should be separated by semi-colon (;) with no spaces (e.g. Ebola;infectious diseases;dengue). The names must match the column “Topics” in the list of topics and the column “Name” in the country/region list from the configuration page.

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/subscribers.png")
```

###  Folder structure

`episomer` stores and aggregates posts. When launching the application, you have to designate the location for storing the data and the configuratio files as the “data folder”.

Within the **data** folder there are three JSON files:

-   properties.json, generated from the information from the General properties of the Shiny app

-   topics.bluesky.json managed by the search loop: it keeps a track of post collection plans and progress in Bluesky

-   tasks.json managed by the detect loop: it keeps information and status of the different tasks done by this process.

There are also the following subfolders:

-   "fs", which contains the Lucene indexes for storing posts and aggregated time series

-   "geo", which contains the GeoNames data as text, index files, posts waiting to be geolocated and settings about the geolocation algorithm.

-   "alert-ml", which contains the machine learning models used for alert classification

-   "hadoop", which contains Spark dependencies for Windows operating systems

-   "jars", which contains collections of Java dependencies needed in the geolocation and aggregate processes

-   "languages", which contains fasttext files indexes and models used to perform geolocation in the post text

-   "stats", which contains JSON files with statistics used to optimise the aggregate process by linking post files and posted dates of posts

-   "alerts", which contains JSON files of alerts detected by the requirements & alerts pipeline

-   "collections", which contains the aggregations produced by `episomer`

-   "jobs", containig information of logs produced by the different running processes 


#### Fs folder &gt; posts
In the fs folder, the subfolder “posts” contains Lucene indexes storing the content of the collected posts and the geolocation information.

The *posts* contains subfolders for each week.

The *geolocated folder* contains compressed JSON files with geolocated information produced by the geolocation algorithm.

#### Fs folder &gt; country\_counts, geolocated, tags, topwords, urls

In the folders *country_counts*,  *geolocated*, *tags*, *topwords*, *urls*, `episomer` stores the aggregated data of the geolocated posts as well as the top elements identified.

Each folder is named matching the respective time series and emitted by ISO week of date of publication. Each weekly folder contains a Lucene index.

This is the aggregate information as described in the section “How does it work? General architecture of episomer > Aggregation”.

## The interactive user application (Shiny app)

You can launch the `episomer` interactive user application (Shiny app) from the R session by typing in the R console (replace “data\_dir” with the desired data directory):

```{r, eval=FALSE, echo=TRUE}
episomer_app("data_dir")
```
You can also reduced version of the `episomer` app dedicadet to the interactive dashboard and the administrative featurs.

```{r, eval=FALSE, echo=TRUE}
# Opens only the interactive dashboard
dashboard_app("data_dir")

# Opens only the rest of the features with the administrative 
admin_app("data_dir")
```

Alternatively, you can use a launcher: Put the following content in an executable bat or sh file, (replacing “data\_dir” with the expected data directory)

R --vanilla -e episomer::episomer\_app('data\_dir')


The `episomer` interactive user application has six pages:

-   The dashboard, where a user can visualise and explore posts

-   The alerts page, where a user can view the current alerts and train machine learning algorithms for alert classification on user's defined categories

-   The geotag evaluation page, where a user can evaluate the geolocation algorithm and provide annotations for improving its performance

-   The data protection page, where a user can search, anonymise and delete posts from the `episomer` database to support data deletion requests 

-   The configuration page, where a user can change settings and check the status of the underlying processes

-   The troubleshoot page, with with automatic checks and hints for using `episomer` with all its functionalities and tools to backup & share your installation


### Dashboard: The interactive user interface for visualisation

The dashboard is where you can interactively explore visualisations of posts. It includes a line graph (trend line) with alerts, a map and top words, URLs, hashtags, entities and contexts of posts for a given topic. After selecting the parameters, you have to click on the 'Run' button in order to see or refresh the report.

In order to interactively explore the data, you can select from several filters, such as topics, countries and regions, time period, time unit, signal confidence and days in baseline. After selecting the filters, you must click on 'Run' to see the outputs in the dashboard.

Note that whatever options/settings you select on the dashboard, will have no effect on the alert detection. The alert detection settings are all selected in the configuration page of the Shiny app.

#### Filters

**Social Media**

You can select one item from the drop-down list of social media to display in the dashboard. Currently only Bluesky is available.

**Topics**

You can select one item from the drop-down list of topics, which is populated by what is specified in the topics on the configuration page. You can also start typing in the text field and select the topics from the filtered drop-down list.

**Countries & regions**

```{r echo=FALSE, fig.cap="", out.width = '40%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/topics_countries.png")
```

If you select **World (all)**, all posts are displayed regardless of their geolocation. You can select an individual country, you can select regions and subregions, and you can select several items at the same time. You can also start typing in the text field and select the geographical item from the drop-down list.

**Period**

```{r echo=FALSE, fig.cap="", out.width = '40%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/period.png")
```

You can select from the past 7 (the default), 30, 60 or 180 days. You can also select "custom" and a calendar option to select study period will appear. These periods will be the time period for inclusion in the visualisations. When selecting custom period, please ensure that the first date is at least one day before the second date.

**Time unit**

```{r echo=FALSE, fig.cap="", out.width = '30%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/time_unit.png")
```

You can display the timeline for the number of posts with weeks or days as units of time. The default is days.

**Include Reposts/quotes**

```{r echo=FALSE, fig.cap="", out.width = '30%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/include_quotes.png")
```

By default, reposts are not included in any of the visualisations. If “include reposts/quotes” is checked, the visualisations display results of posts and reposts/quotes. Otherwise, the visualisations display only posts (without reposts/quotes).

**Signal detection false positive rate**

```{r echo=FALSE, fig.cap="", out.width = '30%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/signal_detection_fpr.png")
```

Using the slider, you can explore the differences in the signals generated when changing the alpha parameter for the false positive rate. Note that this will not change the signal false positive rate for the alert emails. This is just a tool for you to explore this parameter. The default is 0.025. A higher false positive rate will increase the sensitivity and possibly the number of signals detected, and vice versa.

**Outlier false positive rate and outlier downweight strength**

```{r echo=FALSE, fig.cap="", out.width = '40%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/outlier_dashboard.png")
```

The outlier false positive rate relates to the false positive rate for determining what an outlier is when downweighting previous outliers/signals. The lower the value, the fewer previous outliers will potentially be included. A higher value will potentially include more previous outliers.

The outlier downweight strength determines how much an outlier will be downweighted by. The higher the value the greater the downweighting. For more information please see [Annex I](#annexI).

**Bonferroni correction**

```{r echo=FALSE, fig.cap="", out.width = '30%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/Bonferroni_correction.png")
```

The Bonferroni correction is selected by default. It accounts for false positive signal detection through multiple testing. For country-specific signal detection, the alpha is divided by the number of countries. For continent-specific signal detection, the alpha is divided by the number continents.

If you do not wish to use this correction, you can uncheck it.

**Days in baseline**

```{r echo=FALSE, fig.cap="", out.width = '30%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/days_in_baseline.png")
```

The default days in baseline is 7. You can explore the effect of having different days in the baseline. This is only for the visualisation, any changes made for the email alerts have to be made in the configuration page.

**Same weekday baseline**

```{r echo=FALSE, fig.cap="", out.width = '30%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/same_weekday_baseline.png")
```

It is possible that there is a "day of the week effect", where more posts may be posted on a given day of the week (e.g. Monday) than on other days. You can also select to calculate the baseline not on consecutive days, but on the past N days that correspond to the same 24 hour window N days back. This way if N = 7, the baseline is calculated using the “days” from -7, -14, -21, -28, -35, -42, -49 and -56 (if the current "day" is zero).


#### The timeline

The timeline graph is a time series, where you can see the number of posts for a given topic, geographical unit and study period. Signals are indicated as triangles on the graph, with the alpha and baseline days as specified in the filters. The area under the threshold is indicated in the shaded green colour. Note that the signals are related to the choice of alpha and days in baseline in the filters on the dashboard, rather than what is used for the alert emails. This way you can explore the effect of changing these parameters and adapt the settings for the alert emails if needed.

If you hover over the graph, you obtain extra information on country, date, number of posts and the number of posts from the list of known users, the ratio of known users to unknown users, whether the number of posts was associated with a signal and what the threshold and the alpha was.

```{r echo=FALSE, fig.cap="", out.width = '30%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/hover_trendline.png")
```


#### Map
The map shows a proportional symbol map of the posts by country and by topic for the study period. The larger the circle, the greater the number of posts.

The geographical information for the map is based on the choice in the filters: the country/region/subregion and the location type (post, user or both).

```{r echo=FALSE, fig.cap="", out.width = '60%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/hover_map.png")
```

When you hover over the map, you can get information on number of posts and names of the geographical units underlying the circles on the map.

When selecting one country, the symbols show the geographical distribution of posts at subnational level. When selecting two or more countries or other geographical entity (e.g. regions or continents), the symbols show the geographical distribution of posts at national level. Note that if a post has a geotag at country level (e.g. France), it will not be displayed when selecting only that country since no subnational geotagging is available.

#### Most frequent words, tags and URLs found in or extracted from posts

```{r echo=FALSE, fig.cap="", out.width = '50%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/top_words.png")
```

These graphs display the top elements of posts by topic for the study period for the geographical units chosen, and according to the filter on post/repost.

One figure displays the most frequent words. A second figure displays hashtags. A third figure displays the most frequent URLs as a table allowing users to click directly on the links to access these in the browser.

### The alerts page

The alerts page summarises the signals detected within the specified study period and provides functionality for adding annotations for training machine learning algorithms to classify alerts on user's defined topics. This page is splitted on two sections

#### Find alerts

Shows alerts generated by episomer. If you have provided a training set for alert classification, the category evaluated at the time of generation will be displated. If the alert is part of the training setthe vategory displayed would be the one specified by you.

The lisf of alerts can be filtered based on the following elements: date, topics, countries/regions. These three filters will define the scope of alerts to search. Also two modifiers are available for searching:

-   Display: Choose the columns to display on the search results

    -    Posts: Focused on showing relevant information of the alert. Including posts that are more similar to top words of the alert on the associated period and countries. The following columns are displayed for alerts: Date, Hour, Topic, Region, Category, Tops, Posts, Top posts.
    -    Parameters: Focused on showing the parameters of the alerts that were produced. The following columns are displayed for alerts: Date, Hour, Topic Region, Category, Tops, Posts, % from important user, Threshold Baseline, Bonf. corr., Same weekday baseline Day, rank,  With reposts Location,  Alert FPR (alpha), Outlier FPR (alpha). 

-   Limit: Maximum number of alerts to be displayed.

-   Hide Alerts, Search Alerts: Buttons to hide or display the alerts associated with the filters selected by you.

-   Add alerts to annotation: The alerts returned by the search are added to the annotations database so you can annotate and classify them.

#### Alerts annotations

`episomer` classify alerts using top words and top posts using annotations provided by end users through the alert annotation spreadsheet. You can define different algorithms and its parameters using the `runs` spreadsheet on the training database. `episomer` will find the best algorithm to perform the classification. `episomer` will randomly split the annotations database assigning 75% of the alerts for training and 25% of alerts for evaluation. All the algorithms you have provided and/or the three `episomer` predefined algorithms will be tested and the best in terms of F1 score will be selected to classify new alerts. From that moment, the category of new alerts will be predicted using the best algorithm. The algorithm is not used to predict the category retrospectively (historical alerts), only prospectively (new alerts).

You can show, hide, download and upload the alert annotation spreadsheet using the buttons available.

The alert annotation spreadsheet has two sheets that will be used as explained below:

-   Alerts sheet:

    -    Date: Date when the alert was identified
    -    Topic: Topics associated with the alert
    -    Region: Region where the alert was detected
    -    Top words: Top words identified for the alert
    -    Posts: Number of posts onserved on the last 24 hours on the alert
    -    Top Posts: The posts that contains the most of the topwords associated to the alerts on the alert period and geographical location
    -    Given Category: The category you have provided. *This is the only column that you need to update to classify alerts*
    -    Episomer Category: The category associated by episomer using the last version of the alert classification algorithms. Please note that this is only a reference value that include overfitting since the final algoritms is trained with all annotated alerts.

-   Runs sheet: A registry of algorithms and parameters to test including performance metrics on the last evaluation.

    -    Ranking: The ranking of the algorithm in terms of F1 score
    -    Models: The name of the model, referenced by the Apache Spark class. It has no inherit from the classifier class [*://spark.apache.org/docs/latest/api/scala/org/apache/spark/ml/classification/Classifier.html*]
    -    Alerts: The number of alerts used for the training.
    -    Runs: The number of runs to perform to evaluate the algorithm. Each run will make a different random split (if balance classes is set this will be limited to one).
    -    F1Score: The F1 score of all classes as provided by Apache Spark [*MulticlassClassificationEvaluator*](https://spark.apache.org/docs/latest/api/scala/org/apache/spark/ml/evaluation/MulticlassClassificationEvaluator.html)
    -    Accuracy: The Accuracy of all classes as provided by Apache Spark [*MulticlassClassificationEvaluator*](https://spark.apache.org/docs/latest/api/scala/org/apache/spark/ml/evaluation/MulticlassClassificationEvaluator.html)
    -    Precision by Class: The Precision by each category as provided by Apache Spark [*MulticlassClassificationEvaluator*](https://spark.apache.org/docs/latest/api/scala/org/apache/spark/ml/evaluation/MulticlassClassificationEvaluator.html)
    -    Sensitivity by Class: The Sensitivity by each category as provided by Apache Spark [*MulticlassClassificationEvaluator*](https://spark.apache.org/docs/latest/api/scala/org/apache/spark/ml/evaluation/MulticlassClassificationEvaluator.html)
    -    FScore by Class: The FScore by each category as provided by Apache Spark [*MulticlassClassificationEvaluator*](https://spark.apache.org/docs/latest/api/scala/org/apache/spark/ml/evaluation/MulticlassClassificationEvaluator.html)
    -    Last run: The last time and date where this run was evaluated
    -    Balance classses: If set to active (1) then episomer will apply augmentation to the alert training set by adding new sinthetic alerts using other top words with less ranking within the selected period. augmentation will be applied until the less representative class will reach the same amount on elements than the bigger category. If not possible, the categories with more elements will be sampled until all categories are nearly equal. 
    -    Force to use: If set to active (1) episomer will use the selected algorithm configuration to evaluate alerts independently of f1score ranking.
    -    Active: Whether the run is active (1) or not (0). Only active runs will be tryed to choose the best algorithm and parameters.
    -    Documentation: A link to the algorithm documentation
    -    Custom Parameters: A json object providing values for parameters for the given algorithm. The possibility to perform grid search strategies can be achieved by adding the same algoritm on several lines with different parameters.


```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/alerts.png")
```

### The geotag page
This page supports the improvement of the geolocation algorithm. Episomer will allow you to download the data that was used to build the classifiers for location identification. The predefined annotations are based on location vs non location words. Locations are extracted from geonames.org data base. Non location words are obtained from common words on the downloaded models that are not present on GoNames. You will be able to add posts to the annotation datanbase and to manually annotate them until you reach the expected level of performance.

To help you on the evaluation process `episomer` will calculate standard machine learning metrics for evaluating its capacity to identify the right location in texts. These metrics are separately calculated by different type of texts: Post location, Post text, Post user description and total.

The following controls are available:

-   Posts to add: Choose the number of posts to add to the geolocation spreadsheet
-   Geolocate annotations: Pressing this button will add the selected number of annotations to the `episomer` geolocation spreadsheet and evaluate the geolocation algorithm on the current annotations
-   Download annotations: Download the current version of annotations
-   Upload & evaluate annotations: Upload a new version of the geolocation annotations to improve annotations performance. The algorithms for location identification will be trained using the provided data. Also performance metrics will be calculated using 75% of the data for training and 25% for test. The final model is trained using all data.

The geolocation spreadsheet has the following columns:

-   Type: This column is a reference for you to understand the kind of annotation. It can be any of the following.
    -   Text: A text that can contains locations inside it can be a post or a user description.
    -   Location: A text containing a location.
    -   Person: A known person e.g. presidents that can be used by `episomer` to detect countries. This data is not used for training, just for over-writing the algorithm behaviour.
    -   Demonym: A known word to associate to a country e.g. Chilean. This data is not used for training, just for over-writing the algorithm behaviour.
-   Text: The text associated to this annotation.
-   Location in text: The subpart of the text that contains a location. If many locations are present, the same line can be duplicated with different values on this column. If this column is empty `episomer` assumes that the annotation is associated with all the text.
-   Location yes/no: whether the current text contains a location (demonym and people are not considered as location).
-   Associate country code: Whether `episomer` should memorize the annotated text and bypass the machine learning algorithm to always associate the given text to the country (or geonames city) indicated by the provided code.
-   Associate with: Whether `episomer` should memorize the annotated text and bypass the machine learning algorithm to always associate the given text to the country (or geonames city) indicated by the provided name.
-   Source: The source of the annotation, it can be any of: Episomer database (provided by `episomer` team), Episomer model (obtained automatically by `episomer`) or Post (obtained from a post).
-   Post Id: The post id if the text is associated to a post.
-   Lang: The language associated to the annotation.
-   Post part: The post part if associated to a post. It can be any of: "text", "user description" or "user location".
-   Episomer match: The location found by `episomer`
-   Episomer country match: The country of the location found by `episomer`.
-   Episomer country code match: The country code of the location found by `episomer`.


```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/geotag_evaluation.png")
```

### The data protection page
In the data protection page, you can search, anonymise and delete posts from the `episomer` database to support data deletion requests.

The following search filters are available: Topic, period, country & regions, mentioning (for selecting only mentions of the provided users), from users (for selecting only posts from the provided users) and both (for getting posts either mentioning or from the provided user)

The following controls are available:
-   Limit: Limit the perimeter of the search to the first "limit" posts
-   Search: Perform the search and show the results on the screen.
-   Anonymised search: Perform the search and show anonymised results on the screen. User mentions and authors are replaced by a mention USER.
-   Anonymise: Permanently replace all matching posts user mentions and authors  by USER.
-   Delete: Permanently delete all the matching posts.

### The configuration page

In the configuration page, you can change settings of the tool, you can check the status of the various processes/pipelines of the tool and you can add, delete and modify topics and their associated requests, languages for geolocation and the list of the “important users” and email alert subscribers. When changing anything in the “Signal detection” or “General” sections, do not forget to click on the “Update Properties” button at the end of the “General” section. The following sections describe the configuration page in more detail.

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/config.png")
```

**Status**

The status section enables you to quickly assess the latest time point and/or status of the processes for post collection (Post Search), and geolocation, aggregation and signal detection (Detection pipeline).

```{r echo=FALSE, fig.cap="", out.width = '50%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/status.png")
```

In the status section, you can tell if the embedded database, the search pipeline and the detect pipeline processes are running. You can click on “activate” or “stop” the episomer tasks. On Windows the tasks wim be permanently registered as scheduled tasks so they will run automatically or can be manually run from the Windows task scheduler.


**Requirements & alerts pipeline**

You need to run manually the tasks of dependencies, geonames and languages by clicking in the buttons “Run dependencies”, “Run geonames” and “Run languages” the first time you use `episomer`, and then only if you are downloading new versions. 

Geonames and languages relate to the geolocation and language models used by episomer. If you would like to update them (this is not something that needs to be done regularly, more on a yearly basis or so), then you can click on “Run”. When running the 'download dependencies' task, `episomer` will request you to stop the 'episomer database' in order to perform the update. 

The “Run alerts” button can be used to force the start of this task in case there is any error or issue. You can check their status in the “Requirements & alerts pipeline” table.

The 'Requirements & alerts pipeline' table gives more information about the status of the processes of `episomer`. This is useful for troubleshooting any issues arising and for monitoring the progress. It contains the five tasks that are running in the background. GeoNames and languages are tasks that will download and update the local copies of these. This will only be triggered if we add a language or update GeoNames. The start and end dates will generally be much older than those of geotag, aggregate and alerts.

Alerts dates should be more recent if the 'data collection & processing' and 'requirements & alerts' pipelines are active and running. These are scheduled according to the detect span. The status can include running, scheduled, pending, failed or aborted (if it has failed more than three times).

```{r echo=FALSE, fig.cap="", out.width = '70%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/detection_pipeline.png")
```


**Signal detection**

In the signal detection section in the configuration page, you can set the signal false positive rate alpha parameter, which increases (if larger) the the detection interval (more signals are detected), or decreases (if smaller) the detection interval (fewer signals are detected).

```{r echo=FALSE, fig.cap="", out.width = '50%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/signal_detection_config.png")
```

The outlier false positive rate relates to the false positive rate for determining what an outlier is when downweighting previous outliers/signals. The lower the value, the fewer previous outliers will potentially be included. A higher value will potentially include more previous outliers.

The outlier downweight strength determines how much an outlier will be downweighted by. The higher the value the greater the downweighting. For more information please see [Annex I](#annexI).

`episomer` calculates a threshold to determine if the current number of posts for a given 24-hour window exceeds what is expected (see section *"How does it work? General architecture behind episomer > Signal detection"*). This threshold is based on a default of the previous 7 days. In the “default days in baseline” field, you can change the number of days.

You can also change the default of using the previous 7 days for calculating a baseline to the previous 7 same days of the week, in order to avoid a "day of the week effect" (it may be that there are always more posts about this topic on a Monday, for example, which could affect the signal detection).

You can also specify if the signal detection is carried out just with post text, or includes reposts/quotes (check the box "Default with reposts/quotes").

The last checkbox "Default with Bonferroni correction", take multiple testing into account, which can result in false positives. If this box is checked then the signal detection alpha parameter is divided by the number of geographical locations in which signal detection is carried out. For example, at country level, the alpha parameter is divided by the total number of countries. At continent level, the alpha parameter is divided by the total number of continents.

When changing anything in the “Signal detection” section, do not forget to click on the “Update Properties” button at the end of the “General” section.


**General**

```{r echo=FALSE, fig.cap="", out.width = '40%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/general.png")
```

-   In **Data directory** you can view the directory that `episomer` uses to store the post and associated data collected. This is also the directory that the dashboard uses to obtain the datasets for displaying the visualisations. You need to set this folder when you launch `episomer` or set the environment variable 'EPI\_HOME'.

-   The **Search span** relates to how long a search plan is carried out. The default is 60 minutes. This value controls the size of the search window of posts. If you reduce this value you will get posts sooner, but you may 'waste' requests on topics with very few posts. If you increase its value you will take more time to get the posts but you will get more requests for popular posts increasing the chances of being exhaustive. You can see when you are not able to collect the posts on the Shiny configuration page if you have more than one active plan for some topics.

-   The **Detect span** relates to how frequently the processes of the detection pipeline (geotagging, aggregation and alert detection) are carried out. The default is 90 minutes. Email alerts are sent at the end of the detect loop. This value is treated as a lower bound, the detect loop could take more time to finish depending on the volume of tweeets and your system specifications. 

-   The **Launch slots** for the detection pipeline processes will be spaced out according to the "Detect span", with the first one starting at midnight. These values can be used in the subscribers file of the configuration page.

-   To avoid storing credentials in plain files, `episomer` uses a system dependent password store functionality, which is stored in the **Password store**. Depending on your system you can choose the mechanism that suits the environment where `episomer` is running. For details on each implementation see [*https://CRAN.R-project.org/package=keyring*]

     -   wincred: (Windows only) uses the windows credential manager.

     -   macos: (MAC only) uses the Mac OS keychain services

     -   file: Uses password protected encrypted files

     -   secret service: (Linux only) uses Linux secret service

     -   environment: Uses environment variables (extra setup needed, see [*https://CRAN.R-project.org/package=keyring*])

-   **Spark cores and spark memory:** The memory allocation for `episomer` in terms of CPU (Spark cores) and RAM (Spark Memory) is also defined in the “general” section. The default is 6 cores and 6 GB of RAM. This will depend on the CPU and RAM capacity of your machine and it has to be equal or less than that.

-   **Geolocation threshold:** During the geolocation process, sets of words are processed and potential matches to existing locations are determined and given a score. The higher the score, the greater the probability that the geolocation is correct. A threshold is set in `episomer`, under which any matches are not considered good enough for geolocation. The scale goes from 1 to 10, and the default is set at 5.

-   **Geonames URL:** The URL used to download the GeoNames database (used for generating locations) is in the general section. Should this URL ever change, you can make the amendment here.

-   **Simplified geonames:** As GeoNames is a very large file, a simplified version of it is used by default, including only existing geographical locations where the population is known. You can uncheck this option if you wish to use the whole GeoNames database.

-   **Stream database results:** Whether to store files on disk instead of performing streaming evaluation of aggregated data frames coming from the episomer database. If loading the dashboard takes too long, try selecting/unselecting this.

-   **Maven repository:** This is the URL of the maven repository that will be used to download the JAR dependencies for the detect loop, mainly Spark and Lucene.

-   **Winutils URL:** This is the URL that will be used to download winutils.exe. This is a Windows binary necessary for running Spark locally on Windows. If you do not want to use this version you can produce it yourself by downloading Hadoop 2.8.4 or higher and compiling it on a Windows machine.

-   **Region disclaimer:** If you would like to add a disclaimer to the map you are using. This disclaimer is added to the image export of the dashboard map, and also the PDF export of the dashboard.

```{r echo=FALSE, fig.cap="", out.width = '30%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/region_disclaimer.png")
```

**Email authentication (SMTP)**

In this section, you need to specify the email authentication (SMTP) details for the email that will send the alerts.

If **Unsafe certificates** is checked, then `episomer` will use your SMTP server even if the server sends an invalid certificate.

When changing anything in the “general” section, do not forget to click on the “Update Properties” button.

**Social Medias**
This section list the supported social media and allows you to set credentials and use for each social media provider. Actually only Bluesky is supported.

You can set:

-   Which social media are active. This parameter can limit which social media are used for data collection and therefore available in the dashboad.
-   Which social media are considered for alerts. This parameter social media being considered for alert detection. Please note that numer of messaged are sum between all social media before calculating alerts.
-   API credentials for data collection. In the case of Bluesky you have to input a valid username and password.

**Topics**

Topics are what determines what posts `episomer` collects. This is done via an Excel spreadsheet that contains the topics and the associated requests that `episomer` uses to query the API with.

A query consists of keywords and operators that are used to match on post attributes. See section *"How does it work? General architecture behind episomer > Collection of posts > Topics of posts to collect and queries"* for more details about queries.

`episomer` comes with a default list of topics as used by the ECDC Epidemic Intelligence team at the date of package generation (*1^st^ of September, 2020*). You can download this list of topics and upload your own in the “Available Topics” section in the configuration page. See section *"How does it work? General architecture behind episomer > Collection of posts > Topics of posts to collect and queries"* for more details on how to structure the topics list.

In the topics section on the configuration page, you can view the topic, the associated query, the query length and how many active search plans are associated with the query. If more than one search plan is active, this means that `episomer` did not manage to collect all possible posts in the last session. Additionally, you can see the progress and the number of requests from the last search plan.

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/topics.png")
```

**Languages**

In the languages section, you can determine which language models are used to identify text during the geolocation process. The default languages are French, English, Portuguese and Spanish. You can download and upload the language models in the “Available Languages” section and add and delete languages used by `episomer` in the “Active Languages” section. Please consider the computational cost of adding too many languages, depending on the capacity of your machine.

**All default languages (English, French, Spanish and Portuguese) must be downloaded before adding new languages or deleting any of the default languages.**

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/languages.png")
```

### The troubleshoot page
The troubleshoot page has two functionalities "Create a snapshot file" and "Diagnostics"

## Create a snapshot file
Back-ups your `episomer` installation with all its settings and data on a single file. You can choose which kind of data you want to include on your snapshot between: Settings, dependencies, machine learning, aggregation (time series & alerts) posts & logs. This feature has been designed with two purposes:
- Troubleshoot: If you have an issue running `episomer` you can easily create a snapshots to share with your support team.
- Compliance: Sometimes you may need to delete old data from `episomer`. With this feature you can choose the create a snapshot file for last N months of posts and last M months of aggregated data. You can then backup your existing data folder and restore your snapshot file instead. Everything should work as before but only the period you selected should be available. After testing that everything is as you expect you can delete the old data folder.

## Diagnostic
Provides a list of automatic checks and hints for using `episomer` with all its functionalities. Click on "Run diagnostics" to see the list of checks, whether it passed the check ("true") or not ("false"), and hints in case it did not pass the check. More detailed information can be found in [Annex II](#annexII) of this user documentation.

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/troubleshoot.png")
```

### Downloading outputs from the interactive user interface (Shiny app)

Each visualisation on the Shiny app dashboard can be downloaded as an image, using the "image button". A png is a portable network graphic file and is a versatile file format for images that do not need to be of a very high resolution (e.g. professional print graphics).

Note that the png format is not supported in the Internet Explorer browser (but you can download a svg file instead).

You can also download the data of each visualisation by clicking on the data button. This will give you a csv file containing the underlying data that you can use for further analysis or to create your own graphs.

Alternatively, you can use the PDF or the Md button at the bottom of the filters to download a PDF or an HTML file of the dashboard Note that for this you will need to have MiKTeX or TinyTeX installed.

## Annex I: Downweighting the previous signals {#annexI}

## Introduction

In this annex we propose a downweighting approach built as part of the `eears` algorithm used in the `episomer` package and which was described above.

Let the $\mathbf{y}$ denote the vector of historic values which is of length $n$. Part of the computation of the prediction interval at time 0 is the computation of the mean and standard deviation of these historic values, i.e.
$$
\overline{y}_0 = \frac{1}{n}\sum_{t=-n}^{-1} y_{t} \quad\text{and}\quad s_0^2 = \frac{1}{n-1} \sum_{t=-n}^{-1} (y_{t} - \overline{y}_0  )^2
$$
The upper limit of the one-sided $(1-\alpha)\times 100\%$ prediction interval for the observation $y_0$ under an $y_t \stackrel{\text{iid}}{\sim} N(\mu, \sigma^2), t=-n, \ldots, 0$ model is then computed as 
$$
U_0 = \overline{y}_0 + t_{1-\alpha}(n-1) \times s_0 \times \sqrt{1+\frac{1}{n}},
$$
where $t_{1−\alpha}(n − 1)$ denotes the $1 − \alpha$ quantile of the t-distribution with $n − 1$ degrees of freedom. This computation of the threshold corresponds to a statistical sound computation of the threshold [@allevius_prospective_2017].

A desired extension of the above algorithm is the handling of previous signals in the historic values. This problem was already addressed in the quasi-Poisson framework of @farrington_statistical_1996 by first performing a GLM fit and then re-fit the GLM with weights  based on the Anscombe residuals. We follow the same general idea, but adapt it to the Gaussian response used in the EARS algorithm and corresponding residuals from the linear model.

### EARS as a Linear Model

We first observe that the above estimation of $\mu$ and $\sigma^2$ through $\overline{y}_0$ and $s_0^2$ at time 0 can be embedded within a linear regression model, i.e. for $i=1, \ldots, n$ we model
$$
y_i = \mu + \epsilon_i, \quad\text{where}\quad \epsilon_t \stackrel{\text{iid}}{\sim} N(0, \sigma^2).
$$
Note that we, for compatibility with the standard exposition in linear model theory, have indexed the $y$ values s.t. $y_{-n}$ corresponds to $y_1$ and $y_{-1}$ corresponds to $y_n$. In matrix terms let $\mathbf{y} = (y_{1},\ldots,y_n)'$ and for the intercept-only model the design matrix is $\mathbf{X} = (1,\ldots,1)'$, which has rank $k=1$. Thus from standard OLS theory:
$$
\begin{align*}
\hat{\mu} &=  (\mathbf{X}'\mathbf{X})^{-1} \mathbf{X}' \mathbf{y} = \frac{1}{n}\sum_{i=1}^n y_i,
\end{align*}
$$
which corresponds to $\overline{y}_0$. Furthermore, let the raw residuals be defined as $e_i = y_i - \hat{\mu}$ for $i=1,\ldots, n$ and denote by $\mathbf{e}=(e_1,\ldots,e_n)'$ the corresponding vector of residuals. Then
$$
\mathbf{e} = \mathbf{y} - \hat{\mathbf{y}} = \mathbf{y} - \mathbf{P} \mathbf{y} = (\mathbf{I}-\mathbf{P}) \mathbf{y}
$$
where $\mathbf{P} = \mathbf{X} (\mathbf{X}'\mathbf{X})^{-1}\mathbf{X}'$ is the so called hat-matrix known from linear modelling. With this notation we can write up the estimate for $\sigma^2$ as in @chatterjee_sensitivity_1988:

$$
\hat{\sigma}^2 = \frac{\mathbf{e}' \mathbf{e}}{n-k} = \frac{\mathbf{y}'(\mathbf{I}-\mathbf{P})\mathbf{y}}{n-k} = \frac{1}{n-1} \sum_{t=-7}^1 (y_t - \hat{\mu})^2,
$$
which corresponds to the above used expression for $s_0^2$.

### Downweighting

We now compute the so called **externally Studentized residuals** [@chatterjee_sensitivity_1988]
$$
r_i^* = \frac{e_i}{\hat{\sigma}_{(i)}\sqrt{1-p_{ii}}}, \quad i=1, \ldots, n,
$$
where $p_{ii}$ is i'th diagonal element of the hat-matrix $\mathbf{P}$ from the corresponding linear model used above. Furthermore,
$$
\hat{\sigma}_{(i)}^2 = \frac{\mathbf{y}_{(i)}' (\mathbf{I}-\mathbf{P}_{(i)}) \mathbf{y}_{(i)}}{n-k-1}
$$
is the variance estimate obtained from a linear regression, where the i'th observation is removed. Linear modelling theory [@chatterjee_sensitivity_1988] now states that 
$$
r_i^* \stackrel{\text{identical}}{\sim} t(n-k-1).
$$
Note that the residuals are only identically distributed, because they are not independent (see Section 4.2.1. of @chatterjee_sensitivity_1988 for details). However, the above distributional form allows us to assess for each historic value, if it can be considered as an outlier. 
For this purpose define $r_{\text{threshold}}$ as the $1-\alpha_{\text{outlier}}$ quantile of the t-distribution with $n-k-1$ degrees of freedom.
A historic value is an outlier (for which one possible explanation is that it originates from a true increase in posts, e.g. an outbreak situation), if
$r_i^* > r_{\text{threshold}}$. We shall use this to formulate a weighting schemes for the historic values:

Downweight-Outliers:
$$
\begin{align}
w^{(\text{dw})}_i &= \left\{
\begin{array}{ll}
1 & \text{if } r_i^* < r_{\text{threshold}}\\
\left(\frac{r_{\text{threshold}}}{(r_i^*)}\right)^k & \text{otherwise}
\end{array}
\right. \\
&= \min\left\{1,\left(\frac{r_{\text{threshold}}}{r_i^*}\right)^k\right\},
\end{align}
$$
where the decay parameter $k>0$ is a known quantity. In the original @farrington_statistical_1996 algorithm, $k=2$ was used. Furthermore, a threshold value of 1 was used. In the later @noufaily_improved_2013 paper, however, a threshold value of 2.58 was recommended. Note: both values are for the standardized Anscombe residuals, which follow a standard normal distribution. If we take corresponding quantiles for the t-distribution with 6 degrees of freedom the values would be `r sprintf("%.2f",qt(pnorm(1), df=6))` and `r sprintf("%.2f",qt(pnorm(2.58), df=6))`.
Note also that the term $(r_{\text{threshold}}/r_i^*)^k$ is a slight adaptation of @farrington_statistical_1996, which instead uses $1/(r_i^*)^2$. The advantage of our proposal is that it ensures a smooth handling of values around the threshold if the threshold is not 1. It might be worth considering a higher power than 2 to ensure an even larger down-weighting for gross outliers. The current default value for the decay parameter in `episomer` is 4.

Finally, as in @farrington_statistical_1996, we normalise the weights such that they yield a sum of $n$ by
$$
w_i^* = n \times \frac{w_i}{\sum_{i=1}^n w_i}
$$
and then re-fit the linear model with these weights. For this purpose define the weight matrix as $\mathbf{W} = \operatorname{diag}(w_1^*,\ldots,w_n^*)$. We can use a subsequent weighted least squares approach to find
$$
\begin{align*}
\hat{\mu}_W &=  (\mathbf{X}' \mathbf{W} \mathbf{X})^{-1} \mathbf{X}' \mathbf{W} \mathbf{y} = \frac{1}{n}\sum_{i=1}^n w_i^* y_i,
\end{align*}
$$
where the 2nd equal sign is because $(\mathbf{X}' \mathbf{W} \mathbf{X})=\sum_{i=1}^n w_i=n$
and $\mathbf{X}' \mathbf{W} \mathbf{y} = \sum_{i=1}^n w_i^* y_i$.
Furthermore,
$$
s_W^2 = 
\frac{\mathbf{y}'(\mathbf{I}-\mathbf{P}_W)\mathbf{y}}{n-k} =
\frac{\sum_{i=1}^n w_i^*(y_i - \mu_W)^2}{n-1},
$$
where $P_W = \mathbf{X} (\mathbf{X}'\mathbf{W} \mathbf{X})^{-1}\mathbf{X}\mathbf{W}$ is the hat-matrix of the weighted least squares.

The downweighted procedure thus operates with $\mu_{W}$ and $s_W^2$ instead of $\overline{y}_0$ and $s_0^2$, respectively, when computing the upper limit $U_0$ using the above mentioned formula.

### Example of the downweighting approach using Ebola data

Figure 5 below shows the upper limit of the signal detection threshold for `episomer` Ebola data, both the original (in red) with no downweighting and the downweighted upper threshold (in blue) after taking previous signals in the historic values into account. Note that the downweighted upper threshold detects three additional signals, compared to the original threshold. 

<center>Fig 5: Upper limit with and without downweighting for the episomer Ebola data</center>

```{r echo=FALSE, fig.cap="", out.width = '100%'}
knitr::include_graphics("https://raw.githubusercontent.com/EU-ECDC/episomer/main/img/ebola.png")
```

## Annex II: Troubleshooting and tips {#annexII}

This annex contains some tips and common solutions to errors or issues that `episomer` users may encounter, including an explanation on the checks included in the troubleshoot page.

In addition, you can also visit the [general post](https://github.com/EU-ECDC/episomer/discussions/7) in the discussion forum of the GitHub episomer repository for additional materials and training.

### The troubleshoot page
After running the diagnostics in the troubleshoot page, you can see the checks and status on the following aspects:

-   **scheduler**: R package `taskscheduleR` is installed. Only applicable for Windows machines
-   **sm_auth**: Social media token has been created after authenticating either with the provided credentials
-   **search_running**: search task is running
-   **database_running**: embedded database is running
-   **posts**: posts have been collected
-   **os64**: R is 64bits
-   **java**: Java is installed and accessible to episomer
-   **java64**: Java 64bits is installed and accessible to episomer
-   **java_version**: the Java version installed is compatible with `episomer`
-   **winmsvc**: Microsoft Visual C++ 2010 SP1 Redistributable Package is installed. Only applicable for Windows machines
-   **detect_activation**: detect loop has been activated
-   **detection_running**: detection task is running
-   **winutils**: winutils is installed. If false, it can be downloaded by running the update dependencies task. Only applicable for Windows machines
-   **java_deps**: java dependencies are installed 
-   **move_from_temp**: `episomer` can atomically move files from temporary folder to data directory
-   **tar_gz**: checks if tar.gz can be built, which is necessary for creating a compressed snapshot.
-   **geonames**: Geonames.org database is downloaded and indexed
-   **languages**: languages vectors are downloaded and indexed
-   **aggregate**: aggregate task has successfully run
-   **alerts**: alerts have been created
-   **pandoc**: Pandoc is installed and accessible to `episomer`. Necessary for PDF creation.
-   **tex**: a tex distribution is installed and accessible to `episomer`. Necessary for PDF creation.

### Management of episomer pipelines ('episomer database', 'Data collection & processing', 'Requirements & alerts') in Windows 
After activating the three pipelines ('episomer database', 'Data collection & processing', 'Requirements & alerts') from the configuration page of `episomer` (Windows), three tasks will be created in the task scheduler and three terminal windows will be prompted. Please note that if the computer is logged/turned off or the terminal windows are closed, the pipelines will stop.

If you activate these tasks from the configuration page of `episomer` again, the system will overwrite the tasks created in the task scheduler. Instead, after the first successful activation of these tasks from `episomer`, you can easily manage these from the task scheduler. You can stop these tasks by ending and disabling the tasks in the task scheduler, and you can restart these tasks by enabling and running these in the task scheduler.

You can also force stopping these tasks in the configuration page of `episomer` by clicking on 'stop'.

In the task scheduler, you can establish that the tasks “run whether the user is logged on or not” to avoid that these tasks stop when you log off or restart the computer. In this case, you may not see the prompted terminal windows when the tasks are running. 

### Management of episomer pipelines ('episomer database', 'Data collection & processing', 'Requirements & alerts') in Linux and Mac
Since the three pipelines ('episomer database', 'Data collection & processing', 'Requirements & alerts') in Linux or Mac have to be run manually, if the computer is logged/turned off or the terminal windows are closed, the pipelines will stop. Please remember to follow the steps in the section *Setting up post collection and the alert detection loop* to run these tasks again.

The pipelines can be also stopped in the configuration page of `episomer` by clicking on 'stop'.

### Running 'episomer database', 'Data collection and processing' and 'Requirements and alerts' pipelines

#### "Cannot execute task \#####: the task is already running"
Each pipeline creates a file containing their process IDs located in the `episomer` data folder: fs.PID, search.PID and detect.PID. This error arises if `episomer` finds another R process currently running with the same ID. In order to fix this error, you should first verify if the pipelines are already running in another R session. If this is the case, you should not try to start the pipeline since `episomer` only supports one instance of the same pipeline running in the same machine. If the running process is not associated with the task, then you can manually delete the PID file and try to start it again.

#### "Failed while processing alerts"
The error "failed while processing alerts Error in do_next_alerts(tasks): Cannot determine the last aggregated period for alert detection. Please check that series have been aggregated" appears when there are no aggregated series to calculate the alerts. This can happen in the following cases:
* No posts have been collected in the past days so there were no posts available to produce the aggregated series
* Geonames and/or languages are not downloaded so the geolocation cannot be extracted from the posts and, consequently, posts cannot be aggregated.

If it is the first installation, you should wait until geonames and/or languages tasks are completed and collected posts are geotagged and aggregated. Depending on the machine, these steps may take some hours.

You should also checked that posts are being collected.

### Change the user of the Bluesky authentication
1. End and disable the 'Data collection & processing' and 'episomer database' pipelines in the task scheduler (Windows), close the R/terminal window with the pipelines or forcing the pipelines in the configuration page (Windows, Linux and Mac)

2. Search for a file called “.rpost_token” in hidden files. It is usually saved in the Documents folder.

3. Delete that file.

4. Click on “Update properties” in the configuration page of `episomer`.

5. Enable and run the pipelines in the task scheduler or active them in the configuration page (Windows), or run the command in a new R/terminal window with the pipelines (Windows, Linux and Mac). More details are available in the section *“Setting up post collection and the alert detection loop”*

### Downloading GeoNames and/or languages

#### Languages to be added or deleted 
At least one language must be downloaded before adding new languages or deleting any of the default languages.

#### "The specified size exceeds the maximum representable size. Error: Could not create the Java Virtual Machine"

If this error appears when running GeoNames, it means that the machine has Java 32bits. You need to install Java 64bits. And make it accessible to epitwitter either by setting "JAVA\_HOME" environment variable or by setting the right java binary on the system PATH. 

### The "Launch slots" in the configuration page show NAs instead of the time slots

If it is the first time that you install and launch `episomer`, the geotag task of the detection pipeline has to be run at least once in order to see time slots in the “Launch slots” in the configuration page.

### Downloading PDF of the dashboard

#### "Error in: LaTeX failed to compile C:&#92;Users&#92;name~1&#92;...&#92;file######.tex."

This error appears in Windows when clicking on “PDF” in the dashboard and no PDF is saved. The reason is that the path to TEMP and TMP environment variables of the user are too long, Windows shortens the path and `episomer` cannot find this new path. Please follow the next steps to fix this:

1. Open the “environment variable for your account”

2. Change the path for TEMP and TMP to a shorter path (e.g. “C:&#92;Temp”). The same path should be used for both environment variables.

3. Log off and log on

4. You can now download and save the PDF from the dashboard

#### "Error: pandoc document conversion failed with error 6"

1. Downloading this script (https://raw.githubusercontent.com/jgm/pandoc/master/macos/uninstall-pandoc.pl)

2. Uninstall pandoc (https://pandoc.org/installing.html) by running perl uninstall-pandoc.pl

### Different totals in dashboard outputs
When counting the total posts in the dashboard of the Shiny app or in the downloadable data, you might get differences in the total numbers of posts between the three outputs. This might be due to the following reasons:

1. World (all) versus World (geolocated)

    - The default option for the regions in World (all), this means that also non-geolocated posts are included in the trendline, but only geolocated posts can be visualized in the maps and the most frequent words figure, therefore the overall total of posts can differ between these outputs wen selecting World (all) or the empty default.

2. Country specific analysis

    - If you select only one country in the filters, the trendline will show all posts for this country, but the map will show the posts on a subnational level in the map. It could be that some posts might have been geolocated to a certain country, but without further subnational data. These posts will then be visible in the trendline total, but not in the subnational bubbles in the map.

3. Most frequent words/hashtags

    - In contrast to the other outputs in the dashboard, the most frequent words figure is always based on post location regardless of the filter (due to memory capacity). Therefore, if user location or both locations are selected in the location filter, this figure might have a different total then the other two outputs.
    
### Receiving only real-time alerts
This relates to users who have selected topics and/or regions for receiving related alerts in real-time or have selected topics and/or regions for receiving related alerts on a scheduled time span. If, in these cases, you only receive real-time alerts with all topics and regions, it may be that no time slots have been included in the subscribers file from the configuration page. These time slots are used for the scheduled alerts and if no slots are included in the file, alerts from all topics and regions are sent as real-time alerts.

### Not receiving email alerts
If you do not receive email alerts and you see an error in `episomer` referring to denied login, it means that `episomer` could not login to the email account provided in the configuration page. Some of the reasons for that are:

-   The server or port included in the configuration page are incorrect
-   The attempt of `episomer` to login to the email account is being blocked by the server. This can happen with some organisational email accounts. In that case, please contact the IT department of your organisation
-   If using Gmail account, you need to allow less secure apps in the settings of your account

## References

<div id="refs"></div>
  [*https://www.java.com/download/*]: https://www.java.com/download/
  [*https://www.microsoft.com/en-us/download/details.aspx?id=26999*]: https://www.microsoft.com/en-us/download/details.aspx?id=26999
  [*https://pandoc.org/installing.html*]: https://pandoc.org/installing.html
  [*https://yihui.org/tinytex/*]: https://yihui.org/tinytex/
  [*https://miktex.org/download*]: https://miktex.org/download
  [*https://www.openblas.net/*]: https://www.openblas.net/
  [*https://github.com/fommil/netlib-Java*]: https://github.com/fommil/netlib-Java
  [*https://www.intel.com/content/www/us/en/developer/tools/oneapi/onemkl.html*]: https://www.intel.com/content/www/us/en/developer/tools/oneapi/onemkl.html
  [*https://git-scm.com/downloads/*]: https://git-scm.com/downloads/
  [*https://www.scala-sbt.org/download/*]: https://www.scala-sbt.org/download/
  [*https://cran.r-project.org/bin/windows/Rtools/*]: https://cran.r-project.org/bin/windows/Rtools/
  [*https://github.com/steveloughran/winutils/raw/master/hadoop-3.0.0/bin/winutils.exe*]: https://github.com/steveloughran/winutils/raw/master/hadoop-3.0.0/bin/winutils.exe
  [*https://developer.twitter.com/en/apply-for-access*]: https://developer.twitter.com/en/apply-for-access
  [*https://CRAN.R-project.org/package=keyring*]: https://CRAN.R-project.org/package=keyring
  [*Monitoring Count Time Series in R: Aberration Detection in Public Health Surveillance*]: https://www.jstatsoft.org/article/view/v070i10

---
title: "episomer Dashboard"
date: "`r format(Sys.time(), '%d %B, %Y')`"
output: pdf_document
classoption: landscape
params:
  topics: ["COVID-19"]
  countries: ["1", "2"]
  period_type: "created_date"
  period: ["2020-01-01", "2020-03-31"]
  with_quotes: FALSE
  location_type: "post"
  alert_alpha: 0.025
  alert_alpha_outlier: 0.05
  alert_k_decay: 4
  alert_historic: 7
  bonferroni_correction: FALSE
  same_weekday_baseline: FALSE
  top_type1: "hastags"
  top_type2: "urls"
---

```{r include=FALSE}
regions <- get_country_items()
countries <- Reduce(function(l1, l2) {
  unique(c(l1, l2))
}, lapply(as.integer(params$countries), function(i) unlist(regions[[i]]$codes)))
region_names <- lapply(as.integer(params$countries), function(i) regions[[i]]$name)
```

# episomer dashboard
Parameters

* Topics: `r params$topics`
* Regions: `r region_names`
* Period type: `r params$period_type`
* Period: `r params$period`
* With quotes: `r params$with_quotes`
* Location Type: `r params$location_type`
* Signal detection confidence (alpha): `r as.character(params$alert_alpha)`
* Outliers detection confidence (alpha-outlier): `r as.character(params$alert_alpha_outlier)`
* Outliers downweighting strength (k-decay): `r as.character(params$alert_k_decay)`
* Days in baseline: `r params$alert_historic`
* Boferroni correction: `r params$bonferroni_correction`
* Same weekday baseline: `r params$same_weekday_baseline`

## Line chart
```{r, echo = FALSE, message=FALSE, warning=FALSE}
line_chart_from_filters(
  params$topics,
  params$countries,
  params$period_type,
  params$period,
  params$with_quotes,
  params$location_type,
  params$alert_alpha,
  params$alert_alpha_outlier,
  params$alert_k_decay,
  params$alert_historic,
  params$bonferroni_correction,
  params$same_weekday_baseline
)$chart
```

## Map
```{r, echo = FALSE, message=FALSE, warning=FALSE}
map_chart_from_filters(
  params$topics,
  params$countries,
  params$period,
  params$with_quotes,
  params$location_type
)$chart
```

## Top `r params$top_type1 ` chart
```{r, echo = FALSE, message=FALSE, warning=FALSE}
top_chart_from_filters(
  params$topics,
  params$top_type1,
  params$countries,
  params$period,
  params$with_quotes,
  params$location_type,
  20
)$chart
```

## Top `r params$top_type2 ` chart
```{r, echo = FALSE, message=FALSE, warning=FALSE}
top_chart_from_filters(
  params$topics,
  params$top_type2,
  params$countries,
  params$period,
  params$with_quotes,
  params$location_type,
  20
)$chart
```

## Top URLS 

### `r paste("Top URLS of posts mentioning", params$topic, "from", params$periond[[1]], "to", params$period[[2]])`

```{r, echo = FALSE, message=FALSE, warning=FALSE}
df <- top_chart_from_filters(
  params$topics,
  "urls",
  params$countries,
  params$period,
  params$with_quotes,
  params$location_type,
  50
)$data
if (is.data.frame(df)) {
  knitr::kable(df)
} else {
  "No data found for the selected filters"
}
```
Top urls table only considers post location, ignoring the location type parameter



`r ""`

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{bluesky_rerun_after_rate_limit}
\alias{bluesky_rerun_after_rate_limit}
\title{function to extract the number of seconds to wait after a rate limit has been reached}
\usage{
bluesky_rerun_after_rate_limit(resp)
}
\arguments{
\item{resp, }{the httr2 response object}
}
\description{
function to extract the number of seconds to wait after a rate limit has been reached
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{sm_api_set_auth_bluesky}
\alias{sm_api_set_auth_bluesky}
\title{Save bluesky credentials (login and password) and store them securely}
\usage{
sm_api_set_auth_bluesky(shiny_input_list)
}
\arguments{
\item{shiny_input_list}{List of shiny input values containing the bluesky specific credentials}
}
\value{
Nothing. Called for side effects (saving credentials)
}
\description{
Update configuration object and underlying storage with given bluesky username and password. The password is encrypted.
}
\details{
Update authentication tokens in configuration object
}
\examples{
\dontrun{
 #Setting the configuration values
   sm_api_set_auth_bluesky(
     shiny_input_list = list(bluesky_user = "your user here", bluesky_password = "your password"),
   )
}
}
\seealso{
\code{\link{save_config}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alert.R
\name{get_alerts}
\alias{get_alerts}
\title{Getting signals produced by the task \code{\link{generate_alerts}} of \code{\link{detect_loop}}}
\usage{
get_alerts(
  topic = character(),
  countries = numeric(),
  from = "1900-01-01",
  until = "2100-01-01",
  topposts = 0,
  limit = 0,
  duplicates = "all",
  progress = function(a, b) {
 }
)
}
\arguments{
\item{topic}{Character vector. When it is not empty it will limit the returned signals to the provided topics, default: character()}

\item{countries}{Character vector containing the names of countries or regions or a numeric vector containing the indexes of countries
as displayed at the Shiny App to filter the signals to return., default: numeric()}

\item{from}{Date defining the beginning of the period of signals to return, default: '1900-01-01'}

\item{until}{Date defining the end of the period of signals to return, default: '2100-01-01'}

\item{topposts}{Integer number of top posts to be added to the alert. These are obtained from the post index based on topwords and Lucene score, default: 0}

\item{limit}{Maximum number of alerts returned, default: 0}

\item{duplicates}{Character, action to decide what to do with alerts generated on the same day. Options are "all" (keep all alerts), "first" get only first alert and "last" for getting only the last alert}

\item{progress}{Function, function to report progress it should receive two parameter a progress between 0 and 1 and a message, default: empty function}
}
\value{
a data frame containing the calculated alerts for the period. If no alerts are found then NULL is returned
}
\description{
Returns a data frame of signals produced by the \code{\link{detect_loop}}, which are stored on the signal folder.
}
\details{
For more details see the package vignette.
}
\examples{
\dontrun{
   library(episomer)
   # setting up the data folder
   message('Please choose the episomer data directory')
   setup_config(file.choose())

   # Getting signals produced for last 30 days for a particular country
   get_alerts(
     countries = c("Chile", "Australia", "France"),
     from = as.Date(Sys.time())-30,
     until = as.Date(Sys.time())
   )
}
}
\seealso{
\code{\link{generate_alerts}}

\code{\link{detect_loop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataviz.R
\name{trend_line}
\alias{trend_line}
\title{Plot the trendline report of episomer dashboard}
\usage{
trend_line(
  sms,
  topic,
  countries = c(1),
  date_type = "created_date",
  date_min = "1900-01-01",
  date_max = "2100-01-01",
  with_quotes = FALSE,
  alpha = 0.025,
  alpha_outlier = 0.05,
  k_decay = 4,
  no_historic = 7,
  bonferroni_correction = FALSE,
  same_weekday_baseline = FALSE
)
}
\arguments{
\item{sms}{Social media}

\item{topic}{Character(1) containing the topic to use for the report}

\item{countries}{Character vector containing the name of the countries and regions to plot or their respective indexes on the Shiny app select, default: c(1)}

\item{date_type}{Character vector specifying the time granularity of the report either 'created_weeknum' or 'created_date', default: 'created_date'}

\item{date_min}{Date indicating start of the reporting period, default: "1900-01-01"}

\item{date_max}{Date indicating end of the reporting period, default: "2100-01-01"}

\item{with_quotes}{Logical value indicating whether to include reposts in the time series, default: FALSE}

\item{alpha}{Numeric(1) value indicating the alert detection confidence, default: 0.025}

\item{alpha_outlier}{Numeric(1) value indicating the outliers detection confidence for downweighting, default: 0.05}

\item{k_decay}{Strength of outliers downweighting, default: 4}

\item{no_historic}{Number of observations to build the baseline for signal detection, default: 7}

\item{bonferroni_correction}{Logical value indicating whether to apply the Bonferroni correction for signal detection, default: FALSE}

\item{same_weekday_baseline}{Logical value indicating whether to use same day of weeks for building the baseline or consecutive days, default: FALSE}
}
\value{
A named list containing two elements: 'chart' with the ggplot2 figure and 'data' containing the data frame that was used to build the chart.
}
\description{
Generates a trendline chart of number of posts by region for a single topic, including alerts detected using the reweighted version of the EARS algorithm
}
\details{
Produces a multi-region line chart for a particular topic of number of posts collected based on the provided parameters.
Alerts will be calculated using a modified version of the EARS algorithm that applies a Farrington inspired downweighting of previous outliers.

Days in this function are considered as contiguous blocks of 24 hours, starting from the previous hour preceding the last collected post.

This function requires that \code{\link{search_loop}} and \code{\link{detect_loop}} have already run successfully in order to show results.
}
\examples{
\dontrun{
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   #Getting trendline for dengue for South America for the last 30 days
   trend_line(
     topic = "dengue",
     countries = "South America",
     date_min = as.Date(Sys.time())-30,
     date_max=as.Date(Sys.time())
   )
}
}
\seealso{
\code{\link{create_map}}
 \code{\link{create_topwords}}
 \code{\link{generate_alerts}}
 \code{\link{detect_loop}}
 \code{\link{search_loop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{sm_api_translate_query_bluesky}
\alias{sm_api_translate_query_bluesky}
\title{Translate a parsed query to the social media query format}
\usage{
sm_api_translate_query_bluesky(parsed)
}
\arguments{
\item{parsed}{list. Named list containing two attributes 'ors' and 'neg'. 
'ors' is a list containing subqueries that should be contatenated with an OR logic 
  - each subquery in 'ors' is a list containing terms that should be concatecated with an AND logic
      - each term in 'ors subqueries' is a list of synonyms for a particular literal
'neg' is a list of terms that should be removed from any result of the query}
}
\description{
Translate a parsed query to the social media query format
}
\details{
This function receive as a parameter a parsed query from the a topic in the list of topics. 
It can return one or many queries that matches the expected results using the underlying social media API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-plan.R
\name{sm_plan_get_progress_bluesky}
\alias{sm_plan_get_progress_bluesky}
\title{calculates and returns the estimated progress of a plan using custom attributes}
\usage{
sm_plan_get_progress_bluesky(plan)
}
\arguments{
\item{plan}{list, the plan to calculate the progress from}
}
\description{
calculates and returns the estimated progress of a plan using custom attributes
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alert.R
\name{ears_t_reweighted}
\alias{ears_t_reweighted}
\title{algorithm for outbreak detection, extends the EARS algorithm}
\usage{
ears_t_reweighted(
  ts,
  alpha = 0.025,
  alpha_outlier = 0.05,
  k_decay = 4,
  no_historic = 7L,
  same_weekday_baseline = FALSE
)
}
\arguments{
\item{ts}{A numeric vector containing the counts of the univariate
time series to monitor. The last time point in ts is
investigated}

\item{alpha}{The alpha is used to compute the upper limit of the prediction interval:
(1-alpha) * 100\%, default: 0.025}

\item{alpha_outlier}{Residuals beyond 1-alpha_outlier quantile of the
the t(n-k-1) distribution are downweighted, default: 0.05}

\item{k_decay}{Power k in the expression (r_star/r_threshold)^k determining the weight, default: 4}

\item{no_historic}{Number of previous values i.e -1, -2, ..., no_historic
to include when computing baseline parameters, default: 7}

\item{same_weekday_baseline}{whether to calculate baseline using same weekdays or any day, default: FALSE}
}
\value{
A dataframe containing the monitored time point,
        the upper limit and whether a signal is detected or not.
}
\description{
The simple 7 day running mean version of the Early Aberration Reporting System (EARS)
algorithm is extended as follows:
\itemize{
  \item{proper computation of the prediction interval}
  \item{downweighting of previous signals, similar to the approach by Farrington (1996)}
}
}
\details{
for algorithm details see package vignette.
}
\examples{
\dontrun{
   library(episomer)
   #Running the modifies version of the ears algorithm for a particular data series
    ts <- c(150, 130, 122, 160, 155, 128, 144, 125, 300, 319, 289, 277, 500)
    show(ears_t_reweighted(ts))
}
}
\author{
Michael Hoehle <https://www.math.su.se/~hoehle>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm.R
\name{sm_api_get_token}
\alias{sm_api_get_token}
\title{Get social media token}
\usage{
sm_api_get_token(network)
}
\arguments{
\item{network}{char, network the name of the social media}
}
\description{
Get social media token
}
\details{
this function implements the authethication on the respective social media and stores it securely. 
It is called by the search loop before starting message collection
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_new_media.R
\name{create_api_and_plan_files_for_new_social_media}
\alias{create_api_and_plan_files_for_new_social_media}
\title{internal function for producing the code template for a new social media}
\usage{
create_api_and_plan_files_for_new_social_media(
  new_media,
  ref_media = "bluesky"
)
}
\arguments{
\item{new_media}{char, name of the social media to add}

\item{ref_media}{char, name of the social media to use as a reference to produce template}
}
\description{
internal function for producing the code template for a new social media
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo-tagging.R
\name{geolocate_text}
\alias{geolocate_text}
\title{geolocate text in a data frame given a text column and optionally a language column}
\usage{
geolocate_text(df, text_col = "text", lang_col = NULL, min_score = NULL)
}
\arguments{
\item{df}{A data frame containing at least character column with text, a column with the language name can be provided to improve geolocation quality}

\item{text_col}{character, name of the column on the data frame containing the text to geolocalize, default:text}

\item{lang_col}{character, name of the column on the data frame containing the language of texts, default: NULL}

\item{min_score, }{numeric, the minimum score obtained on the Lucene scoring function to accept matches on GeoNames. It has to be empirically set default: NULL}
}
\value{
A new data frame containing the following geolocation columns: geo_code, geo_country_code, geo_country, geo_name, tags
}
\description{
extracts geolocaion information on text on a column of the provided data frame and returns a new data frame with geolocation information
}
\details{
This function perform a call to the episomer database which includes functionality for geolocating for languages activated and successfully processed on the shiny app.

The geolocation process tries to find the best match in GeoNames database \url{https://www.geonames.org/} including all local aliases for words.

If no language is associated to the text, all tokens will be sent as a query to the indexed GeoNames database.

If a language code is associated to the text and this language is trained on episomer, entity recognition techniques will be used to identify the best candidate in text to contain a location
and only these tokens will be sent to the GeoNames query.

A custom scoring function is implemented to grant more weight to cities increasing with population to try to perform disambiguation.

Rules for forcing the geolocation choices of the algorithms and for tuning performance with manual annotations can be performed on the geotag tab of the Shiny app.

A prerequisite to this function is that the tasks \code{\link{download_dependencies}} \code{\link{update_geonames}} and \code{\link{update_languages}} has been run successfully.

This function is called from the Shiny app on geolocation evaluation tab but can also be used for manually evaluating the episomer geolocation algorithm.
}
\examples{
\dontrun{
   library(episomer)
   # setting up the data folder
   message('Please choose the episomer data directory')
   setup_config(file.choose())

   # creating a test dataframe
   df <- data.frame(text = c("Me gusta Santiago de Chile es una linda ciudad"), lang = c("es"))
   geo <- geolocate_text(df = df, text_col = "text", lang_col="lang")

}
}
\seealso{
\code{\link{download_dependencies}}

 \code{\link{update_geonames}}

 \code{\link{detect_loop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm.R
\name{sm_plan_next_search_info}
\alias{sm_plan_next_search_info}
\title{Return a message to log what the next query is going to be for the current plan}
\usage{
sm_plan_next_search_info(plan)
}
\arguments{
\item{plan}{list, the plan to obtain the message}
}
\description{
Return a message to log what the next query is going to be for the current plan
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataviz.R
\name{create_map}
\alias{create_map}
\title{Plot the map report on the episomer dashboard}
\usage{
create_map(
  sms,
  topic = c(),
  countries = c(1),
  date_min = "1900-01-01",
  date_max = "2100-01-01",
  with_quotes = FALSE,
  caption = "",
  proj = NULL,
  forplotly = FALSE
)
}
\arguments{
\item{sms}{Social media}

\item{topic}{Character(1) containing the topic to use for the report}

\item{countries}{Character vector containing the name of the countries and regions to plot or their respective indexes on the Shiny app, default: c(1)}

\item{date_min}{Date indicating start of the reporting period, default: "1900-01-01"}

\item{date_max}{Date indicating end of the reporting period, default: "2100-01-01"}

\item{with_quotes}{Logical value indicating whether to include reposts in the time series, default: FALSE}

\item{caption}{Character(1) vector indicating a caption to print at the bottom of the chart, default: ""}

\item{proj}{Parameter indicating the CRS (Coordinate Reference System) to use on PROJ4 format, default NULL
If null and all countries are selected +proj=robin is used (Robinson projection) otherwise the Lambert azimuthal equal-area projection will be chosen, default: NULL}

\item{forplotly}{Logical(1) parameter indicating whether some hacks are activated to improve plotly rendering, default: FALSE}
}
\value{
A named list containing two elements: 'chart' with the ggplot2 figure and 'data' containing the dataframe that was used to build the map.
}
\description{
Generates a bubble map plot of number of posts by country for a single topic
}
\details{
Produces a bubble chart map for a particular topic on number of posts collected based on the provided parameters.
The map will display information at country level if more than one country is selected, otherwise it will display bubbles at the smallest possible location identified for each post within the period
which could be any administrative level or city level.

Posts associated with a country but with no finer granularity are omitted when displaying a single country.

When an aggregated zone is requested, all countries in that zone are included.

This functions requires that \code{\link{search_loop}} and \code{\link{detect_loop}} have already been run successfully to show results.
}
\examples{
\dontrun{
   #Getting bubble chart for dengue for South America for last 30 days
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   create_map(
     topic = "dengue",
     countries = "South America",
     date_min = as.Date(Sys.time())-30,
     date_max=as.Date(Sys.time())
   )
}
}
\seealso{
\code{\link{trend_line}}
 \code{\link{create_topwords}}
 \code{\link{detect_loop}}
 \code{\link{search_loop}}
 \code{\link[sp]{spTransform}},\code{\link[sp]{coordinates}},\code{\link[sp]{is.projected}},\code{\link[sp]{CRS-class}}
 \code{\link[ggplot2]{fortify}},\code{\link[ggplot2]{geom_polygon}},\code{\link[ggplot2]{geom_point}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alert.R
\name{calculate_region_alerts}
\alias{calculate_region_alerts}
\alias{calculate_regions_alerts}
\title{calculate alerts for for a particular region}
\usage{
calculate_region_alerts(
  sms,
  topic,
  country_codes = list(),
  country_code_cols = "geo_country_code",
  start = NA,
  end = NA,
  with_quotes = FALSE,
  alpha = 0.025,
  alpha_outlier = 0.05,
  k_decay = 4,
  no_historic = 7,
  bonferroni_m = 1,
  same_weekday_baseline = FALSE,
  logenv = NULL
)

calculate_regions_alerts(
  sms,
  topic,
  regions = c(1),
  date_type = c("created_date"),
  date_min = as.Date("1900-01-01"),
  date_max = as.Date("2100-01-01"),
  with_quotes = FALSE,
  alpha = 0.025,
  alpha_outlier = 0.05,
  k_decay = 4,
  no_historic = 7,
  bonferroni_correction = FALSE,
  same_weekday_baseline = FALSE,
  logenv = NULL
)
}
\arguments{
\item{sms}{A given social media for which aggregations have been calculated}

\item{topic}{Limit the alert detection to the given topic}

\item{country_codes}{List with the codes of countries to be aggregated for this alerts}

\item{country_code_cols}{the name of the column containing the country codes}

\item{start}{the start date for the period in consideration}

\item{end}{the end date for the period in consideration}

\item{with_quotes}{wether to consider quotes in the time series to evaluate}

\item{alpha}{The alpha is used to compute the upper limit of the prediction interval:
(1-alpha) * 100\%, default: 0.025}

\item{alpha_outlier}{Residuals beyond 1-alpha_outlier quantile of the
the t(n-k-1) distribution are downweighted, default: 0.05}

\item{k_decay}{Power k in the expression (r_star/r_threshold)^k determining the weight, default: 4}

\item{no_historic}{Number of previous values i.e -1, -2, ..., no_historic
to include when computing baseline parameters, default: 7}

\item{bonferroni_m}{Number the value to apply the Bonferroni correction; it specifies the global number of timeseries being observed}

\item{same_weekday_baseline}{whether to calculate baseline using same weekdays or any day, default: FALSE}

\item{logenv}{an environment variable to store the total number of posts concerned including all country_cols}

\item{regions}{List with the codes of regions to be aggregated for this alerts}

\item{date_type}{the type of date to be used for the period in consideration}

\item{date_min}{the start date for the period in consideration}

\item{date_max}{the end date for the period in consideration}

\item{bonferroni_correction}{whether to apply the bonferroni correction to the alerts, default: FALSE}
}
\value{
A dataframe containing the monitored time point for the given countries and topic,
        the upper limit and whether a signal is detected or not.

A dataframe containing the monitored time point for the given regions and topic,
        the upper limit and whether a signal is detected or not.
}
\description{
calculate alerts for a region and topic using provided parameters to the episomer signal detection function
}
\details{
for algorithm details see package vignette.
}
\examples{
\dontrun{
   library(episomer)
   #Running the alerts for France in topic covid-19
   df <-calculate_region_alerts(sms="bluesky",
                                topic="covid-19",
                                country_codes = "FR",
                                start=as.Date("2025-09-20"),
                                end=as.Date("2025-09-30"))
   show(df)
}
\dontrun{
   library(episomer)
   #Running the alerts for World in topic covid-19
   df <-calculate_regions_alerts(
                                sms="bluesky",
                                topic="covid-19",                                 
                                date_min=as.Date("2025-09-20"),
                                date_max=as.Date("2025-09-30")
)
   show(df)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alert.R
\name{generate_alerts}
\alias{generate_alerts}
\title{Execute the alert task}
\usage{
generate_alerts(tasks = get_tasks())
}
\arguments{
\item{tasks}{Current tasks for reporting purposes, default: get_tasks()}
}
\value{
The list of tasks updated with produced messages
}
\description{
Evaluate alerts for the last collected day for all topics and regions and send email alerts to subscribers
}
\details{
This function calculates alerts for the last aggregated day and then sends emails to subscribers.

The alert calculation is based on the country_counts time series which stores alerts by country, hour and topics.

For each country and region, the process starts by aggregating the last N days. A day is defined as a block of consecutive 24 hours ending before the hour of the last collected post.
N is defined by the alert baseline parameter on the configuration tab of the Shiny application (the default is N=7).

An alert will be produced when the number of posts observed is above the threshold calculated by the modified version of the EARS algorithm (for more details see the package vignette).
The behaviour of the alert detection algorithm is modified by the signal false positive rate (alpha), by downweighting previous alerts and by weekly or daily baseline parameters,
as defined in the configuration tab of the Shiny application and the topics file.

A prerequisite for this function is that the \code{\link{search_loop}} must already have stored posts in the search folder and that the geotagging and aggregation tasks have been completed.
Normally, this function is not called directly by the user, but from the \code{\link{detect_loop}} function.
}
\examples{
\dontrun{
   library(episomer)
   # setting up the data folder
   message('Please choose the episomer data directory')
   setup_config(file.choose())

   # calculating alerts for last day posts and sending emails to subscribers
   generate_alerts()
}
}
\seealso{
\code{\link{detect_loop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{bluesky_parse_date}
\alias{bluesky_parse_date}
\title{parse a date as provided by bluesky into a date_time object}
\usage{
bluesky_parse_date(date_input)
}
\arguments{
\item{date_input}{char, the text of a date}
}
\description{
parse a date as provided by bluesky into a date_time object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scheduler.R
\name{register_search_runner_task}
\alias{register_search_runner_task}
\title{Registers the social media message collection task}
\usage{
register_search_runner_task()
}
\value{
Nothing
}
\description{
registers the social media message collection task or stops if no configuration has been set or if it is already running
}
\details{
Registers the message collextion task or stops if no configuration has been set or if it is already running.
Bluesky authentication needs to be previously set on the shiny app or by calling set_auth().
You can test if authentication is working on the shiny app troubleshot page or by calling (with dplyr): episomer::check_all() %>% filter(check == 'bluesky_auth')
This function will use the task scheduler on windows and will fall back to launching the runner as a separate process (attached to this session) on Linux.
}
\examples{
\dontrun{
   #getting tasks statuses
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   register_search_runner_task()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_new_media.R
\name{add_new_social_media}
\alias{add_new_social_media}
\title{Create new social media template files, so a developer can extend episomer and support a new social media.
this function has to be called within episomer subfolder containing thr R subfolder after executing devtools::load_all()}
\usage{
add_new_social_media(social_media, social_media_ref = "bluesky")
}
\arguments{
\item{social_media}{Name of the new social media}

\item{social_media_ref}{Name of the reference social media}
}
\value{
TRUE, called for side effects (creation of new social media files).
}
\description{
Create new social media template files, so a developer can extend episomer and support a new social media.
this function has to be called within episomer subfolder containing thr R subfolder after executing devtools::load_all()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{save_config}
\alias{save_config}
\title{Save the configuration changes}
\usage{
save_config(data_dir = conf$data_dir, properties = TRUE, sm_topics = NULL)
}
\arguments{
\item{data_dir}{Path to a directory to save configuration settings, Default: conf$data_dir}

\item{properties}{Whether to save the general properties to the properties.json file, default: TRUE}

\item{sm_topics}{Whether to save topic download plans to the topics.json file, default: TRUE}
}
\value{
Nothing
}
\description{
Permanently saves configuration changes to the data folder (excluding social media credentials, but not SMTP credentials)
}
\details{
Permanently saves configuration changes to the data folder (excluding social media credentials, but not SMTP credentials)
to save social media credentials please use \code{\link{sm_api_set_auth_bluesky}}
}
\examples{
\dontrun{
   library(episomer)
   #load configuration
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   #make some changes
   #conf$collect_span = 90
   #saving changes
   save_config()
}
}
\seealso{
\code{\link{setup_config}}
\code{\link{sm_api_set_auth_bluesky}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{bluesky_create_session}
\alias{bluesky_create_session}
\title{Get bearer token}
\usage{
bluesky_create_session(handle = NULL, password = NULL)
}
\arguments{
\item{handle}{bluesky handle}

\item{password}{bluesky password}
}
\value{
List with access_jwt and did
}
\description{
Get bearer token
}
\details{
You can find more information about the bluesky API here: \url{https://docs.bsky.app/docs/api/com-atproto-server-create-session}
}
\examples{
\dontrun{
bluesky_create_session()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm.R
\name{sm_api_translate_query}
\alias{sm_api_translate_query}
\title{Translate a parsed query to the social media query format}
\usage{
sm_api_translate_query(network, parsed)
}
\arguments{
\item{network}{char, the name of the social media}

\item{parsed}{char, parsed query to translate in one or many native equivalent queries}
}
\description{
Translate a parsed query to the social media query format
}
\details{
This function receive as a parameter a parsed query from the a topic in the list of topics. 
It can return one or many queries that matches the expected results using the underlying social media API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fs.R
\name{search_posts}
\alias{search_posts}
\title{perform full text search on posts collected with episomer}
\usage{
search_posts(
  query = NULL,
  topic = NULL,
  from = NULL,
  to = NULL,
  countries = NULL,
  mentioning = NULL,
  users = NULL,
  hide_users = FALSE,
  action = NULL,
  max = 100,
  by_relevance = FALSE
)
}
\arguments{
\item{query}{character. The query to be used if a text it will match the post text. To see how to match particular fields please see details, default:NULL}

\item{topic}{character, Vector of topics to include on the search default:NULL}

\item{from}{an object which can be converted to ‘"POSIXlt"’ only posts posted after or on this date will be included, default:NULL}

\item{to}{an object which can be converted to ‘"POSIXlt"’ only posts posted before or on this date will be included, default:NULL}

\item{countries}{character or numeric, the position or name of episomer regions to be included on the query, default:NULL}

\item{mentioning}{character, limit the search to the posts mentioning the given users, default:NULL}

\item{users}{character, limit the search to the posts created by the provided users, default:NULL}

\item{hide_users}{logical, whether to hide user names on output replacing them by the USER keyword, default:FALSE}

\item{action}{character, an action to be performed on the search results respecting the max parameter. Possible values are 'delete' or 'anonymise' , default:NULL}

\item{max}{integer, maximum number of posts to be included on the search, default:100}

\item{by_relevance}{logical, whether to sort the results by relevance of the matching query or by indexing order, default:FALSE
If not provided the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, default: NA}
}
\value{
a data frame containing the posts matching the selected filters, the data frame contains the following columns: user_name, created_date, is_geo_located, is_quote, text, 
text_loc, user_id, hash, quote_lang, totalCount, created_at, topic_post_id, topic, lang, user_name, quoted_text, id, quopted_text_loc, tags
}
\description{
perform full text search on posts collected with episomer (posts migrated from episomer v<1.0.x are also included)
}
\details{
episomer translate the query provided by all parameters into a single query that will be executed on post indexes which are weekly indexes.
The q parameter should respect the syntax of the Lucene classic parser \url{https://lucene.apache.org/core/8_5_0/queryparser/org/apache/lucene/queryparser/classic/QueryParser.html}
So other than the provided parameters, multi field queries are supported by using the syntax field_name:value1;value2
AND, OR and -(for excluding terms) are supported on q parameter.
Order by week is always applied before relevance so even if you provide by_relevance = TRUE all of the matching posts of the first week will be returned first
}
\examples{
\dontrun{
   #Running the detect loop
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   df <- search_posts(
        q = "vaccination",
        topic="COVID-19",
        countries=c("Chile", "Australia", "France"),
        from = Sys.Date(),
        to = Sys.Date()
   )
   df$text
}
}
\seealso{
\code{\link{search_loop}}

 \code{\link{detect_loop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-plan.R
\name{sm_plan_next_attributes_bluesky}
\alias{sm_plan_next_attributes_bluesky}
\title{returns a list with the custom social media plan attributes with the values for the plan to be run after the plans provided as parameter}
\usage{
sm_plan_next_attributes_bluesky(plans)
}
\arguments{
\item{plans}{list, the list with current plans of the given topic to calculate next plan attributes}
}
\description{
returns a list with the custom social media plan attributes with the values for the plan to be run after the plans provided as parameter
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm.R
\name{sm_plan_parse_attributes}
\alias{sm_plan_parse_attributes}
\title{parse custom attributes for a plan in this specific social media and returns them as a named list with attributes in the proper type}
\usage{
sm_plan_parse_attributes(network, ...)
}
\arguments{
\item{network}{char, the name of the social media}

\item{...}{list, triple dot arguments to be passed as a parameter tu underlying functions}
}
\description{
parse custom attributes for a plan in this specific social media and returns them as a named list with attributes in the proper type
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{bluesky_format_date}
\alias{bluesky_format_date}
\title{formats a date in the expected bluesky format}
\usage{
bluesky_format_date(datetime)
}
\arguments{
\item{datetime, }{the datetime to format}
}
\description{
formats a date in the expected bluesky format
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spark-bridge.R
\name{download_dependencies}
\alias{download_dependencies}
\title{Updates Java dependencies}
\usage{
download_dependencies(tasks = get_tasks())
}
\arguments{
\item{tasks}{Task object for reporting progress and error messages, default: get_tasks()}
}
\value{
The list of tasks updated with produced messages
}
\description{
Download Java dependencies of the application mainly related to Apache SPARK, Apache Lucene, and Apache Pekko
}
\details{
Run a one-shot task consisting of downloading Java and Scala dependencies, this is separated by the following subtasks
\itemize{
  \item{Download jar dependencies from configuration maven repo to project data folder. This includes, scala, spark, lucene. Packages to be downloaded are defined in package file 'sbt-deps.txt'}
  \item{Download winutils from configuration URL to project data folder. For more details on winutils please see
    \url{https://issues.apache.org/jira/browse/HADOOP-13223} and \url{https://issues.apache.org/jira/browse/HADOOP-16816}
  }
}

The URLs to download the JAR dependencies (maven package manager) and Winutils are on the configuration tab of the Shiny app.

Normally this function is not called directly by the user but from the \code{\link{detect_loop}} function.
}
\examples{
\dontrun{
   library(episomer)
   # setting up the data folder
   message('Please choose the episomer data directory')
   setup_config(file.choose())

   # geolocating last posts
   tasks <- download_dependencies()
}
}
\seealso{
\code{\link{detect_loop}}

 \code{\link{get_tasks}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm.R
\name{sm_api_set_auth}
\alias{sm_api_set_auth}
\title{Save credentials for a social media provider and store them securely}
\usage{
sm_api_set_auth(network, shiny_input_list)
}
\arguments{
\item{network}{char, the name of the social media}

\item{shiny_input_list}{list, the shiny input object to extract the information from}
}
\description{
Update configuration object and underlying storage with given bluesky username and password. The password is encrypted.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fs.R
\name{fs_loop}
\alias{fs_loop}
\title{Runs the episomer embedded database loop}
\usage{
fs_loop(data_dir = NA)
}
\arguments{
\item{data_dir}{Path to the 'data directory' containing application settings, models and collected posts.
If not provided, the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, default: NA}
}
\value{
nothing
}
\description{
Infinite loop ensuring that the episomer embedded database is running (Lucene + akka-http)
}
\details{
Launches the episomer embedded database which is accessed via a REST API located on localhost at port 8080. You can test that the database is running by accesssing the local URL and adding "/ping". 
the REST API provide episomer a way to send and retrieve data related with posts and time series and to trigger geolocation or aggregation
The database is implemented using Apache Lucene indexes allowing episomer to access its data as a search engine but also as a tabular database.
\code{\link{health_check}} called each 60 seconds on a background process to send alerts to the administrator if some episomer components fail.
}
\examples{
\dontrun{
   #Running the detect loop
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   fs_loop()
}
}
\seealso{
\code{\link{detect_loop}}

 \code{\link{search_loop}}

 \code{\link{health_check}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scheduler.R
\name{get_tasks}
\alias{get_tasks}
\title{Get the \code{\link{detect_loop}} task status}
\usage{
get_tasks(statuses = list())
}
\arguments{
\item{statuses}{Character vector for limiting the status of the returned tasks, default: list()}
}
\value{
A named list containing all necessary information to run and monitor the detect loop tasks.
}
\description{
Reads the status of the \code{\link{detect_loop}} tasks and update it with changes requested by the Shiny app
}
\details{
After reading the tasks.json file and parsing it with jsonlite, this function will update the necessary fields in the
tasks for executing and monitoring them.
}
\examples{
\dontrun{
   #getting tasks statuses
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   tasks <- get_tasks()
}
}
\seealso{
\code{\link{download_dependencies}}

 \code{\link{update_geonames}}

 \code{\link{update_languages}}

 \code{\link{detect_loop}}

 \code{\link{generate_alerts}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate.R
\name{get_aggregates}
\alias{get_aggregates}
\title{Getting already aggregated time series produced by \code{\link{detect_loop}}}
\usage{
get_aggregates(
  dataset = "country_counts",
  cache = TRUE,
  filter = list(),
  top_field = NULL,
  top_freq = NULL
)
}
\arguments{
\item{dataset}{A character(1) vector with the name of the series to request, it must be one of 'country_counts', 'geolocated', 'topwords', 'hashtags', 'entities', 'urls', 'contexts', default: 'country_counts'}

\item{cache}{Whether to use the cache for lookup and storing the returned dataframe, default: TRUE}

\item{filter}{A named list defining the filter to apply on the requested series, it should be in the form of a named list; e.g., list(post_geo_country_code=list('FR', 'DE')) default: list()}

\item{top_field}{Name of the top field used with top_frequency to enable optimisation for getting only most frequent elements. It will only keep top 500 items after first 50k lines on reverse index order}

\item{top_freq}{character, Name of the frequency fields used with top_field to enable optimisation for retrieving only most frequent elements.
It will only keep top 500 items after first 50k rows on reverse index order}
}
\value{
A data frame containing the requested series for the requested period
}
\description{
Reads and returns the required aggregated dataset for the selected period and topics defined by the filter.
}
\details{
This function returns data aggregated by episomer. The data is found on the 'series' folder, which contains Rds files per weekday and type of series.
starting on v 1.0.x it will also look on Lucene indexes situated on fs folder. Names of files and folders are parsed to limit the files to be read.
When using Lucene indexes, filters are directly applied on read. This is an improvement compared 'series' folder where filters are applied
after read. All returned rows are joined into a single dataframe.
If no filter is provided, the entire data series is returned, which can result in millions of rows depending on the time series.
To limit by period, the filter list must include an element 'period' containing a date vector or list with two dates representing the start and end of the request.

To limit by topic, the filter list must include an element 'topic' containing a non-empty character vector or list with the names of the topics to return.

The available time series are:
\itemize{
  \item{"country_counts" counting posts and reposts by posted date, hour and country}

  \item{"geolocated" counting posts and reposts by posted date and the smallest possible geolocated unit (city, administrative level or country)}

  \item{"topwords" counting posts and reposts by posted date, country and the most popular words, (this excludes words used in the topic search)}
}
The returned dataset can be cached for further calls if requested. Only one dataset per series is cached.
}
\examples{
\dontrun{
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   # Getting all country posts between 2020-jan-10 and 2020-jan-31 for all topics
   df <- get_aggregates(
     dataset = "country_counts",
     filter = list(period = c("2020-01-10", "2020-01-31"))
   )

   # Getting all country posts for the topic dengue
   df <- get_aggregates(dataset = "country_counts", filter = list(topic = "dengue"))

   # Getting all country posts between 2020-jan-10 and 2020-jan-31 for the topic dengue
    df <- get_aggregates(
        dataset = "country_counts",
         filter = list(topic = "dengue", period = c("2020-01-10", "2020-01-31"))
    )
}
}
\seealso{
\code{\link{detect_loop}}
 \code{\link{fs_loop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scheduler.R
\name{register_detect_runner_task}
\alias{register_detect_runner_task}
\title{Registers the alert detection task}
\usage{
register_detect_runner_task()
}
\value{
Nothing
}
\description{
registers the alert detection task or stops if no configuration has been set or if it is already running
}
\details{
Registers the alert detection task or stops if no configuration has been set or if it is already running.
To generate alerts, this task needs the post collection to had successfully run since the last time it ran.
This function will use the task scheduler on windows and will fall back to launching the runner as a separate process (attached to this session) on Linux.
}
\examples{
\dontrun{
   #getting tasks statuses
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   register_detect_runner_task()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{bluesky_parse_quoted}
\alias{bluesky_parse_quoted}
\title{extracts the quotes information from a post in one of the many possible locations found in bluesky format}
\usage{
bluesky_parse_quoted(post)
}
\arguments{
\item{post}{list, the post to to extract data from in bluesky format}
}
\description{
extracts the quotes information from a post in one of the many possible locations found in bluesky format
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{sm_api_search_bluesky}
\alias{sm_api_search_bluesky}
\title{Search posts for a given query and produce standardised results}
\usage{
sm_api_search_bluesky(query, token, plan, max_retries = 10, verbose = TRUE)
}
\arguments{
\item{query}{query to search for}

\item{token}{Access token}

\item{plan}{Plan object}

\item{max_retries}{Maximum number of retries}

\item{verbose}{Whether to print progress messages}
}
\value{
List with posts int the standard post format defined in episomer
}
\description{
Search posts for a given query and produce standardised results
}
\details{
This function impleents the search of messages for a given query and plan.
The query has been already transformed by sm_api_translate_query_xxx so it is expected to be already in the native social media format
The token has been produced by the function sm_api_get_token_xxx and shoulb be used to autenticate the request
The plan should be used to extract the precise time-span to request messages

- Response attributes
  - network: string
  - posts: [...]
  - count: number, 
  - pagination:\{...\}
- post in posts
    - id: string
    - uri: string
    - created_at: datetime
    - user_name: string
    - text: string
    - lang: string
    - quoted_text: string
    - quoted_lang: string
    - tags: [...]
    - urls: [...]
    - categories: [...]
- tag in post.tags: string
- url in post.urls: string
- category in post.categories: string
- pagination: attributes used for paginating requests

This is the implementation of the bluesky search. You can find more information about the bluesky API here: \url{https://docs.bsky.app/docs/api/app-bsky-feed-search-posts}
Additionnaly more information about the "post" object is available here: \url{https://atproto.blue/en/latest/atproto/atproto_client.models.app.bsky.feed.defs.html#atproto_client.models.app.bsky.feed.defs.PostView}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm.R
\name{sm_api_search}
\alias{sm_api_search}
\title{Search posts for a given query and produce standardised results}
\usage{
sm_api_search(query, token, plan)
}
\arguments{
\item{query}{char, the query to use in the search in nattive format}

\item{token}{char, the token to authenticate against the underlying social media api}

\item{plan}{list, the plan containing the targeted time frame}
}
\description{
Search posts for a given query and produce standardised results
}
\details{
This function impleents the search of messages for a given query and plan.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{bluesky_check_token_validity}
\alias{bluesky_check_token_validity}
\title{check the current bluesky token validity}
\usage{
bluesky_check_token_validity(
  access_jwt,
  search_url = "https://bsky.social/xrpc/app.bsky.feed.searchPosts"
)
}
\arguments{
\item{access_jwt}{char, the jwt token provided by the bluesky api}

\item{search_url}{char, the url to test if the current token is still valid}
}
\description{
check the current bluesky token validity
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{setup_config}
\alias{setup_config}
\title{Load episomer application settings}
\usage{
setup_config(
  data_dir = if (exists("data_dir", where = conf)) conf$data_dir else if
    (Sys.getenv("EPI_HOME") != "") Sys.getenv("EPI_HOME") else file.path(tempdir(),
    "episomer"),
  ignore_keyring = FALSE,
  ignore_properties = FALSE,
  ignore_topics = FALSE,
  save_properties_first = FALSE,
  save_topics_first = list()
)
}
\arguments{
\item{data_dir}{Path to the directory containing the application settings (it must exist).
If not provided it takes the value of the latest call to setup_config in the current session, or the value of the EPI_HOME environment variable or episomer subdirectory in the working directory,
default: if (exists("data_dir", where = conf)) conf$data_dir else if (Sys.getenv("EPI_HOME") !=
   "") Sys.getenv("EPI_HOME") else file.path(tempdir(), "episomer")}

\item{ignore_keyring}{Whether to skip loading settings from the keyring (social media and SMTP credentials), default: FALSE}

\item{ignore_properties}{Whether to skip loading settings managed by the Shiny app in properties.json file, Default: FALSE}

\item{ignore_topics}{Whether to skip loading settings defined in the topics.xlsx file and download plans from topics.json file, default: FALSE}

\item{save_properties_first}{Whether to save current settings before loading new ones from disk, default: FALSE}

\item{save_topics_first}{List of topics to save before loading new ones from disk, default: list()}
}
\value{
Nothing
}
\description{
Load episomer application settings from the designated data directory
}
\details{
episomer relies on settings and data stored in a system folder, so before loading the dashboard, collecting posts or detecting alerts the user has to designate this folder.
When a user wants to use episomer from the R console they will need to call this function for initialisation.
The 'data_folder' can also be given as a parameter for program launch functions \code{\link{episomer_app}}, \code{\link{search_loop}} or \code{\link{detect_loop}}, which will internally call this function.

This call will fill (or refresh) a package scoped environment 'conf' that will store the settings. Settings stored in conf are:
\itemize{
  \item{General properties of the Shiny app (stored in properties.json)}
  \item{Download plans from the social media collection process (stored in topics.json merged with data from the topics.xlsx file}
  \item{Credentials for social media APIs and SMTP stored in the defined keyring}
}

When calling this function and the keyring is locked, a password will be prompted to unlock the keyring.
This behaviour can be changed by setting the environment variable 'ecdc_social_tool_kr_password' with the password.

Changes made to conf can be stored permanently (except for 'data_dir') using:
\itemize{
  \item{\code{\link{save_config}}, or}
   \item{\code{\link{sm_api_set_auth_bluesky}}}
}
}
\examples{
\dontrun{
   library(episomer)
   #loading system settings
   message('Please choose the episomer data directory')
   setup_config(file.choose())
}
}
\seealso{
\code{\link{save_config}}
\code{\link{sm_api_set_auth_bluesky}}
\code{\link{episomer_app}}
\code{\link{search_loop}}
\code{\link{detect_loop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataviz.R
\name{create_topwords}
\alias{create_topwords}
\title{Plot the top words report on the episomer dashboard}
\usage{
create_topwords(
  sms,
  topic,
  country_codes = c(),
  date_min = "1900-01-01",
  date_max = "2100-01-01",
  with_quotes = FALSE,
  top = 25
)
}
\arguments{
\item{sms}{Social media}

\item{topic}{Character(1) containing the topic to use for the report}

\item{country_codes}{Character vector containing the ISO 3166-1 alpha-2 countries to plot, default: c()}

\item{date_min}{Date indicating start of the reporting period, default: "1900-01-01"}

\item{date_max}{Date indicating end of the reporting period, default: "2100-01-01"}

\item{with_quotes}{Logical value indicating whether to include reposts in the time series, default: FALSE}

\item{top}{numeric(1) Parameter indicating the number of words to show, default: 25}
}
\value{
A named list containing two elements: 'chart' with the ggplot2 figure and 'data' containing the data frame that was used to build the map.
}
\description{
Generates a bar plot of most popular words in posts, for one topic
}
\details{
Produces a bar chart showing the occurrences of the most popular words in the collected posts based on the provided parameters.

This report may be empty for combinations of countries and topics with very few posts since for performance reasons, the calculation of top words is an approximation using chunks of 10.000 posts.

This function requires that \code{\link{search_loop}} and \code{\link{detect_loop}} have already been run successfully to show results.
}
\examples{
\dontrun{
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   #Getting topword chart for dengue for France, Chile, Australia for last 30 days
   create_topwords(
     topic = "dengue",
     country_codes = c("FR", "CL", "AU"),
     date_min = as.Date(Sys.time())-30,
     date_max=as.Date(Sys.time())
   )
 }
}
\seealso{
\code{\link{trend_line}}
 \code{\link{create_map}}
 \code{\link{detect_loop}}
 \code{\link{search_loop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{bluesky_parse_response}
\alias{bluesky_parse_response}
\title{parses the response of bluesky API into the standardised episomer format. For more details see the vignette.}
\usage{
bluesky_parse_response(response, current_min_date)
}
\arguments{
\item{response}{list, the standardised response}

\item{current_min_date}{datetime, the current min date obtained on the plan to detect if there are messages after this limit and discard them}
}
\description{
parses the response of bluesky API into the standardised episomer format. For more details see the vignette.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{bluesky_rate_limited_check}
\alias{bluesky_rate_limited_check}
\title{function to check if there has been a rate limit error returned by the bluesky API}
\usage{
bluesky_rate_limited_check(resp)
}
\arguments{
\item{resp, }{the httr2 response object}
}
\description{
function to check if there has been a rate limit error returned by the bluesky API
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check.R
\name{check_all}
\alias{check_all}
\title{Run automatic sanity checks}
\usage{
check_all()
}
\value{
Data frame containing the statuses of all realized checks
}
\description{
It runs a set of automated sanity checks for helping the user to troubleshot issues
}
\details{
This function executes a series of sanity checks, concerning, Java, bitness, task status, dependencies and social media authentication.
}
\examples{
\dontrun{
   #importing episomer
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   #running all tests
   check_all()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo-tagging.R
\name{update_languages}
\alias{update_languages}
\title{Updates local copies of languages}
\usage{
update_languages(tasks = get_tasks(), reuse_downloads = FALSE)
}
\arguments{
\item{tasks}{Tasks object for reporting progress and error messages, default: get_tasks()}

\item{reuse_downloads}{logical indicating if the downloads should be reused if they already exist, default: FALSE}
}
\value{
The list of tasks updated with produced messages
}
\description{
Downloading and indexing a fresh version of language models tagged for update on the Shiny app configuration tab
}
\details{
Run a one shot task to download and index a local fasttext \href{https://fasttext.cc/docs/en/crawl-vectors.html}{pretrained models}.
A fasttext model is a collection of vectors for a language automatically produced scrolling a big corpus of text that can be used to capture the semantic of a word.

The URL to download the vectors from are set on the configuration tab of the Shiny app.

This task will also update SVM models to predict whether a word is a location that will be used in the geolocation process.

The indexing is developed in SPARK and Lucene.

A prerequisite to this function is that the \code{\link{search_loop}} must already have stored collected posts in the search folder and that the tasks \code{\link{download_dependencies}}
and \code{\link{update_geonames}} has been run successfully.

Normally this function is not called directly by the user but from the \code{\link{detect_loop}} function.
}
\examples{
\dontrun{
   library(episomer)
   # setting up the data folder
   message('Please choose the episomer data directory')
   setup_config(file.choose())

   # updating language tasks
   tasks <- update_languages()
}
}
\seealso{
\code{\link{download_dependencies}}

 \code{\link{update_geonames}}

 \code{\link{detect_loop}}

 \code{\link{get_tasks}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scheduler.R
\name{stop_search_runner_task}
\alias{stop_search_runner_task}
\title{Stops the post collection task}
\usage{
stop_search_runner_task()
}
\value{
Nothing
}
\description{
stops the post collection task
}
\details{
Stops the post collection task if it is already running
This function will try also deactivate the respective scheduled task on Windows.
}
\examples{
\dontrun{
   #getting tasks statuses
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   stop_search_runner_task()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_new_media.R
\name{keep_roxygen_and_function_declarations}
\alias{keep_roxygen_and_function_declarations}
\title{internal function for exporting code of necessary functions for creating templates for a new social media}
\usage{
keep_roxygen_and_function_declarations(file_content)
}
\arguments{
\item{file_content}{char, lines of the script to extract content from}
}
\description{
internal function for exporting code of necessary functions for creating templates for a new social media
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check.R
\name{create_snapshot}
\alias{create_snapshot}
\title{Snapshot of your episomer installation}
\usage{
create_snapshot(
  destination_dir,
  types = c("settings", "dependencies", "machine-learning", "aggregations", "posts",
    "logs"),
  posts_period = get_aggregated_period(),
  aggregated_period = get_aggregated_period(),
  compress = TRUE,
  progress = function(v, m) message(paste(round(v * 100, 2), m))
)
}
\arguments{
\item{destination_dir, }{character(1) vector with the path of the destination folder to produce the snapshot}

\item{types, }{character vector indicating the types of data to include on a snapshot. Some of: "settings", "dependencies", "machine-learning", "aggregations", "posts", "logs"}

\item{posts_period, }{date(2) start and end dates to filter posts to include on snapshot (if selected)}

\item{aggregated_period, }{date(2) start and end dates to filter time series to include on snapshot (if selected)}

\item{compress, }{logical(1) whether to compress or not the output file}

\item{progress, }{function to report progress during execution.}
}
\value{
Nothing
}
\description{
Creates a snapshot file of your episomer installation folder. This can include all or a subset of the data files.
}
\details{
This function can be used to create a a portable file to move your episomer installation in a single file, to backup your data, to archive your old data or to send information to technical team in order to reproduce an observed issue.
Different kinds of data can be included on the snapshot depending on the type of parameter. Possible values are:
- 'settings': Including all setting files of your installation (excluding passwords)
- 'dependencies': All jars and winutils.exe on windows installations
- 'machine-learning': All trained models and vectors and training data (this can include post text which is personal data)
- 'aggregations': Episomer aggregated time series
- 'posts': Posts collected by episomer
- 'logs': Log files produced automatically on windows task scheduler tasks.
}
\examples{
\dontrun{
   #importing episomer
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   message('Please choose a destination directory')
   dest <- file.choose()
   #creating a compressed snapshot for settings and logs
   create_snapshot(dest, c("settings","dependencies"), compress = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{sm_api_update_plan_after_request_bluesky}
\alias{sm_api_update_plan_after_request_bluesky}
\title{Takes a plan after a request has been performed and the standardised results provided as returned by the function sm_api_search_xxx
it should return a plan with necessary information updated so next request will continue paginating as expected to obtain
targeted messages}
\usage{
sm_api_update_plan_after_request_bluesky(plan, results)
}
\arguments{
\item{plan}{list, the plan to update}

\item{results}{list, the obtained results of the plan}
}
\description{
Takes a plan after a request has been performed and the standardised results provided as returned by the function sm_api_search_xxx
it should return a plan with necessary information updated so next request will continue paginating as expected to obtain
targeted messages
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{search_topic}
\alias{search_topic}
\title{performs a search following the given plan, query and topic.}
\usage{
search_topic(plan, query, topic, sandboxed = FALSE)
}
\arguments{
\item{plan}{list, the plan that defines the timespan to search for}

\item{query}{char, the query to use when searching for the topic}

\item{topic}{char, the topic associated to this search}

\item{sandboxed}{logical, whether to un in sandboxed mode which prevent sending data to the embedded database.}
}
\description{
performs a search following the given plan, query and topic.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataviz.R
\name{create_topchart}
\alias{create_topchart}
\title{Plot the top elements for a specific series on the episomer dashboard}
\usage{
create_topchart(
  sms,
  topic,
  serie,
  country_codes = c(),
  date_min = "1900-01-01",
  date_max = "2100-01-01",
  with_quotes = FALSE,
  top = 25
)
}
\arguments{
\item{sms}{Social media}

\item{topic}{Character(1) containing the topic to use for the report}

\item{serie}{Character(1) name of the series to be used for the report. It should be one of ("topwords", "hashtags", "entities", "contexts", "urls")}

\item{country_codes}{Character vector containing the ISO 3166-1 alpha-2 countries to plot, default: c()}

\item{date_min}{Date indicating start of the reporting period, default: "1900-01-01"}

\item{date_max}{Date indicating end of the reporting period, default: "2100-01-01"}

\item{with_quotes}{Logical value indicating whether to include reposts in the time series, default: FALSE}

\item{top}{numeric(1) Parameter indicating the number of words to show, default: 25}
}
\value{
A named list containing two elements: 'chart' with the ggplot2 figure and 'data' containing the data frame that was used to build the map.
}
\description{
Generates a bar plot of most popular elements in posts, for one topic. Top elements among ("topwords", "hashtags", "entities", "contexts", "urls")
}
\details{
Produces a bar chart showing the occurrences of the most popular words in the collected posts based on the provided parameters.

This functions requires that \code{\link{search_loop}} and \code{\link{detect_loop}} have already been run successfully to show results.
}
\examples{
\dontrun{
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   #Getting topword chart for dengue for France, Chile, Australia for last 30 days
   create_topchart(
     topic = "dengue",
     serie = "topwords",
     country_codes = c("FR", "CL", "AU"),
     date_min = as.Date(Sys.time())-30,
     date_max=as.Date(Sys.time())
   )
 }
}
\seealso{
\code{\link{trend_line}}
 \code{\link{create_map}}
 \code{\link{detect_loop}}
 \code{\link{search_loop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scheduler.R
\name{register_fs_monitor}
\alias{register_fs_monitor}
\title{Registers the fs_monitor for the current process or exits}
\usage{
register_fs_monitor()
}
\value{
Nothing
}
\description{
registers the fs_monitor (by writing fs.monitor.PID file) for the current process or stops if no configuration has been set or if it is already running
}
\details{
Registers the fs_monitor (by writing fs.monitor.PID file) for the current process or stops if no configuration has been set or if it is already running
This function has been exported so it can be properly called from the future package on the database runner, but it is not intended to be directly called by end users.
}
\examples{
\dontrun{
   #getting tasks statuses
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   register_fs_monitor()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{search_loop}
\alias{search_loop}
\alias{search_loop_worker}
\title{Runs the search loop}
\usage{
search_loop(
  data_dir = NA,
  sandboxed = FALSE,
  log_to_file = TRUE,
  max_requests = 0
)

search_loop_worker(network, data_dir = NA, sandboxed = FALSE, max_requests = 0)
}
\arguments{
\item{data_dir}{The data directory to use.}

\item{sandboxed}{Whether to run in sandboxed mode.}

\item{log_to_file}{logical indicating if the search loop should log to a file. Default: TRUE}

\item{max_requests}{The maximum number of requests to perform.}

\item{network}{The social media network to search on.}
}
\value{
Nothing

Nothing
}
\description{
Infinite loop ensuring the permanent collection of social media messages
}
\details{
The detect loop is a pure R function designed for downloading posts from the social media search API. It can handle several topics ensuring that all of them will be downloaded fairly using a
round-robin philosophy and respecting social media API rate-limits.

The progress of this task is reported on the 'topics.*.json' files which are created by this function. This function will try to collect posts respecting a 'collect_span' window
in minutes, which is defined on the Shiny app and defaults to 60 minutes.

To see more details about the collection algorithm please see episomer vignette.

In order to work, this task needs Bluesky credentials, which can be set on the Shiny app or using \code{\link{sm_api_set_auth_bluesky}}

The search loop worker is a function that runs the search loop for a given social media network.
It is used to run the search loop in a separate process.
}
\examples{
\dontrun{
   #Running the search loop
   library(episomer)
   message('Please choose the episomer data directory')
   search_loop(file.choose())
}
}
\seealso{
\code{\link{sm_api_set_auth_bluesky}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-plan.R
\name{sm_plan_format_bluesky}
\alias{sm_plan_format_bluesky}
\title{prepare a plan for serialisation by transforming attributes with non standard serialisation types e.g. dates to string on the format expected by the parse functions}
\usage{
sm_plan_format_bluesky(plan)
}
\arguments{
\item{plan}{list, the plan with the attributes go format}
}
\description{
prepare a plan for serialisation by transforming attributes with non standard serialisation types e.g. dates to string on the format expected by the parse functions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scheduler.R
\name{missing_search_jobs}
\alias{missing_search_jobs}
\title{Missing search jobs if any}
\usage{
missing_search_jobs()
}
\value{
char list of social media with no search job running
}
\description{
check if search jobs for each activated social media are running and return the list names of social media with missing jobs
}
\details{
returns a char vector with the name of missing social media
}
\examples{
\dontrun{
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   missing_search_jobs()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo-tagging.R
\name{update_geonames}
\alias{update_geonames}
\title{Updates the local copy of the GeoNames database}
\usage{
update_geonames(tasks = get_tasks())
}
\arguments{
\item{tasks}{Tasks object for reporting progress and error messages, default: get_tasks()}
}
\value{
The list of tasks updated with produced messages
}
\description{
Downloading and indexing a fresh version of the GeoNames database from the provided URL
}
\details{
Run a one shot task to download and index a local copy of the  \href{http://www.geonames.org/}{GeoNames database}.
The GeoNames geographical database covers all countries and contains over eleven million place names that are available; Creative Commons Attribution 4.0 License.

The URL to download the database from is set on the configuration tab of the Shiny app, in case it changes.

The indexing is developed in Spark and Lucene

A prerequisite to this function is that the \code{\link{search_loop}} must already have stored collected posts in the search folder and that the task \code{\link{download_dependencies}}
has been successfully run.

Normally this function is not called directly by the user but from the \code{\link{detect_loop}} function.
}
\examples{
\dontrun{
   library(episomer)
   # setting up the data folder
   message('Please choose the episomer data directory')
   setup_config(file.choose())

   # geolocating last posts
   tasks <- update_geonames()
}
}
\seealso{
\code{\link{download_dependencies}}

 \code{\link{detect_loop}}

 \code{\link{get_tasks}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shiny-app.R
\name{episomer_app}
\alias{episomer_app}
\alias{dashboard_app}
\alias{admin_app}
\title{Run the episomer Shiny app}
\usage{
episomer_app(
  data_dir = NA,
  profile = c("dashboard", "admin"),
  host = NULL,
  port = NULL
)

dashboard_app(data_dir = NA, host = NULL, port = NULL)

admin_app(data_dir = NA, host = NULL, port = NULL)
}
\arguments{
\item{data_dir}{Path to the 'data directory' containing application settings, models and collected posts.
If not provided the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, default: NA}

\item{profile}{The profileis to run the Shiny app on, default: "dashboard" (can be "dashboard" or "admin")}

\item{host}{The IP if the network interface to run the Shiny app on, default: NULL (will run on 127.0.0.1)}

\item{port}{The port to run the Shiny app on, default: NULL (will run on a random port)}
}
\value{
The Shiny server object containing the launched application

The Shiny server object containing the launched application

The Shiny server object containing the launched application
}
\description{
Open the episomer Shiny app, used to setup the Data collection & processing pipeline, the Requirements & alerts pipeline and to visualise the outputs.
}
\details{
The episomer app is the user entry point to the episomer package. This application will help the user to setup the post collection process, manage all settings,
see the interactive dashboard visualisations, export them to Markdown or PDF, and setup the alert emails.

All its functionality is described on the episomer vignette.

The dashboard app is the user entry point to the episomer package for regular users. This application shows the interactive dashboard with visualisations, 
allow export them to Markdown or PDF.

All its functionality is described on the episomer vignette.

The admin app is contain all functionality for an administrator. It help the user to setup the post collection process, manage all settings,
and setup the alert emails.

All its functionality is described on the episomer vignette.
}
\examples{
\dontrun{
   #Running the episomer app
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   episomer_app()
}
}
\seealso{
\code{\link{search_loop}}

 \code{\link{detect_loop}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scheduler.R
\name{detect_loop}
\alias{detect_loop}
\title{Runs the detect loop}
\usage{
detect_loop(data_dir = NA)
}
\arguments{
\item{data_dir}{Path to the 'data directory' containing application settings, models and collected posts.
If not provided the system will try to reuse the existing one from last session call of \code{\link{setup_config}} or use the EPI_HOME environment variable, default: NA}
}
\value{
nothing
}
\description{
Infinite loop ensuring the daily signal detection and email alerts
}
\details{
The detect loop is composed of three 'one shot tasks' \code{\link{download_dependencies}}, \code{\link{update_geonames}}, \code{\link{update_languages}} ensuring the system has
all necessary components and data to run the three recurrent tasks \code{\link{generate_alerts}}

The loop report progress on the 'tasks.json' file which is read or created by this function.

The recurrent tasks are scheduled to be executed each 'detect span' minutes, which is a parameter set on the Shiny app.

If any of these tasks fails it will be retried three times before going to abort status. Aborted tasks can be relaunched from the Shiny app.
}
\examples{
\dontrun{
   #Running the detect loop
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   detect_loop()
}
}
\seealso{
\code{\link{download_dependencies}}

 \code{\link{update_geonames}}

 \code{\link{update_languages}}

 \code{\link{detect_loop}}

 \code{\link{generate_alerts}}

 \code{\link{get_tasks}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_new_media.R
\name{is_r_folder_present}
\alias{is_r_folder_present}
\title{check whether the R folder is present in the current directory}
\usage{
is_r_folder_present()
}
\description{
check whether the R folder is present in the current directory
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search.R
\name{search_touch}
\alias{search_touch}
\title{updates the modificatin date of a flag file indicating that messages have been sent to the embedded database.}
\usage{
search_touch()
}
\description{
updates the modificatin date of a flag file indicating that messages have been sent to the embedded database.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scheduler.R
\name{register_fs_runner_task}
\alias{register_fs_runner_task}
\title{Registers the episomer database task}
\usage{
register_fs_runner_task()
}
\value{
Nothing
}
\description{
registers the episomer database task or stops if no configuration has been set or if it is already running
}
\details{
Registers the episomer database task or stops if no configuration has been set or if it is already running.
This task need the dependencies, geonames and languages steps to have been successfully ran. This can be done on the shiny app configuration page or by manually running the detect_runner_task.
This function will try to use the task scheduler on windows and will fall back to launching the runner as a separate process (attached to this session) on Linux.
}
\examples{
\dontrun{
   #getting tasks statuses
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   register_fs_runner_task()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm.R
\name{sm_plan_format}
\alias{sm_plan_format}
\title{prepare a plan for serialisation by transforming attributes with non standard serialisation types e.g. dates to string on the format expected by the parse functions}
\usage{
sm_plan_format(plan)
}
\arguments{
\item{plan, }{the plan to format the attributes}
}
\description{
prepare a plan for serialisation by transforming attributes with non standard serialisation types e.g. dates to string on the format expected by the parse functions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check.R
\name{health_check}
\alias{health_check}
\title{Send email to administrator if a failure of episomer is detected}
\usage{
health_check(send_mail = TRUE, one_per_day = TRUE)
}
\arguments{
\item{send_mail}{Boolean. Whether an email should be sent to the defined administrator, default: TRUE}

\item{one_per_day}{Boolean. Whether a limit of one email per day will be applied, default: TRUE}
}
\value{
A list of health check errors found
}
\description{
It validates if episomer is not collecting posts, aggregating posts or not calculating alerts
}
\details{
This function sends an email to the defined administrator if episomer is not collecting posts, aggregating posts or not calculating alerts
}
\examples{
\dontrun{
   #importing episomer
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   #sending the email to the administrator if episomer components are not properly working
   health_check()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{sm_api_got_rows_bluesky}
\alias{sm_api_got_rows_bluesky}
\title{Receive the results provided by a request and return a logical value indicating whether the request returned results.}
\usage{
sm_api_got_rows_bluesky(results)
}
\arguments{
\item{results}{list, results of the request in the standardised episomer format}
}
\description{
Receive the results provided by a request and return a logical value indicating whether the request returned results.
}
\details{
This function needs to be implemented for each social media since it is interpreted as 'No more messages are available for this plan'
When FALSE is returned the underlying plan will be ended
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{bluesky_parse_features}
\alias{bluesky_parse_features}
\title{extracts features from the post structure returned by the bluesky api}
\usage{
bluesky_parse_features(post)
}
\arguments{
\item{post}{list, the post to extract features from in bluesky format}
}
\description{
extracts features from the post structure returned by the bluesky api
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scheduler.R
\name{is_fs_running}
\alias{is_fs_running}
\title{Check whether the database is running}
\usage{
is_fs_running()
}
\value{
logical Whether the database is running
}
\description{
gets the database runner execution status
}
\details{
returns a logical value being TRUE if the database runner is actually running
}
\examples{
\dontrun{
   library(episomer)
   message('Please choose the episomer data directory')
   setup_config(file.choose())
   is_fs_running()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-api.R
\name{sm_api_get_token_bluesky}
\alias{sm_api_get_token_bluesky}
\title{Get social media token}
\usage{
sm_api_get_token_bluesky()
}
\value{
Token
}
\description{
Get social media token
}
\details{
this function implements the authethication on the respective social media and stores it securely. 
It is called by the search loop before starting message collection
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sm-bluesky-plan.R
\name{sm_plan_first_attributes_bluesky}
\alias{sm_plan_first_attributes_bluesky}
\title{returns a list with the custom social media plan attributes with the default values for a first plan in a topic.}
\usage{
sm_plan_first_attributes_bluesky()
}
\description{
returns a list with the custom social media plan attributes with the default values for a first plan in a topic.
}
